;
; Genital68K Version 0.22: Motorola 68000 emulator
; Copyright 2000, 2001 Bart Trzynadlowski, see "README.TXT" for terms of use
; Assemble with NASM (http://www.web-sites.co.uk/nasm) only
;
; Configuration:
; - 24-bit addresses
; - Illegal instruction, Line 1010, and Line 1111 exceptions
; - Idle loop skipping disabled
; - Fetch pointer updated for Bcc, BRA, BSR, and DBcc
; - Data regions used for PC-relative reading
; - Single address space for supervisor and user
; - Definable memory map arrays
; - Stack calling conventions
;
bits 32
section .data
times ($$-$) & 31 nop
global genital68kcontext, _genital68kcontext, genital68kcontext_
genital68kcontext:
_genital68kcontext:
genital68kcontext_:
context_start:
fetch      dd 0
pcfetch    dd 0
read_byte  dd 0
read_word  dd 0
read_long  dd 0
write_byte dd 0
write_word dd 0
write_long dd 0
super_fetch      dd 0
super_pcfetch    dd 0
super_read_byte  dd 0
super_read_word  dd 0
super_read_long  dd 0
super_write_byte dd 0
super_write_word dd 0
super_write_long dd 0
user_fetch       dd 0
user_pcfetch     dd 0
user_read_byte   dd 0
user_read_word   dd 0
user_read_long   dd 0
user_write_byte  dd 0
user_write_word  dd 0
user_write_long  dd 0
intr       times 8 dd 0
cycles     dd 0
remaining  dd 0
d          times 8 dd 0
a          times 8 dd 0
__sp       dd 0
sr         dd 0
pc         dd 0
status     dd 0
Reset      dd 0
context_end:
x          dd 0
memhandler_eax    dd 0
memhandler_ebx    dd 0
memhandler_ecx    dd 0
memhandler_edx    dd 0
memhandler_esi    dd 0
memhandler_edi    dd 0
memhandler_ebp    dd 0
run_eax    dd 0
run_ebx    dd 0
run_ecx    dd 0
run_edx    dd 0
run_esi    dd 0
run_edi    dd 0
run_ebp    dd 0
safety_eax    dd 0
safety_ebx    dd 0
safety_ecx    dd 0
safety_edx    dd 0
safety_esi    dd 0
safety_edi    dd 0
safety_ebp    dd 0
fetch_esi dd 0
section .text
times ($$-$) & 3 nop
global Genital68KInit, _Genital68KInit, Genital68KInit_
Genital68KInit:
_Genital68KInit:
Genital68KInit_:
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
mov  esi, compressed_jmptab
mov  edi, jmptab
xor  ecx, ecx
.l:
mov  cx, [esi]
add  esi, byte 2
mov  edx, ecx
and  edx, 0xc000
cmp  dx, 0xc000
je   .r1
cmp  dx, 0x8000
je   .r8
mov  eax, [esi]
add  esi, byte 4
cld
rep  stosd
.c:
cmp  dword [esi], byte -1
jne  .l
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
xor  eax, eax
ret
times ($$-$) & 3 nop
.r1:
and  ecx, 0x3ff
cld
rep  movsd
jmp  .c
times ($$-$) & 3 nop
.r8:
and  ecx, 0x7ff
.r8_l:
mov  edx, ecx
mov  ecx, 8
mov  eax, [esi]
add  esi, byte 4
cld
rep  stosd
mov  ecx, edx
dec  ecx
jnz  .r8_l
jmp  .c
times ($$-$) & 3 nop
global Genital68KReset, _Genital68KReset, Genital68KReset_
Genital68KReset:
_Genital68KReset:
Genital68KReset_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
or   byte [status], 1
and  byte [status], 0xfd
mov  ebx, 4
call ReadLong
mov  dword [sr], 0x2700
mov  dword [cycles], 0
mov  dword [remaining], 0
mov  esi, edx
xor  eax, eax
mov  edi, d
mov  ecx, 16
cld
rep  stosd
mov  edi, intr
mov  ecx, 8
rep  stosd
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  dword [pc], esi
sub  dword [pc], ebp
xor  ebx, ebx
xor  ecx, ecx
call ReadLong
mov  [a+7*4], edx
mov  [__sp], edx
and  byte [status], 0xfe
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
xor  eax, eax
ret
invalid_error:
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
and  byte [status], 0xfe
mov [remaining], ecx
mov  dword [pc], esi
sub  dword [pc], ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
pop  ebp
mov  eax, 2
ret
fetch_error:
mov  [remaining], ecx
and  byte [status], 0xfe
mov  [pc], esi
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
pop  ebp
mov  eax, 1
ret
times ($$-$) & 3 nop
global Genital68KReadPC, _Genital68KReadPC, Genital68KReadPC_
Genital68KReadPC:
_Genital68KReadPC:
Genital68KReadPC_:
mov  eax, [pc]
ret
times ($$-$) & 3 nop
global Genital68KSetFetch, _Genital68KSetFetch, Genital68KSetFetch_
Genital68KSetFetch:
_Genital68KSetFetch:
Genital68KSetFetch_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 12
mov  [fetch], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetFetch, _Genital68KGetFetch, Genital68KGetFetch_
Genital68KGetFetch:
_Genital68KGetFetch:
Genital68KGetFetch_:
mov  [safety_edx], edx
mov  eax, [fetch]
mov  edx, [safety_edx]
add  eax, 12
ret
times ($$-$) & 3 nop
global Genital68KSetReadByte, _Genital68KSetReadByte, Genital68KSetReadByte_
Genital68KSetReadByte:
_Genital68KSetReadByte:
Genital68KSetReadByte_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [read_byte], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetReadByte, _Genital68KGetReadByte, Genital68KGetReadByte_
Genital68KGetReadByte:
_Genital68KGetReadByte:
Genital68KGetReadByte_:
mov  [safety_edx], edx
mov  eax, [read_byte]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KSetReadWord, _Genital68KSetReadWord, Genital68KSetReadWord_
Genital68KSetReadWord:
_Genital68KSetReadWord:
Genital68KSetReadWord_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [read_word], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetReadWord, _Genital68KGetReadWord, Genital68KGetReadWord_
Genital68KGetReadWord:
_Genital68KGetReadWord:
Genital68KGetReadWord_:
mov  [safety_edx], edx
mov  eax, [read_word]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KSetReadLong, _Genital68KSetReadLong, Genital68KSetReadLong_
Genital68KSetReadLong:
_Genital68KSetReadLong:
Genital68KSetReadLong_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [read_long], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetReadLong, _Genital68KGetReadLong, Genital68KGetReadLong_
Genital68KGetReadLong:
_Genital68KGetReadLong:
Genital68KGetReadLong_:
mov  [safety_edx], edx
mov  eax, [read_long]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KSetWriteByte, _Genital68KSetWriteByte, Genital68KSetWriteByte_
Genital68KSetWriteByte:
_Genital68KSetWriteByte:
Genital68KSetWriteByte_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [write_byte], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetWriteByte, _Genital68KGetWriteByte, Genital68KGetWriteByte_
Genital68KGetWriteByte:
_Genital68KGetWriteByte:
Genital68KGetWriteByte_:
mov  [safety_edx], edx
mov  eax, [write_byte]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KSetWriteWord, _Genital68KSetWriteWord, Genital68KSetWriteWord_
Genital68KSetWriteWord:
_Genital68KSetWriteWord:
Genital68KSetWriteWord_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [write_word], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetWriteWord, _Genital68KGetWriteWord, Genital68KGetWriteWord_
Genital68KGetWriteWord:
_Genital68KGetWriteWord:
Genital68KGetWriteWord_:
mov  [safety_edx], edx
mov  eax, [write_word]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KSetWriteLong, _Genital68KSetWriteLong, Genital68KSetWriteLong_
Genital68KSetWriteLong:
_Genital68KSetWriteLong:
Genital68KSetWriteLong_:
mov  [safety_eax], eax
mov  [safety_edx], edx
mov  eax, [esp+4+(0*4)]
sub  eax, 16
mov  [write_long], eax
mov  edx, [safety_edx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KGetWriteLong, _Genital68KGetWriteLong, Genital68KGetWriteLong_
Genital68KGetWriteLong:
_Genital68KGetWriteLong:
Genital68KGetWriteLong_:
mov  [safety_edx], edx
mov  eax, [write_long]
add  eax, 16
mov  edx, [safety_edx]
ret
times ($$-$) & 3 nop
global Genital68KFetchPtr, _Genital68KFetchPtr, Genital68KFetchPtr_
Genital68KFetchPtr:
_Genital68KFetchPtr:
Genital68KFetchPtr_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
mov  esi, [esp+8+(0*4)]
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop:
add     edx, byte 12
cmp     dword [edx], byte -1
je      short .not_found
cmp     esi, [edx]
jb      short .find_fetch_loop
cmp     esi, [edx+4]
ja      short .find_fetch_loop
mov     ebp, [edx+8]
add     esi, ebp
mov     eax, esi
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
.not_found:
xor  eax, eax
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KReadByte, _Genital68KReadByte, Genital68KReadByte_
Genital68KReadByte:
_Genital68KReadByte:
Genital68KReadByte_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
call ReadByte
mov  eax, edx
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KReadWord, _Genital68KReadWord, Genital68KReadWord_
Genital68KReadWord:
_Genital68KReadWord:
Genital68KReadWord_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
call ReadWord
mov  eax, edx
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KReadLong, _Genital68KReadLong, Genital68KReadLong_
Genital68KReadLong:
_Genital68KReadLong:
Genital68KReadLong_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
call ReadLong
mov  eax, edx
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KWriteByte, _Genital68KWriteByte, Genital68KWriteByte_
Genital68KWriteByte:
_Genital68KWriteByte:
Genital68KWriteByte_:
mov  [safety_eax], eax
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
mov  edx, [esp+8+(1*4)]
call WriteByte
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KWriteWord, _Genital68KWriteWord, Genital68KWriteWord_
Genital68KWriteWord:
_Genital68KWriteWord:
Genital68KWriteWord_:
mov  [safety_eax], eax
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
mov  edx, [esp+8+(1*4)]
call WriteWord
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KWriteLong, _Genital68KWriteLong, Genital68KWriteLong_
Genital68KWriteLong:
_Genital68KWriteLong:
Genital68KWriteLong_:
mov  [safety_eax], eax
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
xor  ebp, ebp
mov  esi, [pc]
mov  ebx, [esp+8+(0*4)]
mov  edx, [esp+8+(1*4)]
call WriteLong
pop  ebp
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
mov  eax, [safety_eax]
ret
times ($$-$) & 3 nop
global Genital68KSetContext, _Genital68KSetContext, Genital68KSetContext_
Genital68KSetContext:
_Genital68KSetContext:
Genital68KSetContext_:
mov  [safety_ecx], ecx
mov  [safety_esi], esi
mov  [safety_edi], edi
mov  esi, [esp+4+(0*4)]
mov  edi, genital68kcontext
mov  ecx, (context_end - context_start) / 4
cld
rep  movsd
sub  dword [read_byte], 16
sub  dword [read_word], 16
sub  dword [read_long], 16
sub  dword [write_byte], 16
sub  dword [write_word], 16
sub  dword [write_long], 16
sub  dword [fetch], 12
sub  dword [pcfetch], 12
sub  dword [super_read_byte], 16
sub  dword [super_read_word], 16
sub  dword [super_read_long], 16
sub  dword [super_write_byte], 16
sub  dword [super_write_word], 16
sub  dword [super_write_long], 16
sub  dword [super_fetch], 12
sub  dword [super_pcfetch], 12
sub  dword [user_read_byte], 16
sub  dword [user_read_word], 16
sub  dword [user_read_long], 16
sub  dword [user_write_byte], 16
sub  dword [user_write_word], 16
sub  dword [user_write_long], 16
sub  dword [user_fetch], 12
sub  dword [user_pcfetch], 12
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  ecx, [safety_ecx]
ret
times ($$-$) & 3 nop
global Genital68KGetContext, _Genital68KGetContext, Genital68KGetContext_
Genital68KGetContext:
_Genital68KGetContext:
Genital68KGetContext_:
mov  [safety_ecx], ecx
mov  [safety_esi], esi
mov  [safety_edi], edi
add  dword [read_byte], 16
add  dword [read_word], 16
add  dword [read_long], 16
add  dword [write_byte], 16
add  dword [write_word], 16
add  dword [write_long], 16
add  dword [fetch], 12
add  dword [pcfetch], 12
add  dword [super_read_byte], 16
add  dword [super_read_word], 16
add  dword [super_read_long], 16
add  dword [super_write_byte], 16
add  dword [super_write_word], 16
add  dword [super_write_long], 16
add  dword [super_fetch], 12
add  dword [super_pcfetch], 12
add  dword [user_read_byte], 16
add  dword [user_read_word], 16
add  dword [user_read_long], 16
add  dword [user_write_byte], 16
add  dword [user_write_word], 16
add  dword [user_write_long], 16
add  dword [user_fetch], 12
add  dword [user_pcfetch], 12
mov  edi, [esp+4+(0*4)]
mov  esi, genital68kcontext
mov  ecx, (context_end - context_start) / 4
cld
rep  movsd
sub  dword [read_byte], 16
sub  dword [read_word], 16
sub  dword [read_long], 16
sub  dword [write_byte], 16
sub  dword [write_word], 16
sub  dword [write_long], 16
sub  dword [fetch], 12
sub  dword [pcfetch], 12
sub  dword [super_read_byte], 16
sub  dword [super_read_word], 16
sub  dword [super_read_long], 16
sub  dword [super_write_byte], 16
sub  dword [super_write_word], 16
sub  dword [super_write_long], 16
sub  dword [super_fetch], 12
sub  dword [super_pcfetch], 12
sub  dword [user_read_byte], 16
sub  dword [user_read_word], 16
sub  dword [user_read_long], 16
sub  dword [user_write_byte], 16
sub  dword [user_write_word], 16
sub  dword [user_write_long], 16
sub  dword [user_fetch], 12
sub  dword [user_pcfetch], 12
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  ecx, [safety_ecx]
ret
times ($$-$) & 3 nop
global Genital68KGetContextSize, _Genital68KGetContextSize, Genital68KGetContextSize_
Genital68KGetContextSize:
_Genital68KGetContextSize:
Genital68KGetContextSize_:
mov  eax, (context_end - context_start)
ret
global Genital68KClearCycles, _Genital68KClearCycles, Genital68KClearCycles_
Genital68KClearCycles:
_Genital68KClearCycles:
Genital68KClearCycles_:
mov  dword [cycles], 0
mov  dword [remaining], 0
ret
global Genital68KFreeTimeSlice, _Genital68KFreeTimeSlice, Genital68KFreeTimeSlice_
Genital68KFreeTimeSlice:
_Genital68KFreeTimeSlice:
Genital68KFreeTimeSlice_:
mov  [safety_eax], eax
mov  eax, [cycles]
sub  eax, [remaining]
mov  [cycles], eax
mov  dword [remaining], 0
mov  eax, [safety_eax]
ret
global Genital68KGetElapsedCycles, _Genital68KGetElapsedCycles, Genital68KGetElapsedCycles_
Genital68KGetElapsedCycles:
_Genital68KGetElapsedCycles:
Genital68KGetElapsedCycles_:
mov  eax, [cycles]
sub  eax, [remaining]
ret
times ($$-$) & 3 nop
global Genital68KProcessInterrupts, _Genital68KProcessInterrupts, Genital68KProcessInterrupts_
Genital68KProcessInterrupts:
_Genital68KProcessInterrupts:
Genital68KProcessInterrupts_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
test byte [status], 1
jnz  near .end
push ebp
mov  ecx, [remaining]
or   byte [status], 1
mov  esi, [pc]
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
call ProcessInterrupts
mov [remaining], ecx
mov  dword [pc], esi
sub  dword [pc], ebp
and  byte [status], 0xfe
pop  ebp
.end:
xor  eax, eax
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KInterrupt, _Genital68KInterrupt, Genital68KInterrupt_
Genital68KInterrupt:
_Genital68KInterrupt:
Genital68KInterrupt_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
mov  eax, [esp+4+(0*4)]
test eax, -1
jz   short .inv_level
cmp  eax, byte 7
ja   short .inv_level
mov  ebx, [esp+4+(1*4)]
cmp  ebx, byte 2
jb   short .inv_vector
cmp  ebx, 256
ja   short .inv_vector
jne  short .not_auto
mov  ebx, eax
add  ebx, byte 24
.not_auto:
dec  eax
shl  eax, byte 2
cmp  [intr+eax], byte 0
jne  short .pending
mov  [intr+eax], ebx
inc  dword [intr+7*4]
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
xor  eax, eax
ret
.inv_level:
mov  eax, 3
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
.inv_vector:
mov  eax, 4
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
.pending:
mov  eax, 5
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 3 nop
global Genital68KCancelInterrupt, _Genital68KCancelInterrupt, Genital68KCancelInterrupt_
Genital68KCancelInterrupt:
_Genital68KCancelInterrupt:
Genital68KCancelInterrupt_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
mov  eax, [esp+4+(0*4)]
test eax, 0xffffffff
jz   short .inv_level
cmp  eax, byte 7
ja   short .inv_level
test byte [status], 0x04
jnz  short .busy
dec  eax
cmp  dword [intr+eax], byte 0
je   .nothing_to_cancel
mov  dword [intr+eax], 0
dec  dword [intr+7*4]
.busy:
.nothing_to_cancel:
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
xor  eax, eax
ret
.inv_level:
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
mov  eax, 3
ret
times ($$-$) & 3 nop
ProcessInterrupts:
or   byte [status], 4
cmp  dword [intr+7*4], byte 0
je   near .no_int_end
mov  [run_eax], eax
mov  al, [sr+1]
and  eax, byte 7
cmp  al, 7
jne  short .no_int7
mov  edi, intr+6*4
jmp  short .l
.no_int7:
mov  edi, eax
shl  edi, byte 2
add  edi, intr
inc  al
.l:
cmp  dword [edi], byte 0
je   near .skip
test byte [sr+1], 0x20
jnz  short .no_sp_magic
mov  edx, [__sp]
xchg [a+7*4], edx
mov  [__sp], edx
.no_sp_magic:
mov  ebx, [a+7*4]
mov  edx, [pc]
sub  ebx, byte 4
mov  [run_ebx], ebx
mov  [run_edi], edi
call WriteLong
mov  edi, [run_edi]
mov  ebx, [run_ebx]
mov  edx, [sr]
sub  ebx, byte 2
mov  [run_edi], edi
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
mov  edi, [run_edi]
mov  dh, al
or   dh, 0x20
and  edx, 0xa71f
mov  [sr], edx
mov  [a+7*4], ebx
mov  ebx, [edi]
shl  ebx, byte 2
mov  [run_edi], edi
call ReadLong
mov  edi, [run_edi]
mov  [pc], edx
mov  dword [edi], 0
dec  dword [intr+7*4]
sub  ecx, byte 44
and  byte [status], 0xfd
.skip:
add  edi, byte 4
inc  al
cmp  edi, intr+7*4
jne  near .l
mov  esi, [pc]
mov  eax, [run_eax]
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
.no_int_end:
and  byte [status], 0xfb
ret
times ($$-$) & 3 nop
global Genital68KRun, _Genital68KRun, Genital68KRun_
Genital68KRun:
_Genital68KRun:
Genital68KRun_:
mov  [safety_ebx], ebx
mov  [safety_ecx], ecx
mov  [safety_edx], edx
mov  [safety_esi], esi
mov  [safety_edi], edi
push ebp
mov  ecx, [esp+8+(0*4)]
mov  [cycles], ecx
mov  [remaining], ecx
test byte [sr+1], 0x20
jz   near .set_user_space
jmp  near .continue
.set_user_space:
.continue:
or   byte [status], 1
mov  esi, [pc]
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
call ProcessInterrupts
test byte [status], 2
jnz  short .stopped
xor  edi, edi
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.stopped:
xor  ecx, ecx
jmp  Genital68KRun_done
times ($$-$) & 3 nop
Genital68KRun_done:
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov [remaining], ecx
mov  dword [pc], esi
sub  dword [pc], ebp
and  byte [status], 0xfe
pop  ebp
xor  eax, eax
mov  edi, [safety_edi]
mov  esi, [safety_esi]
mov  edx, [safety_edx]
mov  ecx, [safety_ecx]
mov  ebx, [safety_ebx]
ret
times ($$-$) & 31 nop
ReadByte:
mov  edi, [read_byte]
mov  [memhandler_ebx], ebx
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .read_from_handler
mov     edx, ebx
xor     dl, 1
add     edx, [edi+8]
mov     dl, [edx]
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.not_found:
xor  edi, edi
xor  edx, edx
dec  edx
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.read_from_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
push ebx
call    dword [edi+12]
add     esp, byte 4
mov     edx, eax
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 31 nop
ReadWord:
mov  edi, [read_word]
mov  [memhandler_ebx], ebx
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .read_from_handler
mov     edx, ebx
add     edx, [edi+8]
mov     edx, [edx]
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.not_found:
xor  edi, edi
xor  edx, edx
dec  edx
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.read_from_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
push ebx
call    dword [edi+12]
add     esp, byte 4
mov     edx, eax
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 31 nop
ReadLong:
mov  edi, [read_long]
mov  [memhandler_ebx], ebx
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .read_from_handler
mov     edx, ebx
add     edx, [edi+8]
mov     edx, [edx]
rol     edx, byte 16
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.not_found:
xor  edi, edi
xor  edx, edx
dec  edx
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.read_from_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
push ebx
call    dword [edi+12]
add     esp, byte 4
mov     edx, eax
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 31 nop
ReadWordSX:
mov  edi, [read_word]
mov  [memhandler_ebx], ebx
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .read_from_handler
mov     edx, ebx
add     edx, [edi+8]
movsx   edx, word [edx]
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.not_found:
xor  edi, edi
xor  edx, edx
dec  edx
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.read_from_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
push ebx
call    dword [edi+12]
add     esp, byte 4
movsx   edx, ax
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 31 nop
WriteByte:
mov  edi, [write_byte]
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .write_with_handler
xor     bl, 1
add     ebx, [edi+8]
mov     [ebx], dl
.not_found:
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.write_with_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
mov  [memhandler_edx], edx
push edx
push ebx
call    dword [edi+12]
add     esp, byte 8
mov  edx, [memhandler_edx]
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
ret
times ($$-$) & 31 nop
WriteWord:
mov  edi, [write_word]
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .write_with_handler
add     ebx, [edi+8]
mov     [ebx], dx
.not_found:
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.write_with_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
mov  [memhandler_edx], edx
push edx
push ebx
call    dword [edi+12]
add     esp, byte 8
mov  edx, [memhandler_edx]
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
ret
times ($$-$) & 31 nop
WriteLong:
mov  edi, [write_long]
and  ebx, 0x00FFFFFF
.loop:
add     edi, byte 16
cmp     dword [edi], byte -1
je      short .not_found
cmp     ebx, [edi]
jb      short .loop
cmp     ebx, [edi+4]
ja      short .loop
cmp     dword [edi+8], byte 0
je      short .write_with_handler
add     ebx, [edi+8]
ror     edx, byte 16
mov     [ebx], edx
ror     edx, byte 16
.not_found:
xor  edi, edi
mov  ebx, [memhandler_ebx]
ret
times ($$-$) & 3 nop
.write_with_handler:
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
mov  [memhandler_edx], edx
push edx
push ebx
call    dword [edi+12]
add     esp, byte 8
mov  edx, [memhandler_edx]
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
ret
times ($$-$) & 3 nop
IC100: ; ABCD		0xC100-0xC107
and  edi, byte 7
mov  al, [d+0*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+0*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC108: ; ABCD		0xC108-0xC10F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+0*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+0*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC300: ; ABCD		0xC300-0xC307
and  edi, byte 7
mov  al, [d+1*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+1*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC308: ; ABCD		0xC308-0xC30F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+1*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+1*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC500: ; ABCD		0xC500-0xC507
and  edi, byte 7
mov  al, [d+2*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+2*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC508: ; ABCD		0xC508-0xC50F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+2*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+2*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC700: ; ABCD		0xC700-0xC707
and  edi, byte 7
mov  al, [d+3*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+3*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC708: ; ABCD		0xC708-0xC70F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+3*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+3*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC900: ; ABCD		0xC900-0xC907
and  edi, byte 7
mov  al, [d+4*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+4*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC908: ; ABCD		0xC908-0xC90F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+4*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+4*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB00: ; ABCD		0xCB00-0xCB07
and  edi, byte 7
mov  al, [d+5*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+5*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB08: ; ABCD		0xCB08-0xCB0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+5*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+5*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD00: ; ABCD		0xCD00-0xCD07
and  edi, byte 7
mov  al, [d+6*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+6*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD08: ; ABCD		0xCD08-0xCD0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+6*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+6*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF00: ; ABCD		0xCF00-0xCF07
and  edi, byte 7
mov  al, [d+7*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
adc  al, bl
daa
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, dl
mov  [d+7*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF08: ; ABCD		0xCF08-0xCF0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+7*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
adc  al, dl
daa
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0xfe
or   ah, [x]
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8000: ; OR		0x8000-0x8007
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8010: ; OR		0x8010-0x8017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8018: ; OR		0x8018-0x801F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8020: ; OR		0x8020-0x8027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8028: ; OR		0x8028-0x802F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8030: ; OR		0x8030-0x8037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8038: ; OR		0x8038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8039: ; OR		0x8039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I803A: ; OR		0x803A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I803B: ; OR		0x803B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I803C: ; OR		0x803C
mov  dl, [esi]
add  esi, byte 2
or   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8040: ; OR		0x8040-0x8047
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8050: ; OR		0x8050-0x8057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8058: ; OR		0x8058-0x805F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8060: ; OR		0x8060-0x8067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8068: ; OR		0x8068-0x806F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8070: ; OR		0x8070-0x8077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8078: ; OR		0x8078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8079: ; OR		0x8079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I807A: ; OR		0x807A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I807B: ; OR		0x807B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I807C: ; OR		0x807C
mov  edx, [esi]
add  esi, byte 2
or   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8080: ; OR		0x8080-0x8087
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8090: ; OR		0x8090-0x8097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8098: ; OR		0x8098-0x809F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80A0: ; OR		0x80A0-0x80A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80A8: ; OR		0x80A8-0x80AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80B0: ; OR		0x80B0-0x80B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80B8: ; OR		0x80B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80B9: ; OR		0x80B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80BA: ; OR		0x80BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80BB: ; OR		0x80BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I80BC: ; OR		0x80BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8110: ; OR		0x8110-0x8117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8118: ; OR		0x8118-0x811F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8120: ; OR		0x8120-0x8127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8128: ; OR		0x8128-0x812F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8130: ; OR		0x8130-0x8137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8138: ; OR		0x8138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8139: ; OR		0x8139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8150: ; OR		0x8150-0x8157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8158: ; OR		0x8158-0x815F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8160: ; OR		0x8160-0x8167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8168: ; OR		0x8168-0x816F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8170: ; OR		0x8170-0x8177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8178: ; OR		0x8178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8179: ; OR		0x8179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8190: ; OR		0x8190-0x8197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8198: ; OR		0x8198-0x819F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81A0: ; OR		0x81A0-0x81A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81A8: ; OR		0x81A8-0x81AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81B0: ; OR		0x81B0-0x81B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81B8: ; OR		0x81B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81B9: ; OR		0x81B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8200: ; OR		0x8200-0x8207
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8210: ; OR		0x8210-0x8217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8218: ; OR		0x8218-0x821F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8220: ; OR		0x8220-0x8227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8228: ; OR		0x8228-0x822F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8230: ; OR		0x8230-0x8237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8238: ; OR		0x8238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8239: ; OR		0x8239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I823A: ; OR		0x823A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I823B: ; OR		0x823B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I823C: ; OR		0x823C
mov  dl, [esi]
add  esi, byte 2
or   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8240: ; OR		0x8240-0x8247
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8250: ; OR		0x8250-0x8257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8258: ; OR		0x8258-0x825F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8260: ; OR		0x8260-0x8267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8268: ; OR		0x8268-0x826F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8270: ; OR		0x8270-0x8277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8278: ; OR		0x8278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8279: ; OR		0x8279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I827A: ; OR		0x827A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I827B: ; OR		0x827B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I827C: ; OR		0x827C
mov  edx, [esi]
add  esi, byte 2
or   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8280: ; OR		0x8280-0x8287
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8290: ; OR		0x8290-0x8297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8298: ; OR		0x8298-0x829F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82A0: ; OR		0x82A0-0x82A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82A8: ; OR		0x82A8-0x82AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82B0: ; OR		0x82B0-0x82B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82B8: ; OR		0x82B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82B9: ; OR		0x82B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82BA: ; OR		0x82BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82BB: ; OR		0x82BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I82BC: ; OR		0x82BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8310: ; OR		0x8310-0x8317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8318: ; OR		0x8318-0x831F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8320: ; OR		0x8320-0x8327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8328: ; OR		0x8328-0x832F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8330: ; OR		0x8330-0x8337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8338: ; OR		0x8338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8339: ; OR		0x8339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8350: ; OR		0x8350-0x8357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8358: ; OR		0x8358-0x835F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8360: ; OR		0x8360-0x8367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8368: ; OR		0x8368-0x836F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8370: ; OR		0x8370-0x8377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8378: ; OR		0x8378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8379: ; OR		0x8379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8390: ; OR		0x8390-0x8397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8398: ; OR		0x8398-0x839F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I83A0: ; OR		0x83A0-0x83A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I83A8: ; OR		0x83A8-0x83AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I83B0: ; OR		0x83B0-0x83B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I83B8: ; OR		0x83B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I83B9: ; OR		0x83B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8400: ; OR		0x8400-0x8407
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8410: ; OR		0x8410-0x8417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8418: ; OR		0x8418-0x841F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8420: ; OR		0x8420-0x8427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8428: ; OR		0x8428-0x842F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8430: ; OR		0x8430-0x8437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8438: ; OR		0x8438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8439: ; OR		0x8439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I843A: ; OR		0x843A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I843B: ; OR		0x843B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I843C: ; OR		0x843C
mov  dl, [esi]
add  esi, byte 2
or   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8440: ; OR		0x8440-0x8447
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8450: ; OR		0x8450-0x8457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8458: ; OR		0x8458-0x845F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8460: ; OR		0x8460-0x8467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8468: ; OR		0x8468-0x846F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8470: ; OR		0x8470-0x8477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8478: ; OR		0x8478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8479: ; OR		0x8479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I847A: ; OR		0x847A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I847B: ; OR		0x847B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I847C: ; OR		0x847C
mov  edx, [esi]
add  esi, byte 2
or   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8480: ; OR		0x8480-0x8487
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8490: ; OR		0x8490-0x8497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8498: ; OR		0x8498-0x849F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84A0: ; OR		0x84A0-0x84A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84A8: ; OR		0x84A8-0x84AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84B0: ; OR		0x84B0-0x84B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84B8: ; OR		0x84B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84B9: ; OR		0x84B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84BA: ; OR		0x84BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84BB: ; OR		0x84BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I84BC: ; OR		0x84BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8510: ; OR		0x8510-0x8517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8518: ; OR		0x8518-0x851F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8520: ; OR		0x8520-0x8527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8528: ; OR		0x8528-0x852F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8530: ; OR		0x8530-0x8537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8538: ; OR		0x8538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8539: ; OR		0x8539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8550: ; OR		0x8550-0x8557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8558: ; OR		0x8558-0x855F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8560: ; OR		0x8560-0x8567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8568: ; OR		0x8568-0x856F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8570: ; OR		0x8570-0x8577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8578: ; OR		0x8578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8579: ; OR		0x8579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8590: ; OR		0x8590-0x8597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8598: ; OR		0x8598-0x859F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I85A0: ; OR		0x85A0-0x85A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I85A8: ; OR		0x85A8-0x85AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I85B0: ; OR		0x85B0-0x85B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I85B8: ; OR		0x85B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I85B9: ; OR		0x85B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8600: ; OR		0x8600-0x8607
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8610: ; OR		0x8610-0x8617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8618: ; OR		0x8618-0x861F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8620: ; OR		0x8620-0x8627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8628: ; OR		0x8628-0x862F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8630: ; OR		0x8630-0x8637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8638: ; OR		0x8638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8639: ; OR		0x8639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I863A: ; OR		0x863A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I863B: ; OR		0x863B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I863C: ; OR		0x863C
mov  dl, [esi]
add  esi, byte 2
or   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8640: ; OR		0x8640-0x8647
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8650: ; OR		0x8650-0x8657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8658: ; OR		0x8658-0x865F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8660: ; OR		0x8660-0x8667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8668: ; OR		0x8668-0x866F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8670: ; OR		0x8670-0x8677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8678: ; OR		0x8678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8679: ; OR		0x8679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I867A: ; OR		0x867A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I867B: ; OR		0x867B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I867C: ; OR		0x867C
mov  edx, [esi]
add  esi, byte 2
or   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8680: ; OR		0x8680-0x8687
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8690: ; OR		0x8690-0x8697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8698: ; OR		0x8698-0x869F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86A0: ; OR		0x86A0-0x86A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86A8: ; OR		0x86A8-0x86AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86B0: ; OR		0x86B0-0x86B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86B8: ; OR		0x86B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86B9: ; OR		0x86B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86BA: ; OR		0x86BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86BB: ; OR		0x86BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I86BC: ; OR		0x86BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8710: ; OR		0x8710-0x8717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8718: ; OR		0x8718-0x871F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8720: ; OR		0x8720-0x8727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8728: ; OR		0x8728-0x872F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8730: ; OR		0x8730-0x8737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8738: ; OR		0x8738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8739: ; OR		0x8739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8750: ; OR		0x8750-0x8757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8758: ; OR		0x8758-0x875F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8760: ; OR		0x8760-0x8767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8768: ; OR		0x8768-0x876F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8770: ; OR		0x8770-0x8777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8778: ; OR		0x8778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8779: ; OR		0x8779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8790: ; OR		0x8790-0x8797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8798: ; OR		0x8798-0x879F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I87A0: ; OR		0x87A0-0x87A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I87A8: ; OR		0x87A8-0x87AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I87B0: ; OR		0x87B0-0x87B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I87B8: ; OR		0x87B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I87B9: ; OR		0x87B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8800: ; OR		0x8800-0x8807
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8810: ; OR		0x8810-0x8817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8818: ; OR		0x8818-0x881F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8820: ; OR		0x8820-0x8827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8828: ; OR		0x8828-0x882F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8830: ; OR		0x8830-0x8837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8838: ; OR		0x8838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8839: ; OR		0x8839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I883A: ; OR		0x883A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I883B: ; OR		0x883B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I883C: ; OR		0x883C
mov  dl, [esi]
add  esi, byte 2
or   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8840: ; OR		0x8840-0x8847
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8850: ; OR		0x8850-0x8857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8858: ; OR		0x8858-0x885F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8860: ; OR		0x8860-0x8867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8868: ; OR		0x8868-0x886F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8870: ; OR		0x8870-0x8877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8878: ; OR		0x8878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8879: ; OR		0x8879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I887A: ; OR		0x887A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I887B: ; OR		0x887B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I887C: ; OR		0x887C
mov  edx, [esi]
add  esi, byte 2
or   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8880: ; OR		0x8880-0x8887
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8890: ; OR		0x8890-0x8897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8898: ; OR		0x8898-0x889F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88A0: ; OR		0x88A0-0x88A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88A8: ; OR		0x88A8-0x88AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88B0: ; OR		0x88B0-0x88B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88B8: ; OR		0x88B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88B9: ; OR		0x88B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88BA: ; OR		0x88BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88BB: ; OR		0x88BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I88BC: ; OR		0x88BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8910: ; OR		0x8910-0x8917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8918: ; OR		0x8918-0x891F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8920: ; OR		0x8920-0x8927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8928: ; OR		0x8928-0x892F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8930: ; OR		0x8930-0x8937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8938: ; OR		0x8938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8939: ; OR		0x8939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8950: ; OR		0x8950-0x8957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8958: ; OR		0x8958-0x895F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8960: ; OR		0x8960-0x8967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8968: ; OR		0x8968-0x896F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8970: ; OR		0x8970-0x8977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8978: ; OR		0x8978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8979: ; OR		0x8979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8990: ; OR		0x8990-0x8997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8998: ; OR		0x8998-0x899F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I89A0: ; OR		0x89A0-0x89A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I89A8: ; OR		0x89A8-0x89AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I89B0: ; OR		0x89B0-0x89B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I89B8: ; OR		0x89B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I89B9: ; OR		0x89B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A00: ; OR		0x8A00-0x8A07
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A10: ; OR		0x8A10-0x8A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A18: ; OR		0x8A18-0x8A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A20: ; OR		0x8A20-0x8A27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A28: ; OR		0x8A28-0x8A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A30: ; OR		0x8A30-0x8A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A38: ; OR		0x8A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A39: ; OR		0x8A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A3A: ; OR		0x8A3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A3B: ; OR		0x8A3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A3C: ; OR		0x8A3C
mov  dl, [esi]
add  esi, byte 2
or   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A40: ; OR		0x8A40-0x8A47
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A50: ; OR		0x8A50-0x8A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A58: ; OR		0x8A58-0x8A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A60: ; OR		0x8A60-0x8A67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A68: ; OR		0x8A68-0x8A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A70: ; OR		0x8A70-0x8A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A78: ; OR		0x8A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A79: ; OR		0x8A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A7A: ; OR		0x8A7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A7B: ; OR		0x8A7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A7C: ; OR		0x8A7C
mov  edx, [esi]
add  esi, byte 2
or   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A80: ; OR		0x8A80-0x8A87
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A90: ; OR		0x8A90-0x8A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8A98: ; OR		0x8A98-0x8A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8AA0: ; OR		0x8AA0-0x8AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8AA8: ; OR		0x8AA8-0x8AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8AB0: ; OR		0x8AB0-0x8AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8AB8: ; OR		0x8AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8AB9: ; OR		0x8AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8ABA: ; OR		0x8ABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8ABB: ; OR		0x8ABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8ABC: ; OR		0x8ABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B10: ; OR		0x8B10-0x8B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B18: ; OR		0x8B18-0x8B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B20: ; OR		0x8B20-0x8B27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B28: ; OR		0x8B28-0x8B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B30: ; OR		0x8B30-0x8B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B38: ; OR		0x8B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B39: ; OR		0x8B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B50: ; OR		0x8B50-0x8B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B58: ; OR		0x8B58-0x8B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B60: ; OR		0x8B60-0x8B67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B68: ; OR		0x8B68-0x8B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B70: ; OR		0x8B70-0x8B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B78: ; OR		0x8B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B79: ; OR		0x8B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B90: ; OR		0x8B90-0x8B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B98: ; OR		0x8B98-0x8B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8BA0: ; OR		0x8BA0-0x8BA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8BA8: ; OR		0x8BA8-0x8BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8BB0: ; OR		0x8BB0-0x8BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8BB8: ; OR		0x8BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8BB9: ; OR		0x8BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C00: ; OR		0x8C00-0x8C07
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C10: ; OR		0x8C10-0x8C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C18: ; OR		0x8C18-0x8C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C20: ; OR		0x8C20-0x8C27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C28: ; OR		0x8C28-0x8C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C30: ; OR		0x8C30-0x8C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C38: ; OR		0x8C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C39: ; OR		0x8C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C3A: ; OR		0x8C3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C3B: ; OR		0x8C3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C3C: ; OR		0x8C3C
mov  dl, [esi]
add  esi, byte 2
or   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C40: ; OR		0x8C40-0x8C47
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C50: ; OR		0x8C50-0x8C57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C58: ; OR		0x8C58-0x8C5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C60: ; OR		0x8C60-0x8C67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C68: ; OR		0x8C68-0x8C6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C70: ; OR		0x8C70-0x8C77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C78: ; OR		0x8C78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C79: ; OR		0x8C79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C7A: ; OR		0x8C7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C7B: ; OR		0x8C7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C7C: ; OR		0x8C7C
mov  edx, [esi]
add  esi, byte 2
or   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C80: ; OR		0x8C80-0x8C87
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C90: ; OR		0x8C90-0x8C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8C98: ; OR		0x8C98-0x8C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CA0: ; OR		0x8CA0-0x8CA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CA8: ; OR		0x8CA8-0x8CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CB0: ; OR		0x8CB0-0x8CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CB8: ; OR		0x8CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CB9: ; OR		0x8CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CBA: ; OR		0x8CBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CBB: ; OR		0x8CBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8CBC: ; OR		0x8CBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D10: ; OR		0x8D10-0x8D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D18: ; OR		0x8D18-0x8D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D20: ; OR		0x8D20-0x8D27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D28: ; OR		0x8D28-0x8D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D30: ; OR		0x8D30-0x8D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D38: ; OR		0x8D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D39: ; OR		0x8D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D50: ; OR		0x8D50-0x8D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D58: ; OR		0x8D58-0x8D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D60: ; OR		0x8D60-0x8D67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D68: ; OR		0x8D68-0x8D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D70: ; OR		0x8D70-0x8D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D78: ; OR		0x8D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D79: ; OR		0x8D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D90: ; OR		0x8D90-0x8D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D98: ; OR		0x8D98-0x8D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8DA0: ; OR		0x8DA0-0x8DA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8DA8: ; OR		0x8DA8-0x8DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8DB0: ; OR		0x8DB0-0x8DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8DB8: ; OR		0x8DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8DB9: ; OR		0x8DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E00: ; OR		0x8E00-0x8E07
and  edi, byte 7
mov  dl, [d+edi*4]
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E10: ; OR		0x8E10-0x8E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E18: ; OR		0x8E18-0x8E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E20: ; OR		0x8E20-0x8E27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E28: ; OR		0x8E28-0x8E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E30: ; OR		0x8E30-0x8E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E38: ; OR		0x8E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E39: ; OR		0x8E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E3A: ; OR		0x8E3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E3B: ; OR		0x8E3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E3C: ; OR		0x8E3C
mov  dl, [esi]
add  esi, byte 2
or   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E40: ; OR		0x8E40-0x8E47
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E50: ; OR		0x8E50-0x8E57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E58: ; OR		0x8E58-0x8E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E60: ; OR		0x8E60-0x8E67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E68: ; OR		0x8E68-0x8E6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E70: ; OR		0x8E70-0x8E77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E78: ; OR		0x8E78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E79: ; OR		0x8E79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E7A: ; OR		0x8E7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E7B: ; OR		0x8E7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E7C: ; OR		0x8E7C
mov  edx, [esi]
add  esi, byte 2
or   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E80: ; OR		0x8E80-0x8E87
and  edi, byte 7
mov  edx, [d+edi*4]
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E90: ; OR		0x8E90-0x8E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8E98: ; OR		0x8E98-0x8E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EA0: ; OR		0x8EA0-0x8EA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EA8: ; OR		0x8EA8-0x8EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EB0: ; OR		0x8EB0-0x8EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EB8: ; OR		0x8EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EB9: ; OR		0x8EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EBA: ; OR		0x8EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EBB: ; OR		0x8EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8EBC: ; OR		0x8EBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
or   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F10: ; OR		0x8F10-0x8F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F18: ; OR		0x8F18-0x8F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F20: ; OR		0x8F20-0x8F27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F28: ; OR		0x8F28-0x8F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F30: ; OR		0x8F30-0x8F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F38: ; OR		0x8F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F39: ; OR		0x8F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
or   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F50: ; OR		0x8F50-0x8F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F58: ; OR		0x8F58-0x8F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F60: ; OR		0x8F60-0x8F67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F68: ; OR		0x8F68-0x8F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F70: ; OR		0x8F70-0x8F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F78: ; OR		0x8F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F79: ; OR		0x8F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
or   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F90: ; OR		0x8F90-0x8F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F98: ; OR		0x8F98-0x8F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8FA0: ; OR		0x8FA0-0x8FA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8FA8: ; OR		0x8FA8-0x8FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8FB0: ; OR		0x8FB0-0x8FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8FB8: ; OR		0x8FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8FB9: ; OR		0x8FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
or   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9000: ; SUB		0x9000-0x9007
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9010: ; SUB		0x9010-0x9017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9018: ; SUB		0x9018-0x901F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9020: ; SUB		0x9020-0x9027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9028: ; SUB		0x9028-0x902F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9030: ; SUB		0x9030-0x9037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9038: ; SUB		0x9038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9039: ; SUB		0x9039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I903A: ; SUB		0x903A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I903B: ; SUB		0x903B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I903C: ; SUB		0x903C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9040: ; SUB		0x9040-0x9047
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9048: ; SUB		0x9048-0x904F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9050: ; SUB		0x9050-0x9057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9058: ; SUB		0x9058-0x905F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9060: ; SUB		0x9060-0x9067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9068: ; SUB		0x9068-0x906F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9070: ; SUB		0x9070-0x9077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9078: ; SUB		0x9078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9079: ; SUB		0x9079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I907A: ; SUB		0x907A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I907B: ; SUB		0x907B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I907C: ; SUB		0x907C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9080: ; SUB		0x9080-0x9087
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9088: ; SUB		0x9088-0x908F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9090: ; SUB		0x9090-0x9097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9098: ; SUB		0x9098-0x909F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90A0: ; SUB		0x90A0-0x90A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90A8: ; SUB		0x90A8-0x90AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90B0: ; SUB		0x90B0-0x90B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90B8: ; SUB		0x90B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90B9: ; SUB		0x90B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90BA: ; SUB		0x90BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90BB: ; SUB		0x90BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90BC: ; SUB		0x90BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9110: ; SUB		0x9110-0x9117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9118: ; SUB		0x9118-0x911F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9120: ; SUB		0x9120-0x9127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9128: ; SUB		0x9128-0x912F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9130: ; SUB		0x9130-0x9137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9138: ; SUB		0x9138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9139: ; SUB		0x9139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9150: ; SUB		0x9150-0x9157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9158: ; SUB		0x9158-0x915F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9160: ; SUB		0x9160-0x9167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9168: ; SUB		0x9168-0x916F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9170: ; SUB		0x9170-0x9177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9178: ; SUB		0x9178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9179: ; SUB		0x9179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9190: ; SUB		0x9190-0x9197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9198: ; SUB		0x9198-0x919F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91A0: ; SUB		0x91A0-0x91A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91A8: ; SUB		0x91A8-0x91AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91B0: ; SUB		0x91B0-0x91B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91B8: ; SUB		0x91B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91B9: ; SUB		0x91B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9200: ; SUB		0x9200-0x9207
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9210: ; SUB		0x9210-0x9217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9218: ; SUB		0x9218-0x921F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9220: ; SUB		0x9220-0x9227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9228: ; SUB		0x9228-0x922F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9230: ; SUB		0x9230-0x9237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9238: ; SUB		0x9238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9239: ; SUB		0x9239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I923A: ; SUB		0x923A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I923B: ; SUB		0x923B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I923C: ; SUB		0x923C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9240: ; SUB		0x9240-0x9247
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9248: ; SUB		0x9248-0x924F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9250: ; SUB		0x9250-0x9257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9258: ; SUB		0x9258-0x925F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9260: ; SUB		0x9260-0x9267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9268: ; SUB		0x9268-0x926F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9270: ; SUB		0x9270-0x9277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9278: ; SUB		0x9278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9279: ; SUB		0x9279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I927A: ; SUB		0x927A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I927B: ; SUB		0x927B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I927C: ; SUB		0x927C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9280: ; SUB		0x9280-0x9287
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9288: ; SUB		0x9288-0x928F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9290: ; SUB		0x9290-0x9297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9298: ; SUB		0x9298-0x929F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92A0: ; SUB		0x92A0-0x92A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92A8: ; SUB		0x92A8-0x92AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92B0: ; SUB		0x92B0-0x92B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92B8: ; SUB		0x92B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92B9: ; SUB		0x92B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92BA: ; SUB		0x92BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92BB: ; SUB		0x92BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92BC: ; SUB		0x92BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9310: ; SUB		0x9310-0x9317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9318: ; SUB		0x9318-0x931F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9320: ; SUB		0x9320-0x9327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9328: ; SUB		0x9328-0x932F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9330: ; SUB		0x9330-0x9337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9338: ; SUB		0x9338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9339: ; SUB		0x9339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9350: ; SUB		0x9350-0x9357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9358: ; SUB		0x9358-0x935F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9360: ; SUB		0x9360-0x9367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9368: ; SUB		0x9368-0x936F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9370: ; SUB		0x9370-0x9377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9378: ; SUB		0x9378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9379: ; SUB		0x9379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9390: ; SUB		0x9390-0x9397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9398: ; SUB		0x9398-0x939F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93A0: ; SUB		0x93A0-0x93A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93A8: ; SUB		0x93A8-0x93AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93B0: ; SUB		0x93B0-0x93B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93B8: ; SUB		0x93B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93B9: ; SUB		0x93B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9400: ; SUB		0x9400-0x9407
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9410: ; SUB		0x9410-0x9417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9418: ; SUB		0x9418-0x941F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9420: ; SUB		0x9420-0x9427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9428: ; SUB		0x9428-0x942F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9430: ; SUB		0x9430-0x9437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9438: ; SUB		0x9438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9439: ; SUB		0x9439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I943A: ; SUB		0x943A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I943B: ; SUB		0x943B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I943C: ; SUB		0x943C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9440: ; SUB		0x9440-0x9447
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9448: ; SUB		0x9448-0x944F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9450: ; SUB		0x9450-0x9457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9458: ; SUB		0x9458-0x945F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9460: ; SUB		0x9460-0x9467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9468: ; SUB		0x9468-0x946F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9470: ; SUB		0x9470-0x9477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9478: ; SUB		0x9478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9479: ; SUB		0x9479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I947A: ; SUB		0x947A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I947B: ; SUB		0x947B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I947C: ; SUB		0x947C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9480: ; SUB		0x9480-0x9487
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9488: ; SUB		0x9488-0x948F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9490: ; SUB		0x9490-0x9497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9498: ; SUB		0x9498-0x949F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94A0: ; SUB		0x94A0-0x94A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94A8: ; SUB		0x94A8-0x94AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94B0: ; SUB		0x94B0-0x94B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94B8: ; SUB		0x94B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94B9: ; SUB		0x94B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94BA: ; SUB		0x94BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94BB: ; SUB		0x94BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94BC: ; SUB		0x94BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9510: ; SUB		0x9510-0x9517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9518: ; SUB		0x9518-0x951F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9520: ; SUB		0x9520-0x9527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9528: ; SUB		0x9528-0x952F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9530: ; SUB		0x9530-0x9537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9538: ; SUB		0x9538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9539: ; SUB		0x9539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9550: ; SUB		0x9550-0x9557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9558: ; SUB		0x9558-0x955F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9560: ; SUB		0x9560-0x9567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9568: ; SUB		0x9568-0x956F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9570: ; SUB		0x9570-0x9577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9578: ; SUB		0x9578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9579: ; SUB		0x9579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9590: ; SUB		0x9590-0x9597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9598: ; SUB		0x9598-0x959F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95A0: ; SUB		0x95A0-0x95A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95A8: ; SUB		0x95A8-0x95AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95B0: ; SUB		0x95B0-0x95B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95B8: ; SUB		0x95B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95B9: ; SUB		0x95B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9600: ; SUB		0x9600-0x9607
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9610: ; SUB		0x9610-0x9617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9618: ; SUB		0x9618-0x961F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9620: ; SUB		0x9620-0x9627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9628: ; SUB		0x9628-0x962F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9630: ; SUB		0x9630-0x9637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9638: ; SUB		0x9638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9639: ; SUB		0x9639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I963A: ; SUB		0x963A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I963B: ; SUB		0x963B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I963C: ; SUB		0x963C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9640: ; SUB		0x9640-0x9647
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9648: ; SUB		0x9648-0x964F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9650: ; SUB		0x9650-0x9657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9658: ; SUB		0x9658-0x965F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9660: ; SUB		0x9660-0x9667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9668: ; SUB		0x9668-0x966F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9670: ; SUB		0x9670-0x9677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9678: ; SUB		0x9678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9679: ; SUB		0x9679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I967A: ; SUB		0x967A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I967B: ; SUB		0x967B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I967C: ; SUB		0x967C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9680: ; SUB		0x9680-0x9687
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9688: ; SUB		0x9688-0x968F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9690: ; SUB		0x9690-0x9697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9698: ; SUB		0x9698-0x969F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96A0: ; SUB		0x96A0-0x96A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96A8: ; SUB		0x96A8-0x96AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96B0: ; SUB		0x96B0-0x96B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96B8: ; SUB		0x96B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96B9: ; SUB		0x96B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96BA: ; SUB		0x96BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96BB: ; SUB		0x96BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96BC: ; SUB		0x96BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9710: ; SUB		0x9710-0x9717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9718: ; SUB		0x9718-0x971F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9720: ; SUB		0x9720-0x9727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9728: ; SUB		0x9728-0x972F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9730: ; SUB		0x9730-0x9737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9738: ; SUB		0x9738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9739: ; SUB		0x9739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9750: ; SUB		0x9750-0x9757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9758: ; SUB		0x9758-0x975F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9760: ; SUB		0x9760-0x9767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9768: ; SUB		0x9768-0x976F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9770: ; SUB		0x9770-0x9777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9778: ; SUB		0x9778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9779: ; SUB		0x9779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9790: ; SUB		0x9790-0x9797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9798: ; SUB		0x9798-0x979F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97A0: ; SUB		0x97A0-0x97A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97A8: ; SUB		0x97A8-0x97AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97B0: ; SUB		0x97B0-0x97B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97B8: ; SUB		0x97B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97B9: ; SUB		0x97B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9800: ; SUB		0x9800-0x9807
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9810: ; SUB		0x9810-0x9817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9818: ; SUB		0x9818-0x981F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9820: ; SUB		0x9820-0x9827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9828: ; SUB		0x9828-0x982F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9830: ; SUB		0x9830-0x9837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9838: ; SUB		0x9838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9839: ; SUB		0x9839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I983A: ; SUB		0x983A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I983B: ; SUB		0x983B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I983C: ; SUB		0x983C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9840: ; SUB		0x9840-0x9847
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9848: ; SUB		0x9848-0x984F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9850: ; SUB		0x9850-0x9857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9858: ; SUB		0x9858-0x985F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9860: ; SUB		0x9860-0x9867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9868: ; SUB		0x9868-0x986F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9870: ; SUB		0x9870-0x9877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9878: ; SUB		0x9878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9879: ; SUB		0x9879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I987A: ; SUB		0x987A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I987B: ; SUB		0x987B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I987C: ; SUB		0x987C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9880: ; SUB		0x9880-0x9887
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9888: ; SUB		0x9888-0x988F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9890: ; SUB		0x9890-0x9897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9898: ; SUB		0x9898-0x989F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98A0: ; SUB		0x98A0-0x98A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98A8: ; SUB		0x98A8-0x98AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98B0: ; SUB		0x98B0-0x98B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98B8: ; SUB		0x98B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98B9: ; SUB		0x98B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98BA: ; SUB		0x98BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98BB: ; SUB		0x98BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98BC: ; SUB		0x98BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9910: ; SUB		0x9910-0x9917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9918: ; SUB		0x9918-0x991F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9920: ; SUB		0x9920-0x9927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9928: ; SUB		0x9928-0x992F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9930: ; SUB		0x9930-0x9937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9938: ; SUB		0x9938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9939: ; SUB		0x9939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9950: ; SUB		0x9950-0x9957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9958: ; SUB		0x9958-0x995F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9960: ; SUB		0x9960-0x9967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9968: ; SUB		0x9968-0x996F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9970: ; SUB		0x9970-0x9977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9978: ; SUB		0x9978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9979: ; SUB		0x9979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9990: ; SUB		0x9990-0x9997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9998: ; SUB		0x9998-0x999F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99A0: ; SUB		0x99A0-0x99A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99A8: ; SUB		0x99A8-0x99AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99B0: ; SUB		0x99B0-0x99B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99B8: ; SUB		0x99B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99B9: ; SUB		0x99B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A00: ; SUB		0x9A00-0x9A07
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A10: ; SUB		0x9A10-0x9A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A18: ; SUB		0x9A18-0x9A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A20: ; SUB		0x9A20-0x9A27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A28: ; SUB		0x9A28-0x9A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A30: ; SUB		0x9A30-0x9A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A38: ; SUB		0x9A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A39: ; SUB		0x9A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A3A: ; SUB		0x9A3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A3B: ; SUB		0x9A3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A3C: ; SUB		0x9A3C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A40: ; SUB		0x9A40-0x9A47
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A48: ; SUB		0x9A48-0x9A4F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A50: ; SUB		0x9A50-0x9A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A58: ; SUB		0x9A58-0x9A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A60: ; SUB		0x9A60-0x9A67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A68: ; SUB		0x9A68-0x9A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A70: ; SUB		0x9A70-0x9A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A78: ; SUB		0x9A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A79: ; SUB		0x9A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A7A: ; SUB		0x9A7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A7B: ; SUB		0x9A7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A7C: ; SUB		0x9A7C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A80: ; SUB		0x9A80-0x9A87
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A88: ; SUB		0x9A88-0x9A8F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A90: ; SUB		0x9A90-0x9A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9A98: ; SUB		0x9A98-0x9A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AA0: ; SUB		0x9AA0-0x9AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AA8: ; SUB		0x9AA8-0x9AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AB0: ; SUB		0x9AB0-0x9AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AB8: ; SUB		0x9AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AB9: ; SUB		0x9AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9ABA: ; SUB		0x9ABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9ABB: ; SUB		0x9ABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9ABC: ; SUB		0x9ABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B10: ; SUB		0x9B10-0x9B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B18: ; SUB		0x9B18-0x9B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B20: ; SUB		0x9B20-0x9B27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B28: ; SUB		0x9B28-0x9B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B30: ; SUB		0x9B30-0x9B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B38: ; SUB		0x9B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B39: ; SUB		0x9B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B50: ; SUB		0x9B50-0x9B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B58: ; SUB		0x9B58-0x9B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B60: ; SUB		0x9B60-0x9B67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B68: ; SUB		0x9B68-0x9B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B70: ; SUB		0x9B70-0x9B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B78: ; SUB		0x9B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B79: ; SUB		0x9B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B90: ; SUB		0x9B90-0x9B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B98: ; SUB		0x9B98-0x9B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BA0: ; SUB		0x9BA0-0x9BA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BA8: ; SUB		0x9BA8-0x9BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BB0: ; SUB		0x9BB0-0x9BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BB8: ; SUB		0x9BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BB9: ; SUB		0x9BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C00: ; SUB		0x9C00-0x9C07
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C10: ; SUB		0x9C10-0x9C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C18: ; SUB		0x9C18-0x9C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C20: ; SUB		0x9C20-0x9C27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C28: ; SUB		0x9C28-0x9C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C30: ; SUB		0x9C30-0x9C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C38: ; SUB		0x9C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C39: ; SUB		0x9C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C3A: ; SUB		0x9C3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C3B: ; SUB		0x9C3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C3C: ; SUB		0x9C3C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C40: ; SUB		0x9C40-0x9C47
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C48: ; SUB		0x9C48-0x9C4F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C50: ; SUB		0x9C50-0x9C57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C58: ; SUB		0x9C58-0x9C5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C60: ; SUB		0x9C60-0x9C67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C68: ; SUB		0x9C68-0x9C6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C70: ; SUB		0x9C70-0x9C77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C78: ; SUB		0x9C78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C79: ; SUB		0x9C79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C7A: ; SUB		0x9C7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C7B: ; SUB		0x9C7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C7C: ; SUB		0x9C7C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C80: ; SUB		0x9C80-0x9C87
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C88: ; SUB		0x9C88-0x9C8F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C90: ; SUB		0x9C90-0x9C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9C98: ; SUB		0x9C98-0x9C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CA0: ; SUB		0x9CA0-0x9CA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CA8: ; SUB		0x9CA8-0x9CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CB0: ; SUB		0x9CB0-0x9CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CB8: ; SUB		0x9CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CB9: ; SUB		0x9CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CBA: ; SUB		0x9CBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CBB: ; SUB		0x9CBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CBC: ; SUB		0x9CBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D10: ; SUB		0x9D10-0x9D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D18: ; SUB		0x9D18-0x9D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D20: ; SUB		0x9D20-0x9D27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D28: ; SUB		0x9D28-0x9D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D30: ; SUB		0x9D30-0x9D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D38: ; SUB		0x9D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D39: ; SUB		0x9D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D50: ; SUB		0x9D50-0x9D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D58: ; SUB		0x9D58-0x9D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D60: ; SUB		0x9D60-0x9D67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D68: ; SUB		0x9D68-0x9D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D70: ; SUB		0x9D70-0x9D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D78: ; SUB		0x9D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D79: ; SUB		0x9D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D90: ; SUB		0x9D90-0x9D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D98: ; SUB		0x9D98-0x9D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DA0: ; SUB		0x9DA0-0x9DA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DA8: ; SUB		0x9DA8-0x9DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DB0: ; SUB		0x9DB0-0x9DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DB8: ; SUB		0x9DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DB9: ; SUB		0x9DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E00: ; SUB		0x9E00-0x9E07
and  edi, byte 7
mov  dl, [d+edi*4]
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E10: ; SUB		0x9E10-0x9E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E18: ; SUB		0x9E18-0x9E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E20: ; SUB		0x9E20-0x9E27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E28: ; SUB		0x9E28-0x9E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E30: ; SUB		0x9E30-0x9E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E38: ; SUB		0x9E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E39: ; SUB		0x9E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E3A: ; SUB		0x9E3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E3B: ; SUB		0x9E3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E3C: ; SUB		0x9E3C
mov  dl, [esi]
add  esi, byte 2
sub   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E40: ; SUB		0x9E40-0x9E47
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E48: ; SUB		0x9E48-0x9E4F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E50: ; SUB		0x9E50-0x9E57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E58: ; SUB		0x9E58-0x9E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E60: ; SUB		0x9E60-0x9E67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E68: ; SUB		0x9E68-0x9E6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E70: ; SUB		0x9E70-0x9E77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E78: ; SUB		0x9E78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E79: ; SUB		0x9E79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E7A: ; SUB		0x9E7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E7B: ; SUB		0x9E7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E7C: ; SUB		0x9E7C
mov  edx, [esi]
add  esi, byte 2
sub   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E80: ; SUB		0x9E80-0x9E87
and  edi, byte 7
mov  edx, [d+edi*4]
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E88: ; SUB		0x9E88-0x9E8F
and  edi, byte 7
mov  edx, [a+edi*4]
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E90: ; SUB		0x9E90-0x9E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9E98: ; SUB		0x9E98-0x9E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EA0: ; SUB		0x9EA0-0x9EA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EA8: ; SUB		0x9EA8-0x9EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EB0: ; SUB		0x9EB0-0x9EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EB8: ; SUB		0x9EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EB9: ; SUB		0x9EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EBA: ; SUB		0x9EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EBB: ; SUB		0x9EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EBC: ; SUB		0x9EBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F10: ; SUB		0x9F10-0x9F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F18: ; SUB		0x9F18-0x9F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F20: ; SUB		0x9F20-0x9F27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F28: ; SUB		0x9F28-0x9F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F30: ; SUB		0x9F30-0x9F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F38: ; SUB		0x9F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F39: ; SUB		0x9F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F50: ; SUB		0x9F50-0x9F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F58: ; SUB		0x9F58-0x9F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F60: ; SUB		0x9F60-0x9F67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F68: ; SUB		0x9F68-0x9F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F70: ; SUB		0x9F70-0x9F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F78: ; SUB		0x9F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F79: ; SUB		0x9F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F90: ; SUB		0x9F90-0x9F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F98: ; SUB		0x9F98-0x9F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FA0: ; SUB		0x9FA0-0x9FA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FA8: ; SUB		0x9FA8-0x9FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FB0: ; SUB		0x9FB0-0x9FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FB8: ; SUB		0x9FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FB9: ; SUB		0x9FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB000: ; CMP		0xB000-0xB007
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB010: ; CMP		0xB010-0xB017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB018: ; CMP		0xB018-0xB01F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB020: ; CMP		0xB020-0xB027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB028: ; CMP		0xB028-0xB02F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB030: ; CMP		0xB030-0xB037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB038: ; CMP		0xB038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB039: ; CMP		0xB039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB03A: ; CMP		0xB03A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB03B: ; CMP		0xB03B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB03C: ; CMP		0xB03C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(0*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB040: ; CMP		0xB040-0xB047
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB048: ; CMP		0xB048-0xB04F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB050: ; CMP		0xB050-0xB057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB058: ; CMP		0xB058-0xB05F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB060: ; CMP		0xB060-0xB067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB068: ; CMP		0xB068-0xB06F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB070: ; CMP		0xB070-0xB077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB078: ; CMP		0xB078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB079: ; CMP		0xB079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB07A: ; CMP		0xB07A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB07B: ; CMP		0xB07B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB07C: ; CMP		0xB07C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(0*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB080: ; CMP		0xB080-0xB087
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB088: ; CMP		0xB088-0xB08F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB090: ; CMP		0xB090-0xB097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB098: ; CMP		0xB098-0xB09F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0A0: ; CMP		0xB0A0-0xB0A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0A8: ; CMP		0xB0A8-0xB0AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0B0: ; CMP		0xB0B0-0xB0B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0B8: ; CMP		0xB0B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0B9: ; CMP		0xB0B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0BA: ; CMP		0xB0BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0BB: ; CMP		0xB0BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0BC: ; CMP		0xB0BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(0*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB100: ; EOR		0xB100-0xB107
and  edi, byte 7
mov  edx, [d+(0*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB110: ; EOR		0xB110-0xB117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB118: ; EOR		0xB118-0xB11F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB120: ; EOR		0xB120-0xB127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB128: ; EOR		0xB128-0xB12F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB130: ; EOR		0xB130-0xB137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB138: ; EOR		0xB138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB139: ; EOR		0xB139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB140: ; EOR		0xB140-0xB147
and  edi, byte 7
mov  edx, [d+(0*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB150: ; EOR		0xB150-0xB157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB158: ; EOR		0xB158-0xB15F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB160: ; EOR		0xB160-0xB167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB168: ; EOR		0xB168-0xB16F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB170: ; EOR		0xB170-0xB177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB178: ; EOR		0xB178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB179: ; EOR		0xB179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB180: ; EOR		0xB180-0xB187
and  edi, byte 7
mov  edx, [d+(0*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB190: ; EOR		0xB190-0xB197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB198: ; EOR		0xB198-0xB19F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1A0: ; EOR		0xB1A0-0xB1A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1A8: ; EOR		0xB1A8-0xB1AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1B0: ; EOR		0xB1B0-0xB1B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1B8: ; EOR		0xB1B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1B9: ; EOR		0xB1B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB200: ; CMP		0xB200-0xB207
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB210: ; CMP		0xB210-0xB217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB218: ; CMP		0xB218-0xB21F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB220: ; CMP		0xB220-0xB227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB228: ; CMP		0xB228-0xB22F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB230: ; CMP		0xB230-0xB237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB238: ; CMP		0xB238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB239: ; CMP		0xB239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB23A: ; CMP		0xB23A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB23B: ; CMP		0xB23B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB23C: ; CMP		0xB23C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(1*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB240: ; CMP		0xB240-0xB247
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB248: ; CMP		0xB248-0xB24F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB250: ; CMP		0xB250-0xB257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB258: ; CMP		0xB258-0xB25F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB260: ; CMP		0xB260-0xB267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB268: ; CMP		0xB268-0xB26F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB270: ; CMP		0xB270-0xB277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB278: ; CMP		0xB278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB279: ; CMP		0xB279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB27A: ; CMP		0xB27A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB27B: ; CMP		0xB27B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB27C: ; CMP		0xB27C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(1*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB280: ; CMP		0xB280-0xB287
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB288: ; CMP		0xB288-0xB28F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB290: ; CMP		0xB290-0xB297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB298: ; CMP		0xB298-0xB29F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2A0: ; CMP		0xB2A0-0xB2A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2A8: ; CMP		0xB2A8-0xB2AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2B0: ; CMP		0xB2B0-0xB2B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2B8: ; CMP		0xB2B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2B9: ; CMP		0xB2B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2BA: ; CMP		0xB2BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2BB: ; CMP		0xB2BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2BC: ; CMP		0xB2BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(1*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB300: ; EOR		0xB300-0xB307
and  edi, byte 7
mov  edx, [d+(1*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB310: ; EOR		0xB310-0xB317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB318: ; EOR		0xB318-0xB31F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB320: ; EOR		0xB320-0xB327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB328: ; EOR		0xB328-0xB32F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB330: ; EOR		0xB330-0xB337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB338: ; EOR		0xB338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB339: ; EOR		0xB339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB340: ; EOR		0xB340-0xB347
and  edi, byte 7
mov  edx, [d+(1*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB350: ; EOR		0xB350-0xB357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB358: ; EOR		0xB358-0xB35F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB360: ; EOR		0xB360-0xB367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB368: ; EOR		0xB368-0xB36F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB370: ; EOR		0xB370-0xB377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB378: ; EOR		0xB378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB379: ; EOR		0xB379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB380: ; EOR		0xB380-0xB387
and  edi, byte 7
mov  edx, [d+(1*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB390: ; EOR		0xB390-0xB397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB398: ; EOR		0xB398-0xB39F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3A0: ; EOR		0xB3A0-0xB3A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3A8: ; EOR		0xB3A8-0xB3AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3B0: ; EOR		0xB3B0-0xB3B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3B8: ; EOR		0xB3B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3B9: ; EOR		0xB3B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB400: ; CMP		0xB400-0xB407
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB410: ; CMP		0xB410-0xB417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB418: ; CMP		0xB418-0xB41F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB420: ; CMP		0xB420-0xB427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB428: ; CMP		0xB428-0xB42F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB430: ; CMP		0xB430-0xB437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB438: ; CMP		0xB438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB439: ; CMP		0xB439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB43A: ; CMP		0xB43A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB43B: ; CMP		0xB43B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB43C: ; CMP		0xB43C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(2*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB440: ; CMP		0xB440-0xB447
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB448: ; CMP		0xB448-0xB44F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB450: ; CMP		0xB450-0xB457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB458: ; CMP		0xB458-0xB45F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB460: ; CMP		0xB460-0xB467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB468: ; CMP		0xB468-0xB46F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB470: ; CMP		0xB470-0xB477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB478: ; CMP		0xB478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB479: ; CMP		0xB479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB47A: ; CMP		0xB47A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB47B: ; CMP		0xB47B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB47C: ; CMP		0xB47C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(2*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB480: ; CMP		0xB480-0xB487
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB488: ; CMP		0xB488-0xB48F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB490: ; CMP		0xB490-0xB497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB498: ; CMP		0xB498-0xB49F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4A0: ; CMP		0xB4A0-0xB4A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4A8: ; CMP		0xB4A8-0xB4AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4B0: ; CMP		0xB4B0-0xB4B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4B8: ; CMP		0xB4B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4B9: ; CMP		0xB4B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4BA: ; CMP		0xB4BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4BB: ; CMP		0xB4BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4BC: ; CMP		0xB4BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(2*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB500: ; EOR		0xB500-0xB507
and  edi, byte 7
mov  edx, [d+(2*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB510: ; EOR		0xB510-0xB517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB518: ; EOR		0xB518-0xB51F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB520: ; EOR		0xB520-0xB527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB528: ; EOR		0xB528-0xB52F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB530: ; EOR		0xB530-0xB537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB538: ; EOR		0xB538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB539: ; EOR		0xB539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB540: ; EOR		0xB540-0xB547
and  edi, byte 7
mov  edx, [d+(2*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB550: ; EOR		0xB550-0xB557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB558: ; EOR		0xB558-0xB55F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB560: ; EOR		0xB560-0xB567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB568: ; EOR		0xB568-0xB56F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB570: ; EOR		0xB570-0xB577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB578: ; EOR		0xB578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB579: ; EOR		0xB579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB580: ; EOR		0xB580-0xB587
and  edi, byte 7
mov  edx, [d+(2*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB590: ; EOR		0xB590-0xB597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB598: ; EOR		0xB598-0xB59F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5A0: ; EOR		0xB5A0-0xB5A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5A8: ; EOR		0xB5A8-0xB5AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5B0: ; EOR		0xB5B0-0xB5B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5B8: ; EOR		0xB5B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5B9: ; EOR		0xB5B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB600: ; CMP		0xB600-0xB607
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB610: ; CMP		0xB610-0xB617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB618: ; CMP		0xB618-0xB61F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB620: ; CMP		0xB620-0xB627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB628: ; CMP		0xB628-0xB62F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB630: ; CMP		0xB630-0xB637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB638: ; CMP		0xB638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB639: ; CMP		0xB639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB63A: ; CMP		0xB63A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB63B: ; CMP		0xB63B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB63C: ; CMP		0xB63C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(3*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB640: ; CMP		0xB640-0xB647
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB648: ; CMP		0xB648-0xB64F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB650: ; CMP		0xB650-0xB657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB658: ; CMP		0xB658-0xB65F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB660: ; CMP		0xB660-0xB667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB668: ; CMP		0xB668-0xB66F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB670: ; CMP		0xB670-0xB677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB678: ; CMP		0xB678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB679: ; CMP		0xB679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB67A: ; CMP		0xB67A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB67B: ; CMP		0xB67B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB67C: ; CMP		0xB67C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(3*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB680: ; CMP		0xB680-0xB687
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB688: ; CMP		0xB688-0xB68F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB690: ; CMP		0xB690-0xB697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB698: ; CMP		0xB698-0xB69F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6A0: ; CMP		0xB6A0-0xB6A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6A8: ; CMP		0xB6A8-0xB6AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6B0: ; CMP		0xB6B0-0xB6B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6B8: ; CMP		0xB6B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6B9: ; CMP		0xB6B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6BA: ; CMP		0xB6BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6BB: ; CMP		0xB6BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6BC: ; CMP		0xB6BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(3*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB700: ; EOR		0xB700-0xB707
and  edi, byte 7
mov  edx, [d+(3*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB710: ; EOR		0xB710-0xB717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB718: ; EOR		0xB718-0xB71F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB720: ; EOR		0xB720-0xB727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB728: ; EOR		0xB728-0xB72F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB730: ; EOR		0xB730-0xB737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB738: ; EOR		0xB738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB739: ; EOR		0xB739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB740: ; EOR		0xB740-0xB747
and  edi, byte 7
mov  edx, [d+(3*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB750: ; EOR		0xB750-0xB757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB758: ; EOR		0xB758-0xB75F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB760: ; EOR		0xB760-0xB767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB768: ; EOR		0xB768-0xB76F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB770: ; EOR		0xB770-0xB777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB778: ; EOR		0xB778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB779: ; EOR		0xB779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB780: ; EOR		0xB780-0xB787
and  edi, byte 7
mov  edx, [d+(3*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB790: ; EOR		0xB790-0xB797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB798: ; EOR		0xB798-0xB79F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7A0: ; EOR		0xB7A0-0xB7A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7A8: ; EOR		0xB7A8-0xB7AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7B0: ; EOR		0xB7B0-0xB7B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7B8: ; EOR		0xB7B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7B9: ; EOR		0xB7B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB800: ; CMP		0xB800-0xB807
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB810: ; CMP		0xB810-0xB817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB818: ; CMP		0xB818-0xB81F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB820: ; CMP		0xB820-0xB827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB828: ; CMP		0xB828-0xB82F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB830: ; CMP		0xB830-0xB837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB838: ; CMP		0xB838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB839: ; CMP		0xB839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB83A: ; CMP		0xB83A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB83B: ; CMP		0xB83B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB83C: ; CMP		0xB83C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(4*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB840: ; CMP		0xB840-0xB847
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB848: ; CMP		0xB848-0xB84F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB850: ; CMP		0xB850-0xB857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB858: ; CMP		0xB858-0xB85F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB860: ; CMP		0xB860-0xB867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB868: ; CMP		0xB868-0xB86F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB870: ; CMP		0xB870-0xB877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB878: ; CMP		0xB878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB879: ; CMP		0xB879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB87A: ; CMP		0xB87A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB87B: ; CMP		0xB87B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB87C: ; CMP		0xB87C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(4*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB880: ; CMP		0xB880-0xB887
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB888: ; CMP		0xB888-0xB88F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB890: ; CMP		0xB890-0xB897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB898: ; CMP		0xB898-0xB89F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8A0: ; CMP		0xB8A0-0xB8A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8A8: ; CMP		0xB8A8-0xB8AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8B0: ; CMP		0xB8B0-0xB8B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8B8: ; CMP		0xB8B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8B9: ; CMP		0xB8B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8BA: ; CMP		0xB8BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8BB: ; CMP		0xB8BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8BC: ; CMP		0xB8BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(4*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB900: ; EOR		0xB900-0xB907
and  edi, byte 7
mov  edx, [d+(4*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB910: ; EOR		0xB910-0xB917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB918: ; EOR		0xB918-0xB91F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB920: ; EOR		0xB920-0xB927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB928: ; EOR		0xB928-0xB92F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB930: ; EOR		0xB930-0xB937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB938: ; EOR		0xB938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB939: ; EOR		0xB939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB940: ; EOR		0xB940-0xB947
and  edi, byte 7
mov  edx, [d+(4*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB950: ; EOR		0xB950-0xB957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB958: ; EOR		0xB958-0xB95F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB960: ; EOR		0xB960-0xB967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB968: ; EOR		0xB968-0xB96F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB970: ; EOR		0xB970-0xB977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB978: ; EOR		0xB978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB979: ; EOR		0xB979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB980: ; EOR		0xB980-0xB987
and  edi, byte 7
mov  edx, [d+(4*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB990: ; EOR		0xB990-0xB997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB998: ; EOR		0xB998-0xB99F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9A0: ; EOR		0xB9A0-0xB9A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9A8: ; EOR		0xB9A8-0xB9AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9B0: ; EOR		0xB9B0-0xB9B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9B8: ; EOR		0xB9B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9B9: ; EOR		0xB9B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA00: ; CMP		0xBA00-0xBA07
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA10: ; CMP		0xBA10-0xBA17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA18: ; CMP		0xBA18-0xBA1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA20: ; CMP		0xBA20-0xBA27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA28: ; CMP		0xBA28-0xBA2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA30: ; CMP		0xBA30-0xBA37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA38: ; CMP		0xBA38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA39: ; CMP		0xBA39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA3A: ; CMP		0xBA3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA3B: ; CMP		0xBA3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA3C: ; CMP		0xBA3C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(5*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA40: ; CMP		0xBA40-0xBA47
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA48: ; CMP		0xBA48-0xBA4F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA50: ; CMP		0xBA50-0xBA57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA58: ; CMP		0xBA58-0xBA5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA60: ; CMP		0xBA60-0xBA67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA68: ; CMP		0xBA68-0xBA6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA70: ; CMP		0xBA70-0xBA77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA78: ; CMP		0xBA78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA79: ; CMP		0xBA79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA7A: ; CMP		0xBA7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA7B: ; CMP		0xBA7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA7C: ; CMP		0xBA7C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(5*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA80: ; CMP		0xBA80-0xBA87
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA88: ; CMP		0xBA88-0xBA8F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA90: ; CMP		0xBA90-0xBA97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBA98: ; CMP		0xBA98-0xBA9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAA0: ; CMP		0xBAA0-0xBAA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAA8: ; CMP		0xBAA8-0xBAAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAB0: ; CMP		0xBAB0-0xBAB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAB8: ; CMP		0xBAB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAB9: ; CMP		0xBAB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBABA: ; CMP		0xBABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBABB: ; CMP		0xBABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBABC: ; CMP		0xBABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(5*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB00: ; EOR		0xBB00-0xBB07
and  edi, byte 7
mov  edx, [d+(5*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB10: ; EOR		0xBB10-0xBB17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB18: ; EOR		0xBB18-0xBB1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB20: ; EOR		0xBB20-0xBB27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB28: ; EOR		0xBB28-0xBB2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB30: ; EOR		0xBB30-0xBB37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB38: ; EOR		0xBB38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB39: ; EOR		0xBB39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB40: ; EOR		0xBB40-0xBB47
and  edi, byte 7
mov  edx, [d+(5*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB50: ; EOR		0xBB50-0xBB57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB58: ; EOR		0xBB58-0xBB5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB60: ; EOR		0xBB60-0xBB67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB68: ; EOR		0xBB68-0xBB6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB70: ; EOR		0xBB70-0xBB77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB78: ; EOR		0xBB78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB79: ; EOR		0xBB79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB80: ; EOR		0xBB80-0xBB87
and  edi, byte 7
mov  edx, [d+(5*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB90: ; EOR		0xBB90-0xBB97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB98: ; EOR		0xBB98-0xBB9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBA0: ; EOR		0xBBA0-0xBBA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBA8: ; EOR		0xBBA8-0xBBAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBB0: ; EOR		0xBBB0-0xBBB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBB8: ; EOR		0xBBB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBB9: ; EOR		0xBBB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC00: ; CMP		0xBC00-0xBC07
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC10: ; CMP		0xBC10-0xBC17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC18: ; CMP		0xBC18-0xBC1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC20: ; CMP		0xBC20-0xBC27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC28: ; CMP		0xBC28-0xBC2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC30: ; CMP		0xBC30-0xBC37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC38: ; CMP		0xBC38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC39: ; CMP		0xBC39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC3A: ; CMP		0xBC3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC3B: ; CMP		0xBC3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC3C: ; CMP		0xBC3C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(6*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC40: ; CMP		0xBC40-0xBC47
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC48: ; CMP		0xBC48-0xBC4F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC50: ; CMP		0xBC50-0xBC57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC58: ; CMP		0xBC58-0xBC5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC60: ; CMP		0xBC60-0xBC67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC68: ; CMP		0xBC68-0xBC6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC70: ; CMP		0xBC70-0xBC77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC78: ; CMP		0xBC78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC79: ; CMP		0xBC79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC7A: ; CMP		0xBC7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC7B: ; CMP		0xBC7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC7C: ; CMP		0xBC7C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(6*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC80: ; CMP		0xBC80-0xBC87
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC88: ; CMP		0xBC88-0xBC8F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC90: ; CMP		0xBC90-0xBC97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBC98: ; CMP		0xBC98-0xBC9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCA0: ; CMP		0xBCA0-0xBCA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCA8: ; CMP		0xBCA8-0xBCAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCB0: ; CMP		0xBCB0-0xBCB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCB8: ; CMP		0xBCB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCB9: ; CMP		0xBCB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCBA: ; CMP		0xBCBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCBB: ; CMP		0xBCBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCBC: ; CMP		0xBCBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(6*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD00: ; EOR		0xBD00-0xBD07
and  edi, byte 7
mov  edx, [d+(6*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD10: ; EOR		0xBD10-0xBD17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD18: ; EOR		0xBD18-0xBD1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD20: ; EOR		0xBD20-0xBD27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD28: ; EOR		0xBD28-0xBD2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD30: ; EOR		0xBD30-0xBD37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD38: ; EOR		0xBD38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD39: ; EOR		0xBD39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD40: ; EOR		0xBD40-0xBD47
and  edi, byte 7
mov  edx, [d+(6*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD50: ; EOR		0xBD50-0xBD57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD58: ; EOR		0xBD58-0xBD5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD60: ; EOR		0xBD60-0xBD67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD68: ; EOR		0xBD68-0xBD6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD70: ; EOR		0xBD70-0xBD77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD78: ; EOR		0xBD78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD79: ; EOR		0xBD79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD80: ; EOR		0xBD80-0xBD87
and  edi, byte 7
mov  edx, [d+(6*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD90: ; EOR		0xBD90-0xBD97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD98: ; EOR		0xBD98-0xBD9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDA0: ; EOR		0xBDA0-0xBDA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDA8: ; EOR		0xBDA8-0xBDAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDB0: ; EOR		0xBDB0-0xBDB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDB8: ; EOR		0xBDB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDB9: ; EOR		0xBDB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE00: ; CMP		0xBE00-0xBE07
and  edi, byte 7
mov  dl, [d+edi*4]
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE10: ; CMP		0xBE10-0xBE17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE18: ; CMP		0xBE18-0xBE1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE20: ; CMP		0xBE20-0xBE27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE28: ; CMP		0xBE28-0xBE2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE30: ; CMP		0xBE30-0xBE37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE38: ; CMP		0xBE38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE39: ; CMP		0xBE39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE3A: ; CMP		0xBE3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE3B: ; CMP		0xBE3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE3C: ; CMP		0xBE3C
mov  dl, [esi]
add  esi, byte 2
cmp   [d+(7*4)], dl
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE40: ; CMP		0xBE40-0xBE47
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE48: ; CMP		0xBE48-0xBE4F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE50: ; CMP		0xBE50-0xBE57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE58: ; CMP		0xBE58-0xBE5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE60: ; CMP		0xBE60-0xBE67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE68: ; CMP		0xBE68-0xBE6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE70: ; CMP		0xBE70-0xBE77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE78: ; CMP		0xBE78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE79: ; CMP		0xBE79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE7A: ; CMP		0xBE7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE7B: ; CMP		0xBE7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE7C: ; CMP		0xBE7C
mov  edx, [esi]
add  esi, byte 2
cmp   [d+(7*4)], dx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE80: ; CMP		0xBE80-0xBE87
and  edi, byte 7
mov  edx, [d+edi*4]
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE88: ; CMP		0xBE88-0xBE8F
and  edi, byte 7
mov  edx, [a+edi*4]
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE90: ; CMP		0xBE90-0xBE97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBE98: ; CMP		0xBE98-0xBE9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEA0: ; CMP		0xBEA0-0xBEA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEA8: ; CMP		0xBEA8-0xBEAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEB0: ; CMP		0xBEB0-0xBEB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEB8: ; CMP		0xBEB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEB9: ; CMP		0xBEB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEBA: ; CMP		0xBEBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEBB: ; CMP		0xBEBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEBC: ; CMP		0xBEBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp   [d+(7*4)], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF00: ; EOR		0xBF00-0xBF07
and  edi, byte 7
mov  edx, [d+(7*4)]
xor   byte [d+(edi*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF10: ; EOR		0xBF10-0xBF17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF18: ; EOR		0xBF18-0xBF1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF20: ; EOR		0xBF20-0xBF27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF28: ; EOR		0xBF28-0xBF2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF30: ; EOR		0xBF30-0xBF37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF38: ; EOR		0xBF38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF39: ; EOR		0xBF39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF40: ; EOR		0xBF40-0xBF47
and  edi, byte 7
mov  edx, [d+(7*4)]
xor   word [d+(edi*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF50: ; EOR		0xBF50-0xBF57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF58: ; EOR		0xBF58-0xBF5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF60: ; EOR		0xBF60-0xBF67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF68: ; EOR		0xBF68-0xBF6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF70: ; EOR		0xBF70-0xBF77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF78: ; EOR		0xBF78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF79: ; EOR		0xBF79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
xor   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF80: ; EOR		0xBF80-0xBF87
and  edi, byte 7
mov  edx, [d+(7*4)]
xor   dword [d+(edi*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF90: ; EOR		0xBF90-0xBF97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF98: ; EOR		0xBF98-0xBF9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFA0: ; EOR		0xBFA0-0xBFA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFA8: ; EOR		0xBFA8-0xBFAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFB0: ; EOR		0xBFB0-0xBFB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFB8: ; EOR		0xBFB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFB9: ; EOR		0xBFB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
xor   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC000: ; AND		0xC000-0xC007
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC010: ; AND		0xC010-0xC017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC018: ; AND		0xC018-0xC01F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC020: ; AND		0xC020-0xC027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC028: ; AND		0xC028-0xC02F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC030: ; AND		0xC030-0xC037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC038: ; AND		0xC038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC039: ; AND		0xC039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC03A: ; AND		0xC03A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC03B: ; AND		0xC03B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC03C: ; AND		0xC03C
mov  dl, [esi]
add  esi, byte 2
and   [d+(0*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC040: ; AND		0xC040-0xC047
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC050: ; AND		0xC050-0xC057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC058: ; AND		0xC058-0xC05F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC060: ; AND		0xC060-0xC067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC068: ; AND		0xC068-0xC06F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC070: ; AND		0xC070-0xC077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC078: ; AND		0xC078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC079: ; AND		0xC079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC07A: ; AND		0xC07A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC07B: ; AND		0xC07B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC07C: ; AND		0xC07C
mov  edx, [esi]
add  esi, byte 2
and   [d+(0*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC080: ; AND		0xC080-0xC087
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC090: ; AND		0xC090-0xC097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC098: ; AND		0xC098-0xC09F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0A0: ; AND		0xC0A0-0xC0A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0A8: ; AND		0xC0A8-0xC0AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0B0: ; AND		0xC0B0-0xC0B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0B8: ; AND		0xC0B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0B9: ; AND		0xC0B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0BA: ; AND		0xC0BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0BB: ; AND		0xC0BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0BC: ; AND		0xC0BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(0*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC110: ; AND		0xC110-0xC117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC118: ; AND		0xC118-0xC11F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC120: ; AND		0xC120-0xC127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC128: ; AND		0xC128-0xC12F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC130: ; AND		0xC130-0xC137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC138: ; AND		0xC138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC139: ; AND		0xC139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(0*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC150: ; AND		0xC150-0xC157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC158: ; AND		0xC158-0xC15F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC160: ; AND		0xC160-0xC167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC168: ; AND		0xC168-0xC16F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC170: ; AND		0xC170-0xC177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC178: ; AND		0xC178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC179: ; AND		0xC179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(0*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC190: ; AND		0xC190-0xC197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC198: ; AND		0xC198-0xC19F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1A0: ; AND		0xC1A0-0xC1A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1A8: ; AND		0xC1A8-0xC1AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1B0: ; AND		0xC1B0-0xC1B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1B8: ; AND		0xC1B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1B9: ; AND		0xC1B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(0*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC200: ; AND		0xC200-0xC207
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC210: ; AND		0xC210-0xC217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC218: ; AND		0xC218-0xC21F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC220: ; AND		0xC220-0xC227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC228: ; AND		0xC228-0xC22F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC230: ; AND		0xC230-0xC237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC238: ; AND		0xC238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC239: ; AND		0xC239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC23A: ; AND		0xC23A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC23B: ; AND		0xC23B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC23C: ; AND		0xC23C
mov  dl, [esi]
add  esi, byte 2
and   [d+(1*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC240: ; AND		0xC240-0xC247
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC250: ; AND		0xC250-0xC257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC258: ; AND		0xC258-0xC25F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC260: ; AND		0xC260-0xC267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC268: ; AND		0xC268-0xC26F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC270: ; AND		0xC270-0xC277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC278: ; AND		0xC278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC279: ; AND		0xC279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC27A: ; AND		0xC27A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC27B: ; AND		0xC27B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC27C: ; AND		0xC27C
mov  edx, [esi]
add  esi, byte 2
and   [d+(1*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC280: ; AND		0xC280-0xC287
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC290: ; AND		0xC290-0xC297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC298: ; AND		0xC298-0xC29F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2A0: ; AND		0xC2A0-0xC2A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2A8: ; AND		0xC2A8-0xC2AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2B0: ; AND		0xC2B0-0xC2B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2B8: ; AND		0xC2B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2B9: ; AND		0xC2B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2BA: ; AND		0xC2BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2BB: ; AND		0xC2BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2BC: ; AND		0xC2BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(1*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC310: ; AND		0xC310-0xC317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC318: ; AND		0xC318-0xC31F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC320: ; AND		0xC320-0xC327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC328: ; AND		0xC328-0xC32F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC330: ; AND		0xC330-0xC337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC338: ; AND		0xC338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC339: ; AND		0xC339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(1*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC350: ; AND		0xC350-0xC357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC358: ; AND		0xC358-0xC35F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC360: ; AND		0xC360-0xC367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC368: ; AND		0xC368-0xC36F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC370: ; AND		0xC370-0xC377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC378: ; AND		0xC378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC379: ; AND		0xC379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(1*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC390: ; AND		0xC390-0xC397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC398: ; AND		0xC398-0xC39F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3A0: ; AND		0xC3A0-0xC3A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3A8: ; AND		0xC3A8-0xC3AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3B0: ; AND		0xC3B0-0xC3B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3B8: ; AND		0xC3B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3B9: ; AND		0xC3B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(1*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC400: ; AND		0xC400-0xC407
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC410: ; AND		0xC410-0xC417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC418: ; AND		0xC418-0xC41F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC420: ; AND		0xC420-0xC427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC428: ; AND		0xC428-0xC42F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC430: ; AND		0xC430-0xC437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC438: ; AND		0xC438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC439: ; AND		0xC439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC43A: ; AND		0xC43A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC43B: ; AND		0xC43B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC43C: ; AND		0xC43C
mov  dl, [esi]
add  esi, byte 2
and   [d+(2*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC440: ; AND		0xC440-0xC447
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC450: ; AND		0xC450-0xC457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC458: ; AND		0xC458-0xC45F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC460: ; AND		0xC460-0xC467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC468: ; AND		0xC468-0xC46F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC470: ; AND		0xC470-0xC477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC478: ; AND		0xC478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC479: ; AND		0xC479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC47A: ; AND		0xC47A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC47B: ; AND		0xC47B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC47C: ; AND		0xC47C
mov  edx, [esi]
add  esi, byte 2
and   [d+(2*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC480: ; AND		0xC480-0xC487
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC490: ; AND		0xC490-0xC497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC498: ; AND		0xC498-0xC49F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4A0: ; AND		0xC4A0-0xC4A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4A8: ; AND		0xC4A8-0xC4AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4B0: ; AND		0xC4B0-0xC4B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4B8: ; AND		0xC4B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4B9: ; AND		0xC4B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4BA: ; AND		0xC4BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4BB: ; AND		0xC4BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4BC: ; AND		0xC4BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(2*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC510: ; AND		0xC510-0xC517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC518: ; AND		0xC518-0xC51F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC520: ; AND		0xC520-0xC527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC528: ; AND		0xC528-0xC52F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC530: ; AND		0xC530-0xC537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC538: ; AND		0xC538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC539: ; AND		0xC539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(2*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC550: ; AND		0xC550-0xC557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC558: ; AND		0xC558-0xC55F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC560: ; AND		0xC560-0xC567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC568: ; AND		0xC568-0xC56F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC570: ; AND		0xC570-0xC577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC578: ; AND		0xC578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC579: ; AND		0xC579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(2*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC590: ; AND		0xC590-0xC597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC598: ; AND		0xC598-0xC59F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5A0: ; AND		0xC5A0-0xC5A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5A8: ; AND		0xC5A8-0xC5AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5B0: ; AND		0xC5B0-0xC5B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5B8: ; AND		0xC5B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5B9: ; AND		0xC5B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(2*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC600: ; AND		0xC600-0xC607
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC610: ; AND		0xC610-0xC617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC618: ; AND		0xC618-0xC61F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC620: ; AND		0xC620-0xC627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC628: ; AND		0xC628-0xC62F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC630: ; AND		0xC630-0xC637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC638: ; AND		0xC638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC639: ; AND		0xC639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC63A: ; AND		0xC63A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC63B: ; AND		0xC63B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC63C: ; AND		0xC63C
mov  dl, [esi]
add  esi, byte 2
and   [d+(3*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC640: ; AND		0xC640-0xC647
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC650: ; AND		0xC650-0xC657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC658: ; AND		0xC658-0xC65F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC660: ; AND		0xC660-0xC667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC668: ; AND		0xC668-0xC66F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC670: ; AND		0xC670-0xC677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC678: ; AND		0xC678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC679: ; AND		0xC679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC67A: ; AND		0xC67A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC67B: ; AND		0xC67B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC67C: ; AND		0xC67C
mov  edx, [esi]
add  esi, byte 2
and   [d+(3*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC680: ; AND		0xC680-0xC687
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC690: ; AND		0xC690-0xC697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC698: ; AND		0xC698-0xC69F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6A0: ; AND		0xC6A0-0xC6A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6A8: ; AND		0xC6A8-0xC6AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6B0: ; AND		0xC6B0-0xC6B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6B8: ; AND		0xC6B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6B9: ; AND		0xC6B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6BA: ; AND		0xC6BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6BB: ; AND		0xC6BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6BC: ; AND		0xC6BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(3*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC710: ; AND		0xC710-0xC717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC718: ; AND		0xC718-0xC71F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC720: ; AND		0xC720-0xC727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC728: ; AND		0xC728-0xC72F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC730: ; AND		0xC730-0xC737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC738: ; AND		0xC738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC739: ; AND		0xC739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(3*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC750: ; AND		0xC750-0xC757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC758: ; AND		0xC758-0xC75F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC760: ; AND		0xC760-0xC767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC768: ; AND		0xC768-0xC76F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC770: ; AND		0xC770-0xC777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC778: ; AND		0xC778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC779: ; AND		0xC779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(3*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC790: ; AND		0xC790-0xC797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC798: ; AND		0xC798-0xC79F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7A0: ; AND		0xC7A0-0xC7A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7A8: ; AND		0xC7A8-0xC7AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7B0: ; AND		0xC7B0-0xC7B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7B8: ; AND		0xC7B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7B9: ; AND		0xC7B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(3*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC800: ; AND		0xC800-0xC807
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC810: ; AND		0xC810-0xC817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC818: ; AND		0xC818-0xC81F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC820: ; AND		0xC820-0xC827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC828: ; AND		0xC828-0xC82F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC830: ; AND		0xC830-0xC837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC838: ; AND		0xC838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC839: ; AND		0xC839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC83A: ; AND		0xC83A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC83B: ; AND		0xC83B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC83C: ; AND		0xC83C
mov  dl, [esi]
add  esi, byte 2
and   [d+(4*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC840: ; AND		0xC840-0xC847
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC850: ; AND		0xC850-0xC857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC858: ; AND		0xC858-0xC85F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC860: ; AND		0xC860-0xC867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC868: ; AND		0xC868-0xC86F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC870: ; AND		0xC870-0xC877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC878: ; AND		0xC878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC879: ; AND		0xC879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC87A: ; AND		0xC87A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC87B: ; AND		0xC87B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC87C: ; AND		0xC87C
mov  edx, [esi]
add  esi, byte 2
and   [d+(4*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC880: ; AND		0xC880-0xC887
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC890: ; AND		0xC890-0xC897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC898: ; AND		0xC898-0xC89F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8A0: ; AND		0xC8A0-0xC8A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8A8: ; AND		0xC8A8-0xC8AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8B0: ; AND		0xC8B0-0xC8B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8B8: ; AND		0xC8B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8B9: ; AND		0xC8B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8BA: ; AND		0xC8BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8BB: ; AND		0xC8BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8BC: ; AND		0xC8BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(4*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC910: ; AND		0xC910-0xC917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC918: ; AND		0xC918-0xC91F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC920: ; AND		0xC920-0xC927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC928: ; AND		0xC928-0xC92F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC930: ; AND		0xC930-0xC937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC938: ; AND		0xC938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC939: ; AND		0xC939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(4*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC950: ; AND		0xC950-0xC957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC958: ; AND		0xC958-0xC95F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC960: ; AND		0xC960-0xC967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC968: ; AND		0xC968-0xC96F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC970: ; AND		0xC970-0xC977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC978: ; AND		0xC978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC979: ; AND		0xC979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(4*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC990: ; AND		0xC990-0xC997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC998: ; AND		0xC998-0xC99F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9A0: ; AND		0xC9A0-0xC9A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9A8: ; AND		0xC9A8-0xC9AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9B0: ; AND		0xC9B0-0xC9B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9B8: ; AND		0xC9B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9B9: ; AND		0xC9B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(4*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA00: ; AND		0xCA00-0xCA07
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA10: ; AND		0xCA10-0xCA17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA18: ; AND		0xCA18-0xCA1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA20: ; AND		0xCA20-0xCA27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA28: ; AND		0xCA28-0xCA2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA30: ; AND		0xCA30-0xCA37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA38: ; AND		0xCA38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA39: ; AND		0xCA39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA3A: ; AND		0xCA3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA3B: ; AND		0xCA3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA3C: ; AND		0xCA3C
mov  dl, [esi]
add  esi, byte 2
and   [d+(5*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA40: ; AND		0xCA40-0xCA47
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA50: ; AND		0xCA50-0xCA57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA58: ; AND		0xCA58-0xCA5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA60: ; AND		0xCA60-0xCA67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA68: ; AND		0xCA68-0xCA6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA70: ; AND		0xCA70-0xCA77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA78: ; AND		0xCA78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA79: ; AND		0xCA79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA7A: ; AND		0xCA7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA7B: ; AND		0xCA7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA7C: ; AND		0xCA7C
mov  edx, [esi]
add  esi, byte 2
and   [d+(5*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA80: ; AND		0xCA80-0xCA87
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA90: ; AND		0xCA90-0xCA97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICA98: ; AND		0xCA98-0xCA9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAA0: ; AND		0xCAA0-0xCAA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAA8: ; AND		0xCAA8-0xCAAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAB0: ; AND		0xCAB0-0xCAB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAB8: ; AND		0xCAB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAB9: ; AND		0xCAB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICABA: ; AND		0xCABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICABB: ; AND		0xCABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICABC: ; AND		0xCABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(5*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB10: ; AND		0xCB10-0xCB17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB18: ; AND		0xCB18-0xCB1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB20: ; AND		0xCB20-0xCB27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB28: ; AND		0xCB28-0xCB2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB30: ; AND		0xCB30-0xCB37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB38: ; AND		0xCB38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB39: ; AND		0xCB39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(5*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB50: ; AND		0xCB50-0xCB57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB58: ; AND		0xCB58-0xCB5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB60: ; AND		0xCB60-0xCB67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB68: ; AND		0xCB68-0xCB6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB70: ; AND		0xCB70-0xCB77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB78: ; AND		0xCB78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB79: ; AND		0xCB79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(5*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB90: ; AND		0xCB90-0xCB97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB98: ; AND		0xCB98-0xCB9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBA0: ; AND		0xCBA0-0xCBA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBA8: ; AND		0xCBA8-0xCBAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBB0: ; AND		0xCBB0-0xCBB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBB8: ; AND		0xCBB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBB9: ; AND		0xCBB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(5*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC00: ; AND		0xCC00-0xCC07
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC10: ; AND		0xCC10-0xCC17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC18: ; AND		0xCC18-0xCC1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC20: ; AND		0xCC20-0xCC27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC28: ; AND		0xCC28-0xCC2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC30: ; AND		0xCC30-0xCC37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC38: ; AND		0xCC38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC39: ; AND		0xCC39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC3A: ; AND		0xCC3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC3B: ; AND		0xCC3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC3C: ; AND		0xCC3C
mov  dl, [esi]
add  esi, byte 2
and   [d+(6*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC40: ; AND		0xCC40-0xCC47
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC50: ; AND		0xCC50-0xCC57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC58: ; AND		0xCC58-0xCC5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC60: ; AND		0xCC60-0xCC67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC68: ; AND		0xCC68-0xCC6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC70: ; AND		0xCC70-0xCC77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC78: ; AND		0xCC78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC79: ; AND		0xCC79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC7A: ; AND		0xCC7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC7B: ; AND		0xCC7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC7C: ; AND		0xCC7C
mov  edx, [esi]
add  esi, byte 2
and   [d+(6*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC80: ; AND		0xCC80-0xCC87
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC90: ; AND		0xCC90-0xCC97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICC98: ; AND		0xCC98-0xCC9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCA0: ; AND		0xCCA0-0xCCA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCA8: ; AND		0xCCA8-0xCCAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCB0: ; AND		0xCCB0-0xCCB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCB8: ; AND		0xCCB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCB9: ; AND		0xCCB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCBA: ; AND		0xCCBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCBB: ; AND		0xCCBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCBC: ; AND		0xCCBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(6*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD10: ; AND		0xCD10-0xCD17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD18: ; AND		0xCD18-0xCD1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD20: ; AND		0xCD20-0xCD27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD28: ; AND		0xCD28-0xCD2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD30: ; AND		0xCD30-0xCD37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD38: ; AND		0xCD38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD39: ; AND		0xCD39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(6*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD50: ; AND		0xCD50-0xCD57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD58: ; AND		0xCD58-0xCD5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD60: ; AND		0xCD60-0xCD67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD68: ; AND		0xCD68-0xCD6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD70: ; AND		0xCD70-0xCD77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD78: ; AND		0xCD78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD79: ; AND		0xCD79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(6*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD90: ; AND		0xCD90-0xCD97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD98: ; AND		0xCD98-0xCD9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDA0: ; AND		0xCDA0-0xCDA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDA8: ; AND		0xCDA8-0xCDAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDB0: ; AND		0xCDB0-0xCDB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDB8: ; AND		0xCDB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDB9: ; AND		0xCDB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(6*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE00: ; AND		0xCE00-0xCE07
and  edi, byte 7
mov  dl, [d+edi*4]
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE10: ; AND		0xCE10-0xCE17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE18: ; AND		0xCE18-0xCE1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE20: ; AND		0xCE20-0xCE27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE28: ; AND		0xCE28-0xCE2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE30: ; AND		0xCE30-0xCE37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE38: ; AND		0xCE38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE39: ; AND		0xCE39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE3A: ; AND		0xCE3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE3B: ; AND		0xCE3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE3C: ; AND		0xCE3C
mov  dl, [esi]
add  esi, byte 2
and   [d+(7*4)], dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE40: ; AND		0xCE40-0xCE47
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE50: ; AND		0xCE50-0xCE57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE58: ; AND		0xCE58-0xCE5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE60: ; AND		0xCE60-0xCE67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE68: ; AND		0xCE68-0xCE6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE70: ; AND		0xCE70-0xCE77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE78: ; AND		0xCE78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE79: ; AND		0xCE79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE7A: ; AND		0xCE7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE7B: ; AND		0xCE7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE7C: ; AND		0xCE7C
mov  edx, [esi]
add  esi, byte 2
and   [d+(7*4)], dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE80: ; AND		0xCE80-0xCE87
and  edi, byte 7
mov  edx, [d+edi*4]
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE90: ; AND		0xCE90-0xCE97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICE98: ; AND		0xCE98-0xCE9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEA0: ; AND		0xCEA0-0xCEA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEA8: ; AND		0xCEA8-0xCEAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEB0: ; AND		0xCEB0-0xCEB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEB8: ; AND		0xCEB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEB9: ; AND		0xCEB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEBA: ; AND		0xCEBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEBB: ; AND		0xCEBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEBC: ; AND		0xCEBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
and   [d+(7*4)], edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF10: ; AND		0xCF10-0xCF17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF18: ; AND		0xCF18-0xCF1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF20: ; AND		0xCF20-0xCF27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF28: ; AND		0xCF28-0xCF2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF30: ; AND		0xCF30-0xCF37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF38: ; AND		0xCF38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF39: ; AND		0xCF39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
and   dl, [d+(7*4)]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF50: ; AND		0xCF50-0xCF57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF58: ; AND		0xCF58-0xCF5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF60: ; AND		0xCF60-0xCF67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF68: ; AND		0xCF68-0xCF6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF70: ; AND		0xCF70-0xCF77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF78: ; AND		0xCF78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF79: ; AND		0xCF79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
and   dx, [d+(7*4)]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF90: ; AND		0xCF90-0xCF97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF98: ; AND		0xCF98-0xCF9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFA0: ; AND		0xCFA0-0xCFA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFA8: ; AND		0xCFA8-0xCFAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFB0: ; AND		0xCFB0-0xCFB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFB8: ; AND		0xCFB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFB9: ; AND		0xCFB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
and   edx, [d+(7*4)]
lahf
xor  al, al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID000: ; ADD		0xD000-0xD007
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID010: ; ADD		0xD010-0xD017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID018: ; ADD		0xD018-0xD01F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID020: ; ADD		0xD020-0xD027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID028: ; ADD		0xD028-0xD02F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID030: ; ADD		0xD030-0xD037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID038: ; ADD		0xD038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID039: ; ADD		0xD039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID03A: ; ADD		0xD03A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID03B: ; ADD		0xD03B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID03C: ; ADD		0xD03C
mov  dl, [esi]
add  esi, byte 2
add   [d+(0*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID040: ; ADD		0xD040-0xD047
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID048: ; ADD		0xD048-0xD04F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID050: ; ADD		0xD050-0xD057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID058: ; ADD		0xD058-0xD05F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID060: ; ADD		0xD060-0xD067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID068: ; ADD		0xD068-0xD06F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID070: ; ADD		0xD070-0xD077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID078: ; ADD		0xD078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID079: ; ADD		0xD079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID07A: ; ADD		0xD07A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID07B: ; ADD		0xD07B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID07C: ; ADD		0xD07C
mov  edx, [esi]
add  esi, byte 2
add   [d+(0*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID080: ; ADD		0xD080-0xD087
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID088: ; ADD		0xD088-0xD08F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID090: ; ADD		0xD090-0xD097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID098: ; ADD		0xD098-0xD09F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0A0: ; ADD		0xD0A0-0xD0A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0A8: ; ADD		0xD0A8-0xD0AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0B0: ; ADD		0xD0B0-0xD0B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0B8: ; ADD		0xD0B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0B9: ; ADD		0xD0B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0BA: ; ADD		0xD0BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0BB: ; ADD		0xD0BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0BC: ; ADD		0xD0BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(0*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID110: ; ADD		0xD110-0xD117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID118: ; ADD		0xD118-0xD11F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID120: ; ADD		0xD120-0xD127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID128: ; ADD		0xD128-0xD12F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID130: ; ADD		0xD130-0xD137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID138: ; ADD		0xD138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID139: ; ADD		0xD139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID150: ; ADD		0xD150-0xD157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID158: ; ADD		0xD158-0xD15F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID160: ; ADD		0xD160-0xD167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID168: ; ADD		0xD168-0xD16F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID170: ; ADD		0xD170-0xD177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID178: ; ADD		0xD178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID179: ; ADD		0xD179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID190: ; ADD		0xD190-0xD197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID198: ; ADD		0xD198-0xD19F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1A0: ; ADD		0xD1A0-0xD1A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1A8: ; ADD		0xD1A8-0xD1AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1B0: ; ADD		0xD1B0-0xD1B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1B8: ; ADD		0xD1B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1B9: ; ADD		0xD1B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(0*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID200: ; ADD		0xD200-0xD207
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID210: ; ADD		0xD210-0xD217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID218: ; ADD		0xD218-0xD21F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID220: ; ADD		0xD220-0xD227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID228: ; ADD		0xD228-0xD22F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID230: ; ADD		0xD230-0xD237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID238: ; ADD		0xD238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID239: ; ADD		0xD239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID23A: ; ADD		0xD23A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID23B: ; ADD		0xD23B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID23C: ; ADD		0xD23C
mov  dl, [esi]
add  esi, byte 2
add   [d+(1*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID240: ; ADD		0xD240-0xD247
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID248: ; ADD		0xD248-0xD24F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID250: ; ADD		0xD250-0xD257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID258: ; ADD		0xD258-0xD25F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID260: ; ADD		0xD260-0xD267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID268: ; ADD		0xD268-0xD26F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID270: ; ADD		0xD270-0xD277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID278: ; ADD		0xD278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID279: ; ADD		0xD279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID27A: ; ADD		0xD27A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID27B: ; ADD		0xD27B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID27C: ; ADD		0xD27C
mov  edx, [esi]
add  esi, byte 2
add   [d+(1*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID280: ; ADD		0xD280-0xD287
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID288: ; ADD		0xD288-0xD28F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID290: ; ADD		0xD290-0xD297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID298: ; ADD		0xD298-0xD29F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2A0: ; ADD		0xD2A0-0xD2A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2A8: ; ADD		0xD2A8-0xD2AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2B0: ; ADD		0xD2B0-0xD2B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2B8: ; ADD		0xD2B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2B9: ; ADD		0xD2B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2BA: ; ADD		0xD2BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2BB: ; ADD		0xD2BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2BC: ; ADD		0xD2BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(1*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID310: ; ADD		0xD310-0xD317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID318: ; ADD		0xD318-0xD31F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID320: ; ADD		0xD320-0xD327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID328: ; ADD		0xD328-0xD32F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID330: ; ADD		0xD330-0xD337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID338: ; ADD		0xD338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID339: ; ADD		0xD339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID350: ; ADD		0xD350-0xD357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID358: ; ADD		0xD358-0xD35F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID360: ; ADD		0xD360-0xD367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID368: ; ADD		0xD368-0xD36F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID370: ; ADD		0xD370-0xD377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID378: ; ADD		0xD378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID379: ; ADD		0xD379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID390: ; ADD		0xD390-0xD397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID398: ; ADD		0xD398-0xD39F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3A0: ; ADD		0xD3A0-0xD3A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3A8: ; ADD		0xD3A8-0xD3AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3B0: ; ADD		0xD3B0-0xD3B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3B8: ; ADD		0xD3B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3B9: ; ADD		0xD3B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(1*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID400: ; ADD		0xD400-0xD407
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID410: ; ADD		0xD410-0xD417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID418: ; ADD		0xD418-0xD41F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID420: ; ADD		0xD420-0xD427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID428: ; ADD		0xD428-0xD42F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID430: ; ADD		0xD430-0xD437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID438: ; ADD		0xD438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID439: ; ADD		0xD439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID43A: ; ADD		0xD43A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID43B: ; ADD		0xD43B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID43C: ; ADD		0xD43C
mov  dl, [esi]
add  esi, byte 2
add   [d+(2*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID440: ; ADD		0xD440-0xD447
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID448: ; ADD		0xD448-0xD44F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID450: ; ADD		0xD450-0xD457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID458: ; ADD		0xD458-0xD45F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID460: ; ADD		0xD460-0xD467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID468: ; ADD		0xD468-0xD46F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID470: ; ADD		0xD470-0xD477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID478: ; ADD		0xD478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID479: ; ADD		0xD479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID47A: ; ADD		0xD47A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID47B: ; ADD		0xD47B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID47C: ; ADD		0xD47C
mov  edx, [esi]
add  esi, byte 2
add   [d+(2*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID480: ; ADD		0xD480-0xD487
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID488: ; ADD		0xD488-0xD48F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID490: ; ADD		0xD490-0xD497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID498: ; ADD		0xD498-0xD49F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4A0: ; ADD		0xD4A0-0xD4A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4A8: ; ADD		0xD4A8-0xD4AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4B0: ; ADD		0xD4B0-0xD4B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4B8: ; ADD		0xD4B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4B9: ; ADD		0xD4B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4BA: ; ADD		0xD4BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4BB: ; ADD		0xD4BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4BC: ; ADD		0xD4BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(2*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID510: ; ADD		0xD510-0xD517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID518: ; ADD		0xD518-0xD51F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID520: ; ADD		0xD520-0xD527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID528: ; ADD		0xD528-0xD52F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID530: ; ADD		0xD530-0xD537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID538: ; ADD		0xD538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID539: ; ADD		0xD539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID550: ; ADD		0xD550-0xD557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID558: ; ADD		0xD558-0xD55F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID560: ; ADD		0xD560-0xD567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID568: ; ADD		0xD568-0xD56F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID570: ; ADD		0xD570-0xD577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID578: ; ADD		0xD578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID579: ; ADD		0xD579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID590: ; ADD		0xD590-0xD597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID598: ; ADD		0xD598-0xD59F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5A0: ; ADD		0xD5A0-0xD5A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5A8: ; ADD		0xD5A8-0xD5AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5B0: ; ADD		0xD5B0-0xD5B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5B8: ; ADD		0xD5B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5B9: ; ADD		0xD5B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(2*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID600: ; ADD		0xD600-0xD607
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID610: ; ADD		0xD610-0xD617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID618: ; ADD		0xD618-0xD61F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID620: ; ADD		0xD620-0xD627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID628: ; ADD		0xD628-0xD62F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID630: ; ADD		0xD630-0xD637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID638: ; ADD		0xD638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID639: ; ADD		0xD639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID63A: ; ADD		0xD63A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID63B: ; ADD		0xD63B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID63C: ; ADD		0xD63C
mov  dl, [esi]
add  esi, byte 2
add   [d+(3*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID640: ; ADD		0xD640-0xD647
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID648: ; ADD		0xD648-0xD64F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID650: ; ADD		0xD650-0xD657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID658: ; ADD		0xD658-0xD65F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID660: ; ADD		0xD660-0xD667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID668: ; ADD		0xD668-0xD66F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID670: ; ADD		0xD670-0xD677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID678: ; ADD		0xD678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID679: ; ADD		0xD679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID67A: ; ADD		0xD67A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID67B: ; ADD		0xD67B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID67C: ; ADD		0xD67C
mov  edx, [esi]
add  esi, byte 2
add   [d+(3*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID680: ; ADD		0xD680-0xD687
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID688: ; ADD		0xD688-0xD68F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID690: ; ADD		0xD690-0xD697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID698: ; ADD		0xD698-0xD69F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6A0: ; ADD		0xD6A0-0xD6A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6A8: ; ADD		0xD6A8-0xD6AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6B0: ; ADD		0xD6B0-0xD6B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6B8: ; ADD		0xD6B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6B9: ; ADD		0xD6B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6BA: ; ADD		0xD6BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6BB: ; ADD		0xD6BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6BC: ; ADD		0xD6BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(3*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID710: ; ADD		0xD710-0xD717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID718: ; ADD		0xD718-0xD71F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID720: ; ADD		0xD720-0xD727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID728: ; ADD		0xD728-0xD72F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID730: ; ADD		0xD730-0xD737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID738: ; ADD		0xD738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID739: ; ADD		0xD739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID750: ; ADD		0xD750-0xD757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID758: ; ADD		0xD758-0xD75F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID760: ; ADD		0xD760-0xD767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID768: ; ADD		0xD768-0xD76F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID770: ; ADD		0xD770-0xD777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID778: ; ADD		0xD778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID779: ; ADD		0xD779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID790: ; ADD		0xD790-0xD797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID798: ; ADD		0xD798-0xD79F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7A0: ; ADD		0xD7A0-0xD7A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7A8: ; ADD		0xD7A8-0xD7AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7B0: ; ADD		0xD7B0-0xD7B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7B8: ; ADD		0xD7B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7B9: ; ADD		0xD7B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(3*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID800: ; ADD		0xD800-0xD807
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID810: ; ADD		0xD810-0xD817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID818: ; ADD		0xD818-0xD81F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID820: ; ADD		0xD820-0xD827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID828: ; ADD		0xD828-0xD82F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID830: ; ADD		0xD830-0xD837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID838: ; ADD		0xD838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID839: ; ADD		0xD839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID83A: ; ADD		0xD83A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID83B: ; ADD		0xD83B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID83C: ; ADD		0xD83C
mov  dl, [esi]
add  esi, byte 2
add   [d+(4*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID840: ; ADD		0xD840-0xD847
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID848: ; ADD		0xD848-0xD84F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID850: ; ADD		0xD850-0xD857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID858: ; ADD		0xD858-0xD85F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID860: ; ADD		0xD860-0xD867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID868: ; ADD		0xD868-0xD86F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID870: ; ADD		0xD870-0xD877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID878: ; ADD		0xD878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID879: ; ADD		0xD879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID87A: ; ADD		0xD87A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID87B: ; ADD		0xD87B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID87C: ; ADD		0xD87C
mov  edx, [esi]
add  esi, byte 2
add   [d+(4*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID880: ; ADD		0xD880-0xD887
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID888: ; ADD		0xD888-0xD88F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID890: ; ADD		0xD890-0xD897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID898: ; ADD		0xD898-0xD89F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8A0: ; ADD		0xD8A0-0xD8A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8A8: ; ADD		0xD8A8-0xD8AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8B0: ; ADD		0xD8B0-0xD8B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8B8: ; ADD		0xD8B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8B9: ; ADD		0xD8B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8BA: ; ADD		0xD8BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8BB: ; ADD		0xD8BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8BC: ; ADD		0xD8BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(4*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID910: ; ADD		0xD910-0xD917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID918: ; ADD		0xD918-0xD91F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID920: ; ADD		0xD920-0xD927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID928: ; ADD		0xD928-0xD92F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID930: ; ADD		0xD930-0xD937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID938: ; ADD		0xD938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID939: ; ADD		0xD939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID950: ; ADD		0xD950-0xD957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID958: ; ADD		0xD958-0xD95F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID960: ; ADD		0xD960-0xD967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID968: ; ADD		0xD968-0xD96F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID970: ; ADD		0xD970-0xD977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID978: ; ADD		0xD978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID979: ; ADD		0xD979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID990: ; ADD		0xD990-0xD997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID998: ; ADD		0xD998-0xD99F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9A0: ; ADD		0xD9A0-0xD9A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9A8: ; ADD		0xD9A8-0xD9AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9B0: ; ADD		0xD9B0-0xD9B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9B8: ; ADD		0xD9B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9B9: ; ADD		0xD9B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(4*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA00: ; ADD		0xDA00-0xDA07
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA10: ; ADD		0xDA10-0xDA17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA18: ; ADD		0xDA18-0xDA1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA20: ; ADD		0xDA20-0xDA27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA28: ; ADD		0xDA28-0xDA2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA30: ; ADD		0xDA30-0xDA37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA38: ; ADD		0xDA38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA39: ; ADD		0xDA39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA3A: ; ADD		0xDA3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA3B: ; ADD		0xDA3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA3C: ; ADD		0xDA3C
mov  dl, [esi]
add  esi, byte 2
add   [d+(5*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA40: ; ADD		0xDA40-0xDA47
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA48: ; ADD		0xDA48-0xDA4F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA50: ; ADD		0xDA50-0xDA57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA58: ; ADD		0xDA58-0xDA5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA60: ; ADD		0xDA60-0xDA67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA68: ; ADD		0xDA68-0xDA6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA70: ; ADD		0xDA70-0xDA77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA78: ; ADD		0xDA78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA79: ; ADD		0xDA79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA7A: ; ADD		0xDA7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA7B: ; ADD		0xDA7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA7C: ; ADD		0xDA7C
mov  edx, [esi]
add  esi, byte 2
add   [d+(5*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA80: ; ADD		0xDA80-0xDA87
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA88: ; ADD		0xDA88-0xDA8F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA90: ; ADD		0xDA90-0xDA97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDA98: ; ADD		0xDA98-0xDA9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAA0: ; ADD		0xDAA0-0xDAA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAA8: ; ADD		0xDAA8-0xDAAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAB0: ; ADD		0xDAB0-0xDAB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAB8: ; ADD		0xDAB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAB9: ; ADD		0xDAB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDABA: ; ADD		0xDABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDABB: ; ADD		0xDABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDABC: ; ADD		0xDABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(5*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB10: ; ADD		0xDB10-0xDB17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB18: ; ADD		0xDB18-0xDB1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB20: ; ADD		0xDB20-0xDB27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB28: ; ADD		0xDB28-0xDB2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB30: ; ADD		0xDB30-0xDB37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB38: ; ADD		0xDB38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB39: ; ADD		0xDB39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB50: ; ADD		0xDB50-0xDB57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB58: ; ADD		0xDB58-0xDB5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB60: ; ADD		0xDB60-0xDB67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB68: ; ADD		0xDB68-0xDB6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB70: ; ADD		0xDB70-0xDB77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB78: ; ADD		0xDB78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB79: ; ADD		0xDB79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB90: ; ADD		0xDB90-0xDB97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB98: ; ADD		0xDB98-0xDB9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBA0: ; ADD		0xDBA0-0xDBA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBA8: ; ADD		0xDBA8-0xDBAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBB0: ; ADD		0xDBB0-0xDBB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBB8: ; ADD		0xDBB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBB9: ; ADD		0xDBB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(5*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC00: ; ADD		0xDC00-0xDC07
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC10: ; ADD		0xDC10-0xDC17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC18: ; ADD		0xDC18-0xDC1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC20: ; ADD		0xDC20-0xDC27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC28: ; ADD		0xDC28-0xDC2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC30: ; ADD		0xDC30-0xDC37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC38: ; ADD		0xDC38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC39: ; ADD		0xDC39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC3A: ; ADD		0xDC3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC3B: ; ADD		0xDC3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC3C: ; ADD		0xDC3C
mov  dl, [esi]
add  esi, byte 2
add   [d+(6*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC40: ; ADD		0xDC40-0xDC47
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC48: ; ADD		0xDC48-0xDC4F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC50: ; ADD		0xDC50-0xDC57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC58: ; ADD		0xDC58-0xDC5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC60: ; ADD		0xDC60-0xDC67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC68: ; ADD		0xDC68-0xDC6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC70: ; ADD		0xDC70-0xDC77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC78: ; ADD		0xDC78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC79: ; ADD		0xDC79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC7A: ; ADD		0xDC7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC7B: ; ADD		0xDC7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC7C: ; ADD		0xDC7C
mov  edx, [esi]
add  esi, byte 2
add   [d+(6*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC80: ; ADD		0xDC80-0xDC87
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC88: ; ADD		0xDC88-0xDC8F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC90: ; ADD		0xDC90-0xDC97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDC98: ; ADD		0xDC98-0xDC9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCA0: ; ADD		0xDCA0-0xDCA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCA8: ; ADD		0xDCA8-0xDCAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCB0: ; ADD		0xDCB0-0xDCB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCB8: ; ADD		0xDCB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCB9: ; ADD		0xDCB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCBA: ; ADD		0xDCBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCBB: ; ADD		0xDCBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCBC: ; ADD		0xDCBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(6*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD10: ; ADD		0xDD10-0xDD17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD18: ; ADD		0xDD18-0xDD1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD20: ; ADD		0xDD20-0xDD27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD28: ; ADD		0xDD28-0xDD2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD30: ; ADD		0xDD30-0xDD37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD38: ; ADD		0xDD38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD39: ; ADD		0xDD39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD50: ; ADD		0xDD50-0xDD57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD58: ; ADD		0xDD58-0xDD5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD60: ; ADD		0xDD60-0xDD67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD68: ; ADD		0xDD68-0xDD6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD70: ; ADD		0xDD70-0xDD77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD78: ; ADD		0xDD78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD79: ; ADD		0xDD79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD90: ; ADD		0xDD90-0xDD97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD98: ; ADD		0xDD98-0xDD9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDA0: ; ADD		0xDDA0-0xDDA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDA8: ; ADD		0xDDA8-0xDDAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDB0: ; ADD		0xDDB0-0xDDB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDB8: ; ADD		0xDDB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDB9: ; ADD		0xDDB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(6*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE00: ; ADD		0xDE00-0xDE07
and  edi, byte 7
mov  dl, [d+edi*4]
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE10: ; ADD		0xDE10-0xDE17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE18: ; ADD		0xDE18-0xDE1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE20: ; ADD		0xDE20-0xDE27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE28: ; ADD		0xDE28-0xDE2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE30: ; ADD		0xDE30-0xDE37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE38: ; ADD		0xDE38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE39: ; ADD		0xDE39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE3A: ; ADD		0xDE3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE3B: ; ADD		0xDE3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE3C: ; ADD		0xDE3C
mov  dl, [esi]
add  esi, byte 2
add   [d+(7*4)], dl
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE40: ; ADD		0xDE40-0xDE47
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE48: ; ADD		0xDE48-0xDE4F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE50: ; ADD		0xDE50-0xDE57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE58: ; ADD		0xDE58-0xDE5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE60: ; ADD		0xDE60-0xDE67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE68: ; ADD		0xDE68-0xDE6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE70: ; ADD		0xDE70-0xDE77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE78: ; ADD		0xDE78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE79: ; ADD		0xDE79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE7A: ; ADD		0xDE7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE7B: ; ADD		0xDE7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE7C: ; ADD		0xDE7C
mov  edx, [esi]
add  esi, byte 2
add   [d+(7*4)], dx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE80: ; ADD		0xDE80-0xDE87
and  edi, byte 7
mov  edx, [d+edi*4]
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE88: ; ADD		0xDE88-0xDE8F
and  edi, byte 7
mov  edx, [a+edi*4]
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE90: ; ADD		0xDE90-0xDE97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDE98: ; ADD		0xDE98-0xDE9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEA0: ; ADD		0xDEA0-0xDEA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEA8: ; ADD		0xDEA8-0xDEAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEB0: ; ADD		0xDEB0-0xDEB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEB8: ; ADD		0xDEB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEB9: ; ADD		0xDEB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEBA: ; ADD		0xDEBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEBB: ; ADD		0xDEBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEBC: ; ADD		0xDEBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add   [d+(7*4)], edx
lahf
setc byte [x]
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF10: ; ADD		0xDF10-0xDF17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF18: ; ADD		0xDF18-0xDF1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF20: ; ADD		0xDF20-0xDF27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF28: ; ADD		0xDF28-0xDF2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF30: ; ADD		0xDF30-0xDF37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF38: ; ADD		0xDF38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF39: ; ADD		0xDF39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF50: ; ADD		0xDF50-0xDF57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF58: ; ADD		0xDF58-0xDF5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF60: ; ADD		0xDF60-0xDF67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF68: ; ADD		0xDF68-0xDF6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF70: ; ADD		0xDF70-0xDF77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF78: ; ADD		0xDF78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF79: ; ADD		0xDF79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF90: ; ADD		0xDF90-0xDF97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF98: ; ADD		0xDF98-0xDF9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFA0: ; ADD		0xDFA0-0xDFA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFA8: ; ADD		0xDFA8-0xDFAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFB0: ; ADD		0xDFB0-0xDFB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFB8: ; ADD		0xDFB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFB9: ; ADD		0xDFB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, [d+(7*4)]
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0000: ; ORI		0x0000-0x0007
and  edi, byte 7
mov  edx, [esi]
or   byte [d+edi*4], dl
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0010: ; ORI		0x0010-0x0017
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0018: ; ORI		0x0018-0x001F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0020: ; ORI		0x0020-0x0027
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0028: ; ORI		0x0028-0x002F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0030: ; ORI		0x0030-0x0037
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0038: ; ORI		0x0038
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0039: ; ORI		0x0039
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
or   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0040: ; ORI		0x0040-0x0047
and  edi, byte 7
mov  edx, [esi]
or   word [d+edi*4], dx
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0050: ; ORI		0x0050-0x0057
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0058: ; ORI		0x0058-0x005F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0060: ; ORI		0x0060-0x0067
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0068: ; ORI		0x0068-0x006F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0070: ; ORI		0x0070-0x0077
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0078: ; ORI		0x0078
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0079: ; ORI		0x0079
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
or   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0080: ; ORI		0x0080-0x0087
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
or   dword [d+edi*4], edx
lahf
xor  al, al
add  esi, byte 4
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0090: ; ORI		0x0090-0x0097
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0098: ; ORI		0x0098-0x009F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I00A0: ; ORI		0x00A0-0x00A7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I00A8: ; ORI		0x00A8-0x00AF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I00B0: ; ORI		0x00B0-0x00B7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I00B8: ; ORI		0x00B8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I00B9: ; ORI		0x00B9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
or   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0200: ; ANDI		0x0200-0x0207
and  edi, byte 7
mov  edx, [esi]
and   byte [d+edi*4], dl
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0210: ; ANDI		0x0210-0x0217
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0218: ; ANDI		0x0218-0x021F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0220: ; ANDI		0x0220-0x0227
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0228: ; ANDI		0x0228-0x022F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0230: ; ANDI		0x0230-0x0237
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0238: ; ANDI		0x0238
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0239: ; ANDI		0x0239
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
and   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0240: ; ANDI		0x0240-0x0247
and  edi, byte 7
mov  edx, [esi]
and   word [d+edi*4], dx
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0250: ; ANDI		0x0250-0x0257
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0258: ; ANDI		0x0258-0x025F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0260: ; ANDI		0x0260-0x0267
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0268: ; ANDI		0x0268-0x026F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0270: ; ANDI		0x0270-0x0277
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0278: ; ANDI		0x0278
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0279: ; ANDI		0x0279
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
and   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0280: ; ANDI		0x0280-0x0287
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
and   dword [d+edi*4], edx
lahf
xor  al, al
add  esi, byte 4
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0290: ; ANDI		0x0290-0x0297
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0298: ; ANDI		0x0298-0x029F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I02A0: ; ANDI		0x02A0-0x02A7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I02A8: ; ANDI		0x02A8-0x02AF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I02B0: ; ANDI		0x02B0-0x02B7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I02B8: ; ANDI		0x02B8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I02B9: ; ANDI		0x02B9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
and   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0400: ; SUBI		0x0400-0x0407
and  edi, byte 7
mov  edx, [esi]
sub   byte [d+edi*4], dl
lahf
setc byte [x]
seto al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0410: ; SUBI		0x0410-0x0417
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0418: ; SUBI		0x0418-0x041F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0420: ; SUBI		0x0420-0x0427
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0428: ; SUBI		0x0428-0x042F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0430: ; SUBI		0x0430-0x0437
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0438: ; SUBI		0x0438
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0439: ; SUBI		0x0439
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
sub   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0440: ; SUBI		0x0440-0x0447
and  edi, byte 7
mov  edx, [esi]
sub   word [d+edi*4], dx
lahf
setc byte [x]
seto al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0450: ; SUBI		0x0450-0x0457
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0458: ; SUBI		0x0458-0x045F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0460: ; SUBI		0x0460-0x0467
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0468: ; SUBI		0x0468-0x046F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0470: ; SUBI		0x0470-0x0477
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0478: ; SUBI		0x0478
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0479: ; SUBI		0x0479
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
sub   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0480: ; SUBI		0x0480-0x0487
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
sub   dword [d+edi*4], edx
lahf
setc byte [x]
seto al
add  esi, byte 4
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0490: ; SUBI		0x0490-0x0497
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0498: ; SUBI		0x0498-0x049F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I04A0: ; SUBI		0x04A0-0x04A7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I04A8: ; SUBI		0x04A8-0x04AF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I04B0: ; SUBI		0x04B0-0x04B7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I04B8: ; SUBI		0x04B8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I04B9: ; SUBI		0x04B9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
sub   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0600: ; ADDI		0x0600-0x0607
and  edi, byte 7
mov  edx, [esi]
add   byte [d+edi*4], dl
lahf
setc byte [x]
seto al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0610: ; ADDI		0x0610-0x0617
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0618: ; ADDI		0x0618-0x061F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0620: ; ADDI		0x0620-0x0627
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0628: ; ADDI		0x0628-0x062F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0630: ; ADDI		0x0630-0x0637
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0638: ; ADDI		0x0638
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0639: ; ADDI		0x0639
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
add   dl, [edi]
lahf
setc byte [x]
seto al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0640: ; ADDI		0x0640-0x0647
and  edi, byte 7
mov  edx, [esi]
add   word [d+edi*4], dx
lahf
setc byte [x]
seto al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0650: ; ADDI		0x0650-0x0657
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0658: ; ADDI		0x0658-0x065F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0660: ; ADDI		0x0660-0x0667
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0668: ; ADDI		0x0668-0x066F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0670: ; ADDI		0x0670-0x0677
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0678: ; ADDI		0x0678
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0679: ; ADDI		0x0679
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
add   dx, [edi]
lahf
setc byte [x]
seto al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0680: ; ADDI		0x0680-0x0687
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
add   dword [d+edi*4], edx
lahf
setc byte [x]
seto al
add  esi, byte 4
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0690: ; ADDI		0x0690-0x0697
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0698: ; ADDI		0x0698-0x069F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I06A0: ; ADDI		0x06A0-0x06A7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I06A8: ; ADDI		0x06A8-0x06AF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I06B0: ; ADDI		0x06B0-0x06B7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I06B8: ; ADDI		0x06B8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I06B9: ; ADDI		0x06B9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
add   edx, edi
lahf
setc byte [x]
seto al
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A00: ; EORI		0x0A00-0x0A07
and  edi, byte 7
mov  edx, [esi]
xor   byte [d+edi*4], dl
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A10: ; EORI		0x0A10-0x0A17
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A18: ; EORI		0x0A18-0x0A1F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A20: ; EORI		0x0A20-0x0A27
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A28: ; EORI		0x0A28-0x0A2F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A30: ; EORI		0x0A30-0x0A37
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A38: ; EORI		0x0A38
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A39: ; EORI		0x0A39
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
xor   dl, [edi]
lahf
xor  al, al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A40: ; EORI		0x0A40-0x0A47
and  edi, byte 7
mov  edx, [esi]
xor   word [d+edi*4], dx
lahf
xor  al, al
add  esi, byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A50: ; EORI		0x0A50-0x0A57
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A58: ; EORI		0x0A58-0x0A5F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A60: ; EORI		0x0A60-0x0A67
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A68: ; EORI		0x0A68-0x0A6F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A70: ; EORI		0x0A70-0x0A77
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A78: ; EORI		0x0A78
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A79: ; EORI		0x0A79
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
xor   dx, [edi]
lahf
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A80: ; EORI		0x0A80-0x0A87
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
xor   dword [d+edi*4], edx
lahf
xor  al, al
add  esi, byte 4
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A90: ; EORI		0x0A90-0x0A97
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A98: ; EORI		0x0A98-0x0A9F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0AA0: ; EORI		0x0AA0-0x0AA7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0AA8: ; EORI		0x0AA8-0x0AAF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0AB0: ; EORI		0x0AB0-0x0AB7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0AB8: ; EORI		0x0AB8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0AB9: ; EORI		0x0AB9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
xor   edx, edi
lahf
xor  al, al
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C00: ; CMPI		0x0C00-0x0C07
and  edi, byte 7
mov  edx, [esi]
cmp   byte [d+edi*4], dl
lahf
seto al
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C10: ; CMPI		0x0C10-0x0C17
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C18: ; CMPI		0x0C18-0x0C1F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C20: ; CMPI		0x0C20-0x0C27
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C28: ; CMPI		0x0C28-0x0C2F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C30: ; CMPI		0x0C30-0x0C37
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C38: ; CMPI		0x0C38
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C39: ; CMPI		0x0C39
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_esi]
cmp   dl, [edi]
lahf
seto al
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C40: ; CMPI		0x0C40-0x0C47
and  edi, byte 7
mov  edx, [esi]
cmp   word [d+edi*4], dx
lahf
seto al
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C50: ; CMPI		0x0C50-0x0C57
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C58: ; CMPI		0x0C58-0x0C5F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C60: ; CMPI		0x0C60-0x0C67
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C68: ; CMPI		0x0C68-0x0C6F
mov  [run_esi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C70: ; CMPI		0x0C70-0x0C77
mov  [run_esi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C78: ; CMPI		0x0C78
mov  [run_esi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C79: ; CMPI		0x0C79
mov  [run_esi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, [run_esi]
cmp   dx, [edi]
lahf
seto al
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C80: ; CMPI		0x0C80-0x0C87
and  edi, byte 7
mov  edx, [esi]
ror  edx, byte 16
cmp   dword [d+edi*4], edx
lahf
seto al
add  esi, byte 4
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C90: ; CMPI		0x0C90-0x0C97
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0C98: ; CMPI		0x0C98-0x0C9F
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0CA0: ; CMPI		0x0CA0-0x0CA7
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0CA8: ; CMPI		0x0CA8-0x0CAF
mov  [run_esi], esi
add  esi, byte 4
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0CB0: ; CMPI		0x0CB0-0x0CB7
mov  [run_esi], esi
add  esi, byte 4
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0CB8: ; CMPI		0x0CB8
mov  [run_esi], esi
add  esi, byte 4
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0CB9: ; CMPI		0x0CB9
mov  [run_esi], esi
add  esi, byte 4
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, [run_esi]
mov  edi, [edi]
ror  edi, byte 16
cmp   edx, edi
lahf
seto al
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0C0: ; ADDA		0xD0C0-0xD0C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0C8: ; ADDA		0xD0C8-0xD0CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0D0: ; ADDA		0xD0D0-0xD0D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0D8: ; ADDA		0xD0D8-0xD0DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0E0: ; ADDA		0xD0E0-0xD0E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0E8: ; ADDA		0xD0E8-0xD0EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0F0: ; ADDA		0xD0F0-0xD0F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0F8: ; ADDA		0xD0F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0F9: ; ADDA		0xD0F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0FA: ; ADDA		0xD0FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0FB: ; ADDA		0xD0FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID0FC: ; ADDA		0xD0FC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2C0: ; ADDA		0xD2C0-0xD2C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2C8: ; ADDA		0xD2C8-0xD2CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2D0: ; ADDA		0xD2D0-0xD2D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2D8: ; ADDA		0xD2D8-0xD2DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2E0: ; ADDA		0xD2E0-0xD2E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2E8: ; ADDA		0xD2E8-0xD2EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2F0: ; ADDA		0xD2F0-0xD2F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2F8: ; ADDA		0xD2F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2F9: ; ADDA		0xD2F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2FA: ; ADDA		0xD2FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2FB: ; ADDA		0xD2FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID2FC: ; ADDA		0xD2FC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4C0: ; ADDA		0xD4C0-0xD4C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4C8: ; ADDA		0xD4C8-0xD4CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4D0: ; ADDA		0xD4D0-0xD4D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4D8: ; ADDA		0xD4D8-0xD4DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4E0: ; ADDA		0xD4E0-0xD4E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4E8: ; ADDA		0xD4E8-0xD4EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4F0: ; ADDA		0xD4F0-0xD4F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4F8: ; ADDA		0xD4F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4F9: ; ADDA		0xD4F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4FA: ; ADDA		0xD4FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4FB: ; ADDA		0xD4FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID4FC: ; ADDA		0xD4FC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6C0: ; ADDA		0xD6C0-0xD6C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6C8: ; ADDA		0xD6C8-0xD6CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6D0: ; ADDA		0xD6D0-0xD6D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6D8: ; ADDA		0xD6D8-0xD6DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6E0: ; ADDA		0xD6E0-0xD6E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6E8: ; ADDA		0xD6E8-0xD6EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6F0: ; ADDA		0xD6F0-0xD6F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6F8: ; ADDA		0xD6F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6F9: ; ADDA		0xD6F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6FA: ; ADDA		0xD6FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6FB: ; ADDA		0xD6FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID6FC: ; ADDA		0xD6FC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8C0: ; ADDA		0xD8C0-0xD8C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8C8: ; ADDA		0xD8C8-0xD8CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8D0: ; ADDA		0xD8D0-0xD8D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8D8: ; ADDA		0xD8D8-0xD8DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8E0: ; ADDA		0xD8E0-0xD8E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8E8: ; ADDA		0xD8E8-0xD8EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8F0: ; ADDA		0xD8F0-0xD8F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8F8: ; ADDA		0xD8F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8F9: ; ADDA		0xD8F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8FA: ; ADDA		0xD8FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8FB: ; ADDA		0xD8FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID8FC: ; ADDA		0xD8FC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAC0: ; ADDA		0xDAC0-0xDAC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAC8: ; ADDA		0xDAC8-0xDACF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAD0: ; ADDA		0xDAD0-0xDAD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAD8: ; ADDA		0xDAD8-0xDADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAE0: ; ADDA		0xDAE0-0xDAE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAE8: ; ADDA		0xDAE8-0xDAEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAF0: ; ADDA		0xDAF0-0xDAF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAF8: ; ADDA		0xDAF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAF9: ; ADDA		0xDAF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAFA: ; ADDA		0xDAFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAFB: ; ADDA		0xDAFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDAFC: ; ADDA		0xDAFC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCC0: ; ADDA		0xDCC0-0xDCC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCC8: ; ADDA		0xDCC8-0xDCCF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCD0: ; ADDA		0xDCD0-0xDCD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCD8: ; ADDA		0xDCD8-0xDCDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCE0: ; ADDA		0xDCE0-0xDCE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCE8: ; ADDA		0xDCE8-0xDCEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCF0: ; ADDA		0xDCF0-0xDCF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCF8: ; ADDA		0xDCF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCF9: ; ADDA		0xDCF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCFA: ; ADDA		0xDCFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCFB: ; ADDA		0xDCFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDCFC: ; ADDA		0xDCFC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEC0: ; ADDA		0xDEC0-0xDEC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
add  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEC8: ; ADDA		0xDEC8-0xDECF
and  edi, byte 7
movsx    edx, word [a+edi*4]
add  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDED0: ; ADDA		0xDED0-0xDED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDED8: ; ADDA		0xDED8-0xDEDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEE0: ; ADDA		0xDEE0-0xDEE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEE8: ; ADDA		0xDEE8-0xDEEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEF0: ; ADDA		0xDEF0-0xDEF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEF8: ; ADDA		0xDEF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEF9: ; ADDA		0xDEF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEFA: ; ADDA		0xDEFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEFB: ; ADDA		0xDEFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
add  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDEFC: ; ADDA		0xDEFC
movsx    edx, word [esi]
add  esi, byte 2
add  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1C0: ; ADDA		0xD1C0-0xD1C7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1C8: ; ADDA		0xD1C8-0xD1CF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1D0: ; ADDA		0xD1D0-0xD1D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1D8: ; ADDA		0xD1D8-0xD1DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1E0: ; ADDA		0xD1E0-0xD1E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1E8: ; ADDA		0xD1E8-0xD1EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1F0: ; ADDA		0xD1F0-0xD1F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1F8: ; ADDA		0xD1F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1F9: ; ADDA		0xD1F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1FA: ; ADDA		0xD1FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1FB: ; ADDA		0xD1FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID1FC: ; ADDA		0xD1FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3C0: ; ADDA		0xD3C0-0xD3C7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3C8: ; ADDA		0xD3C8-0xD3CF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3D0: ; ADDA		0xD3D0-0xD3D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3D8: ; ADDA		0xD3D8-0xD3DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3E0: ; ADDA		0xD3E0-0xD3E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3E8: ; ADDA		0xD3E8-0xD3EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3F0: ; ADDA		0xD3F0-0xD3F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3F8: ; ADDA		0xD3F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3F9: ; ADDA		0xD3F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3FA: ; ADDA		0xD3FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3FB: ; ADDA		0xD3FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID3FC: ; ADDA		0xD3FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5C0: ; ADDA		0xD5C0-0xD5C7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5C8: ; ADDA		0xD5C8-0xD5CF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5D0: ; ADDA		0xD5D0-0xD5D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5D8: ; ADDA		0xD5D8-0xD5DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5E0: ; ADDA		0xD5E0-0xD5E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5E8: ; ADDA		0xD5E8-0xD5EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5F0: ; ADDA		0xD5F0-0xD5F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5F8: ; ADDA		0xD5F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5F9: ; ADDA		0xD5F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5FA: ; ADDA		0xD5FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5FB: ; ADDA		0xD5FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID5FC: ; ADDA		0xD5FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7C0: ; ADDA		0xD7C0-0xD7C7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7C8: ; ADDA		0xD7C8-0xD7CF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7D0: ; ADDA		0xD7D0-0xD7D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7D8: ; ADDA		0xD7D8-0xD7DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7E0: ; ADDA		0xD7E0-0xD7E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7E8: ; ADDA		0xD7E8-0xD7EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7F0: ; ADDA		0xD7F0-0xD7F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7F8: ; ADDA		0xD7F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7F9: ; ADDA		0xD7F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7FA: ; ADDA		0xD7FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7FB: ; ADDA		0xD7FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID7FC: ; ADDA		0xD7FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9C0: ; ADDA		0xD9C0-0xD9C7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9C8: ; ADDA		0xD9C8-0xD9CF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9D0: ; ADDA		0xD9D0-0xD9D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9D8: ; ADDA		0xD9D8-0xD9DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9E0: ; ADDA		0xD9E0-0xD9E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9E8: ; ADDA		0xD9E8-0xD9EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9F0: ; ADDA		0xD9F0-0xD9F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9F8: ; ADDA		0xD9F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9F9: ; ADDA		0xD9F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9FA: ; ADDA		0xD9FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9FB: ; ADDA		0xD9FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID9FC: ; ADDA		0xD9FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBC0: ; ADDA		0xDBC0-0xDBC7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBC8: ; ADDA		0xDBC8-0xDBCF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBD0: ; ADDA		0xDBD0-0xDBD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBD8: ; ADDA		0xDBD8-0xDBDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBE0: ; ADDA		0xDBE0-0xDBE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBE8: ; ADDA		0xDBE8-0xDBEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBF0: ; ADDA		0xDBF0-0xDBF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBF8: ; ADDA		0xDBF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBF9: ; ADDA		0xDBF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBFA: ; ADDA		0xDBFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBFB: ; ADDA		0xDBFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDBFC: ; ADDA		0xDBFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDC0: ; ADDA		0xDDC0-0xDDC7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDC8: ; ADDA		0xDDC8-0xDDCF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDD0: ; ADDA		0xDDD0-0xDDD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDD8: ; ADDA		0xDDD8-0xDDDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDE0: ; ADDA		0xDDE0-0xDDE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDE8: ; ADDA		0xDDE8-0xDDEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDF0: ; ADDA		0xDDF0-0xDDF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDF8: ; ADDA		0xDDF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDF9: ; ADDA		0xDDF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDFA: ; ADDA		0xDDFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDFB: ; ADDA		0xDDFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDDFC: ; ADDA		0xDDFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFC0: ; ADDA		0xDFC0-0xDFC7
and  edi, byte 7
mov  edx, [d+edi*4]
add  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFC8: ; ADDA		0xDFC8-0xDFCF
and  edi, byte 7
mov  edx, [a+edi*4]
add  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFD0: ; ADDA		0xDFD0-0xDFD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFD8: ; ADDA		0xDFD8-0xDFDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFE0: ; ADDA		0xDFE0-0xDFE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFE8: ; ADDA		0xDFE8-0xDFEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFF0: ; ADDA		0xDFF0-0xDFF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFF8: ; ADDA		0xDFF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFF9: ; ADDA		0xDFF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFFA: ; ADDA		0xDFFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFFB: ; ADDA		0xDFFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
add  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDFFC: ; ADDA		0xDFFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
add  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90C0: ; SUBA		0x90C0-0x90C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90C8: ; SUBA		0x90C8-0x90CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90D0: ; SUBA		0x90D0-0x90D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90D8: ; SUBA		0x90D8-0x90DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90E0: ; SUBA		0x90E0-0x90E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90E8: ; SUBA		0x90E8-0x90EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90F0: ; SUBA		0x90F0-0x90F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90F8: ; SUBA		0x90F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90F9: ; SUBA		0x90F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90FA: ; SUBA		0x90FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90FB: ; SUBA		0x90FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I90FC: ; SUBA		0x90FC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92C0: ; SUBA		0x92C0-0x92C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92C8: ; SUBA		0x92C8-0x92CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92D0: ; SUBA		0x92D0-0x92D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92D8: ; SUBA		0x92D8-0x92DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92E0: ; SUBA		0x92E0-0x92E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92E8: ; SUBA		0x92E8-0x92EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92F0: ; SUBA		0x92F0-0x92F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92F8: ; SUBA		0x92F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92F9: ; SUBA		0x92F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92FA: ; SUBA		0x92FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92FB: ; SUBA		0x92FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I92FC: ; SUBA		0x92FC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94C0: ; SUBA		0x94C0-0x94C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94C8: ; SUBA		0x94C8-0x94CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94D0: ; SUBA		0x94D0-0x94D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94D8: ; SUBA		0x94D8-0x94DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94E0: ; SUBA		0x94E0-0x94E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94E8: ; SUBA		0x94E8-0x94EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94F0: ; SUBA		0x94F0-0x94F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94F8: ; SUBA		0x94F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94F9: ; SUBA		0x94F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94FA: ; SUBA		0x94FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94FB: ; SUBA		0x94FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I94FC: ; SUBA		0x94FC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96C0: ; SUBA		0x96C0-0x96C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96C8: ; SUBA		0x96C8-0x96CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96D0: ; SUBA		0x96D0-0x96D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96D8: ; SUBA		0x96D8-0x96DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96E0: ; SUBA		0x96E0-0x96E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96E8: ; SUBA		0x96E8-0x96EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96F0: ; SUBA		0x96F0-0x96F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96F8: ; SUBA		0x96F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96F9: ; SUBA		0x96F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96FA: ; SUBA		0x96FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96FB: ; SUBA		0x96FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I96FC: ; SUBA		0x96FC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98C0: ; SUBA		0x98C0-0x98C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98C8: ; SUBA		0x98C8-0x98CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98D0: ; SUBA		0x98D0-0x98D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98D8: ; SUBA		0x98D8-0x98DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98E0: ; SUBA		0x98E0-0x98E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98E8: ; SUBA		0x98E8-0x98EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98F0: ; SUBA		0x98F0-0x98F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98F8: ; SUBA		0x98F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98F9: ; SUBA		0x98F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98FA: ; SUBA		0x98FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98FB: ; SUBA		0x98FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I98FC: ; SUBA		0x98FC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AC0: ; SUBA		0x9AC0-0x9AC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AC8: ; SUBA		0x9AC8-0x9ACF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AD0: ; SUBA		0x9AD0-0x9AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AD8: ; SUBA		0x9AD8-0x9ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AE0: ; SUBA		0x9AE0-0x9AE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AE8: ; SUBA		0x9AE8-0x9AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AF0: ; SUBA		0x9AF0-0x9AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AF8: ; SUBA		0x9AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AF9: ; SUBA		0x9AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AFA: ; SUBA		0x9AFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AFB: ; SUBA		0x9AFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9AFC: ; SUBA		0x9AFC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CC0: ; SUBA		0x9CC0-0x9CC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CC8: ; SUBA		0x9CC8-0x9CCF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CD0: ; SUBA		0x9CD0-0x9CD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CD8: ; SUBA		0x9CD8-0x9CDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CE0: ; SUBA		0x9CE0-0x9CE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CE8: ; SUBA		0x9CE8-0x9CEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CF0: ; SUBA		0x9CF0-0x9CF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CF8: ; SUBA		0x9CF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CF9: ; SUBA		0x9CF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CFA: ; SUBA		0x9CFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CFB: ; SUBA		0x9CFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9CFC: ; SUBA		0x9CFC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EC0: ; SUBA		0x9EC0-0x9EC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
sub  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EC8: ; SUBA		0x9EC8-0x9ECF
and  edi, byte 7
movsx    edx, word [a+edi*4]
sub  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9ED0: ; SUBA		0x9ED0-0x9ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9ED8: ; SUBA		0x9ED8-0x9EDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EE0: ; SUBA		0x9EE0-0x9EE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EE8: ; SUBA		0x9EE8-0x9EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EF0: ; SUBA		0x9EF0-0x9EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EF8: ; SUBA		0x9EF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EF9: ; SUBA		0x9EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EFA: ; SUBA		0x9EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EFB: ; SUBA		0x9EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
sub  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9EFC: ; SUBA		0x9EFC
movsx    edx, word [esi]
add  esi, byte 2
sub  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91C0: ; SUBA		0x91C0-0x91C7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91C8: ; SUBA		0x91C8-0x91CF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91D0: ; SUBA		0x91D0-0x91D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91D8: ; SUBA		0x91D8-0x91DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91E0: ; SUBA		0x91E0-0x91E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91E8: ; SUBA		0x91E8-0x91EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91F0: ; SUBA		0x91F0-0x91F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91F8: ; SUBA		0x91F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91F9: ; SUBA		0x91F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91FA: ; SUBA		0x91FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91FB: ; SUBA		0x91FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I91FC: ; SUBA		0x91FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93C0: ; SUBA		0x93C0-0x93C7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93C8: ; SUBA		0x93C8-0x93CF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93D0: ; SUBA		0x93D0-0x93D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93D8: ; SUBA		0x93D8-0x93DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93E0: ; SUBA		0x93E0-0x93E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93E8: ; SUBA		0x93E8-0x93EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93F0: ; SUBA		0x93F0-0x93F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93F8: ; SUBA		0x93F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93F9: ; SUBA		0x93F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93FA: ; SUBA		0x93FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93FB: ; SUBA		0x93FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I93FC: ; SUBA		0x93FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95C0: ; SUBA		0x95C0-0x95C7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95C8: ; SUBA		0x95C8-0x95CF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95D0: ; SUBA		0x95D0-0x95D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95D8: ; SUBA		0x95D8-0x95DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95E0: ; SUBA		0x95E0-0x95E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95E8: ; SUBA		0x95E8-0x95EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95F0: ; SUBA		0x95F0-0x95F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95F8: ; SUBA		0x95F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95F9: ; SUBA		0x95F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95FA: ; SUBA		0x95FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95FB: ; SUBA		0x95FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I95FC: ; SUBA		0x95FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97C0: ; SUBA		0x97C0-0x97C7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97C8: ; SUBA		0x97C8-0x97CF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97D0: ; SUBA		0x97D0-0x97D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97D8: ; SUBA		0x97D8-0x97DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97E0: ; SUBA		0x97E0-0x97E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97E8: ; SUBA		0x97E8-0x97EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97F0: ; SUBA		0x97F0-0x97F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97F8: ; SUBA		0x97F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97F9: ; SUBA		0x97F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97FA: ; SUBA		0x97FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97FB: ; SUBA		0x97FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I97FC: ; SUBA		0x97FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99C0: ; SUBA		0x99C0-0x99C7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99C8: ; SUBA		0x99C8-0x99CF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99D0: ; SUBA		0x99D0-0x99D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99D8: ; SUBA		0x99D8-0x99DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99E0: ; SUBA		0x99E0-0x99E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99E8: ; SUBA		0x99E8-0x99EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99F0: ; SUBA		0x99F0-0x99F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99F8: ; SUBA		0x99F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99F9: ; SUBA		0x99F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99FA: ; SUBA		0x99FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99FB: ; SUBA		0x99FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I99FC: ; SUBA		0x99FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BC0: ; SUBA		0x9BC0-0x9BC7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BC8: ; SUBA		0x9BC8-0x9BCF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BD0: ; SUBA		0x9BD0-0x9BD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BD8: ; SUBA		0x9BD8-0x9BDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BE0: ; SUBA		0x9BE0-0x9BE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BE8: ; SUBA		0x9BE8-0x9BEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BF0: ; SUBA		0x9BF0-0x9BF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BF8: ; SUBA		0x9BF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BF9: ; SUBA		0x9BF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BFA: ; SUBA		0x9BFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BFB: ; SUBA		0x9BFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9BFC: ; SUBA		0x9BFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DC0: ; SUBA		0x9DC0-0x9DC7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DC8: ; SUBA		0x9DC8-0x9DCF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DD0: ; SUBA		0x9DD0-0x9DD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DD8: ; SUBA		0x9DD8-0x9DDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DE0: ; SUBA		0x9DE0-0x9DE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DE8: ; SUBA		0x9DE8-0x9DEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DF0: ; SUBA		0x9DF0-0x9DF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DF8: ; SUBA		0x9DF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DF9: ; SUBA		0x9DF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DFA: ; SUBA		0x9DFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DFB: ; SUBA		0x9DFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9DFC: ; SUBA		0x9DFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FC0: ; SUBA		0x9FC0-0x9FC7
and  edi, byte 7
mov  edx, [d+edi*4]
sub  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FC8: ; SUBA		0x9FC8-0x9FCF
and  edi, byte 7
mov  edx, [a+edi*4]
sub  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FD0: ; SUBA		0x9FD0-0x9FD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FD8: ; SUBA		0x9FD8-0x9FDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FE0: ; SUBA		0x9FE0-0x9FE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FE8: ; SUBA		0x9FE8-0x9FEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FF0: ; SUBA		0x9FF0-0x9FF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FF8: ; SUBA		0x9FF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FF9: ; SUBA		0x9FF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FFA: ; SUBA		0x9FFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FFB: ; SUBA		0x9FFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
sub  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9FFC: ; SUBA		0x9FFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
sub  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5000: ; ADDQ		0x5000-0x5007
and  edi, byte 7
add   byte [d+edi*4], 8
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5010: ; ADDQ		0x5010-0x5017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5018: ; ADDQ		0x5018-0x501F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5020: ; ADDQ		0x5020-0x5027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5028: ; ADDQ		0x5028-0x502F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5030: ; ADDQ		0x5030-0x5037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5038: ; ADDQ		0x5038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5039: ; ADDQ		0x5039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5040: ; ADDQ		0x5040-0x5047
and  edi, byte 7
add   word [d+edi*4], 8
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5048: ; ADDQ		0x5048-0x504F
and  edi, byte 7
add   dword [a+edi*4], byte 8
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5050: ; ADDQ		0x5050-0x5057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5058: ; ADDQ		0x5058-0x505F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5060: ; ADDQ		0x5060-0x5067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5068: ; ADDQ		0x5068-0x506F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5070: ; ADDQ		0x5070-0x5077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5078: ; ADDQ		0x5078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5079: ; ADDQ		0x5079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5080: ; ADDQ		0x5080-0x5087
and  edi, byte 7
add   dword [d+edi*4], byte 8
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5088: ; ADDQ		0x5088-0x508F
and  edi, byte 7
add   dword [a+edi*4], byte 8
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5090: ; ADDQ		0x5090-0x5097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5098: ; ADDQ		0x5098-0x509F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50A0: ; ADDQ		0x50A0-0x50A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50A8: ; ADDQ		0x50A8-0x50AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50B0: ; ADDQ		0x50B0-0x50B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50B8: ; ADDQ		0x50B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50B9: ; ADDQ		0x50B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5100: ; SUBQ		0x5100-0x5107
and  edi, byte 7
sub   byte [d+edi*4], 8
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5110: ; SUBQ		0x5110-0x5117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5118: ; SUBQ		0x5118-0x511F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5120: ; SUBQ		0x5120-0x5127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5128: ; SUBQ		0x5128-0x512F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5130: ; SUBQ		0x5130-0x5137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5138: ; SUBQ		0x5138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5139: ; SUBQ		0x5139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 8
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5140: ; SUBQ		0x5140-0x5147
and  edi, byte 7
sub   word [d+edi*4], 8
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5148: ; SUBQ		0x5148-0x514F
and  edi, byte 7
sub   dword [a+edi*4], byte 8
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5150: ; SUBQ		0x5150-0x5157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5158: ; SUBQ		0x5158-0x515F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5160: ; SUBQ		0x5160-0x5167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5168: ; SUBQ		0x5168-0x516F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5170: ; SUBQ		0x5170-0x5177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5178: ; SUBQ		0x5178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5179: ; SUBQ		0x5179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 8
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5180: ; SUBQ		0x5180-0x5187
and  edi, byte 7
sub   dword [d+edi*4], byte 8
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5188: ; SUBQ		0x5188-0x518F
and  edi, byte 7
sub   dword [a+edi*4], byte 8
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5190: ; SUBQ		0x5190-0x5197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5198: ; SUBQ		0x5198-0x519F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51A0: ; SUBQ		0x51A0-0x51A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51A8: ; SUBQ		0x51A8-0x51AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51B0: ; SUBQ		0x51B0-0x51B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51B8: ; SUBQ		0x51B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51B9: ; SUBQ		0x51B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 8
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5200: ; ADDQ		0x5200-0x5207
and  edi, byte 7
add   byte [d+edi*4], 1
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5210: ; ADDQ		0x5210-0x5217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5218: ; ADDQ		0x5218-0x521F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5220: ; ADDQ		0x5220-0x5227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5228: ; ADDQ		0x5228-0x522F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5230: ; ADDQ		0x5230-0x5237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5238: ; ADDQ		0x5238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5239: ; ADDQ		0x5239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5240: ; ADDQ		0x5240-0x5247
and  edi, byte 7
add   word [d+edi*4], 1
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5248: ; ADDQ		0x5248-0x524F
and  edi, byte 7
add   dword [a+edi*4], byte 1
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5250: ; ADDQ		0x5250-0x5257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5258: ; ADDQ		0x5258-0x525F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5260: ; ADDQ		0x5260-0x5267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5268: ; ADDQ		0x5268-0x526F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5270: ; ADDQ		0x5270-0x5277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5278: ; ADDQ		0x5278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5279: ; ADDQ		0x5279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5280: ; ADDQ		0x5280-0x5287
and  edi, byte 7
add   dword [d+edi*4], byte 1
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5288: ; ADDQ		0x5288-0x528F
and  edi, byte 7
add   dword [a+edi*4], byte 1
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5290: ; ADDQ		0x5290-0x5297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5298: ; ADDQ		0x5298-0x529F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52A0: ; ADDQ		0x52A0-0x52A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52A8: ; ADDQ		0x52A8-0x52AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52B0: ; ADDQ		0x52B0-0x52B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52B8: ; ADDQ		0x52B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52B9: ; ADDQ		0x52B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5300: ; SUBQ		0x5300-0x5307
and  edi, byte 7
sub   byte [d+edi*4], 1
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5310: ; SUBQ		0x5310-0x5317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5318: ; SUBQ		0x5318-0x531F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5320: ; SUBQ		0x5320-0x5327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5328: ; SUBQ		0x5328-0x532F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5330: ; SUBQ		0x5330-0x5337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5338: ; SUBQ		0x5338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5339: ; SUBQ		0x5339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 1
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5340: ; SUBQ		0x5340-0x5347
and  edi, byte 7
sub   word [d+edi*4], 1
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5348: ; SUBQ		0x5348-0x534F
and  edi, byte 7
sub   dword [a+edi*4], byte 1
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5350: ; SUBQ		0x5350-0x5357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5358: ; SUBQ		0x5358-0x535F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5360: ; SUBQ		0x5360-0x5367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5368: ; SUBQ		0x5368-0x536F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5370: ; SUBQ		0x5370-0x5377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5378: ; SUBQ		0x5378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5379: ; SUBQ		0x5379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 1
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5380: ; SUBQ		0x5380-0x5387
and  edi, byte 7
sub   dword [d+edi*4], byte 1
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5388: ; SUBQ		0x5388-0x538F
and  edi, byte 7
sub   dword [a+edi*4], byte 1
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5390: ; SUBQ		0x5390-0x5397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5398: ; SUBQ		0x5398-0x539F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53A0: ; SUBQ		0x53A0-0x53A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53A8: ; SUBQ		0x53A8-0x53AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53B0: ; SUBQ		0x53B0-0x53B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53B8: ; SUBQ		0x53B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53B9: ; SUBQ		0x53B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 1
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5400: ; ADDQ		0x5400-0x5407
and  edi, byte 7
add   byte [d+edi*4], 2
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5410: ; ADDQ		0x5410-0x5417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5418: ; ADDQ		0x5418-0x541F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5420: ; ADDQ		0x5420-0x5427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5428: ; ADDQ		0x5428-0x542F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5430: ; ADDQ		0x5430-0x5437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5438: ; ADDQ		0x5438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5439: ; ADDQ		0x5439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5440: ; ADDQ		0x5440-0x5447
and  edi, byte 7
add   word [d+edi*4], 2
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5448: ; ADDQ		0x5448-0x544F
and  edi, byte 7
add   dword [a+edi*4], byte 2
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5450: ; ADDQ		0x5450-0x5457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5458: ; ADDQ		0x5458-0x545F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5460: ; ADDQ		0x5460-0x5467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5468: ; ADDQ		0x5468-0x546F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5470: ; ADDQ		0x5470-0x5477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5478: ; ADDQ		0x5478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5479: ; ADDQ		0x5479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5480: ; ADDQ		0x5480-0x5487
and  edi, byte 7
add   dword [d+edi*4], byte 2
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5488: ; ADDQ		0x5488-0x548F
and  edi, byte 7
add   dword [a+edi*4], byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5490: ; ADDQ		0x5490-0x5497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5498: ; ADDQ		0x5498-0x549F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54A0: ; ADDQ		0x54A0-0x54A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54A8: ; ADDQ		0x54A8-0x54AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54B0: ; ADDQ		0x54B0-0x54B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54B8: ; ADDQ		0x54B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54B9: ; ADDQ		0x54B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5500: ; SUBQ		0x5500-0x5507
and  edi, byte 7
sub   byte [d+edi*4], 2
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5510: ; SUBQ		0x5510-0x5517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5518: ; SUBQ		0x5518-0x551F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5520: ; SUBQ		0x5520-0x5527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5528: ; SUBQ		0x5528-0x552F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5530: ; SUBQ		0x5530-0x5537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5538: ; SUBQ		0x5538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5539: ; SUBQ		0x5539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 2
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5540: ; SUBQ		0x5540-0x5547
and  edi, byte 7
sub   word [d+edi*4], 2
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5548: ; SUBQ		0x5548-0x554F
and  edi, byte 7
sub   dword [a+edi*4], byte 2
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5550: ; SUBQ		0x5550-0x5557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5558: ; SUBQ		0x5558-0x555F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5560: ; SUBQ		0x5560-0x5567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5568: ; SUBQ		0x5568-0x556F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5570: ; SUBQ		0x5570-0x5577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5578: ; SUBQ		0x5578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5579: ; SUBQ		0x5579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 2
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5580: ; SUBQ		0x5580-0x5587
and  edi, byte 7
sub   dword [d+edi*4], byte 2
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5588: ; SUBQ		0x5588-0x558F
and  edi, byte 7
sub   dword [a+edi*4], byte 2
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5590: ; SUBQ		0x5590-0x5597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5598: ; SUBQ		0x5598-0x559F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55A0: ; SUBQ		0x55A0-0x55A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55A8: ; SUBQ		0x55A8-0x55AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55B0: ; SUBQ		0x55B0-0x55B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55B8: ; SUBQ		0x55B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55B9: ; SUBQ		0x55B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 2
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5600: ; ADDQ		0x5600-0x5607
and  edi, byte 7
add   byte [d+edi*4], 3
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5610: ; ADDQ		0x5610-0x5617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5618: ; ADDQ		0x5618-0x561F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5620: ; ADDQ		0x5620-0x5627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5628: ; ADDQ		0x5628-0x562F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5630: ; ADDQ		0x5630-0x5637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5638: ; ADDQ		0x5638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5639: ; ADDQ		0x5639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5640: ; ADDQ		0x5640-0x5647
and  edi, byte 7
add   word [d+edi*4], 3
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5648: ; ADDQ		0x5648-0x564F
and  edi, byte 7
add   dword [a+edi*4], byte 3
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5650: ; ADDQ		0x5650-0x5657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5658: ; ADDQ		0x5658-0x565F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5660: ; ADDQ		0x5660-0x5667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5668: ; ADDQ		0x5668-0x566F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5670: ; ADDQ		0x5670-0x5677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5678: ; ADDQ		0x5678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5679: ; ADDQ		0x5679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5680: ; ADDQ		0x5680-0x5687
and  edi, byte 7
add   dword [d+edi*4], byte 3
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5688: ; ADDQ		0x5688-0x568F
and  edi, byte 7
add   dword [a+edi*4], byte 3
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5690: ; ADDQ		0x5690-0x5697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5698: ; ADDQ		0x5698-0x569F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56A0: ; ADDQ		0x56A0-0x56A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56A8: ; ADDQ		0x56A8-0x56AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56B0: ; ADDQ		0x56B0-0x56B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56B8: ; ADDQ		0x56B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56B9: ; ADDQ		0x56B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5700: ; SUBQ		0x5700-0x5707
and  edi, byte 7
sub   byte [d+edi*4], 3
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5710: ; SUBQ		0x5710-0x5717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5718: ; SUBQ		0x5718-0x571F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5720: ; SUBQ		0x5720-0x5727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5728: ; SUBQ		0x5728-0x572F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5730: ; SUBQ		0x5730-0x5737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5738: ; SUBQ		0x5738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5739: ; SUBQ		0x5739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 3
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5740: ; SUBQ		0x5740-0x5747
and  edi, byte 7
sub   word [d+edi*4], 3
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5748: ; SUBQ		0x5748-0x574F
and  edi, byte 7
sub   dword [a+edi*4], byte 3
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5750: ; SUBQ		0x5750-0x5757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5758: ; SUBQ		0x5758-0x575F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5760: ; SUBQ		0x5760-0x5767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5768: ; SUBQ		0x5768-0x576F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5770: ; SUBQ		0x5770-0x5777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5778: ; SUBQ		0x5778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5779: ; SUBQ		0x5779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 3
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5780: ; SUBQ		0x5780-0x5787
and  edi, byte 7
sub   dword [d+edi*4], byte 3
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5788: ; SUBQ		0x5788-0x578F
and  edi, byte 7
sub   dword [a+edi*4], byte 3
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5790: ; SUBQ		0x5790-0x5797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5798: ; SUBQ		0x5798-0x579F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57A0: ; SUBQ		0x57A0-0x57A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57A8: ; SUBQ		0x57A8-0x57AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57B0: ; SUBQ		0x57B0-0x57B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57B8: ; SUBQ		0x57B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57B9: ; SUBQ		0x57B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 3
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5800: ; ADDQ		0x5800-0x5807
and  edi, byte 7
add   byte [d+edi*4], 4
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5810: ; ADDQ		0x5810-0x5817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5818: ; ADDQ		0x5818-0x581F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5820: ; ADDQ		0x5820-0x5827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5828: ; ADDQ		0x5828-0x582F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5830: ; ADDQ		0x5830-0x5837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5838: ; ADDQ		0x5838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5839: ; ADDQ		0x5839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5840: ; ADDQ		0x5840-0x5847
and  edi, byte 7
add   word [d+edi*4], 4
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5848: ; ADDQ		0x5848-0x584F
and  edi, byte 7
add   dword [a+edi*4], byte 4
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5850: ; ADDQ		0x5850-0x5857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5858: ; ADDQ		0x5858-0x585F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5860: ; ADDQ		0x5860-0x5867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5868: ; ADDQ		0x5868-0x586F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5870: ; ADDQ		0x5870-0x5877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5878: ; ADDQ		0x5878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5879: ; ADDQ		0x5879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5880: ; ADDQ		0x5880-0x5887
and  edi, byte 7
add   dword [d+edi*4], byte 4
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5888: ; ADDQ		0x5888-0x588F
and  edi, byte 7
add   dword [a+edi*4], byte 4
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5890: ; ADDQ		0x5890-0x5897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5898: ; ADDQ		0x5898-0x589F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58A0: ; ADDQ		0x58A0-0x58A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58A8: ; ADDQ		0x58A8-0x58AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58B0: ; ADDQ		0x58B0-0x58B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58B8: ; ADDQ		0x58B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58B9: ; ADDQ		0x58B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5900: ; SUBQ		0x5900-0x5907
and  edi, byte 7
sub   byte [d+edi*4], 4
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5910: ; SUBQ		0x5910-0x5917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5918: ; SUBQ		0x5918-0x591F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5920: ; SUBQ		0x5920-0x5927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5928: ; SUBQ		0x5928-0x592F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5930: ; SUBQ		0x5930-0x5937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5938: ; SUBQ		0x5938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5939: ; SUBQ		0x5939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 4
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5940: ; SUBQ		0x5940-0x5947
and  edi, byte 7
sub   word [d+edi*4], 4
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5948: ; SUBQ		0x5948-0x594F
and  edi, byte 7
sub   dword [a+edi*4], byte 4
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5950: ; SUBQ		0x5950-0x5957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5958: ; SUBQ		0x5958-0x595F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5960: ; SUBQ		0x5960-0x5967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5968: ; SUBQ		0x5968-0x596F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5970: ; SUBQ		0x5970-0x5977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5978: ; SUBQ		0x5978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5979: ; SUBQ		0x5979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 4
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5980: ; SUBQ		0x5980-0x5987
and  edi, byte 7
sub   dword [d+edi*4], byte 4
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5988: ; SUBQ		0x5988-0x598F
and  edi, byte 7
sub   dword [a+edi*4], byte 4
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5990: ; SUBQ		0x5990-0x5997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5998: ; SUBQ		0x5998-0x599F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59A0: ; SUBQ		0x59A0-0x59A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59A8: ; SUBQ		0x59A8-0x59AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59B0: ; SUBQ		0x59B0-0x59B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59B8: ; SUBQ		0x59B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59B9: ; SUBQ		0x59B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 4
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A00: ; ADDQ		0x5A00-0x5A07
and  edi, byte 7
add   byte [d+edi*4], 5
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A10: ; ADDQ		0x5A10-0x5A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A18: ; ADDQ		0x5A18-0x5A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A20: ; ADDQ		0x5A20-0x5A27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A28: ; ADDQ		0x5A28-0x5A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A30: ; ADDQ		0x5A30-0x5A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A38: ; ADDQ		0x5A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A39: ; ADDQ		0x5A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A40: ; ADDQ		0x5A40-0x5A47
and  edi, byte 7
add   word [d+edi*4], 5
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A48: ; ADDQ		0x5A48-0x5A4F
and  edi, byte 7
add   dword [a+edi*4], byte 5
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A50: ; ADDQ		0x5A50-0x5A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A58: ; ADDQ		0x5A58-0x5A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A60: ; ADDQ		0x5A60-0x5A67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A68: ; ADDQ		0x5A68-0x5A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A70: ; ADDQ		0x5A70-0x5A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A78: ; ADDQ		0x5A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A79: ; ADDQ		0x5A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A80: ; ADDQ		0x5A80-0x5A87
and  edi, byte 7
add   dword [d+edi*4], byte 5
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A88: ; ADDQ		0x5A88-0x5A8F
and  edi, byte 7
add   dword [a+edi*4], byte 5
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A90: ; ADDQ		0x5A90-0x5A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5A98: ; ADDQ		0x5A98-0x5A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AA0: ; ADDQ		0x5AA0-0x5AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AA8: ; ADDQ		0x5AA8-0x5AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AB0: ; ADDQ		0x5AB0-0x5AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AB8: ; ADDQ		0x5AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AB9: ; ADDQ		0x5AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B00: ; SUBQ		0x5B00-0x5B07
and  edi, byte 7
sub   byte [d+edi*4], 5
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B10: ; SUBQ		0x5B10-0x5B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B18: ; SUBQ		0x5B18-0x5B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B20: ; SUBQ		0x5B20-0x5B27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B28: ; SUBQ		0x5B28-0x5B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B30: ; SUBQ		0x5B30-0x5B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B38: ; SUBQ		0x5B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B39: ; SUBQ		0x5B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 5
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B40: ; SUBQ		0x5B40-0x5B47
and  edi, byte 7
sub   word [d+edi*4], 5
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B48: ; SUBQ		0x5B48-0x5B4F
and  edi, byte 7
sub   dword [a+edi*4], byte 5
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B50: ; SUBQ		0x5B50-0x5B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B58: ; SUBQ		0x5B58-0x5B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B60: ; SUBQ		0x5B60-0x5B67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B68: ; SUBQ		0x5B68-0x5B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B70: ; SUBQ		0x5B70-0x5B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B78: ; SUBQ		0x5B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B79: ; SUBQ		0x5B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 5
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B80: ; SUBQ		0x5B80-0x5B87
and  edi, byte 7
sub   dword [d+edi*4], byte 5
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B88: ; SUBQ		0x5B88-0x5B8F
and  edi, byte 7
sub   dword [a+edi*4], byte 5
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B90: ; SUBQ		0x5B90-0x5B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5B98: ; SUBQ		0x5B98-0x5B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BA0: ; SUBQ		0x5BA0-0x5BA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BA8: ; SUBQ		0x5BA8-0x5BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BB0: ; SUBQ		0x5BB0-0x5BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BB8: ; SUBQ		0x5BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BB9: ; SUBQ		0x5BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 5
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C00: ; ADDQ		0x5C00-0x5C07
and  edi, byte 7
add   byte [d+edi*4], 6
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C10: ; ADDQ		0x5C10-0x5C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C18: ; ADDQ		0x5C18-0x5C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C20: ; ADDQ		0x5C20-0x5C27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C28: ; ADDQ		0x5C28-0x5C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C30: ; ADDQ		0x5C30-0x5C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C38: ; ADDQ		0x5C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C39: ; ADDQ		0x5C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C40: ; ADDQ		0x5C40-0x5C47
and  edi, byte 7
add   word [d+edi*4], 6
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C48: ; ADDQ		0x5C48-0x5C4F
and  edi, byte 7
add   dword [a+edi*4], byte 6
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C50: ; ADDQ		0x5C50-0x5C57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C58: ; ADDQ		0x5C58-0x5C5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C60: ; ADDQ		0x5C60-0x5C67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C68: ; ADDQ		0x5C68-0x5C6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C70: ; ADDQ		0x5C70-0x5C77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C78: ; ADDQ		0x5C78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C79: ; ADDQ		0x5C79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C80: ; ADDQ		0x5C80-0x5C87
and  edi, byte 7
add   dword [d+edi*4], byte 6
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C88: ; ADDQ		0x5C88-0x5C8F
and  edi, byte 7
add   dword [a+edi*4], byte 6
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C90: ; ADDQ		0x5C90-0x5C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5C98: ; ADDQ		0x5C98-0x5C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CA0: ; ADDQ		0x5CA0-0x5CA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CA8: ; ADDQ		0x5CA8-0x5CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CB0: ; ADDQ		0x5CB0-0x5CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CB8: ; ADDQ		0x5CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CB9: ; ADDQ		0x5CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D00: ; SUBQ		0x5D00-0x5D07
and  edi, byte 7
sub   byte [d+edi*4], 6
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D10: ; SUBQ		0x5D10-0x5D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D18: ; SUBQ		0x5D18-0x5D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D20: ; SUBQ		0x5D20-0x5D27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D28: ; SUBQ		0x5D28-0x5D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D30: ; SUBQ		0x5D30-0x5D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D38: ; SUBQ		0x5D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D39: ; SUBQ		0x5D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 6
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D40: ; SUBQ		0x5D40-0x5D47
and  edi, byte 7
sub   word [d+edi*4], 6
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D48: ; SUBQ		0x5D48-0x5D4F
and  edi, byte 7
sub   dword [a+edi*4], byte 6
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D50: ; SUBQ		0x5D50-0x5D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D58: ; SUBQ		0x5D58-0x5D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D60: ; SUBQ		0x5D60-0x5D67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D68: ; SUBQ		0x5D68-0x5D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D70: ; SUBQ		0x5D70-0x5D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D78: ; SUBQ		0x5D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D79: ; SUBQ		0x5D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 6
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D80: ; SUBQ		0x5D80-0x5D87
and  edi, byte 7
sub   dword [d+edi*4], byte 6
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D88: ; SUBQ		0x5D88-0x5D8F
and  edi, byte 7
sub   dword [a+edi*4], byte 6
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D90: ; SUBQ		0x5D90-0x5D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5D98: ; SUBQ		0x5D98-0x5D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DA0: ; SUBQ		0x5DA0-0x5DA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DA8: ; SUBQ		0x5DA8-0x5DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DB0: ; SUBQ		0x5DB0-0x5DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DB8: ; SUBQ		0x5DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DB9: ; SUBQ		0x5DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 6
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E00: ; ADDQ		0x5E00-0x5E07
and  edi, byte 7
add   byte [d+edi*4], 7
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E10: ; ADDQ		0x5E10-0x5E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E18: ; ADDQ		0x5E18-0x5E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E20: ; ADDQ		0x5E20-0x5E27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E28: ; ADDQ		0x5E28-0x5E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E30: ; ADDQ		0x5E30-0x5E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E38: ; ADDQ		0x5E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E39: ; ADDQ		0x5E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
add   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E40: ; ADDQ		0x5E40-0x5E47
and  edi, byte 7
add   word [d+edi*4], 7
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E48: ; ADDQ		0x5E48-0x5E4F
and  edi, byte 7
add   dword [a+edi*4], byte 7
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E50: ; ADDQ		0x5E50-0x5E57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E58: ; ADDQ		0x5E58-0x5E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E60: ; ADDQ		0x5E60-0x5E67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E68: ; ADDQ		0x5E68-0x5E6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E70: ; ADDQ		0x5E70-0x5E77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E78: ; ADDQ		0x5E78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E79: ; ADDQ		0x5E79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
add   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E80: ; ADDQ		0x5E80-0x5E87
and  edi, byte 7
add   dword [d+edi*4], byte 7
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E88: ; ADDQ		0x5E88-0x5E8F
and  edi, byte 7
add   dword [a+edi*4], byte 7
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E90: ; ADDQ		0x5E90-0x5E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5E98: ; ADDQ		0x5E98-0x5E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EA0: ; ADDQ		0x5EA0-0x5EA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EA8: ; ADDQ		0x5EA8-0x5EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EB0: ; ADDQ		0x5EB0-0x5EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EB8: ; ADDQ		0x5EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EB9: ; ADDQ		0x5EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
add   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F00: ; SUBQ		0x5F00-0x5F07
and  edi, byte 7
sub   byte [d+edi*4], 7
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F10: ; SUBQ		0x5F10-0x5F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F18: ; SUBQ		0x5F18-0x5F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F20: ; SUBQ		0x5F20-0x5F27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F28: ; SUBQ		0x5F28-0x5F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F30: ; SUBQ		0x5F30-0x5F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F38: ; SUBQ		0x5F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F39: ; SUBQ		0x5F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
sub   dl, byte 7
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F40: ; SUBQ		0x5F40-0x5F47
and  edi, byte 7
sub   word [d+edi*4], 7
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F48: ; SUBQ		0x5F48-0x5F4F
and  edi, byte 7
sub   dword [a+edi*4], byte 7
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F50: ; SUBQ		0x5F50-0x5F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F58: ; SUBQ		0x5F58-0x5F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F60: ; SUBQ		0x5F60-0x5F67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F68: ; SUBQ		0x5F68-0x5F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F70: ; SUBQ		0x5F70-0x5F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F78: ; SUBQ		0x5F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F79: ; SUBQ		0x5F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub   dx, 7
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F80: ; SUBQ		0x5F80-0x5F87
and  edi, byte 7
sub   dword [d+edi*4], byte 7
lahf
seto al
setc byte [x]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F88: ; SUBQ		0x5F88-0x5F8F
and  edi, byte 7
sub   dword [a+edi*4], byte 7
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F90: ; SUBQ		0x5F90-0x5F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5F98: ; SUBQ		0x5F98-0x5F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FA0: ; SUBQ		0x5FA0-0x5FA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FA8: ; SUBQ		0x5FA8-0x5FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FB0: ; SUBQ		0x5FB0-0x5FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FB8: ; SUBQ		0x5FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FB9: ; SUBQ		0x5FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
sub   edx, byte 7
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID100: ; ADDX		0xD100-0xD107
and  edi, byte 7
mov  bl, [d+0*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID108: ; ADDX		0xD108-0xD10F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+0*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID140: ; ADDX		0xD140-0xD147
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID148: ; ADDX		0xD148-0xD14F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+0*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID180: ; ADDX		0xD180-0xD187
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID188: ; ADDX		0xD188-0xD18F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+0*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID300: ; ADDX		0xD300-0xD307
and  edi, byte 7
mov  bl, [d+1*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID308: ; ADDX		0xD308-0xD30F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+1*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID340: ; ADDX		0xD340-0xD347
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID348: ; ADDX		0xD348-0xD34F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+1*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID380: ; ADDX		0xD380-0xD387
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID388: ; ADDX		0xD388-0xD38F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+1*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID500: ; ADDX		0xD500-0xD507
and  edi, byte 7
mov  bl, [d+2*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID508: ; ADDX		0xD508-0xD50F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+2*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID540: ; ADDX		0xD540-0xD547
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID548: ; ADDX		0xD548-0xD54F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+2*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID580: ; ADDX		0xD580-0xD587
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID588: ; ADDX		0xD588-0xD58F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+2*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID700: ; ADDX		0xD700-0xD707
and  edi, byte 7
mov  bl, [d+3*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID708: ; ADDX		0xD708-0xD70F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+3*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID740: ; ADDX		0xD740-0xD747
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID748: ; ADDX		0xD748-0xD74F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+3*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID780: ; ADDX		0xD780-0xD787
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID788: ; ADDX		0xD788-0xD78F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+3*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID900: ; ADDX		0xD900-0xD907
and  edi, byte 7
mov  bl, [d+4*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID908: ; ADDX		0xD908-0xD90F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+4*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID940: ; ADDX		0xD940-0xD947
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID948: ; ADDX		0xD948-0xD94F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+4*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID980: ; ADDX		0xD980-0xD987
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ID988: ; ADDX		0xD988-0xD98F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+4*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB00: ; ADDX		0xDB00-0xDB07
and  edi, byte 7
mov  bl, [d+5*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB08: ; ADDX		0xDB08-0xDB0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+5*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB40: ; ADDX		0xDB40-0xDB47
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB48: ; ADDX		0xDB48-0xDB4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+5*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB80: ; ADDX		0xDB80-0xDB87
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDB88: ; ADDX		0xDB88-0xDB8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+5*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD00: ; ADDX		0xDD00-0xDD07
and  edi, byte 7
mov  bl, [d+6*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD08: ; ADDX		0xDD08-0xDD0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+6*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD40: ; ADDX		0xDD40-0xDD47
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD48: ; ADDX		0xDD48-0xDD4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+6*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD80: ; ADDX		0xDD80-0xDD87
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDD88: ; ADDX		0xDD88-0xDD8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+6*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF00: ; ADDX		0xDF00-0xDF07
and  edi, byte 7
mov  bl, [d+7*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF08: ; ADDX		0xDF08-0xDF0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF40: ; ADDX		0xDF40-0xDF47
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF48: ; ADDX		0xDF48-0xDF4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF80: ; ADDX		0xDF80-0xDF87
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IDF88: ; ADDX		0xDF88-0xDF8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+7*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
adc  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I023C: ; ANDI to CCR		0x023C
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  dl, [esi]
and  dl, 0x1f
and  [sr], dl
add  esi, byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I027C: ; ANDI to SR		0x027C
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [esi]
and  edx, 0xa71f
and  [sr], dx
add  esi, byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
test byte [sr+1], 0x20
jnz  short .in_s
mov  ebx, [a+7*4]
mov  edx, [__sp]
mov  [a+7*4], edx
mov  [__sp], ebx
.in_s:
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE000: ; ASR		0xE000-0xE007
and  edi, byte 7
sar  byte [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE020: ; ASR		0xE020-0xE027
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE040: ; ASR		0xE040-0xE047
and  edi, byte 7
sar  word [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE060: ; ASR		0xE060-0xE067
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE080: ; ASR		0xE080-0xE087
and  edi, byte 7
sar  dword [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0A0: ; ASR		0xE0A0-0xE0A7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE100: ; ASL		0xE100-0xE107
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE120: ; ASL		0xE120-0xE127
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+0*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE140: ; ASL		0xE140-0xE147
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE160: ; ASL		0xE160-0xE167
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+0*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE180: ; ASL		0xE180-0xE187
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1A0: ; ASL		0xE1A0-0xE1A7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+0*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE200: ; ASR		0xE200-0xE207
and  edi, byte 7
sar  byte [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE220: ; ASR		0xE220-0xE227
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE240: ; ASR		0xE240-0xE247
and  edi, byte 7
sar  word [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE260: ; ASR		0xE260-0xE267
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE280: ; ASR		0xE280-0xE287
and  edi, byte 7
sar  dword [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2A0: ; ASR		0xE2A0-0xE2A7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE300: ; ASL		0xE300-0xE307
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE320: ; ASL		0xE320-0xE327
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+1*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE340: ; ASL		0xE340-0xE347
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE360: ; ASL		0xE360-0xE367
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+1*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE380: ; ASL		0xE380-0xE387
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3A0: ; ASL		0xE3A0-0xE3A7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+1*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE400: ; ASR		0xE400-0xE407
and  edi, byte 7
sar  byte [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE420: ; ASR		0xE420-0xE427
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE440: ; ASR		0xE440-0xE447
and  edi, byte 7
sar  word [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE460: ; ASR		0xE460-0xE467
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE480: ; ASR		0xE480-0xE487
and  edi, byte 7
sar  dword [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4A0: ; ASR		0xE4A0-0xE4A7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE500: ; ASL		0xE500-0xE507
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE520: ; ASL		0xE520-0xE527
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+2*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE540: ; ASL		0xE540-0xE547
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE560: ; ASL		0xE560-0xE567
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+2*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE580: ; ASL		0xE580-0xE587
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5A0: ; ASL		0xE5A0-0xE5A7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+2*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE600: ; ASR		0xE600-0xE607
and  edi, byte 7
sar  byte [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE620: ; ASR		0xE620-0xE627
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE640: ; ASR		0xE640-0xE647
and  edi, byte 7
sar  word [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE660: ; ASR		0xE660-0xE667
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE680: ; ASR		0xE680-0xE687
and  edi, byte 7
sar  dword [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6A0: ; ASR		0xE6A0-0xE6A7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE700: ; ASL		0xE700-0xE707
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE720: ; ASL		0xE720-0xE727
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+3*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE740: ; ASL		0xE740-0xE747
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE760: ; ASL		0xE760-0xE767
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+3*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE780: ; ASL		0xE780-0xE787
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7A0: ; ASL		0xE7A0-0xE7A7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+3*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE800: ; ASR		0xE800-0xE807
and  edi, byte 7
sar  byte [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE820: ; ASR		0xE820-0xE827
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE840: ; ASR		0xE840-0xE847
and  edi, byte 7
sar  word [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE860: ; ASR		0xE860-0xE867
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE880: ; ASR		0xE880-0xE887
and  edi, byte 7
sar  dword [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE8A0: ; ASR		0xE8A0-0xE8A7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE900: ; ASL		0xE900-0xE907
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE920: ; ASL		0xE920-0xE927
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+4*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE940: ; ASL		0xE940-0xE947
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE960: ; ASL		0xE960-0xE967
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+4*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE980: ; ASL		0xE980-0xE987
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE9A0: ; ASL		0xE9A0-0xE9A7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+4*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA00: ; ASR		0xEA00-0xEA07
and  edi, byte 7
sar  byte [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA20: ; ASR		0xEA20-0xEA27
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA40: ; ASR		0xEA40-0xEA47
and  edi, byte 7
sar  word [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA60: ; ASR		0xEA60-0xEA67
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA80: ; ASR		0xEA80-0xEA87
and  edi, byte 7
sar  dword [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEAA0: ; ASR		0xEAA0-0xEAA7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB00: ; ASL		0xEB00-0xEB07
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB20: ; ASL		0xEB20-0xEB27
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+5*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB40: ; ASL		0xEB40-0xEB47
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB60: ; ASL		0xEB60-0xEB67
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+5*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB80: ; ASL		0xEB80-0xEB87
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEBA0: ; ASL		0xEBA0-0xEBA7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+5*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC00: ; ASR		0xEC00-0xEC07
and  edi, byte 7
sar  byte [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC20: ; ASR		0xEC20-0xEC27
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC40: ; ASR		0xEC40-0xEC47
and  edi, byte 7
sar  word [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC60: ; ASR		0xEC60-0xEC67
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC80: ; ASR		0xEC80-0xEC87
and  edi, byte 7
sar  dword [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IECA0: ; ASR		0xECA0-0xECA7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED00: ; ASL		0xED00-0xED07
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED20: ; ASL		0xED20-0xED27
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+6*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED40: ; ASL		0xED40-0xED47
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED60: ; ASL		0xED60-0xED67
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+6*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED80: ; ASL		0xED80-0xED87
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEDA0: ; ASL		0xEDA0-0xEDA7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+6*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE00: ; ASR		0xEE00-0xEE07
and  edi, byte 7
sar  byte [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE20: ; ASR		0xEE20-0xEE27
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [d+edi*4]
sar  dl, cl
setc bl
test dl, dl
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dl
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE40: ; ASR		0xEE40-0xEE47
and  edi, byte 7
sar  word [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE60: ; ASR		0xEE60-0xEE67
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  dx, cl
setc bl
test dx, dx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], dx
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE80: ; ASR		0xEE80-0xEE87
and  edi, byte 7
sar  dword [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEEA0: ; ASR		0xEEA0-0xEEA7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
sar  edx, cl
setc bl
test edx, edx
lahf
test cl, cl
jz   short .no_x
or   ah, bl
test     ah, 1
setnz    byte [x]
.no_x:
mov  [d+edi*4], edx
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF00: ; ASL		0xEF00-0xEF07
and  edi, byte 7
xor  al, al
mov  dl, [d+edi*4]
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dl, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF20: ; ASL		0xEF20-0xEF27
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+7*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dl, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dl
test dl, dl
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dl, dl
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF40: ; ASL		0xEF40-0xEF47
and  edi, byte 7
xor  al, al
mov  dx, [d+edi*4]
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  dx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], dx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF60: ; ASL		0xEF60-0xEF67
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+7*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  dx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], dx
test dx, dx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test dx, dx
lahf
xor  al, al
.end:
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF80: ; ASL		0xEF80-0xEF87
and  edi, byte 7
xor  al, al
mov  edx, [d+edi*4]
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
shl  edx, 1
lahf
setc byte [x]
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
mov  [d+edi*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEFA0: ; ASL		0xEFA0-0xEFA7
and  edi, byte 7
xor  al, al
mov  [run_edi], edi
mov  edx, [d+edi*4]
mov  edi, [d+7*4]
and  edi, byte 0x3f
jz   short .zero_shift
.loop:
shl  edx, 1
setc bl
sets bh
xor  bh, bl
or   al, bh
sub  ecx, byte 2
dec  edi
jnz  short .loop
mov  edi, [run_edi]
mov  [d+edi*4], edx
test edx, edx
lahf
or   ah, bl
test bl, bl
setnz byte [x]
jmp  short .end
.zero_shift:
test edx, edx
lahf
xor  al, al
.end:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0D0: ; ASR		0xE0D0-0xE0D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0D8: ; ASR		0xE0D8-0xE0DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0E0: ; ASR		0xE0E0-0xE0E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0E8: ; ASR		0xE0E8-0xE0EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0F0: ; ASR		0xE0F0-0xE0F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0F8: ; ASR		0xE0F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0F9: ; ASR		0xE0F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sar  dx, byte 1
lahf
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1D0: ; ASL		0xE1D0-0xE1D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1D8: ; ASL		0xE1D8-0xE1DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1E0: ; ASL		0xE1E0-0xE1E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1E8: ; ASL		0xE1E8-0xE1EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1F0: ; ASL		0xE1F0-0xE1F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1F8: ; ASL		0xE1F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1F9: ; ASL		0xE1F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
sets al
adc  al, byte 0
and  al, 1
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6200: ; BHI		0x6200
test ah, byte 0x41
jz   short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6201: ; BHI		0x6201-0x62FF
test ah, byte 0x41
jz   short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6300: ; BLS		0x6300
test ah, byte 0x41
jnz  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6301: ; BLS		0x6301-0x63FF
test ah, byte 0x41
jnz  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6400: ; BCC		0x6400
test ah, byte 1
jz   short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6401: ; BCC		0x6401-0x64FF
test ah, byte 1
jz   short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6500: ; BCS		0x6500
test ah, byte 1
jnz  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6501: ; BCS		0x6501-0x65FF
test ah, byte 1
jnz  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6600: ; BNE		0x6600
test ah, byte 0x40
jz   short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6601: ; BNE		0x6601-0x66FF
test ah, byte 0x40
jz   short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6700: ; BEQ		0x6700
test ah, byte 0x40
jnz  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6701: ; BEQ		0x6701-0x67FF
test ah, byte 0x40
jnz  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6800: ; BVC		0x6800
test al, byte 1
jz   short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6801: ; BVC		0x6801-0x68FF
test al, byte 1
jz   short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6900: ; BVS		0x6900
test al, byte 1
jnz  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6901: ; BVS		0x6901-0x69FF
test al, byte 1
jnz  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6A00: ; BPL		0x6A00
test ah, byte 0x80
jz   short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6A01: ; BPL		0x6A01-0x6AFF
test ah, byte 0x80
jz   short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6B00: ; BMI		0x6B00
test ah, byte 0x80
jnz  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6B01: ; BMI		0x6B01-0x6BFF
test ah, byte 0x80
jnz  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6C00: ; BGE		0x6C00
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6C01: ; BGE		0x6C01-0x6CFF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6D00: ; BLT		0x6D00
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
.dont_do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6D01: ; BLT		0x6D01-0x6DFF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
.dont_do:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6E00: ; BGT		0x6E00
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
.dont_do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6E01: ; BGT		0x6E01-0x6EFF
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
.dont_do:
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6F00: ; BLE		0x6F00
test ah, 0x40
jnz  short .do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6F01: ; BLE		0x6F01-0x6FFF
test ah, 0x40
jnz  short .do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6000: ; BRA		0x6000
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6001: ; BRA		0x6001-0x60FF
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6100: ; BSR		0x6100
mov  edx, 2
add  edx, esi
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [run_edx], edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
sub  edx, ebp
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I6101: ; BSR		0x6101-0x61FF
mov  edx, esi
and      edi, 0xff
mov      ebx, edi
movsx    ebx, bl
add      esi, ebx
sub  esi, ebp
mov  [run_edx], edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
sub  edx, ebp
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0140: ; BCHG		0x0140-0x0147
and  edi, byte 7
mov  ebx, [d+0*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0150: ; BCHG		0x0150-0x0157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0158: ; BCHG		0x0158-0x015F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0160: ; BCHG		0x0160-0x0167
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0168: ; BCHG		0x0168-0x016F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0170: ; BCHG		0x0170-0x0177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0178: ; BCHG		0x0178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0179: ; BCHG		0x0179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0340: ; BCHG		0x0340-0x0347
and  edi, byte 7
mov  ebx, [d+1*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0350: ; BCHG		0x0350-0x0357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0358: ; BCHG		0x0358-0x035F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0360: ; BCHG		0x0360-0x0367
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0368: ; BCHG		0x0368-0x036F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0370: ; BCHG		0x0370-0x0377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0378: ; BCHG		0x0378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0379: ; BCHG		0x0379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0540: ; BCHG		0x0540-0x0547
and  edi, byte 7
mov  ebx, [d+2*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0550: ; BCHG		0x0550-0x0557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0558: ; BCHG		0x0558-0x055F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0560: ; BCHG		0x0560-0x0567
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0568: ; BCHG		0x0568-0x056F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0570: ; BCHG		0x0570-0x0577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0578: ; BCHG		0x0578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0579: ; BCHG		0x0579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0740: ; BCHG		0x0740-0x0747
and  edi, byte 7
mov  ebx, [d+3*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0750: ; BCHG		0x0750-0x0757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0758: ; BCHG		0x0758-0x075F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0760: ; BCHG		0x0760-0x0767
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0768: ; BCHG		0x0768-0x076F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0770: ; BCHG		0x0770-0x0777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0778: ; BCHG		0x0778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0779: ; BCHG		0x0779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0940: ; BCHG		0x0940-0x0947
and  edi, byte 7
mov  ebx, [d+4*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0950: ; BCHG		0x0950-0x0957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0958: ; BCHG		0x0958-0x095F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0960: ; BCHG		0x0960-0x0967
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0968: ; BCHG		0x0968-0x096F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0970: ; BCHG		0x0970-0x0977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0978: ; BCHG		0x0978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0979: ; BCHG		0x0979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B40: ; BCHG		0x0B40-0x0B47
and  edi, byte 7
mov  ebx, [d+5*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B50: ; BCHG		0x0B50-0x0B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B58: ; BCHG		0x0B58-0x0B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B60: ; BCHG		0x0B60-0x0B67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B68: ; BCHG		0x0B68-0x0B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B70: ; BCHG		0x0B70-0x0B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B78: ; BCHG		0x0B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B79: ; BCHG		0x0B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D40: ; BCHG		0x0D40-0x0D47
and  edi, byte 7
mov  ebx, [d+6*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D50: ; BCHG		0x0D50-0x0D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D58: ; BCHG		0x0D58-0x0D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D60: ; BCHG		0x0D60-0x0D67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D68: ; BCHG		0x0D68-0x0D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D70: ; BCHG		0x0D70-0x0D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D78: ; BCHG		0x0D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D79: ; BCHG		0x0D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F40: ; BCHG		0x0F40-0x0F47
and  edi, byte 7
mov  ebx, [d+7*4]
and  ebx, byte 31
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F50: ; BCHG		0x0F50-0x0F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F58: ; BCHG		0x0F58-0x0F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F60: ; BCHG		0x0F60-0x0F67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F68: ; BCHG		0x0F68-0x0F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F70: ; BCHG		0x0F70-0x0F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F78: ; BCHG		0x0F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F79: ; BCHG		0x0F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0840: ; BCHG		0x0840-0x0847
and  edi, byte 7
mov  ebx, [esi]
and  ebx, byte 31
add  esi, byte 2
btc  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0850: ; BCHG		0x0850-0x0857
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0858: ; BCHG		0x0858-0x085F
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0860: ; BCHG		0x0860-0x0867
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0868: ; BCHG		0x0868-0x086F
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0870: ; BCHG		0x0870-0x0877
mov  [run_edi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0878: ; BCHG		0x0878
mov  [run_edi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0879: ; BCHG		0x0879
mov  [run_edi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btc  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0180: ; BCLR		0x0180-0x0187
and  edi, byte 7
mov  ebx, [d+0*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0190: ; BCLR		0x0190-0x0197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0198: ; BCLR		0x0198-0x019F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01A0: ; BCLR		0x01A0-0x01A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01A8: ; BCLR		0x01A8-0x01AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01B0: ; BCLR		0x01B0-0x01B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01B8: ; BCLR		0x01B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01B9: ; BCLR		0x01B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0380: ; BCLR		0x0380-0x0387
and  edi, byte 7
mov  ebx, [d+1*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0390: ; BCLR		0x0390-0x0397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0398: ; BCLR		0x0398-0x039F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03A0: ; BCLR		0x03A0-0x03A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03A8: ; BCLR		0x03A8-0x03AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03B0: ; BCLR		0x03B0-0x03B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03B8: ; BCLR		0x03B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03B9: ; BCLR		0x03B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0580: ; BCLR		0x0580-0x0587
and  edi, byte 7
mov  ebx, [d+2*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0590: ; BCLR		0x0590-0x0597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0598: ; BCLR		0x0598-0x059F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05A0: ; BCLR		0x05A0-0x05A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05A8: ; BCLR		0x05A8-0x05AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05B0: ; BCLR		0x05B0-0x05B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05B8: ; BCLR		0x05B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05B9: ; BCLR		0x05B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0780: ; BCLR		0x0780-0x0787
and  edi, byte 7
mov  ebx, [d+3*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0790: ; BCLR		0x0790-0x0797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0798: ; BCLR		0x0798-0x079F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07A0: ; BCLR		0x07A0-0x07A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07A8: ; BCLR		0x07A8-0x07AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07B0: ; BCLR		0x07B0-0x07B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07B8: ; BCLR		0x07B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07B9: ; BCLR		0x07B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0980: ; BCLR		0x0980-0x0987
and  edi, byte 7
mov  ebx, [d+4*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0990: ; BCLR		0x0990-0x0997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0998: ; BCLR		0x0998-0x099F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09A0: ; BCLR		0x09A0-0x09A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09A8: ; BCLR		0x09A8-0x09AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09B0: ; BCLR		0x09B0-0x09B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09B8: ; BCLR		0x09B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09B9: ; BCLR		0x09B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B80: ; BCLR		0x0B80-0x0B87
and  edi, byte 7
mov  ebx, [d+5*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B90: ; BCLR		0x0B90-0x0B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B98: ; BCLR		0x0B98-0x0B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BA0: ; BCLR		0x0BA0-0x0BA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BA8: ; BCLR		0x0BA8-0x0BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BB0: ; BCLR		0x0BB0-0x0BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BB8: ; BCLR		0x0BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BB9: ; BCLR		0x0BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D80: ; BCLR		0x0D80-0x0D87
and  edi, byte 7
mov  ebx, [d+6*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D90: ; BCLR		0x0D90-0x0D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D98: ; BCLR		0x0D98-0x0D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DA0: ; BCLR		0x0DA0-0x0DA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DA8: ; BCLR		0x0DA8-0x0DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DB0: ; BCLR		0x0DB0-0x0DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DB8: ; BCLR		0x0DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DB9: ; BCLR		0x0DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F80: ; BCLR		0x0F80-0x0F87
and  edi, byte 7
mov  ebx, [d+7*4]
and  ebx, byte 31
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F90: ; BCLR		0x0F90-0x0F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F98: ; BCLR		0x0F98-0x0F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FA0: ; BCLR		0x0FA0-0x0FA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FA8: ; BCLR		0x0FA8-0x0FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FB0: ; BCLR		0x0FB0-0x0FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FB8: ; BCLR		0x0FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FB9: ; BCLR		0x0FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0880: ; BCLR		0x0880-0x0887
and  edi, byte 7
mov  ebx, [esi]
and  ebx, byte 31
add  esi, byte 2
btr  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0890: ; BCLR		0x0890-0x0897
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0898: ; BCLR		0x0898-0x089F
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08A0: ; BCLR		0x08A0-0x08A7
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08A8: ; BCLR		0x08A8-0x08AF
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08B0: ; BCLR		0x08B0-0x08B7
mov  [run_edi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08B8: ; BCLR		0x08B8
mov  [run_edi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08B9: ; BCLR		0x08B9
mov  [run_edi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
btr  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01C0: ; BSET		0x01C0-0x01C7
and  edi, byte 7
mov  ebx, [d+0*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01D0: ; BSET		0x01D0-0x01D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01D8: ; BSET		0x01D8-0x01DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01E0: ; BSET		0x01E0-0x01E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01E8: ; BSET		0x01E8-0x01EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01F0: ; BSET		0x01F0-0x01F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01F8: ; BSET		0x01F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01F9: ; BSET		0x01F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+0*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03C0: ; BSET		0x03C0-0x03C7
and  edi, byte 7
mov  ebx, [d+1*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03D0: ; BSET		0x03D0-0x03D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03D8: ; BSET		0x03D8-0x03DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03E0: ; BSET		0x03E0-0x03E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03E8: ; BSET		0x03E8-0x03EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03F0: ; BSET		0x03F0-0x03F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03F8: ; BSET		0x03F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03F9: ; BSET		0x03F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+1*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05C0: ; BSET		0x05C0-0x05C7
and  edi, byte 7
mov  ebx, [d+2*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05D0: ; BSET		0x05D0-0x05D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05D8: ; BSET		0x05D8-0x05DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05E0: ; BSET		0x05E0-0x05E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05E8: ; BSET		0x05E8-0x05EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05F0: ; BSET		0x05F0-0x05F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05F8: ; BSET		0x05F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05F9: ; BSET		0x05F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+2*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07C0: ; BSET		0x07C0-0x07C7
and  edi, byte 7
mov  ebx, [d+3*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07D0: ; BSET		0x07D0-0x07D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07D8: ; BSET		0x07D8-0x07DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07E0: ; BSET		0x07E0-0x07E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07E8: ; BSET		0x07E8-0x07EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07F0: ; BSET		0x07F0-0x07F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07F8: ; BSET		0x07F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07F9: ; BSET		0x07F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+3*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09C0: ; BSET		0x09C0-0x09C7
and  edi, byte 7
mov  ebx, [d+4*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09D0: ; BSET		0x09D0-0x09D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09D8: ; BSET		0x09D8-0x09DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09E0: ; BSET		0x09E0-0x09E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09E8: ; BSET		0x09E8-0x09EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09F0: ; BSET		0x09F0-0x09F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09F8: ; BSET		0x09F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09F9: ; BSET		0x09F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+4*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BC0: ; BSET		0x0BC0-0x0BC7
and  edi, byte 7
mov  ebx, [d+5*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BD0: ; BSET		0x0BD0-0x0BD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BD8: ; BSET		0x0BD8-0x0BDF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BE0: ; BSET		0x0BE0-0x0BE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BE8: ; BSET		0x0BE8-0x0BEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BF0: ; BSET		0x0BF0-0x0BF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BF8: ; BSET		0x0BF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BF9: ; BSET		0x0BF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+5*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DC0: ; BSET		0x0DC0-0x0DC7
and  edi, byte 7
mov  ebx, [d+6*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DD0: ; BSET		0x0DD0-0x0DD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DD8: ; BSET		0x0DD8-0x0DDF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DE0: ; BSET		0x0DE0-0x0DE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DE8: ; BSET		0x0DE8-0x0DEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DF0: ; BSET		0x0DF0-0x0DF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DF8: ; BSET		0x0DF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DF9: ; BSET		0x0DF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+6*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FC0: ; BSET		0x0FC0-0x0FC7
and  edi, byte 7
mov  ebx, [d+7*4]
and  ebx, byte 31
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FD0: ; BSET		0x0FD0-0x0FD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FD8: ; BSET		0x0FD8-0x0FDF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FE0: ; BSET		0x0FE0-0x0FE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FE8: ; BSET		0x0FE8-0x0FEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FF0: ; BSET		0x0FF0-0x0FF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FF8: ; BSET		0x0FF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FF9: ; BSET		0x0FF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [d+7*4]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08C0: ; BSET		0x08C0-0x08C7
and  edi, byte 7
mov  ebx, [esi]
and  ebx, byte 31
add  esi, byte 2
bts  dword [d+edi*4], ebx
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08D0: ; BSET		0x08D0-0x08D7
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08D8: ; BSET		0x08D8-0x08DF
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08E0: ; BSET		0x08E0-0x08E7
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08E8: ; BSET		0x08E8-0x08EF
mov  [run_edi], esi
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08F0: ; BSET		0x08F0-0x08F7
mov  [run_edi], esi
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08F8: ; BSET		0x08F8
mov  [run_edi], esi
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I08F9: ; BSET		0x08F9
mov  [run_edi], esi
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, [run_edi]
mov  edi, [edi]
and  edi, byte 7
bts  edx, edi
setnc dh
and  ah, 0xbf
shl  dh, 6
or   ah, dh
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0100: ; BTST		0x0100-0x0107
and  edi, byte 7
mov  ebx, [d+0*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0110: ; BTST		0x0110-0x0117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0118: ; BTST		0x0118-0x011F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0120: ; BTST		0x0120-0x0127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0128: ; BTST		0x0128-0x012F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0130: ; BTST		0x0130-0x0137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0138: ; BTST		0x0138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0139: ; BTST		0x0139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I013A: ; BTST		0x013A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I013B: ; BTST		0x013B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I013C: ; BTST		0x013C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+0*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0300: ; BTST		0x0300-0x0307
and  edi, byte 7
mov  ebx, [d+1*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0310: ; BTST		0x0310-0x0317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0318: ; BTST		0x0318-0x031F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0320: ; BTST		0x0320-0x0327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0328: ; BTST		0x0328-0x032F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0330: ; BTST		0x0330-0x0337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0338: ; BTST		0x0338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0339: ; BTST		0x0339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I033A: ; BTST		0x033A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I033B: ; BTST		0x033B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I033C: ; BTST		0x033C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+1*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0500: ; BTST		0x0500-0x0507
and  edi, byte 7
mov  ebx, [d+2*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0510: ; BTST		0x0510-0x0517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0518: ; BTST		0x0518-0x051F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0520: ; BTST		0x0520-0x0527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0528: ; BTST		0x0528-0x052F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0530: ; BTST		0x0530-0x0537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0538: ; BTST		0x0538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0539: ; BTST		0x0539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I053A: ; BTST		0x053A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I053B: ; BTST		0x053B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I053C: ; BTST		0x053C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+2*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0700: ; BTST		0x0700-0x0707
and  edi, byte 7
mov  ebx, [d+3*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0710: ; BTST		0x0710-0x0717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0718: ; BTST		0x0718-0x071F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0720: ; BTST		0x0720-0x0727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0728: ; BTST		0x0728-0x072F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0730: ; BTST		0x0730-0x0737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0738: ; BTST		0x0738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0739: ; BTST		0x0739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I073A: ; BTST		0x073A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I073B: ; BTST		0x073B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I073C: ; BTST		0x073C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+3*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0900: ; BTST		0x0900-0x0907
and  edi, byte 7
mov  ebx, [d+4*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0910: ; BTST		0x0910-0x0917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0918: ; BTST		0x0918-0x091F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0920: ; BTST		0x0920-0x0927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0928: ; BTST		0x0928-0x092F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0930: ; BTST		0x0930-0x0937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0938: ; BTST		0x0938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0939: ; BTST		0x0939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I093A: ; BTST		0x093A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I093B: ; BTST		0x093B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I093C: ; BTST		0x093C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+4*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B00: ; BTST		0x0B00-0x0B07
and  edi, byte 7
mov  ebx, [d+5*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B10: ; BTST		0x0B10-0x0B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B18: ; BTST		0x0B18-0x0B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B20: ; BTST		0x0B20-0x0B27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B28: ; BTST		0x0B28-0x0B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B30: ; BTST		0x0B30-0x0B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B38: ; BTST		0x0B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B39: ; BTST		0x0B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B3A: ; BTST		0x0B3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B3B: ; BTST		0x0B3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B3C: ; BTST		0x0B3C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+5*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D00: ; BTST		0x0D00-0x0D07
and  edi, byte 7
mov  ebx, [d+6*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D10: ; BTST		0x0D10-0x0D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D18: ; BTST		0x0D18-0x0D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D20: ; BTST		0x0D20-0x0D27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D28: ; BTST		0x0D28-0x0D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D30: ; BTST		0x0D30-0x0D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D38: ; BTST		0x0D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D39: ; BTST		0x0D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D3A: ; BTST		0x0D3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D3B: ; BTST		0x0D3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D3C: ; BTST		0x0D3C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+6*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F00: ; BTST		0x0F00-0x0F07
and  edi, byte 7
mov  ebx, [d+7*4]
and  ebx, byte 31
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F10: ; BTST		0x0F10-0x0F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F18: ; BTST		0x0F18-0x0F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F20: ; BTST		0x0F20-0x0F27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F28: ; BTST		0x0F28-0x0F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F30: ; BTST		0x0F30-0x0F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F38: ; BTST		0x0F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F39: ; BTST		0x0F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F3A: ; BTST		0x0F3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F3B: ; BTST		0x0F3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F3C: ; BTST		0x0F3C
mov  dl, [esi]
add  esi, byte 2
mov  ebx, [d+7*4]
and  ebx, byte 7
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0800: ; BTST		0x0800-0x0807
and  edi, byte 7
mov  ebx, [esi]
and  ebx, byte 31
add  esi, byte 2
bt   dword [d+edi*4], ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0810: ; BTST		0x0810-0x0817
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0818: ; BTST		0x0818-0x081F
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0820: ; BTST		0x0820-0x0827
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0828: ; BTST		0x0828-0x082F
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0830: ; BTST		0x0830-0x0837
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0838: ; BTST		0x0838
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0839: ; BTST		0x0839
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I083A: ; BTST		0x083A
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I083B: ; BTST		0x083B
mov  ebx, [esi]
and  ebx, byte 7
add  esi, byte 2
mov  [run_ebx], ebx
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
mov  ebx, [run_ebx]
bt   edx, ebx
setnc bl
and  ah, 0xbf
shl  bl, 6
or   ah, bl
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4180: ; CHK		0x4180-0x4187
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4190: ; CHK		0x4190-0x4197
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4198: ; CHK		0x4198-0x419F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41A0: ; CHK		0x41A0-0x41A7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41A8: ; CHK		0x41A8-0x41AF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41B0: ; CHK		0x41B0-0x41B7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41B8: ; CHK		0x41B8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41B9: ; CHK		0x41B9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41BA: ; CHK		0x41BA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41BB: ; CHK		0x41BB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I41BC: ; CHK		0x41BC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+0*4+1], 0x80
jnz  near .trap_dn
cmp  [d+0*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4380: ; CHK		0x4380-0x4387
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4390: ; CHK		0x4390-0x4397
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4398: ; CHK		0x4398-0x439F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43A0: ; CHK		0x43A0-0x43A7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43A8: ; CHK		0x43A8-0x43AF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43B0: ; CHK		0x43B0-0x43B7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43B8: ; CHK		0x43B8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43B9: ; CHK		0x43B9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43BA: ; CHK		0x43BA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43BB: ; CHK		0x43BB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I43BC: ; CHK		0x43BC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+1*4+1], 0x80
jnz  near .trap_dn
cmp  [d+1*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4580: ; CHK		0x4580-0x4587
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4590: ; CHK		0x4590-0x4597
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4598: ; CHK		0x4598-0x459F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45A0: ; CHK		0x45A0-0x45A7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45A8: ; CHK		0x45A8-0x45AF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45B0: ; CHK		0x45B0-0x45B7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45B8: ; CHK		0x45B8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45B9: ; CHK		0x45B9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45BA: ; CHK		0x45BA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45BB: ; CHK		0x45BB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I45BC: ; CHK		0x45BC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+2*4+1], 0x80
jnz  near .trap_dn
cmp  [d+2*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4780: ; CHK		0x4780-0x4787
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4790: ; CHK		0x4790-0x4797
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4798: ; CHK		0x4798-0x479F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47A0: ; CHK		0x47A0-0x47A7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47A8: ; CHK		0x47A8-0x47AF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47B0: ; CHK		0x47B0-0x47B7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47B8: ; CHK		0x47B8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47B9: ; CHK		0x47B9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47BA: ; CHK		0x47BA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47BB: ; CHK		0x47BB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I47BC: ; CHK		0x47BC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+3*4+1], 0x80
jnz  near .trap_dn
cmp  [d+3*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4980: ; CHK		0x4980-0x4987
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4990: ; CHK		0x4990-0x4997
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4998: ; CHK		0x4998-0x499F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49A0: ; CHK		0x49A0-0x49A7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49A8: ; CHK		0x49A8-0x49AF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49B0: ; CHK		0x49B0-0x49B7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49B8: ; CHK		0x49B8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49B9: ; CHK		0x49B9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49BA: ; CHK		0x49BA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49BB: ; CHK		0x49BB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I49BC: ; CHK		0x49BC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+4*4+1], 0x80
jnz  near .trap_dn
cmp  [d+4*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4B80: ; CHK		0x4B80-0x4B87
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4B90: ; CHK		0x4B90-0x4B97
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4B98: ; CHK		0x4B98-0x4B9F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BA0: ; CHK		0x4BA0-0x4BA7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BA8: ; CHK		0x4BA8-0x4BAF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BB0: ; CHK		0x4BB0-0x4BB7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BB8: ; CHK		0x4BB8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BB9: ; CHK		0x4BB9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BBA: ; CHK		0x4BBA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BBB: ; CHK		0x4BBB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4BBC: ; CHK		0x4BBC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+5*4+1], 0x80
jnz  near .trap_dn
cmp  [d+5*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4D80: ; CHK		0x4D80-0x4D87
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4D90: ; CHK		0x4D90-0x4D97
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4D98: ; CHK		0x4D98-0x4D9F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DA0: ; CHK		0x4DA0-0x4DA7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DA8: ; CHK		0x4DA8-0x4DAF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DB0: ; CHK		0x4DB0-0x4DB7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DB8: ; CHK		0x4DB8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DB9: ; CHK		0x4DB9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DBA: ; CHK		0x4DBA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DBB: ; CHK		0x4DBB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4DBC: ; CHK		0x4DBC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+6*4+1], 0x80
jnz  near .trap_dn
cmp  [d+6*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4F80: ; CHK		0x4F80-0x4F87
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4F90: ; CHK		0x4F90-0x4F97
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4F98: ; CHK		0x4F98-0x4F9F
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FA0: ; CHK		0x4FA0-0x4FA7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FA8: ; CHK		0x4FA8-0x4FAF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FB0: ; CHK		0x4FB0-0x4FB7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FB8: ; CHK		0x4FB8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FB9: ; CHK		0x4FB9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FBA: ; CHK		0x4FBA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FBB: ; CHK		0x4FBB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4FBC: ; CHK		0x4FBC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test byte [d+7*4+1], 0x80
jnz  near .trap_dn
cmp  [d+7*4], dx
ja   near .trap_ea
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.trap_dn:
or   ah, 0x80
jmp  short .trap
.trap_ea:
and  ah, 0x7f
.trap:
jmp  near exception_chk
times ($$-$) & 3 nop
I4200: ; CLR		0x4200-0x4207
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  [d+edi*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4210: ; CLR		0x4210-0x4217
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4218: ; CLR		0x4218-0x421F
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4220: ; CLR		0x4220-0x4227
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4228: ; CLR		0x4228-0x422F
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4230: ; CLR		0x4230-0x4237
xor  edx, edx
mov  eax, 0x4000
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4238: ; CLR		0x4238
xor  edx, edx
mov  eax, 0x4000
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4239: ; CLR		0x4239
xor  edx, edx
mov  eax, 0x4000
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4240: ; CLR		0x4240-0x4247
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  [d+edi*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4250: ; CLR		0x4250-0x4257
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4258: ; CLR		0x4258-0x425F
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4260: ; CLR		0x4260-0x4267
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4268: ; CLR		0x4268-0x426F
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4270: ; CLR		0x4270-0x4277
xor  edx, edx
mov  eax, 0x4000
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4278: ; CLR		0x4278
xor  edx, edx
mov  eax, 0x4000
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4279: ; CLR		0x4279
xor  edx, edx
mov  eax, 0x4000
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4280: ; CLR		0x4280-0x4287
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  [d+edi*4], edx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4290: ; CLR		0x4290-0x4297
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4298: ; CLR		0x4298-0x429F
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I42A0: ; CLR		0x42A0-0x42A7
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I42A8: ; CLR		0x42A8-0x42AF
xor  edx, edx
mov  eax, 0x4000
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I42B0: ; CLR		0x42B0-0x42B7
xor  edx, edx
mov  eax, 0x4000
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I42B8: ; CLR		0x42B8
xor  edx, edx
mov  eax, 0x4000
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I42B9: ; CLR		0x42B9
xor  edx, edx
mov  eax, 0x4000
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0C0: ; CMPA		0xB0C0-0xB0C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0C8: ; CMPA		0xB0C8-0xB0CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0D0: ; CMPA		0xB0D0-0xB0D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0D8: ; CMPA		0xB0D8-0xB0DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0E0: ; CMPA		0xB0E0-0xB0E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0E8: ; CMPA		0xB0E8-0xB0EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0F0: ; CMPA		0xB0F0-0xB0F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0F8: ; CMPA		0xB0F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0F9: ; CMPA		0xB0F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0FA: ; CMPA		0xB0FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0FB: ; CMPA		0xB0FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB0FC: ; CMPA		0xB0FC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1C0: ; CMPA		0xB1C0-0xB1C7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1C8: ; CMPA		0xB1C8-0xB1CF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1D0: ; CMPA		0xB1D0-0xB1D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1D8: ; CMPA		0xB1D8-0xB1DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1E0: ; CMPA		0xB1E0-0xB1E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1E8: ; CMPA		0xB1E8-0xB1EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1F0: ; CMPA		0xB1F0-0xB1F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1F8: ; CMPA		0xB1F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1F9: ; CMPA		0xB1F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1FA: ; CMPA		0xB1FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1FB: ; CMPA		0xB1FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB1FC: ; CMPA		0xB1FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+0*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2C0: ; CMPA		0xB2C0-0xB2C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2C8: ; CMPA		0xB2C8-0xB2CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2D0: ; CMPA		0xB2D0-0xB2D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2D8: ; CMPA		0xB2D8-0xB2DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2E0: ; CMPA		0xB2E0-0xB2E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2E8: ; CMPA		0xB2E8-0xB2EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2F0: ; CMPA		0xB2F0-0xB2F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2F8: ; CMPA		0xB2F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2F9: ; CMPA		0xB2F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2FA: ; CMPA		0xB2FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2FB: ; CMPA		0xB2FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB2FC: ; CMPA		0xB2FC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3C0: ; CMPA		0xB3C0-0xB3C7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3C8: ; CMPA		0xB3C8-0xB3CF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3D0: ; CMPA		0xB3D0-0xB3D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3D8: ; CMPA		0xB3D8-0xB3DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3E0: ; CMPA		0xB3E0-0xB3E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3E8: ; CMPA		0xB3E8-0xB3EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3F0: ; CMPA		0xB3F0-0xB3F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3F8: ; CMPA		0xB3F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3F9: ; CMPA		0xB3F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3FA: ; CMPA		0xB3FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3FB: ; CMPA		0xB3FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB3FC: ; CMPA		0xB3FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+1*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4C0: ; CMPA		0xB4C0-0xB4C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4C8: ; CMPA		0xB4C8-0xB4CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4D0: ; CMPA		0xB4D0-0xB4D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4D8: ; CMPA		0xB4D8-0xB4DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4E0: ; CMPA		0xB4E0-0xB4E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4E8: ; CMPA		0xB4E8-0xB4EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4F0: ; CMPA		0xB4F0-0xB4F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4F8: ; CMPA		0xB4F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4F9: ; CMPA		0xB4F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4FA: ; CMPA		0xB4FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4FB: ; CMPA		0xB4FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB4FC: ; CMPA		0xB4FC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5C0: ; CMPA		0xB5C0-0xB5C7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5C8: ; CMPA		0xB5C8-0xB5CF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5D0: ; CMPA		0xB5D0-0xB5D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5D8: ; CMPA		0xB5D8-0xB5DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5E0: ; CMPA		0xB5E0-0xB5E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5E8: ; CMPA		0xB5E8-0xB5EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5F0: ; CMPA		0xB5F0-0xB5F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5F8: ; CMPA		0xB5F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5F9: ; CMPA		0xB5F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5FA: ; CMPA		0xB5FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5FB: ; CMPA		0xB5FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB5FC: ; CMPA		0xB5FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+2*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6C0: ; CMPA		0xB6C0-0xB6C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6C8: ; CMPA		0xB6C8-0xB6CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6D0: ; CMPA		0xB6D0-0xB6D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6D8: ; CMPA		0xB6D8-0xB6DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6E0: ; CMPA		0xB6E0-0xB6E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6E8: ; CMPA		0xB6E8-0xB6EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6F0: ; CMPA		0xB6F0-0xB6F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6F8: ; CMPA		0xB6F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6F9: ; CMPA		0xB6F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6FA: ; CMPA		0xB6FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6FB: ; CMPA		0xB6FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB6FC: ; CMPA		0xB6FC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7C0: ; CMPA		0xB7C0-0xB7C7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7C8: ; CMPA		0xB7C8-0xB7CF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7D0: ; CMPA		0xB7D0-0xB7D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7D8: ; CMPA		0xB7D8-0xB7DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7E0: ; CMPA		0xB7E0-0xB7E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7E8: ; CMPA		0xB7E8-0xB7EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7F0: ; CMPA		0xB7F0-0xB7F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7F8: ; CMPA		0xB7F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7F9: ; CMPA		0xB7F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7FA: ; CMPA		0xB7FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7FB: ; CMPA		0xB7FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB7FC: ; CMPA		0xB7FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+3*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8C0: ; CMPA		0xB8C0-0xB8C7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8C8: ; CMPA		0xB8C8-0xB8CF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8D0: ; CMPA		0xB8D0-0xB8D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8D8: ; CMPA		0xB8D8-0xB8DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8E0: ; CMPA		0xB8E0-0xB8E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8E8: ; CMPA		0xB8E8-0xB8EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8F0: ; CMPA		0xB8F0-0xB8F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8F8: ; CMPA		0xB8F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8F9: ; CMPA		0xB8F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8FA: ; CMPA		0xB8FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8FB: ; CMPA		0xB8FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB8FC: ; CMPA		0xB8FC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9C0: ; CMPA		0xB9C0-0xB9C7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9C8: ; CMPA		0xB9C8-0xB9CF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9D0: ; CMPA		0xB9D0-0xB9D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9D8: ; CMPA		0xB9D8-0xB9DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9E0: ; CMPA		0xB9E0-0xB9E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9E8: ; CMPA		0xB9E8-0xB9EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9F0: ; CMPA		0xB9F0-0xB9F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9F8: ; CMPA		0xB9F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9F9: ; CMPA		0xB9F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9FA: ; CMPA		0xB9FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9FB: ; CMPA		0xB9FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB9FC: ; CMPA		0xB9FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+4*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAC0: ; CMPA		0xBAC0-0xBAC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAC8: ; CMPA		0xBAC8-0xBACF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAD0: ; CMPA		0xBAD0-0xBAD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAD8: ; CMPA		0xBAD8-0xBADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAE0: ; CMPA		0xBAE0-0xBAE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAE8: ; CMPA		0xBAE8-0xBAEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAF0: ; CMPA		0xBAF0-0xBAF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAF8: ; CMPA		0xBAF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAF9: ; CMPA		0xBAF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAFA: ; CMPA		0xBAFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAFB: ; CMPA		0xBAFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBAFC: ; CMPA		0xBAFC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBC0: ; CMPA		0xBBC0-0xBBC7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBC8: ; CMPA		0xBBC8-0xBBCF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBD0: ; CMPA		0xBBD0-0xBBD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBD8: ; CMPA		0xBBD8-0xBBDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBE0: ; CMPA		0xBBE0-0xBBE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBE8: ; CMPA		0xBBE8-0xBBEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBF0: ; CMPA		0xBBF0-0xBBF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBF8: ; CMPA		0xBBF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBF9: ; CMPA		0xBBF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBFA: ; CMPA		0xBBFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBFB: ; CMPA		0xBBFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBBFC: ; CMPA		0xBBFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+5*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCC0: ; CMPA		0xBCC0-0xBCC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCC8: ; CMPA		0xBCC8-0xBCCF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCD0: ; CMPA		0xBCD0-0xBCD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCD8: ; CMPA		0xBCD8-0xBCDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCE0: ; CMPA		0xBCE0-0xBCE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCE8: ; CMPA		0xBCE8-0xBCEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCF0: ; CMPA		0xBCF0-0xBCF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCF8: ; CMPA		0xBCF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCF9: ; CMPA		0xBCF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCFA: ; CMPA		0xBCFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCFB: ; CMPA		0xBCFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBCFC: ; CMPA		0xBCFC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDC0: ; CMPA		0xBDC0-0xBDC7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDC8: ; CMPA		0xBDC8-0xBDCF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDD0: ; CMPA		0xBDD0-0xBDD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDD8: ; CMPA		0xBDD8-0xBDDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDE0: ; CMPA		0xBDE0-0xBDE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDE8: ; CMPA		0xBDE8-0xBDEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDF0: ; CMPA		0xBDF0-0xBDF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDF8: ; CMPA		0xBDF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDF9: ; CMPA		0xBDF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDFA: ; CMPA		0xBDFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDFB: ; CMPA		0xBDFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBDFC: ; CMPA		0xBDFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+6*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEC0: ; CMPA		0xBEC0-0xBEC7
and  edi, byte 7
movsx    edx, word [d+edi*4]
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEC8: ; CMPA		0xBEC8-0xBECF
and  edi, byte 7
movsx    edx, word [a+edi*4]
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBED0: ; CMPA		0xBED0-0xBED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBED8: ; CMPA		0xBED8-0xBEDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEE0: ; CMPA		0xBEE0-0xBEE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEE8: ; CMPA		0xBEE8-0xBEEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEF0: ; CMPA		0xBEF0-0xBEF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEF8: ; CMPA		0xBEF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEF9: ; CMPA		0xBEF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEFA: ; CMPA		0xBEFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEFB: ; CMPA		0xBEFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBEFC: ; CMPA		0xBEFC
movsx    edx, word [esi]
add  esi, byte 2
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFC0: ; CMPA		0xBFC0-0xBFC7
and  edi, byte 7
mov  edx, [d+edi*4]
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFC8: ; CMPA		0xBFC8-0xBFCF
and  edi, byte 7
mov  edx, [a+edi*4]
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFD0: ; CMPA		0xBFD0-0xBFD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFD8: ; CMPA		0xBFD8-0xBFDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFE0: ; CMPA		0xBFE0-0xBFE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFE8: ; CMPA		0xBFE8-0xBFEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFF0: ; CMPA		0xBFF0-0xBFF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFF8: ; CMPA		0xBFF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFF9: ; CMPA		0xBFF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFFA: ; CMPA		0xBFFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFFB: ; CMPA		0xBFFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBFFC: ; CMPA		0xBFFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
cmp  [a+7*4], edx
lahf
seto al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB108: ; CMPM		0xB108-0xB10F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB148: ; CMPM		0xB148-0xB14F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB188: ; CMPM		0xB188-0xB18F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB308: ; CMPM		0xB308-0xB30F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB348: ; CMPM		0xB348-0xB34F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB388: ; CMPM		0xB388-0xB38F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB508: ; CMPM		0xB508-0xB50F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB548: ; CMPM		0xB548-0xB54F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB588: ; CMPM		0xB588-0xB58F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB708: ; CMPM		0xB708-0xB70F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB748: ; CMPM		0xB748-0xB74F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB788: ; CMPM		0xB788-0xB78F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB908: ; CMPM		0xB908-0xB90F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB948: ; CMPM		0xB948-0xB94F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IB988: ; CMPM		0xB988-0xB98F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB08: ; CMPM		0xBB08-0xBB0F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB48: ; CMPM		0xBB48-0xBB4F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBB88: ; CMPM		0xBB88-0xBB8F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD08: ; CMPM		0xBD08-0xBD0F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD48: ; CMPM		0xBD48-0xBD4F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBD88: ; CMPM		0xBD88-0xBD8F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF08: ; CMPM		0xBF08-0xBF0F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call ReadByte
mov  bl, [run_edx]
cmp  dl, bl
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF48: ; CMPM		0xBF48-0xBF4F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call ReadWord
mov  ebx, [run_edx]
cmp  dx, bx
lahf
seto al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IBF88: ; CMPM		0xBF88-0xBF8F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call ReadLong
mov  ebx, [run_edx]
cmp  edx, ebx
lahf
seto al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50C8: ; DBT		0x50C8-0x50CF
add  esi, byte 2
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51C8: ; DBF		0x51C8-0x51CF
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52C8: ; DBHI		0x52C8-0x52CF
test ah, byte 0x41
jz   short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53C8: ; DBLS		0x53C8-0x53CF
test ah, byte 0x41
jnz  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54C8: ; DBCC		0x54C8-0x54CF
test ah, byte 1
jz   short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55C8: ; DBCS		0x55C8-0x55CF
test ah, byte 1
jnz  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56C8: ; DBNE		0x56C8-0x56CF
test ah, byte 0x40
jz   short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57C8: ; DBEQ		0x57C8-0x57CF
test ah, byte 0x40
jnz  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58C8: ; DBVC		0x58C8-0x58CF
test al, byte 1
jz   short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59C8: ; DBVS		0x59C8-0x59CF
test al, byte 1
jnz  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AC8: ; DBPL		0x5AC8-0x5ACF
test ah, byte 0x80
jz   short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BC8: ; DBMI		0x5BC8-0x5BCF
test ah, byte 0x80
jnz  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CC8: ; DBGE		0x5CC8-0x5CCF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DC8: ; DBLT		0x5DC8-0x5DCF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
.dont_do:
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EC8: ; DBGT		0x5EC8-0x5ECF
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short .do
.dont_do:
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FC8: ; DBLE		0x5FC8-0x5FCF
test ah, 0x40
jnz  short .do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short .do
and  edi, byte 7
sub  word [d+edi*4], 1
jc   short .do_loop_expired
movsx    ebx, word [esi]
add      esi, ebx
sub  esi, ebp
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do:
add  esi, byte 2
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
.do_loop_expired:
add  esi, byte 2
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I81C0: ; DIVS		0x81C0-0x81C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81D0: ; DIVS		0x81D0-0x81D7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81D8: ; DIVS		0x81D8-0x81DF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81E0: ; DIVS		0x81E0-0x81E7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81E8: ; DIVS		0x81E8-0x81EF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81F0: ; DIVS		0x81F0-0x81F7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81F8: ; DIVS		0x81F8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81F9: ; DIVS		0x81F9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81FA: ; DIVS		0x81FA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81FB: ; DIVS		0x81FB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I81FC: ; DIVS		0x81FC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+0*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83C0: ; DIVS		0x83C0-0x83C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83D0: ; DIVS		0x83D0-0x83D7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83D8: ; DIVS		0x83D8-0x83DF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83E0: ; DIVS		0x83E0-0x83E7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83E8: ; DIVS		0x83E8-0x83EF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83F0: ; DIVS		0x83F0-0x83F7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83F8: ; DIVS		0x83F8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83F9: ; DIVS		0x83F9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83FA: ; DIVS		0x83FA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83FB: ; DIVS		0x83FB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I83FC: ; DIVS		0x83FC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+1*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85C0: ; DIVS		0x85C0-0x85C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85D0: ; DIVS		0x85D0-0x85D7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85D8: ; DIVS		0x85D8-0x85DF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85E0: ; DIVS		0x85E0-0x85E7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85E8: ; DIVS		0x85E8-0x85EF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85F0: ; DIVS		0x85F0-0x85F7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85F8: ; DIVS		0x85F8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85F9: ; DIVS		0x85F9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85FA: ; DIVS		0x85FA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85FB: ; DIVS		0x85FB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I85FC: ; DIVS		0x85FC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+2*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87C0: ; DIVS		0x87C0-0x87C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87D0: ; DIVS		0x87D0-0x87D7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87D8: ; DIVS		0x87D8-0x87DF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87E0: ; DIVS		0x87E0-0x87E7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87E8: ; DIVS		0x87E8-0x87EF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87F0: ; DIVS		0x87F0-0x87F7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87F8: ; DIVS		0x87F8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87F9: ; DIVS		0x87F9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87FA: ; DIVS		0x87FA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87FB: ; DIVS		0x87FB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I87FC: ; DIVS		0x87FC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+3*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89C0: ; DIVS		0x89C0-0x89C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89D0: ; DIVS		0x89D0-0x89D7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89D8: ; DIVS		0x89D8-0x89DF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89E0: ; DIVS		0x89E0-0x89E7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89E8: ; DIVS		0x89E8-0x89EF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89F0: ; DIVS		0x89F0-0x89F7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89F8: ; DIVS		0x89F8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89F9: ; DIVS		0x89F9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89FA: ; DIVS		0x89FA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89FB: ; DIVS		0x89FB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I89FC: ; DIVS		0x89FC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+4*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BC0: ; DIVS		0x8BC0-0x8BC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BD0: ; DIVS		0x8BD0-0x8BD7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BD8: ; DIVS		0x8BD8-0x8BDF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BE0: ; DIVS		0x8BE0-0x8BE7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BE8: ; DIVS		0x8BE8-0x8BEF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BF0: ; DIVS		0x8BF0-0x8BF7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BF8: ; DIVS		0x8BF8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BF9: ; DIVS		0x8BF9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BFA: ; DIVS		0x8BFA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BFB: ; DIVS		0x8BFB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8BFC: ; DIVS		0x8BFC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+5*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DC0: ; DIVS		0x8DC0-0x8DC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DD0: ; DIVS		0x8DD0-0x8DD7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DD8: ; DIVS		0x8DD8-0x8DDF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DE0: ; DIVS		0x8DE0-0x8DE7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DE8: ; DIVS		0x8DE8-0x8DEF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DF0: ; DIVS		0x8DF0-0x8DF7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DF8: ; DIVS		0x8DF8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DF9: ; DIVS		0x8DF9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DFA: ; DIVS		0x8DFA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DFB: ; DIVS		0x8DFB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8DFC: ; DIVS		0x8DFC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+6*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FC0: ; DIVS		0x8FC0-0x8FC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FD0: ; DIVS		0x8FD0-0x8FD7
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FD8: ; DIVS		0x8FD8-0x8FDF
sub     ecx, byte 4
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FE0: ; DIVS		0x8FE0-0x8FE7
sub     ecx, byte 6
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FE8: ; DIVS		0x8FE8-0x8FEF
sub     ecx, byte 8
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FF0: ; DIVS		0x8FF0-0x8FF7
sub     ecx, byte 10
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FF8: ; DIVS		0x8FF8
sub     ecx, byte 8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FF9: ; DIVS		0x8FF9
sub     ecx, byte 12
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FFA: ; DIVS		0x8FFA
sub     ecx, byte 8
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FFB: ; DIVS		0x8FFB
sub     ecx, byte 10
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8FFC: ; DIVS		0x8FFC
sub     ecx, byte 4
mov  edx, [esi]
add  esi, byte 2
test dx, dx
jz   .divide_by_zero
movsx    ebx, dx
mov  [run_eax], eax
mov  eax, [d+7*4]
cdq
idiv ebx
cmp  eax, -32768
jl   .overflow
cmp  eax, 32767
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 158
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80C0: ; DIVU		0x80C0-0x80C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80D0: ; DIVU		0x80D0-0x80D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80D8: ; DIVU		0x80D8-0x80DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80E0: ; DIVU		0x80E0-0x80E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80E8: ; DIVU		0x80E8-0x80EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80F0: ; DIVU		0x80F0-0x80F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80F8: ; DIVU		0x80F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80F9: ; DIVU		0x80F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80FA: ; DIVU		0x80FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80FB: ; DIVU		0x80FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I80FC: ; DIVU		0x80FC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+0*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(0*4)+2], dx
mov     [d+(0*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82C0: ; DIVU		0x82C0-0x82C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82D0: ; DIVU		0x82D0-0x82D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82D8: ; DIVU		0x82D8-0x82DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82E0: ; DIVU		0x82E0-0x82E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82E8: ; DIVU		0x82E8-0x82EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82F0: ; DIVU		0x82F0-0x82F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82F8: ; DIVU		0x82F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82F9: ; DIVU		0x82F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82FA: ; DIVU		0x82FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82FB: ; DIVU		0x82FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I82FC: ; DIVU		0x82FC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+1*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(1*4)+2], dx
mov     [d+(1*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84C0: ; DIVU		0x84C0-0x84C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84D0: ; DIVU		0x84D0-0x84D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84D8: ; DIVU		0x84D8-0x84DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84E0: ; DIVU		0x84E0-0x84E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84E8: ; DIVU		0x84E8-0x84EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84F0: ; DIVU		0x84F0-0x84F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84F8: ; DIVU		0x84F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84F9: ; DIVU		0x84F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84FA: ; DIVU		0x84FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84FB: ; DIVU		0x84FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I84FC: ; DIVU		0x84FC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+2*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(2*4)+2], dx
mov     [d+(2*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86C0: ; DIVU		0x86C0-0x86C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86D0: ; DIVU		0x86D0-0x86D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86D8: ; DIVU		0x86D8-0x86DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86E0: ; DIVU		0x86E0-0x86E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86E8: ; DIVU		0x86E8-0x86EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86F0: ; DIVU		0x86F0-0x86F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86F8: ; DIVU		0x86F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86F9: ; DIVU		0x86F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86FA: ; DIVU		0x86FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86FB: ; DIVU		0x86FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I86FC: ; DIVU		0x86FC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+3*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(3*4)+2], dx
mov     [d+(3*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88C0: ; DIVU		0x88C0-0x88C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88D0: ; DIVU		0x88D0-0x88D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88D8: ; DIVU		0x88D8-0x88DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88E0: ; DIVU		0x88E0-0x88E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88E8: ; DIVU		0x88E8-0x88EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88F0: ; DIVU		0x88F0-0x88F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88F8: ; DIVU		0x88F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88F9: ; DIVU		0x88F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88FA: ; DIVU		0x88FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88FB: ; DIVU		0x88FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I88FC: ; DIVU		0x88FC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+4*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(4*4)+2], dx
mov     [d+(4*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AC0: ; DIVU		0x8AC0-0x8AC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AD0: ; DIVU		0x8AD0-0x8AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AD8: ; DIVU		0x8AD8-0x8ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AE0: ; DIVU		0x8AE0-0x8AE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AE8: ; DIVU		0x8AE8-0x8AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AF0: ; DIVU		0x8AF0-0x8AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AF8: ; DIVU		0x8AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AF9: ; DIVU		0x8AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AFA: ; DIVU		0x8AFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AFB: ; DIVU		0x8AFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8AFC: ; DIVU		0x8AFC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+5*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(5*4)+2], dx
mov     [d+(5*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CC0: ; DIVU		0x8CC0-0x8CC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CD0: ; DIVU		0x8CD0-0x8CD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CD8: ; DIVU		0x8CD8-0x8CDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CE0: ; DIVU		0x8CE0-0x8CE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CE8: ; DIVU		0x8CE8-0x8CEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CF0: ; DIVU		0x8CF0-0x8CF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CF8: ; DIVU		0x8CF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CF9: ; DIVU		0x8CF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CFA: ; DIVU		0x8CFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CFB: ; DIVU		0x8CFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8CFC: ; DIVU		0x8CFC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+6*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(6*4)+2], dx
mov     [d+(6*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EC0: ; DIVU		0x8EC0-0x8EC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8ED0: ; DIVU		0x8ED0-0x8ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8ED8: ; DIVU		0x8ED8-0x8EDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EE0: ; DIVU		0x8EE0-0x8EE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 6
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EE8: ; DIVU		0x8EE8-0x8EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EF0: ; DIVU		0x8EF0-0x8EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EF8: ; DIVU		0x8EF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EF9: ; DIVU		0x8EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
sub     ecx, byte 12
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EFA: ; DIVU		0x8EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 8
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EFB: ; DIVU		0x8EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
sub     ecx, byte 10
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I8EFC: ; DIVU		0x8EFC
mov  edx, [esi]
add  esi, byte 2
sub     ecx, byte 4
test dx, dx
jz   .divide_by_zero
mov      ebx, edx
and      ebx, 0xffff
mov  [run_eax], eax
mov  eax, [d+7*4]
xor  edx, edx
div  ebx
cmp  eax, 0xffff
jg   .overflow
test ax, ax
mov     [d+(7*4)+2], dx
mov     [d+(7*4)+0], ax
lahf
xor     al, al
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.overflow:
mov  eax, [run_eax]
mov  eax, 1
sub  ecx, 144
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.divide_by_zero:
xor  eax, eax
jmp     near exception_divide_by_zero
times ($$-$) & 3 nop
I0A3C: ; EORI to CCR		0x0A3C
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  dl, [esi]
and  dl, 0x1f
xor  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
add  esi, byte 2
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0A7C: ; EORI to SR		0x0A7C
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [esi]
and  edx, 0xa71f
xor  [sr], dx
add  esi, byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
test byte [sr+1], 0x20
jnz  short .in_s
mov  ebx, [a+7*4]
mov  edx, [__sp]
mov  [a+7*4], edx
mov  [__sp], ebx
.in_s:
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC140: ; EXG		0xC140-0xC147
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [d+edi*4]
mov  [d+0*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC148: ; EXG		0xC148-0xC14F
and  edi, byte 7
mov  ebx, [a+0*4]
mov  edx, [a+edi*4]
mov  [a+0*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC188: ; EXG		0xC188-0xC18F
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [a+edi*4]
mov  [d+0*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC340: ; EXG		0xC340-0xC347
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [d+edi*4]
mov  [d+1*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC348: ; EXG		0xC348-0xC34F
and  edi, byte 7
mov  ebx, [a+1*4]
mov  edx, [a+edi*4]
mov  [a+1*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC388: ; EXG		0xC388-0xC38F
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [a+edi*4]
mov  [d+1*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC540: ; EXG		0xC540-0xC547
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [d+edi*4]
mov  [d+2*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC548: ; EXG		0xC548-0xC54F
and  edi, byte 7
mov  ebx, [a+2*4]
mov  edx, [a+edi*4]
mov  [a+2*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC588: ; EXG		0xC588-0xC58F
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [a+edi*4]
mov  [d+2*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC740: ; EXG		0xC740-0xC747
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [d+edi*4]
mov  [d+3*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC748: ; EXG		0xC748-0xC74F
and  edi, byte 7
mov  ebx, [a+3*4]
mov  edx, [a+edi*4]
mov  [a+3*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC788: ; EXG		0xC788-0xC78F
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [a+edi*4]
mov  [d+3*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC940: ; EXG		0xC940-0xC947
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [d+edi*4]
mov  [d+4*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC948: ; EXG		0xC948-0xC94F
and  edi, byte 7
mov  ebx, [a+4*4]
mov  edx, [a+edi*4]
mov  [a+4*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC988: ; EXG		0xC988-0xC98F
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [a+edi*4]
mov  [d+4*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB40: ; EXG		0xCB40-0xCB47
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [d+edi*4]
mov  [d+5*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB48: ; EXG		0xCB48-0xCB4F
and  edi, byte 7
mov  ebx, [a+5*4]
mov  edx, [a+edi*4]
mov  [a+5*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICB88: ; EXG		0xCB88-0xCB8F
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [a+edi*4]
mov  [d+5*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD40: ; EXG		0xCD40-0xCD47
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [d+edi*4]
mov  [d+6*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD48: ; EXG		0xCD48-0xCD4F
and  edi, byte 7
mov  ebx, [a+6*4]
mov  edx, [a+edi*4]
mov  [a+6*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICD88: ; EXG		0xCD88-0xCD8F
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [a+edi*4]
mov  [d+6*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF40: ; EXG		0xCF40-0xCF47
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [d+edi*4]
mov  [d+7*4], edx
mov  [d+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF48: ; EXG		0xCF48-0xCF4F
and  edi, byte 7
mov  ebx, [a+7*4]
mov  edx, [a+edi*4]
mov  [a+7*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICF88: ; EXG		0xCF88-0xCF8F
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [a+edi*4]
mov  [d+7*4], edx
mov  [a+edi*4], ebx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4880: ; EXT		0x4880-0x4887
and  edi, byte 7
mov  edx, [d+edi*4]
movsx    dx, dl
mov      [d+edi*4], dx
test     dx, dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48C0: ; EXT		0x48C0-0x48C7
and  edi, byte 7
mov  edx, [d+edi*4]
movsx    edx, dx
mov      [d+edi*4], edx
test     edx, edx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AFC: ; ILLEGAL		0x4AFC
jmp  near exception_illegal_instruction
times ($$-$) & 3 nop
I4ED0: ; JMP		0x4ED0-0x4ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EE8: ; JMP		0x4EE8-0x4EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EF0: ; JMP		0x4EF0-0x4EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EF8: ; JMP		0x4EF8
movsx    ebx, word [esi]
add      esi, byte 2
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EF9: ; JMP		0x4EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EFA: ; JMP		0x4EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EFB: ; JMP		0x4EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E90: ; JSR		0x4E90-0x4E97
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EA8: ; JSR		0x4EA8-0x4EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EB0: ; JSR		0x4EB0-0x4EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EB8: ; JSR		0x4EB8
movsx    ebx, word [esi]
add      esi, byte 2
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EB9: ; JSR		0x4EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EBA: ; JSR		0x4EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4EBB: ; JSR		0x4EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  edx, esi
sub  edx, ebp
mov  [run_edx], edx
mov  esi, ebx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
mov  edx, [run_edx]
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41D0: ; LEA		0x41D0-0x41D7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+0*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41E8: ; LEA		0x41E8-0x41EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+0*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41F0: ; LEA		0x41F0-0x41F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+0*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41F8: ; LEA		0x41F8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+0*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41F9: ; LEA		0x41F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+0*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41FA: ; LEA		0x41FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+0*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I41FB: ; LEA		0x41FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+0*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43D0: ; LEA		0x43D0-0x43D7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+1*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43E8: ; LEA		0x43E8-0x43EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+1*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43F0: ; LEA		0x43F0-0x43F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+1*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43F8: ; LEA		0x43F8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+1*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43F9: ; LEA		0x43F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+1*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43FA: ; LEA		0x43FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+1*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I43FB: ; LEA		0x43FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+1*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45D0: ; LEA		0x45D0-0x45D7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+2*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45E8: ; LEA		0x45E8-0x45EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+2*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45F0: ; LEA		0x45F0-0x45F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+2*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45F8: ; LEA		0x45F8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+2*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45F9: ; LEA		0x45F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+2*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45FA: ; LEA		0x45FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+2*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I45FB: ; LEA		0x45FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+2*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47D0: ; LEA		0x47D0-0x47D7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+3*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47E8: ; LEA		0x47E8-0x47EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+3*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47F0: ; LEA		0x47F0-0x47F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+3*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47F8: ; LEA		0x47F8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+3*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47F9: ; LEA		0x47F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+3*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47FA: ; LEA		0x47FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+3*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I47FB: ; LEA		0x47FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+3*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49D0: ; LEA		0x49D0-0x49D7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+4*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49E8: ; LEA		0x49E8-0x49EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+4*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49F0: ; LEA		0x49F0-0x49F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+4*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49F8: ; LEA		0x49F8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+4*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49F9: ; LEA		0x49F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+4*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49FA: ; LEA		0x49FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+4*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I49FB: ; LEA		0x49FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+4*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BD0: ; LEA		0x4BD0-0x4BD7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+5*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BE8: ; LEA		0x4BE8-0x4BEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+5*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BF0: ; LEA		0x4BF0-0x4BF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+5*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BF8: ; LEA		0x4BF8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+5*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BF9: ; LEA		0x4BF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+5*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BFA: ; LEA		0x4BFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+5*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4BFB: ; LEA		0x4BFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+5*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DD0: ; LEA		0x4DD0-0x4DD7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+6*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DE8: ; LEA		0x4DE8-0x4DEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+6*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DF0: ; LEA		0x4DF0-0x4DF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+6*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DF8: ; LEA		0x4DF8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+6*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DF9: ; LEA		0x4DF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+6*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DFA: ; LEA		0x4DFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+6*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4DFB: ; LEA		0x4DFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+6*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FD0: ; LEA		0x4FD0-0x4FD7
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [a+7*4], ebx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FE8: ; LEA		0x4FE8-0x4FEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [a+7*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FF0: ; LEA		0x4FF0-0x4FF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [a+7*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FF8: ; LEA		0x4FF8
movsx    ebx, word [esi]
add      esi, byte 2
mov  [a+7*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FF9: ; LEA		0x4FF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [a+7*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FFA: ; LEA		0x4FFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+7*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4FFB: ; LEA		0x4FFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [a+7*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E50: ; LINK		0x4E50-0x4E57
and  edi, byte 7
mov  ebx, [a+7*4]
sub  ebx, byte 4
mov  edx, [a+edi*4]
mov  [run_ebx], ebx
mov  [run_edi], edi
call WriteLong
mov  edi, [run_edi]
mov  ebx, [run_ebx]
mov  [a+edi*4], ebx
movsx    edx, word [esi]
add      ebx, edx
mov      [a+7*4], ebx
add      esi, byte 2
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE008: ; LSR		0xE008-0xE00F
and  edi, byte 7
shr   byte [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE028: ; LSR		0xE028-0xE02F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE048: ; LSR		0xE048-0xE04F
and  edi, byte 7
shr   word [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE068: ; LSR		0xE068-0xE06F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE088: ; LSR		0xE088-0xE08F
and  edi, byte 7
shr   dword [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0A8: ; LSR		0xE0A8-0xE0AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE108: ; LSL		0xE108-0xE10F
and  edi, byte 7
shl   byte [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE128: ; LSL		0xE128-0xE12F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE148: ; LSL		0xE148-0xE14F
and  edi, byte 7
shl   word [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE168: ; LSL		0xE168-0xE16F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE188: ; LSL		0xE188-0xE18F
and  edi, byte 7
shl   dword [d+edi*4], 8
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1A8: ; LSL		0xE1A8-0xE1AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE208: ; LSR		0xE208-0xE20F
and  edi, byte 7
shr   byte [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE228: ; LSR		0xE228-0xE22F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE248: ; LSR		0xE248-0xE24F
and  edi, byte 7
shr   word [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE268: ; LSR		0xE268-0xE26F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE288: ; LSR		0xE288-0xE28F
and  edi, byte 7
shr   dword [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2A8: ; LSR		0xE2A8-0xE2AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE308: ; LSL		0xE308-0xE30F
and  edi, byte 7
shl   byte [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE328: ; LSL		0xE328-0xE32F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE348: ; LSL		0xE348-0xE34F
and  edi, byte 7
shl   word [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE368: ; LSL		0xE368-0xE36F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE388: ; LSL		0xE388-0xE38F
and  edi, byte 7
shl   dword [d+edi*4], 1
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3A8: ; LSL		0xE3A8-0xE3AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE408: ; LSR		0xE408-0xE40F
and  edi, byte 7
shr   byte [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE428: ; LSR		0xE428-0xE42F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE448: ; LSR		0xE448-0xE44F
and  edi, byte 7
shr   word [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE468: ; LSR		0xE468-0xE46F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE488: ; LSR		0xE488-0xE48F
and  edi, byte 7
shr   dword [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4A8: ; LSR		0xE4A8-0xE4AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE508: ; LSL		0xE508-0xE50F
and  edi, byte 7
shl   byte [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE528: ; LSL		0xE528-0xE52F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE548: ; LSL		0xE548-0xE54F
and  edi, byte 7
shl   word [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE568: ; LSL		0xE568-0xE56F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE588: ; LSL		0xE588-0xE58F
and  edi, byte 7
shl   dword [d+edi*4], 2
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5A8: ; LSL		0xE5A8-0xE5AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE608: ; LSR		0xE608-0xE60F
and  edi, byte 7
shr   byte [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE628: ; LSR		0xE628-0xE62F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE648: ; LSR		0xE648-0xE64F
and  edi, byte 7
shr   word [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE668: ; LSR		0xE668-0xE66F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE688: ; LSR		0xE688-0xE68F
and  edi, byte 7
shr   dword [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6A8: ; LSR		0xE6A8-0xE6AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE708: ; LSL		0xE708-0xE70F
and  edi, byte 7
shl   byte [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE728: ; LSL		0xE728-0xE72F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE748: ; LSL		0xE748-0xE74F
and  edi, byte 7
shl   word [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE768: ; LSL		0xE768-0xE76F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE788: ; LSL		0xE788-0xE78F
and  edi, byte 7
shl   dword [d+edi*4], 3
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7A8: ; LSL		0xE7A8-0xE7AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE808: ; LSR		0xE808-0xE80F
and  edi, byte 7
shr   byte [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE828: ; LSR		0xE828-0xE82F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE848: ; LSR		0xE848-0xE84F
and  edi, byte 7
shr   word [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE868: ; LSR		0xE868-0xE86F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE888: ; LSR		0xE888-0xE88F
and  edi, byte 7
shr   dword [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE8A8: ; LSR		0xE8A8-0xE8AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE908: ; LSL		0xE908-0xE90F
and  edi, byte 7
shl   byte [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE928: ; LSL		0xE928-0xE92F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE948: ; LSL		0xE948-0xE94F
and  edi, byte 7
shl   word [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE968: ; LSL		0xE968-0xE96F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE988: ; LSL		0xE988-0xE98F
and  edi, byte 7
shl   dword [d+edi*4], 4
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE9A8: ; LSL		0xE9A8-0xE9AF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA08: ; LSR		0xEA08-0xEA0F
and  edi, byte 7
shr   byte [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA28: ; LSR		0xEA28-0xEA2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA48: ; LSR		0xEA48-0xEA4F
and  edi, byte 7
shr   word [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA68: ; LSR		0xEA68-0xEA6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA88: ; LSR		0xEA88-0xEA8F
and  edi, byte 7
shr   dword [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEAA8: ; LSR		0xEAA8-0xEAAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB08: ; LSL		0xEB08-0xEB0F
and  edi, byte 7
shl   byte [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB28: ; LSL		0xEB28-0xEB2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB48: ; LSL		0xEB48-0xEB4F
and  edi, byte 7
shl   word [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB68: ; LSL		0xEB68-0xEB6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB88: ; LSL		0xEB88-0xEB8F
and  edi, byte 7
shl   dword [d+edi*4], 5
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEBA8: ; LSL		0xEBA8-0xEBAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC08: ; LSR		0xEC08-0xEC0F
and  edi, byte 7
shr   byte [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC28: ; LSR		0xEC28-0xEC2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC48: ; LSR		0xEC48-0xEC4F
and  edi, byte 7
shr   word [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC68: ; LSR		0xEC68-0xEC6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC88: ; LSR		0xEC88-0xEC8F
and  edi, byte 7
shr   dword [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IECA8: ; LSR		0xECA8-0xECAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED08: ; LSL		0xED08-0xED0F
and  edi, byte 7
shl   byte [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED28: ; LSL		0xED28-0xED2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED48: ; LSL		0xED48-0xED4F
and  edi, byte 7
shl   word [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED68: ; LSL		0xED68-0xED6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED88: ; LSL		0xED88-0xED8F
and  edi, byte 7
shl   dword [d+edi*4], 6
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEDA8: ; LSL		0xEDA8-0xEDAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE08: ; LSR		0xEE08-0xEE0F
and  edi, byte 7
shr   byte [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE28: ; LSR		0xEE28-0xEE2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE48: ; LSR		0xEE48-0xEE4F
and  edi, byte 7
shr   word [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE68: ; LSR		0xEE68-0xEE6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE88: ; LSR		0xEE88-0xEE8F
and  edi, byte 7
shr   dword [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEEA8: ; LSR		0xEEA8-0xEEAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shr   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF08: ; LSL		0xEF08-0xEF0F
and  edi, byte 7
shl   byte [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF28: ; LSL		0xEF28-0xEF2F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dl, cl
setc al
test dl, dl
mov  [d+edi*4], dl
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF48: ; LSL		0xEF48-0xEF4F
and  edi, byte 7
shl   word [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF68: ; LSL		0xEF68-0xEF6F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   dx, cl
setc al
test dx, dx
mov  [d+edi*4], dx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF88: ; LSL		0xEF88-0xEF8F
and  edi, byte 7
shl   dword [d+edi*4], 7
lahf
setc byte [x]
xor  al, al
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEFA8: ; LSL		0xEFA8-0xEFAF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
shl   edx, cl
setc al
test edx, edx
mov  [d+edi*4], edx
lahf
test cl, cl
jz   short .zero
or   ah, al
test al, 1
setnz    byte [x]
.zero:
xor  al, al
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2D0: ; LSR		0xE2D0-0xE2D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2D8: ; LSR		0xE2D8-0xE2DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2E0: ; LSR		0xE2E0-0xE2E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2E8: ; LSR		0xE2E8-0xE2EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2F0: ; LSR		0xE2F0-0xE2F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2F8: ; LSR		0xE2F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2F9: ; LSR		0xE2F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
shr  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3D0: ; LSL		0xE3D0-0xE3D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3D8: ; LSL		0xE3D8-0xE3DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3E0: ; LSL		0xE3E0-0xE3E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3E8: ; LSL		0xE3E8-0xE3EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3F0: ; LSL		0xE3F0-0xE3F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3F8: ; LSL		0xE3F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3F9: ; LSL		0xE3F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
shl  dx, byte 1
lahf
setc byte [x]
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1000: ; MOVE		0x1000-0x1007
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1010: ; MOVE		0x1010-0x1017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1018: ; MOVE		0x1018-0x101F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1020: ; MOVE		0x1020-0x1027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1028: ; MOVE		0x1028-0x102F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1030: ; MOVE		0x1030-0x1037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1038: ; MOVE		0x1038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1039: ; MOVE		0x1039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I103A: ; MOVE		0x103A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I103B: ; MOVE		0x103B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I103C: ; MOVE		0x103C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+0*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1080: ; MOVE		0x1080-0x1087
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1090: ; MOVE		0x1090-0x1097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1098: ; MOVE		0x1098-0x109F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10A0: ; MOVE		0x10A0-0x10A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10A8: ; MOVE		0x10A8-0x10AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10B0: ; MOVE		0x10B0-0x10B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10B8: ; MOVE		0x10B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10B9: ; MOVE		0x10B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10BA: ; MOVE		0x10BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10BB: ; MOVE		0x10BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10BC: ; MOVE		0x10BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10C0: ; MOVE		0x10C0-0x10C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10D0: ; MOVE		0x10D0-0x10D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10D8: ; MOVE		0x10D8-0x10DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10E0: ; MOVE		0x10E0-0x10E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10E8: ; MOVE		0x10E8-0x10EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10F0: ; MOVE		0x10F0-0x10F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10F8: ; MOVE		0x10F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10F9: ; MOVE		0x10F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10FA: ; MOVE		0x10FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10FB: ; MOVE		0x10FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I10FC: ; MOVE		0x10FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+0*4]
inc  dword [a+0*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1100: ; MOVE		0x1100-0x1107
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1110: ; MOVE		0x1110-0x1117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1118: ; MOVE		0x1118-0x111F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1120: ; MOVE		0x1120-0x1127
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1128: ; MOVE		0x1128-0x112F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1130: ; MOVE		0x1130-0x1137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1138: ; MOVE		0x1138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1139: ; MOVE		0x1139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I113A: ; MOVE		0x113A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I113B: ; MOVE		0x113B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I113C: ; MOVE		0x113C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+0*4]
mov  ebx, [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1140: ; MOVE		0x1140-0x1147
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1150: ; MOVE		0x1150-0x1157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1158: ; MOVE		0x1158-0x115F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1160: ; MOVE		0x1160-0x1167
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1168: ; MOVE		0x1168-0x116F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1170: ; MOVE		0x1170-0x1177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1178: ; MOVE		0x1178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1179: ; MOVE		0x1179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I117A: ; MOVE		0x117A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I117B: ; MOVE		0x117B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I117C: ; MOVE		0x117C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1180: ; MOVE		0x1180-0x1187
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1190: ; MOVE		0x1190-0x1197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1198: ; MOVE		0x1198-0x119F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11A0: ; MOVE		0x11A0-0x11A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11A8: ; MOVE		0x11A8-0x11AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11B0: ; MOVE		0x11B0-0x11B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11B8: ; MOVE		0x11B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11B9: ; MOVE		0x11B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11BA: ; MOVE		0x11BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11BB: ; MOVE		0x11BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11BC: ; MOVE		0x11BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11C0: ; MOVE		0x11C0-0x11C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11D0: ; MOVE		0x11D0-0x11D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11D8: ; MOVE		0x11D8-0x11DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11E0: ; MOVE		0x11E0-0x11E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11E8: ; MOVE		0x11E8-0x11EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11F0: ; MOVE		0x11F0-0x11F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11F8: ; MOVE		0x11F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11F9: ; MOVE		0x11F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11FA: ; MOVE		0x11FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11FB: ; MOVE		0x11FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I11FC: ; MOVE		0x11FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1200: ; MOVE		0x1200-0x1207
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1210: ; MOVE		0x1210-0x1217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1218: ; MOVE		0x1218-0x121F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1220: ; MOVE		0x1220-0x1227
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1228: ; MOVE		0x1228-0x122F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1230: ; MOVE		0x1230-0x1237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1238: ; MOVE		0x1238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1239: ; MOVE		0x1239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I123A: ; MOVE		0x123A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I123B: ; MOVE		0x123B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I123C: ; MOVE		0x123C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+1*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1280: ; MOVE		0x1280-0x1287
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1290: ; MOVE		0x1290-0x1297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1298: ; MOVE		0x1298-0x129F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12A0: ; MOVE		0x12A0-0x12A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12A8: ; MOVE		0x12A8-0x12AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12B0: ; MOVE		0x12B0-0x12B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12B8: ; MOVE		0x12B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12B9: ; MOVE		0x12B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12BA: ; MOVE		0x12BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12BB: ; MOVE		0x12BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12BC: ; MOVE		0x12BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12C0: ; MOVE		0x12C0-0x12C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12D0: ; MOVE		0x12D0-0x12D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12D8: ; MOVE		0x12D8-0x12DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12E0: ; MOVE		0x12E0-0x12E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12E8: ; MOVE		0x12E8-0x12EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12F0: ; MOVE		0x12F0-0x12F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12F8: ; MOVE		0x12F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12F9: ; MOVE		0x12F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12FA: ; MOVE		0x12FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12FB: ; MOVE		0x12FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I12FC: ; MOVE		0x12FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+1*4]
inc  dword [a+1*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1300: ; MOVE		0x1300-0x1307
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1310: ; MOVE		0x1310-0x1317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1318: ; MOVE		0x1318-0x131F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1320: ; MOVE		0x1320-0x1327
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1328: ; MOVE		0x1328-0x132F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1330: ; MOVE		0x1330-0x1337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1338: ; MOVE		0x1338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1339: ; MOVE		0x1339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I133A: ; MOVE		0x133A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I133B: ; MOVE		0x133B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I133C: ; MOVE		0x133C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+1*4]
mov  ebx, [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1340: ; MOVE		0x1340-0x1347
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1350: ; MOVE		0x1350-0x1357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1358: ; MOVE		0x1358-0x135F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1360: ; MOVE		0x1360-0x1367
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1368: ; MOVE		0x1368-0x136F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1370: ; MOVE		0x1370-0x1377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1378: ; MOVE		0x1378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1379: ; MOVE		0x1379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I137A: ; MOVE		0x137A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I137B: ; MOVE		0x137B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I137C: ; MOVE		0x137C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1380: ; MOVE		0x1380-0x1387
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1390: ; MOVE		0x1390-0x1397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1398: ; MOVE		0x1398-0x139F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13A0: ; MOVE		0x13A0-0x13A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13A8: ; MOVE		0x13A8-0x13AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13B0: ; MOVE		0x13B0-0x13B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13B8: ; MOVE		0x13B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13B9: ; MOVE		0x13B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13BA: ; MOVE		0x13BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13BB: ; MOVE		0x13BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13BC: ; MOVE		0x13BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13C0: ; MOVE		0x13C0-0x13C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13D0: ; MOVE		0x13D0-0x13D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13D8: ; MOVE		0x13D8-0x13DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13E0: ; MOVE		0x13E0-0x13E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13E8: ; MOVE		0x13E8-0x13EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13F0: ; MOVE		0x13F0-0x13F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13F8: ; MOVE		0x13F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13F9: ; MOVE		0x13F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13FA: ; MOVE		0x13FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13FB: ; MOVE		0x13FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I13FC: ; MOVE		0x13FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1400: ; MOVE		0x1400-0x1407
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1410: ; MOVE		0x1410-0x1417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1418: ; MOVE		0x1418-0x141F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1420: ; MOVE		0x1420-0x1427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1428: ; MOVE		0x1428-0x142F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1430: ; MOVE		0x1430-0x1437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1438: ; MOVE		0x1438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1439: ; MOVE		0x1439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I143A: ; MOVE		0x143A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I143B: ; MOVE		0x143B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I143C: ; MOVE		0x143C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+2*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1480: ; MOVE		0x1480-0x1487
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1490: ; MOVE		0x1490-0x1497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1498: ; MOVE		0x1498-0x149F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14A0: ; MOVE		0x14A0-0x14A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14A8: ; MOVE		0x14A8-0x14AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14B0: ; MOVE		0x14B0-0x14B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14B8: ; MOVE		0x14B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14B9: ; MOVE		0x14B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14BA: ; MOVE		0x14BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14BB: ; MOVE		0x14BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14BC: ; MOVE		0x14BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14C0: ; MOVE		0x14C0-0x14C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14D0: ; MOVE		0x14D0-0x14D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14D8: ; MOVE		0x14D8-0x14DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14E0: ; MOVE		0x14E0-0x14E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14E8: ; MOVE		0x14E8-0x14EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14F0: ; MOVE		0x14F0-0x14F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14F8: ; MOVE		0x14F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14F9: ; MOVE		0x14F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14FA: ; MOVE		0x14FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14FB: ; MOVE		0x14FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I14FC: ; MOVE		0x14FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+2*4]
inc  dword [a+2*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1500: ; MOVE		0x1500-0x1507
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1510: ; MOVE		0x1510-0x1517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1518: ; MOVE		0x1518-0x151F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1520: ; MOVE		0x1520-0x1527
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1528: ; MOVE		0x1528-0x152F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1530: ; MOVE		0x1530-0x1537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1538: ; MOVE		0x1538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1539: ; MOVE		0x1539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I153A: ; MOVE		0x153A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I153B: ; MOVE		0x153B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I153C: ; MOVE		0x153C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+2*4]
mov  ebx, [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1540: ; MOVE		0x1540-0x1547
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1550: ; MOVE		0x1550-0x1557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1558: ; MOVE		0x1558-0x155F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1560: ; MOVE		0x1560-0x1567
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1568: ; MOVE		0x1568-0x156F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1570: ; MOVE		0x1570-0x1577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1578: ; MOVE		0x1578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1579: ; MOVE		0x1579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I157A: ; MOVE		0x157A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I157B: ; MOVE		0x157B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I157C: ; MOVE		0x157C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1580: ; MOVE		0x1580-0x1587
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1590: ; MOVE		0x1590-0x1597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1598: ; MOVE		0x1598-0x159F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15A0: ; MOVE		0x15A0-0x15A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15A8: ; MOVE		0x15A8-0x15AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15B0: ; MOVE		0x15B0-0x15B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15B8: ; MOVE		0x15B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15B9: ; MOVE		0x15B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15BA: ; MOVE		0x15BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15BB: ; MOVE		0x15BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I15BC: ; MOVE		0x15BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1600: ; MOVE		0x1600-0x1607
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1610: ; MOVE		0x1610-0x1617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1618: ; MOVE		0x1618-0x161F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1620: ; MOVE		0x1620-0x1627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1628: ; MOVE		0x1628-0x162F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1630: ; MOVE		0x1630-0x1637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1638: ; MOVE		0x1638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1639: ; MOVE		0x1639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I163A: ; MOVE		0x163A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I163B: ; MOVE		0x163B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I163C: ; MOVE		0x163C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+3*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1680: ; MOVE		0x1680-0x1687
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1690: ; MOVE		0x1690-0x1697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1698: ; MOVE		0x1698-0x169F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16A0: ; MOVE		0x16A0-0x16A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16A8: ; MOVE		0x16A8-0x16AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16B0: ; MOVE		0x16B0-0x16B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16B8: ; MOVE		0x16B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16B9: ; MOVE		0x16B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16BA: ; MOVE		0x16BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16BB: ; MOVE		0x16BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16BC: ; MOVE		0x16BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16C0: ; MOVE		0x16C0-0x16C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16D0: ; MOVE		0x16D0-0x16D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16D8: ; MOVE		0x16D8-0x16DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16E0: ; MOVE		0x16E0-0x16E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16E8: ; MOVE		0x16E8-0x16EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16F0: ; MOVE		0x16F0-0x16F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16F8: ; MOVE		0x16F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16F9: ; MOVE		0x16F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16FA: ; MOVE		0x16FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16FB: ; MOVE		0x16FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I16FC: ; MOVE		0x16FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+3*4]
inc  dword [a+3*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1700: ; MOVE		0x1700-0x1707
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1710: ; MOVE		0x1710-0x1717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1718: ; MOVE		0x1718-0x171F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1720: ; MOVE		0x1720-0x1727
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1728: ; MOVE		0x1728-0x172F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1730: ; MOVE		0x1730-0x1737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1738: ; MOVE		0x1738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1739: ; MOVE		0x1739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I173A: ; MOVE		0x173A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I173B: ; MOVE		0x173B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I173C: ; MOVE		0x173C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+3*4]
mov  ebx, [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1740: ; MOVE		0x1740-0x1747
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1750: ; MOVE		0x1750-0x1757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1758: ; MOVE		0x1758-0x175F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1760: ; MOVE		0x1760-0x1767
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1768: ; MOVE		0x1768-0x176F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1770: ; MOVE		0x1770-0x1777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1778: ; MOVE		0x1778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1779: ; MOVE		0x1779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I177A: ; MOVE		0x177A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I177B: ; MOVE		0x177B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I177C: ; MOVE		0x177C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1780: ; MOVE		0x1780-0x1787
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1790: ; MOVE		0x1790-0x1797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1798: ; MOVE		0x1798-0x179F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17A0: ; MOVE		0x17A0-0x17A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17A8: ; MOVE		0x17A8-0x17AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17B0: ; MOVE		0x17B0-0x17B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17B8: ; MOVE		0x17B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17B9: ; MOVE		0x17B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17BA: ; MOVE		0x17BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17BB: ; MOVE		0x17BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I17BC: ; MOVE		0x17BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1800: ; MOVE		0x1800-0x1807
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1810: ; MOVE		0x1810-0x1817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1818: ; MOVE		0x1818-0x181F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1820: ; MOVE		0x1820-0x1827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1828: ; MOVE		0x1828-0x182F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1830: ; MOVE		0x1830-0x1837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1838: ; MOVE		0x1838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1839: ; MOVE		0x1839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I183A: ; MOVE		0x183A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I183B: ; MOVE		0x183B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I183C: ; MOVE		0x183C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+4*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1880: ; MOVE		0x1880-0x1887
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1890: ; MOVE		0x1890-0x1897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1898: ; MOVE		0x1898-0x189F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18A0: ; MOVE		0x18A0-0x18A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18A8: ; MOVE		0x18A8-0x18AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18B0: ; MOVE		0x18B0-0x18B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18B8: ; MOVE		0x18B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18B9: ; MOVE		0x18B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18BA: ; MOVE		0x18BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18BB: ; MOVE		0x18BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18BC: ; MOVE		0x18BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18C0: ; MOVE		0x18C0-0x18C7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18D0: ; MOVE		0x18D0-0x18D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18D8: ; MOVE		0x18D8-0x18DF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18E0: ; MOVE		0x18E0-0x18E7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18E8: ; MOVE		0x18E8-0x18EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18F0: ; MOVE		0x18F0-0x18F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18F8: ; MOVE		0x18F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18F9: ; MOVE		0x18F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18FA: ; MOVE		0x18FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18FB: ; MOVE		0x18FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I18FC: ; MOVE		0x18FC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+4*4]
inc  dword [a+4*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1900: ; MOVE		0x1900-0x1907
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1910: ; MOVE		0x1910-0x1917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1918: ; MOVE		0x1918-0x191F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1920: ; MOVE		0x1920-0x1927
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1928: ; MOVE		0x1928-0x192F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1930: ; MOVE		0x1930-0x1937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1938: ; MOVE		0x1938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1939: ; MOVE		0x1939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I193A: ; MOVE		0x193A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I193B: ; MOVE		0x193B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I193C: ; MOVE		0x193C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+4*4]
mov  ebx, [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1940: ; MOVE		0x1940-0x1947
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1950: ; MOVE		0x1950-0x1957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1958: ; MOVE		0x1958-0x195F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1960: ; MOVE		0x1960-0x1967
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1968: ; MOVE		0x1968-0x196F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1970: ; MOVE		0x1970-0x1977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1978: ; MOVE		0x1978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1979: ; MOVE		0x1979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I197A: ; MOVE		0x197A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I197B: ; MOVE		0x197B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I197C: ; MOVE		0x197C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1980: ; MOVE		0x1980-0x1987
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1990: ; MOVE		0x1990-0x1997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1998: ; MOVE		0x1998-0x199F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19A0: ; MOVE		0x19A0-0x19A7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19A8: ; MOVE		0x19A8-0x19AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19B0: ; MOVE		0x19B0-0x19B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19B8: ; MOVE		0x19B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19B9: ; MOVE		0x19B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19BA: ; MOVE		0x19BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19BB: ; MOVE		0x19BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I19BC: ; MOVE		0x19BC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A00: ; MOVE		0x1A00-0x1A07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A10: ; MOVE		0x1A10-0x1A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A18: ; MOVE		0x1A18-0x1A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A20: ; MOVE		0x1A20-0x1A27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A28: ; MOVE		0x1A28-0x1A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A30: ; MOVE		0x1A30-0x1A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A38: ; MOVE		0x1A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A39: ; MOVE		0x1A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A3A: ; MOVE		0x1A3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A3B: ; MOVE		0x1A3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A3C: ; MOVE		0x1A3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+5*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A80: ; MOVE		0x1A80-0x1A87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A90: ; MOVE		0x1A90-0x1A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1A98: ; MOVE		0x1A98-0x1A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AA0: ; MOVE		0x1AA0-0x1AA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AA8: ; MOVE		0x1AA8-0x1AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AB0: ; MOVE		0x1AB0-0x1AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AB8: ; MOVE		0x1AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AB9: ; MOVE		0x1AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1ABA: ; MOVE		0x1ABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1ABB: ; MOVE		0x1ABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1ABC: ; MOVE		0x1ABC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AC0: ; MOVE		0x1AC0-0x1AC7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AD0: ; MOVE		0x1AD0-0x1AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AD8: ; MOVE		0x1AD8-0x1ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AE0: ; MOVE		0x1AE0-0x1AE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AE8: ; MOVE		0x1AE8-0x1AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AF0: ; MOVE		0x1AF0-0x1AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AF8: ; MOVE		0x1AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AF9: ; MOVE		0x1AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AFA: ; MOVE		0x1AFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AFB: ; MOVE		0x1AFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1AFC: ; MOVE		0x1AFC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+5*4]
inc  dword [a+5*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B00: ; MOVE		0x1B00-0x1B07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B10: ; MOVE		0x1B10-0x1B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B18: ; MOVE		0x1B18-0x1B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B20: ; MOVE		0x1B20-0x1B27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B28: ; MOVE		0x1B28-0x1B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B30: ; MOVE		0x1B30-0x1B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B38: ; MOVE		0x1B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B39: ; MOVE		0x1B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B3A: ; MOVE		0x1B3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B3B: ; MOVE		0x1B3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B3C: ; MOVE		0x1B3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+5*4]
mov  ebx, [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B40: ; MOVE		0x1B40-0x1B47
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B50: ; MOVE		0x1B50-0x1B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B58: ; MOVE		0x1B58-0x1B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B60: ; MOVE		0x1B60-0x1B67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B68: ; MOVE		0x1B68-0x1B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B70: ; MOVE		0x1B70-0x1B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B78: ; MOVE		0x1B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B79: ; MOVE		0x1B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B7A: ; MOVE		0x1B7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B7B: ; MOVE		0x1B7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B7C: ; MOVE		0x1B7C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B80: ; MOVE		0x1B80-0x1B87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B90: ; MOVE		0x1B90-0x1B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1B98: ; MOVE		0x1B98-0x1B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BA0: ; MOVE		0x1BA0-0x1BA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BA8: ; MOVE		0x1BA8-0x1BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BB0: ; MOVE		0x1BB0-0x1BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BB8: ; MOVE		0x1BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BB9: ; MOVE		0x1BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BBA: ; MOVE		0x1BBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BBB: ; MOVE		0x1BBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1BBC: ; MOVE		0x1BBC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C00: ; MOVE		0x1C00-0x1C07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C10: ; MOVE		0x1C10-0x1C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C18: ; MOVE		0x1C18-0x1C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C20: ; MOVE		0x1C20-0x1C27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C28: ; MOVE		0x1C28-0x1C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C30: ; MOVE		0x1C30-0x1C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C38: ; MOVE		0x1C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C39: ; MOVE		0x1C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C3A: ; MOVE		0x1C3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C3B: ; MOVE		0x1C3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C3C: ; MOVE		0x1C3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+6*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C80: ; MOVE		0x1C80-0x1C87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C90: ; MOVE		0x1C90-0x1C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1C98: ; MOVE		0x1C98-0x1C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CA0: ; MOVE		0x1CA0-0x1CA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CA8: ; MOVE		0x1CA8-0x1CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CB0: ; MOVE		0x1CB0-0x1CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CB8: ; MOVE		0x1CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CB9: ; MOVE		0x1CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CBA: ; MOVE		0x1CBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CBB: ; MOVE		0x1CBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CBC: ; MOVE		0x1CBC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CC0: ; MOVE		0x1CC0-0x1CC7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CD0: ; MOVE		0x1CD0-0x1CD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CD8: ; MOVE		0x1CD8-0x1CDF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CE0: ; MOVE		0x1CE0-0x1CE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CE8: ; MOVE		0x1CE8-0x1CEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CF0: ; MOVE		0x1CF0-0x1CF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CF8: ; MOVE		0x1CF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CF9: ; MOVE		0x1CF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CFA: ; MOVE		0x1CFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CFB: ; MOVE		0x1CFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1CFC: ; MOVE		0x1CFC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+6*4]
inc  dword [a+6*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D00: ; MOVE		0x1D00-0x1D07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D10: ; MOVE		0x1D10-0x1D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D18: ; MOVE		0x1D18-0x1D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D20: ; MOVE		0x1D20-0x1D27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D28: ; MOVE		0x1D28-0x1D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D30: ; MOVE		0x1D30-0x1D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D38: ; MOVE		0x1D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D39: ; MOVE		0x1D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D3A: ; MOVE		0x1D3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D3B: ; MOVE		0x1D3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D3C: ; MOVE		0x1D3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
dec  dword [a+6*4]
mov  ebx, [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D40: ; MOVE		0x1D40-0x1D47
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D50: ; MOVE		0x1D50-0x1D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D58: ; MOVE		0x1D58-0x1D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D60: ; MOVE		0x1D60-0x1D67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D68: ; MOVE		0x1D68-0x1D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D70: ; MOVE		0x1D70-0x1D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D78: ; MOVE		0x1D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D79: ; MOVE		0x1D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D7A: ; MOVE		0x1D7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D7B: ; MOVE		0x1D7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D7C: ; MOVE		0x1D7C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D80: ; MOVE		0x1D80-0x1D87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D90: ; MOVE		0x1D90-0x1D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1D98: ; MOVE		0x1D98-0x1D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DA0: ; MOVE		0x1DA0-0x1DA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DA8: ; MOVE		0x1DA8-0x1DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DB0: ; MOVE		0x1DB0-0x1DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DB8: ; MOVE		0x1DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DB9: ; MOVE		0x1DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DBA: ; MOVE		0x1DBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DBB: ; MOVE		0x1DBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1DBC: ; MOVE		0x1DBC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E00: ; MOVE		0x1E00-0x1E07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E10: ; MOVE		0x1E10-0x1E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E18: ; MOVE		0x1E18-0x1E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E20: ; MOVE		0x1E20-0x1E27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E28: ; MOVE		0x1E28-0x1E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E30: ; MOVE		0x1E30-0x1E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E38: ; MOVE		0x1E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E39: ; MOVE		0x1E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E3A: ; MOVE		0x1E3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E3B: ; MOVE		0x1E3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E3C: ; MOVE		0x1E3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  [d+7*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E80: ; MOVE		0x1E80-0x1E87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E90: ; MOVE		0x1E90-0x1E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1E98: ; MOVE		0x1E98-0x1E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EA0: ; MOVE		0x1EA0-0x1EA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EA8: ; MOVE		0x1EA8-0x1EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EB0: ; MOVE		0x1EB0-0x1EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EB8: ; MOVE		0x1EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EB9: ; MOVE		0x1EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EBA: ; MOVE		0x1EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EBB: ; MOVE		0x1EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EBC: ; MOVE		0x1EBC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EC0: ; MOVE		0x1EC0-0x1EC7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1ED0: ; MOVE		0x1ED0-0x1ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1ED8: ; MOVE		0x1ED8-0x1EDF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EE0: ; MOVE		0x1EE0-0x1EE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EE8: ; MOVE		0x1EE8-0x1EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EF0: ; MOVE		0x1EF0-0x1EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EF8: ; MOVE		0x1EF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EF9: ; MOVE		0x1EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EFA: ; MOVE		0x1EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EFB: ; MOVE		0x1EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1EFC: ; MOVE		0x1EFC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F00: ; MOVE		0x1F00-0x1F07
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F10: ; MOVE		0x1F10-0x1F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F18: ; MOVE		0x1F18-0x1F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F20: ; MOVE		0x1F20-0x1F27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F28: ; MOVE		0x1F28-0x1F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F30: ; MOVE		0x1F30-0x1F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F38: ; MOVE		0x1F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F39: ; MOVE		0x1F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F3A: ; MOVE		0x1F3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F3B: ; MOVE		0x1F3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F3C: ; MOVE		0x1F3C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F40: ; MOVE		0x1F40-0x1F47
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F50: ; MOVE		0x1F50-0x1F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F58: ; MOVE		0x1F58-0x1F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F60: ; MOVE		0x1F60-0x1F67
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F68: ; MOVE		0x1F68-0x1F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F70: ; MOVE		0x1F70-0x1F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F78: ; MOVE		0x1F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F79: ; MOVE		0x1F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F7A: ; MOVE		0x1F7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F7B: ; MOVE		0x1F7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F7C: ; MOVE		0x1F7C
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F80: ; MOVE		0x1F80-0x1F87
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F90: ; MOVE		0x1F90-0x1F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1F98: ; MOVE		0x1F98-0x1F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FA0: ; MOVE		0x1FA0-0x1FA7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FA8: ; MOVE		0x1FA8-0x1FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FB0: ; MOVE		0x1FB0-0x1FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FB8: ; MOVE		0x1FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FB9: ; MOVE		0x1FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FBA: ; MOVE		0x1FBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FBB: ; MOVE		0x1FBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I1FBC: ; MOVE		0x1FBC
mov  dl, [esi]
add  esi, byte 2
test dl, dl
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2000: ; MOVE		0x2000-0x2007
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2008: ; MOVE		0x2008-0x200F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2010: ; MOVE		0x2010-0x2017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2018: ; MOVE		0x2018-0x201F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2020: ; MOVE		0x2020-0x2027
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2028: ; MOVE		0x2028-0x202F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2030: ; MOVE		0x2030-0x2037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2038: ; MOVE		0x2038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2039: ; MOVE		0x2039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I203A: ; MOVE		0x203A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I203B: ; MOVE		0x203B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I203C: ; MOVE		0x203C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2080: ; MOVE		0x2080-0x2087
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2088: ; MOVE		0x2088-0x208F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2090: ; MOVE		0x2090-0x2097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2098: ; MOVE		0x2098-0x209F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20A0: ; MOVE		0x20A0-0x20A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20A8: ; MOVE		0x20A8-0x20AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20B0: ; MOVE		0x20B0-0x20B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20B8: ; MOVE		0x20B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20B9: ; MOVE		0x20B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20BA: ; MOVE		0x20BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20BB: ; MOVE		0x20BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20BC: ; MOVE		0x20BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20C0: ; MOVE		0x20C0-0x20C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20C8: ; MOVE		0x20C8-0x20CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20D0: ; MOVE		0x20D0-0x20D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20D8: ; MOVE		0x20D8-0x20DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20E0: ; MOVE		0x20E0-0x20E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20E8: ; MOVE		0x20E8-0x20EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20F0: ; MOVE		0x20F0-0x20F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20F8: ; MOVE		0x20F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20F9: ; MOVE		0x20F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20FA: ; MOVE		0x20FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20FB: ; MOVE		0x20FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I20FC: ; MOVE		0x20FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2100: ; MOVE		0x2100-0x2107
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2108: ; MOVE		0x2108-0x210F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2110: ; MOVE		0x2110-0x2117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2118: ; MOVE		0x2118-0x211F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2120: ; MOVE		0x2120-0x2127
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2128: ; MOVE		0x2128-0x212F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2130: ; MOVE		0x2130-0x2137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2138: ; MOVE		0x2138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2139: ; MOVE		0x2139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I213A: ; MOVE		0x213A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I213B: ; MOVE		0x213B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I213C: ; MOVE		0x213C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+0*4], byte 4
mov  ebx, [a+0*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2140: ; MOVE		0x2140-0x2147
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2148: ; MOVE		0x2148-0x214F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2150: ; MOVE		0x2150-0x2157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2158: ; MOVE		0x2158-0x215F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2160: ; MOVE		0x2160-0x2167
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2168: ; MOVE		0x2168-0x216F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2170: ; MOVE		0x2170-0x2177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2178: ; MOVE		0x2178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2179: ; MOVE		0x2179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I217A: ; MOVE		0x217A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I217B: ; MOVE		0x217B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I217C: ; MOVE		0x217C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2180: ; MOVE		0x2180-0x2187
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2188: ; MOVE		0x2188-0x218F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2190: ; MOVE		0x2190-0x2197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2198: ; MOVE		0x2198-0x219F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21A0: ; MOVE		0x21A0-0x21A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21A8: ; MOVE		0x21A8-0x21AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21B0: ; MOVE		0x21B0-0x21B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21B8: ; MOVE		0x21B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21B9: ; MOVE		0x21B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21BA: ; MOVE		0x21BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21BB: ; MOVE		0x21BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21BC: ; MOVE		0x21BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21C0: ; MOVE		0x21C0-0x21C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21C8: ; MOVE		0x21C8-0x21CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21D0: ; MOVE		0x21D0-0x21D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21D8: ; MOVE		0x21D8-0x21DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21E0: ; MOVE		0x21E0-0x21E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21E8: ; MOVE		0x21E8-0x21EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21F0: ; MOVE		0x21F0-0x21F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21F8: ; MOVE		0x21F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21F9: ; MOVE		0x21F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21FA: ; MOVE		0x21FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21FB: ; MOVE		0x21FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I21FC: ; MOVE		0x21FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2200: ; MOVE		0x2200-0x2207
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2208: ; MOVE		0x2208-0x220F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2210: ; MOVE		0x2210-0x2217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2218: ; MOVE		0x2218-0x221F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2220: ; MOVE		0x2220-0x2227
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2228: ; MOVE		0x2228-0x222F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2230: ; MOVE		0x2230-0x2237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2238: ; MOVE		0x2238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2239: ; MOVE		0x2239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I223A: ; MOVE		0x223A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I223B: ; MOVE		0x223B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I223C: ; MOVE		0x223C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2280: ; MOVE		0x2280-0x2287
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2288: ; MOVE		0x2288-0x228F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2290: ; MOVE		0x2290-0x2297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2298: ; MOVE		0x2298-0x229F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22A0: ; MOVE		0x22A0-0x22A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22A8: ; MOVE		0x22A8-0x22AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22B0: ; MOVE		0x22B0-0x22B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22B8: ; MOVE		0x22B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22B9: ; MOVE		0x22B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22BA: ; MOVE		0x22BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22BB: ; MOVE		0x22BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22BC: ; MOVE		0x22BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22C0: ; MOVE		0x22C0-0x22C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22C8: ; MOVE		0x22C8-0x22CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22D0: ; MOVE		0x22D0-0x22D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22D8: ; MOVE		0x22D8-0x22DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22E0: ; MOVE		0x22E0-0x22E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22E8: ; MOVE		0x22E8-0x22EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22F0: ; MOVE		0x22F0-0x22F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22F8: ; MOVE		0x22F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22F9: ; MOVE		0x22F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22FA: ; MOVE		0x22FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22FB: ; MOVE		0x22FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I22FC: ; MOVE		0x22FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2300: ; MOVE		0x2300-0x2307
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2308: ; MOVE		0x2308-0x230F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2310: ; MOVE		0x2310-0x2317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2318: ; MOVE		0x2318-0x231F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2320: ; MOVE		0x2320-0x2327
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2328: ; MOVE		0x2328-0x232F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2330: ; MOVE		0x2330-0x2337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2338: ; MOVE		0x2338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2339: ; MOVE		0x2339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I233A: ; MOVE		0x233A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I233B: ; MOVE		0x233B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I233C: ; MOVE		0x233C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+1*4], byte 4
mov  ebx, [a+1*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2340: ; MOVE		0x2340-0x2347
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2348: ; MOVE		0x2348-0x234F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2350: ; MOVE		0x2350-0x2357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2358: ; MOVE		0x2358-0x235F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2360: ; MOVE		0x2360-0x2367
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2368: ; MOVE		0x2368-0x236F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2370: ; MOVE		0x2370-0x2377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2378: ; MOVE		0x2378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2379: ; MOVE		0x2379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I237A: ; MOVE		0x237A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I237B: ; MOVE		0x237B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I237C: ; MOVE		0x237C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2380: ; MOVE		0x2380-0x2387
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2388: ; MOVE		0x2388-0x238F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2390: ; MOVE		0x2390-0x2397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2398: ; MOVE		0x2398-0x239F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23A0: ; MOVE		0x23A0-0x23A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23A8: ; MOVE		0x23A8-0x23AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23B0: ; MOVE		0x23B0-0x23B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23B8: ; MOVE		0x23B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23B9: ; MOVE		0x23B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23BA: ; MOVE		0x23BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23BB: ; MOVE		0x23BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23BC: ; MOVE		0x23BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23C0: ; MOVE		0x23C0-0x23C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23C8: ; MOVE		0x23C8-0x23CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23D0: ; MOVE		0x23D0-0x23D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23D8: ; MOVE		0x23D8-0x23DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23E0: ; MOVE		0x23E0-0x23E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23E8: ; MOVE		0x23E8-0x23EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23F0: ; MOVE		0x23F0-0x23F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23F8: ; MOVE		0x23F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23F9: ; MOVE		0x23F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 36
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23FA: ; MOVE		0x23FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23FB: ; MOVE		0x23FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I23FC: ; MOVE		0x23FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2400: ; MOVE		0x2400-0x2407
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2408: ; MOVE		0x2408-0x240F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2410: ; MOVE		0x2410-0x2417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2418: ; MOVE		0x2418-0x241F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2420: ; MOVE		0x2420-0x2427
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2428: ; MOVE		0x2428-0x242F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2430: ; MOVE		0x2430-0x2437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2438: ; MOVE		0x2438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2439: ; MOVE		0x2439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I243A: ; MOVE		0x243A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I243B: ; MOVE		0x243B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I243C: ; MOVE		0x243C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2480: ; MOVE		0x2480-0x2487
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2488: ; MOVE		0x2488-0x248F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2490: ; MOVE		0x2490-0x2497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2498: ; MOVE		0x2498-0x249F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24A0: ; MOVE		0x24A0-0x24A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24A8: ; MOVE		0x24A8-0x24AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24B0: ; MOVE		0x24B0-0x24B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24B8: ; MOVE		0x24B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24B9: ; MOVE		0x24B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24BA: ; MOVE		0x24BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24BB: ; MOVE		0x24BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24BC: ; MOVE		0x24BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24C0: ; MOVE		0x24C0-0x24C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24C8: ; MOVE		0x24C8-0x24CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24D0: ; MOVE		0x24D0-0x24D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24D8: ; MOVE		0x24D8-0x24DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24E0: ; MOVE		0x24E0-0x24E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24E8: ; MOVE		0x24E8-0x24EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24F0: ; MOVE		0x24F0-0x24F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24F8: ; MOVE		0x24F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24F9: ; MOVE		0x24F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24FA: ; MOVE		0x24FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24FB: ; MOVE		0x24FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I24FC: ; MOVE		0x24FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2500: ; MOVE		0x2500-0x2507
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2508: ; MOVE		0x2508-0x250F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2510: ; MOVE		0x2510-0x2517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2518: ; MOVE		0x2518-0x251F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2520: ; MOVE		0x2520-0x2527
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2528: ; MOVE		0x2528-0x252F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2530: ; MOVE		0x2530-0x2537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2538: ; MOVE		0x2538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2539: ; MOVE		0x2539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I253A: ; MOVE		0x253A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I253B: ; MOVE		0x253B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I253C: ; MOVE		0x253C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+2*4], byte 4
mov  ebx, [a+2*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2540: ; MOVE		0x2540-0x2547
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2548: ; MOVE		0x2548-0x254F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2550: ; MOVE		0x2550-0x2557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2558: ; MOVE		0x2558-0x255F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2560: ; MOVE		0x2560-0x2567
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2568: ; MOVE		0x2568-0x256F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2570: ; MOVE		0x2570-0x2577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2578: ; MOVE		0x2578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2579: ; MOVE		0x2579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I257A: ; MOVE		0x257A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I257B: ; MOVE		0x257B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I257C: ; MOVE		0x257C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2580: ; MOVE		0x2580-0x2587
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2588: ; MOVE		0x2588-0x258F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2590: ; MOVE		0x2590-0x2597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2598: ; MOVE		0x2598-0x259F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25A0: ; MOVE		0x25A0-0x25A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25A8: ; MOVE		0x25A8-0x25AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25B0: ; MOVE		0x25B0-0x25B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25B8: ; MOVE		0x25B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25B9: ; MOVE		0x25B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25BA: ; MOVE		0x25BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25BB: ; MOVE		0x25BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I25BC: ; MOVE		0x25BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2600: ; MOVE		0x2600-0x2607
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2608: ; MOVE		0x2608-0x260F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2610: ; MOVE		0x2610-0x2617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2618: ; MOVE		0x2618-0x261F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2620: ; MOVE		0x2620-0x2627
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2628: ; MOVE		0x2628-0x262F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2630: ; MOVE		0x2630-0x2637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2638: ; MOVE		0x2638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2639: ; MOVE		0x2639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I263A: ; MOVE		0x263A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I263B: ; MOVE		0x263B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I263C: ; MOVE		0x263C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2680: ; MOVE		0x2680-0x2687
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2688: ; MOVE		0x2688-0x268F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2690: ; MOVE		0x2690-0x2697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2698: ; MOVE		0x2698-0x269F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26A0: ; MOVE		0x26A0-0x26A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26A8: ; MOVE		0x26A8-0x26AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26B0: ; MOVE		0x26B0-0x26B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26B8: ; MOVE		0x26B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26B9: ; MOVE		0x26B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26BA: ; MOVE		0x26BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26BB: ; MOVE		0x26BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26BC: ; MOVE		0x26BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26C0: ; MOVE		0x26C0-0x26C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26C8: ; MOVE		0x26C8-0x26CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26D0: ; MOVE		0x26D0-0x26D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26D8: ; MOVE		0x26D8-0x26DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26E0: ; MOVE		0x26E0-0x26E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26E8: ; MOVE		0x26E8-0x26EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26F0: ; MOVE		0x26F0-0x26F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26F8: ; MOVE		0x26F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26F9: ; MOVE		0x26F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26FA: ; MOVE		0x26FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26FB: ; MOVE		0x26FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I26FC: ; MOVE		0x26FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2700: ; MOVE		0x2700-0x2707
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2708: ; MOVE		0x2708-0x270F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2710: ; MOVE		0x2710-0x2717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2718: ; MOVE		0x2718-0x271F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2720: ; MOVE		0x2720-0x2727
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2728: ; MOVE		0x2728-0x272F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2730: ; MOVE		0x2730-0x2737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2738: ; MOVE		0x2738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2739: ; MOVE		0x2739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I273A: ; MOVE		0x273A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I273B: ; MOVE		0x273B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I273C: ; MOVE		0x273C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+3*4], byte 4
mov  ebx, [a+3*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2740: ; MOVE		0x2740-0x2747
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2748: ; MOVE		0x2748-0x274F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2750: ; MOVE		0x2750-0x2757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2758: ; MOVE		0x2758-0x275F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2760: ; MOVE		0x2760-0x2767
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2768: ; MOVE		0x2768-0x276F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2770: ; MOVE		0x2770-0x2777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2778: ; MOVE		0x2778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2779: ; MOVE		0x2779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I277A: ; MOVE		0x277A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I277B: ; MOVE		0x277B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I277C: ; MOVE		0x277C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2780: ; MOVE		0x2780-0x2787
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2788: ; MOVE		0x2788-0x278F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2790: ; MOVE		0x2790-0x2797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2798: ; MOVE		0x2798-0x279F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27A0: ; MOVE		0x27A0-0x27A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27A8: ; MOVE		0x27A8-0x27AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27B0: ; MOVE		0x27B0-0x27B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27B8: ; MOVE		0x27B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27B9: ; MOVE		0x27B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27BA: ; MOVE		0x27BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27BB: ; MOVE		0x27BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I27BC: ; MOVE		0x27BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2800: ; MOVE		0x2800-0x2807
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2808: ; MOVE		0x2808-0x280F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2810: ; MOVE		0x2810-0x2817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2818: ; MOVE		0x2818-0x281F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2820: ; MOVE		0x2820-0x2827
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2828: ; MOVE		0x2828-0x282F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2830: ; MOVE		0x2830-0x2837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2838: ; MOVE		0x2838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2839: ; MOVE		0x2839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I283A: ; MOVE		0x283A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I283B: ; MOVE		0x283B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I283C: ; MOVE		0x283C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2880: ; MOVE		0x2880-0x2887
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2888: ; MOVE		0x2888-0x288F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2890: ; MOVE		0x2890-0x2897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2898: ; MOVE		0x2898-0x289F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28A0: ; MOVE		0x28A0-0x28A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28A8: ; MOVE		0x28A8-0x28AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28B0: ; MOVE		0x28B0-0x28B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28B8: ; MOVE		0x28B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28B9: ; MOVE		0x28B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28BA: ; MOVE		0x28BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28BB: ; MOVE		0x28BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28BC: ; MOVE		0x28BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28C0: ; MOVE		0x28C0-0x28C7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28C8: ; MOVE		0x28C8-0x28CF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28D0: ; MOVE		0x28D0-0x28D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28D8: ; MOVE		0x28D8-0x28DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28E0: ; MOVE		0x28E0-0x28E7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28E8: ; MOVE		0x28E8-0x28EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28F0: ; MOVE		0x28F0-0x28F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28F8: ; MOVE		0x28F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28F9: ; MOVE		0x28F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28FA: ; MOVE		0x28FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28FB: ; MOVE		0x28FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I28FC: ; MOVE		0x28FC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2900: ; MOVE		0x2900-0x2907
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2908: ; MOVE		0x2908-0x290F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2910: ; MOVE		0x2910-0x2917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2918: ; MOVE		0x2918-0x291F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2920: ; MOVE		0x2920-0x2927
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2928: ; MOVE		0x2928-0x292F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2930: ; MOVE		0x2930-0x2937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2938: ; MOVE		0x2938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2939: ; MOVE		0x2939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I293A: ; MOVE		0x293A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I293B: ; MOVE		0x293B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I293C: ; MOVE		0x293C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+4*4], byte 4
mov  ebx, [a+4*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2940: ; MOVE		0x2940-0x2947
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2948: ; MOVE		0x2948-0x294F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2950: ; MOVE		0x2950-0x2957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2958: ; MOVE		0x2958-0x295F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2960: ; MOVE		0x2960-0x2967
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2968: ; MOVE		0x2968-0x296F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2970: ; MOVE		0x2970-0x2977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2978: ; MOVE		0x2978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2979: ; MOVE		0x2979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I297A: ; MOVE		0x297A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I297B: ; MOVE		0x297B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I297C: ; MOVE		0x297C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2980: ; MOVE		0x2980-0x2987
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2988: ; MOVE		0x2988-0x298F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2990: ; MOVE		0x2990-0x2997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2998: ; MOVE		0x2998-0x299F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29A0: ; MOVE		0x29A0-0x29A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29A8: ; MOVE		0x29A8-0x29AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29B0: ; MOVE		0x29B0-0x29B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29B8: ; MOVE		0x29B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29B9: ; MOVE		0x29B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29BA: ; MOVE		0x29BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29BB: ; MOVE		0x29BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I29BC: ; MOVE		0x29BC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A00: ; MOVE		0x2A00-0x2A07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A08: ; MOVE		0x2A08-0x2A0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A10: ; MOVE		0x2A10-0x2A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A18: ; MOVE		0x2A18-0x2A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A20: ; MOVE		0x2A20-0x2A27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A28: ; MOVE		0x2A28-0x2A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A30: ; MOVE		0x2A30-0x2A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A38: ; MOVE		0x2A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A39: ; MOVE		0x2A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A3A: ; MOVE		0x2A3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A3B: ; MOVE		0x2A3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A3C: ; MOVE		0x2A3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A80: ; MOVE		0x2A80-0x2A87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A88: ; MOVE		0x2A88-0x2A8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A90: ; MOVE		0x2A90-0x2A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A98: ; MOVE		0x2A98-0x2A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AA0: ; MOVE		0x2AA0-0x2AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AA8: ; MOVE		0x2AA8-0x2AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AB0: ; MOVE		0x2AB0-0x2AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AB8: ; MOVE		0x2AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AB9: ; MOVE		0x2AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2ABA: ; MOVE		0x2ABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2ABB: ; MOVE		0x2ABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2ABC: ; MOVE		0x2ABC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AC0: ; MOVE		0x2AC0-0x2AC7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AC8: ; MOVE		0x2AC8-0x2ACF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AD0: ; MOVE		0x2AD0-0x2AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AD8: ; MOVE		0x2AD8-0x2ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AE0: ; MOVE		0x2AE0-0x2AE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AE8: ; MOVE		0x2AE8-0x2AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AF0: ; MOVE		0x2AF0-0x2AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AF8: ; MOVE		0x2AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AF9: ; MOVE		0x2AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AFA: ; MOVE		0x2AFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AFB: ; MOVE		0x2AFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2AFC: ; MOVE		0x2AFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B00: ; MOVE		0x2B00-0x2B07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B08: ; MOVE		0x2B08-0x2B0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B10: ; MOVE		0x2B10-0x2B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B18: ; MOVE		0x2B18-0x2B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B20: ; MOVE		0x2B20-0x2B27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B28: ; MOVE		0x2B28-0x2B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B30: ; MOVE		0x2B30-0x2B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B38: ; MOVE		0x2B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B39: ; MOVE		0x2B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B3A: ; MOVE		0x2B3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B3B: ; MOVE		0x2B3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B3C: ; MOVE		0x2B3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+5*4], byte 4
mov  ebx, [a+5*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B40: ; MOVE		0x2B40-0x2B47
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B48: ; MOVE		0x2B48-0x2B4F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B50: ; MOVE		0x2B50-0x2B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B58: ; MOVE		0x2B58-0x2B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B60: ; MOVE		0x2B60-0x2B67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B68: ; MOVE		0x2B68-0x2B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B70: ; MOVE		0x2B70-0x2B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B78: ; MOVE		0x2B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B79: ; MOVE		0x2B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B7A: ; MOVE		0x2B7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B7B: ; MOVE		0x2B7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B7C: ; MOVE		0x2B7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B80: ; MOVE		0x2B80-0x2B87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B88: ; MOVE		0x2B88-0x2B8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B90: ; MOVE		0x2B90-0x2B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2B98: ; MOVE		0x2B98-0x2B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BA0: ; MOVE		0x2BA0-0x2BA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BA8: ; MOVE		0x2BA8-0x2BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BB0: ; MOVE		0x2BB0-0x2BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BB8: ; MOVE		0x2BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BB9: ; MOVE		0x2BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BBA: ; MOVE		0x2BBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BBB: ; MOVE		0x2BBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2BBC: ; MOVE		0x2BBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C00: ; MOVE		0x2C00-0x2C07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C08: ; MOVE		0x2C08-0x2C0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C10: ; MOVE		0x2C10-0x2C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C18: ; MOVE		0x2C18-0x2C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C20: ; MOVE		0x2C20-0x2C27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C28: ; MOVE		0x2C28-0x2C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C30: ; MOVE		0x2C30-0x2C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C38: ; MOVE		0x2C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C39: ; MOVE		0x2C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C3A: ; MOVE		0x2C3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C3B: ; MOVE		0x2C3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C3C: ; MOVE		0x2C3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C80: ; MOVE		0x2C80-0x2C87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C88: ; MOVE		0x2C88-0x2C8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C90: ; MOVE		0x2C90-0x2C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C98: ; MOVE		0x2C98-0x2C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CA0: ; MOVE		0x2CA0-0x2CA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CA8: ; MOVE		0x2CA8-0x2CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CB0: ; MOVE		0x2CB0-0x2CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CB8: ; MOVE		0x2CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CB9: ; MOVE		0x2CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CBA: ; MOVE		0x2CBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CBB: ; MOVE		0x2CBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CBC: ; MOVE		0x2CBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CC0: ; MOVE		0x2CC0-0x2CC7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CC8: ; MOVE		0x2CC8-0x2CCF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CD0: ; MOVE		0x2CD0-0x2CD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CD8: ; MOVE		0x2CD8-0x2CDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CE0: ; MOVE		0x2CE0-0x2CE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CE8: ; MOVE		0x2CE8-0x2CEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CF0: ; MOVE		0x2CF0-0x2CF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CF8: ; MOVE		0x2CF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CF9: ; MOVE		0x2CF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CFA: ; MOVE		0x2CFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CFB: ; MOVE		0x2CFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2CFC: ; MOVE		0x2CFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D00: ; MOVE		0x2D00-0x2D07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D08: ; MOVE		0x2D08-0x2D0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D10: ; MOVE		0x2D10-0x2D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D18: ; MOVE		0x2D18-0x2D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D20: ; MOVE		0x2D20-0x2D27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D28: ; MOVE		0x2D28-0x2D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D30: ; MOVE		0x2D30-0x2D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D38: ; MOVE		0x2D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D39: ; MOVE		0x2D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D3A: ; MOVE		0x2D3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D3B: ; MOVE		0x2D3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D3C: ; MOVE		0x2D3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+6*4], byte 4
mov  ebx, [a+6*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D40: ; MOVE		0x2D40-0x2D47
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D48: ; MOVE		0x2D48-0x2D4F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D50: ; MOVE		0x2D50-0x2D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D58: ; MOVE		0x2D58-0x2D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D60: ; MOVE		0x2D60-0x2D67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D68: ; MOVE		0x2D68-0x2D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D70: ; MOVE		0x2D70-0x2D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D78: ; MOVE		0x2D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D79: ; MOVE		0x2D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D7A: ; MOVE		0x2D7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D7B: ; MOVE		0x2D7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D7C: ; MOVE		0x2D7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D80: ; MOVE		0x2D80-0x2D87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D88: ; MOVE		0x2D88-0x2D8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D90: ; MOVE		0x2D90-0x2D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2D98: ; MOVE		0x2D98-0x2D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DA0: ; MOVE		0x2DA0-0x2DA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DA8: ; MOVE		0x2DA8-0x2DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DB0: ; MOVE		0x2DB0-0x2DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DB8: ; MOVE		0x2DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DB9: ; MOVE		0x2DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DBA: ; MOVE		0x2DBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DBB: ; MOVE		0x2DBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2DBC: ; MOVE		0x2DBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E00: ; MOVE		0x2E00-0x2E07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E08: ; MOVE		0x2E08-0x2E0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E10: ; MOVE		0x2E10-0x2E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E18: ; MOVE		0x2E18-0x2E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E20: ; MOVE		0x2E20-0x2E27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E28: ; MOVE		0x2E28-0x2E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E30: ; MOVE		0x2E30-0x2E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E38: ; MOVE		0x2E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E39: ; MOVE		0x2E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E3A: ; MOVE		0x2E3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E3B: ; MOVE		0x2E3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E3C: ; MOVE		0x2E3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E80: ; MOVE		0x2E80-0x2E87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E88: ; MOVE		0x2E88-0x2E8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E90: ; MOVE		0x2E90-0x2E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E98: ; MOVE		0x2E98-0x2E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EA0: ; MOVE		0x2EA0-0x2EA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EA8: ; MOVE		0x2EA8-0x2EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EB0: ; MOVE		0x2EB0-0x2EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EB8: ; MOVE		0x2EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EB9: ; MOVE		0x2EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EBA: ; MOVE		0x2EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EBB: ; MOVE		0x2EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EBC: ; MOVE		0x2EBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EC0: ; MOVE		0x2EC0-0x2EC7
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EC8: ; MOVE		0x2EC8-0x2ECF
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2ED0: ; MOVE		0x2ED0-0x2ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2ED8: ; MOVE		0x2ED8-0x2EDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EE0: ; MOVE		0x2EE0-0x2EE7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EE8: ; MOVE		0x2EE8-0x2EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EF0: ; MOVE		0x2EF0-0x2EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EF8: ; MOVE		0x2EF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EF9: ; MOVE		0x2EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EFA: ; MOVE		0x2EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EFB: ; MOVE		0x2EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2EFC: ; MOVE		0x2EFC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 4
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F00: ; MOVE		0x2F00-0x2F07
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F08: ; MOVE		0x2F08-0x2F0F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F10: ; MOVE		0x2F10-0x2F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F18: ; MOVE		0x2F18-0x2F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F20: ; MOVE		0x2F20-0x2F27
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F28: ; MOVE		0x2F28-0x2F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F30: ; MOVE		0x2F30-0x2F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F38: ; MOVE		0x2F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F39: ; MOVE		0x2F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F3A: ; MOVE		0x2F3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F3B: ; MOVE		0x2F3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F3C: ; MOVE		0x2F3C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
sub  dword [a+7*4], byte 4
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F40: ; MOVE		0x2F40-0x2F47
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F48: ; MOVE		0x2F48-0x2F4F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F50: ; MOVE		0x2F50-0x2F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F58: ; MOVE		0x2F58-0x2F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F60: ; MOVE		0x2F60-0x2F67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F68: ; MOVE		0x2F68-0x2F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F70: ; MOVE		0x2F70-0x2F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F78: ; MOVE		0x2F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F79: ; MOVE		0x2F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F7A: ; MOVE		0x2F7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F7B: ; MOVE		0x2F7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F7C: ; MOVE		0x2F7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F80: ; MOVE		0x2F80-0x2F87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F88: ; MOVE		0x2F88-0x2F8F
and  edi, byte 7
mov  edx, [a+edi*4]
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F90: ; MOVE		0x2F90-0x2F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2F98: ; MOVE		0x2F98-0x2F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FA0: ; MOVE		0x2FA0-0x2FA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FA8: ; MOVE		0x2FA8-0x2FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FB0: ; MOVE		0x2FB0-0x2FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FB8: ; MOVE		0x2FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FB9: ; MOVE		0x2FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FBA: ; MOVE		0x2FBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FBB: ; MOVE		0x2FBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 32
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2FBC: ; MOVE		0x2FBC
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
test edx, edx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3000: ; MOVE		0x3000-0x3007
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3008: ; MOVE		0x3008-0x300F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3010: ; MOVE		0x3010-0x3017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3018: ; MOVE		0x3018-0x301F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3020: ; MOVE		0x3020-0x3027
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3028: ; MOVE		0x3028-0x302F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3030: ; MOVE		0x3030-0x3037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3038: ; MOVE		0x3038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3039: ; MOVE		0x3039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I303A: ; MOVE		0x303A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I303B: ; MOVE		0x303B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I303C: ; MOVE		0x303C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+0*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3080: ; MOVE		0x3080-0x3087
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3088: ; MOVE		0x3088-0x308F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3090: ; MOVE		0x3090-0x3097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3098: ; MOVE		0x3098-0x309F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30A0: ; MOVE		0x30A0-0x30A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30A8: ; MOVE		0x30A8-0x30AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30B0: ; MOVE		0x30B0-0x30B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30B8: ; MOVE		0x30B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30B9: ; MOVE		0x30B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30BA: ; MOVE		0x30BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30BB: ; MOVE		0x30BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30BC: ; MOVE		0x30BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30C0: ; MOVE		0x30C0-0x30C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30C8: ; MOVE		0x30C8-0x30CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30D0: ; MOVE		0x30D0-0x30D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30D8: ; MOVE		0x30D8-0x30DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30E0: ; MOVE		0x30E0-0x30E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30E8: ; MOVE		0x30E8-0x30EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30F0: ; MOVE		0x30F0-0x30F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30F8: ; MOVE		0x30F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30F9: ; MOVE		0x30F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30FA: ; MOVE		0x30FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30FB: ; MOVE		0x30FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I30FC: ; MOVE		0x30FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+0*4]
add  dword [a+0*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3100: ; MOVE		0x3100-0x3107
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3108: ; MOVE		0x3108-0x310F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3110: ; MOVE		0x3110-0x3117
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3118: ; MOVE		0x3118-0x311F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3120: ; MOVE		0x3120-0x3127
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3128: ; MOVE		0x3128-0x312F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3130: ; MOVE		0x3130-0x3137
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3138: ; MOVE		0x3138
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3139: ; MOVE		0x3139
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I313A: ; MOVE		0x313A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I313B: ; MOVE		0x313B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I313C: ; MOVE		0x313C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+0*4], byte 2
mov  ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3140: ; MOVE		0x3140-0x3147
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3148: ; MOVE		0x3148-0x314F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3150: ; MOVE		0x3150-0x3157
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3158: ; MOVE		0x3158-0x315F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3160: ; MOVE		0x3160-0x3167
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3168: ; MOVE		0x3168-0x316F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3170: ; MOVE		0x3170-0x3177
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3178: ; MOVE		0x3178
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3179: ; MOVE		0x3179
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I317A: ; MOVE		0x317A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I317B: ; MOVE		0x317B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I317C: ; MOVE		0x317C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+0*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3180: ; MOVE		0x3180-0x3187
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3188: ; MOVE		0x3188-0x318F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3190: ; MOVE		0x3190-0x3197
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3198: ; MOVE		0x3198-0x319F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31A0: ; MOVE		0x31A0-0x31A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31A8: ; MOVE		0x31A8-0x31AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31B0: ; MOVE		0x31B0-0x31B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31B8: ; MOVE		0x31B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31B9: ; MOVE		0x31B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31BA: ; MOVE		0x31BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31BB: ; MOVE		0x31BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31BC: ; MOVE		0x31BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+0*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31C0: ; MOVE		0x31C0-0x31C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31C8: ; MOVE		0x31C8-0x31CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31D0: ; MOVE		0x31D0-0x31D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31D8: ; MOVE		0x31D8-0x31DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31E0: ; MOVE		0x31E0-0x31E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31E8: ; MOVE		0x31E8-0x31EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31F0: ; MOVE		0x31F0-0x31F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31F8: ; MOVE		0x31F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31F9: ; MOVE		0x31F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31FA: ; MOVE		0x31FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31FB: ; MOVE		0x31FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I31FC: ; MOVE		0x31FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3200: ; MOVE		0x3200-0x3207
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3208: ; MOVE		0x3208-0x320F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3210: ; MOVE		0x3210-0x3217
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3218: ; MOVE		0x3218-0x321F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3220: ; MOVE		0x3220-0x3227
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3228: ; MOVE		0x3228-0x322F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3230: ; MOVE		0x3230-0x3237
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3238: ; MOVE		0x3238
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3239: ; MOVE		0x3239
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I323A: ; MOVE		0x323A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I323B: ; MOVE		0x323B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I323C: ; MOVE		0x323C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+1*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3280: ; MOVE		0x3280-0x3287
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3288: ; MOVE		0x3288-0x328F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3290: ; MOVE		0x3290-0x3297
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3298: ; MOVE		0x3298-0x329F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32A0: ; MOVE		0x32A0-0x32A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32A8: ; MOVE		0x32A8-0x32AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32B0: ; MOVE		0x32B0-0x32B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32B8: ; MOVE		0x32B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32B9: ; MOVE		0x32B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32BA: ; MOVE		0x32BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32BB: ; MOVE		0x32BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32BC: ; MOVE		0x32BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32C0: ; MOVE		0x32C0-0x32C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32C8: ; MOVE		0x32C8-0x32CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32D0: ; MOVE		0x32D0-0x32D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32D8: ; MOVE		0x32D8-0x32DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32E0: ; MOVE		0x32E0-0x32E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32E8: ; MOVE		0x32E8-0x32EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32F0: ; MOVE		0x32F0-0x32F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32F8: ; MOVE		0x32F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32F9: ; MOVE		0x32F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32FA: ; MOVE		0x32FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32FB: ; MOVE		0x32FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I32FC: ; MOVE		0x32FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+1*4]
add  dword [a+1*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3300: ; MOVE		0x3300-0x3307
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3308: ; MOVE		0x3308-0x330F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3310: ; MOVE		0x3310-0x3317
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3318: ; MOVE		0x3318-0x331F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3320: ; MOVE		0x3320-0x3327
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3328: ; MOVE		0x3328-0x332F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3330: ; MOVE		0x3330-0x3337
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3338: ; MOVE		0x3338
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3339: ; MOVE		0x3339
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I333A: ; MOVE		0x333A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I333B: ; MOVE		0x333B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I333C: ; MOVE		0x333C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+1*4], byte 2
mov  ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3340: ; MOVE		0x3340-0x3347
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3348: ; MOVE		0x3348-0x334F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3350: ; MOVE		0x3350-0x3357
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3358: ; MOVE		0x3358-0x335F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3360: ; MOVE		0x3360-0x3367
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3368: ; MOVE		0x3368-0x336F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3370: ; MOVE		0x3370-0x3377
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3378: ; MOVE		0x3378
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3379: ; MOVE		0x3379
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I337A: ; MOVE		0x337A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I337B: ; MOVE		0x337B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I337C: ; MOVE		0x337C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+1*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3380: ; MOVE		0x3380-0x3387
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3388: ; MOVE		0x3388-0x338F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3390: ; MOVE		0x3390-0x3397
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3398: ; MOVE		0x3398-0x339F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33A0: ; MOVE		0x33A0-0x33A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33A8: ; MOVE		0x33A8-0x33AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33B0: ; MOVE		0x33B0-0x33B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33B8: ; MOVE		0x33B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33B9: ; MOVE		0x33B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33BA: ; MOVE		0x33BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33BB: ; MOVE		0x33BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33BC: ; MOVE		0x33BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+1*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33C0: ; MOVE		0x33C0-0x33C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33C8: ; MOVE		0x33C8-0x33CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33D0: ; MOVE		0x33D0-0x33D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33D8: ; MOVE		0x33D8-0x33DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33E0: ; MOVE		0x33E0-0x33E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33E8: ; MOVE		0x33E8-0x33EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33F0: ; MOVE		0x33F0-0x33F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33F8: ; MOVE		0x33F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33F9: ; MOVE		0x33F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33FA: ; MOVE		0x33FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33FB: ; MOVE		0x33FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I33FC: ; MOVE		0x33FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3400: ; MOVE		0x3400-0x3407
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3408: ; MOVE		0x3408-0x340F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3410: ; MOVE		0x3410-0x3417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3418: ; MOVE		0x3418-0x341F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3420: ; MOVE		0x3420-0x3427
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3428: ; MOVE		0x3428-0x342F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3430: ; MOVE		0x3430-0x3437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3438: ; MOVE		0x3438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3439: ; MOVE		0x3439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I343A: ; MOVE		0x343A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I343B: ; MOVE		0x343B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I343C: ; MOVE		0x343C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+2*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3480: ; MOVE		0x3480-0x3487
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3488: ; MOVE		0x3488-0x348F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3490: ; MOVE		0x3490-0x3497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3498: ; MOVE		0x3498-0x349F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34A0: ; MOVE		0x34A0-0x34A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34A8: ; MOVE		0x34A8-0x34AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34B0: ; MOVE		0x34B0-0x34B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34B8: ; MOVE		0x34B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34B9: ; MOVE		0x34B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34BA: ; MOVE		0x34BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34BB: ; MOVE		0x34BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34BC: ; MOVE		0x34BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34C0: ; MOVE		0x34C0-0x34C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34C8: ; MOVE		0x34C8-0x34CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34D0: ; MOVE		0x34D0-0x34D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34D8: ; MOVE		0x34D8-0x34DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34E0: ; MOVE		0x34E0-0x34E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34E8: ; MOVE		0x34E8-0x34EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34F0: ; MOVE		0x34F0-0x34F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34F8: ; MOVE		0x34F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34F9: ; MOVE		0x34F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34FA: ; MOVE		0x34FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34FB: ; MOVE		0x34FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I34FC: ; MOVE		0x34FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+2*4]
add  dword [a+2*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3500: ; MOVE		0x3500-0x3507
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3508: ; MOVE		0x3508-0x350F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3510: ; MOVE		0x3510-0x3517
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3518: ; MOVE		0x3518-0x351F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3520: ; MOVE		0x3520-0x3527
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3528: ; MOVE		0x3528-0x352F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3530: ; MOVE		0x3530-0x3537
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3538: ; MOVE		0x3538
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3539: ; MOVE		0x3539
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I353A: ; MOVE		0x353A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I353B: ; MOVE		0x353B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I353C: ; MOVE		0x353C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+2*4], byte 2
mov  ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3540: ; MOVE		0x3540-0x3547
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3548: ; MOVE		0x3548-0x354F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3550: ; MOVE		0x3550-0x3557
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3558: ; MOVE		0x3558-0x355F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3560: ; MOVE		0x3560-0x3567
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3568: ; MOVE		0x3568-0x356F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3570: ; MOVE		0x3570-0x3577
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3578: ; MOVE		0x3578
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3579: ; MOVE		0x3579
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I357A: ; MOVE		0x357A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I357B: ; MOVE		0x357B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I357C: ; MOVE		0x357C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+2*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3580: ; MOVE		0x3580-0x3587
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3588: ; MOVE		0x3588-0x358F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3590: ; MOVE		0x3590-0x3597
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3598: ; MOVE		0x3598-0x359F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35A0: ; MOVE		0x35A0-0x35A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35A8: ; MOVE		0x35A8-0x35AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35B0: ; MOVE		0x35B0-0x35B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35B8: ; MOVE		0x35B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35B9: ; MOVE		0x35B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35BA: ; MOVE		0x35BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35BB: ; MOVE		0x35BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I35BC: ; MOVE		0x35BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+2*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3600: ; MOVE		0x3600-0x3607
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3608: ; MOVE		0x3608-0x360F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3610: ; MOVE		0x3610-0x3617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3618: ; MOVE		0x3618-0x361F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3620: ; MOVE		0x3620-0x3627
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3628: ; MOVE		0x3628-0x362F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3630: ; MOVE		0x3630-0x3637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3638: ; MOVE		0x3638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3639: ; MOVE		0x3639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I363A: ; MOVE		0x363A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I363B: ; MOVE		0x363B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I363C: ; MOVE		0x363C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+3*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3680: ; MOVE		0x3680-0x3687
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3688: ; MOVE		0x3688-0x368F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3690: ; MOVE		0x3690-0x3697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3698: ; MOVE		0x3698-0x369F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36A0: ; MOVE		0x36A0-0x36A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36A8: ; MOVE		0x36A8-0x36AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36B0: ; MOVE		0x36B0-0x36B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36B8: ; MOVE		0x36B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36B9: ; MOVE		0x36B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36BA: ; MOVE		0x36BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36BB: ; MOVE		0x36BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36BC: ; MOVE		0x36BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36C0: ; MOVE		0x36C0-0x36C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36C8: ; MOVE		0x36C8-0x36CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36D0: ; MOVE		0x36D0-0x36D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36D8: ; MOVE		0x36D8-0x36DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36E0: ; MOVE		0x36E0-0x36E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36E8: ; MOVE		0x36E8-0x36EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36F0: ; MOVE		0x36F0-0x36F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36F8: ; MOVE		0x36F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36F9: ; MOVE		0x36F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36FA: ; MOVE		0x36FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36FB: ; MOVE		0x36FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I36FC: ; MOVE		0x36FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+3*4]
add  dword [a+3*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3700: ; MOVE		0x3700-0x3707
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3708: ; MOVE		0x3708-0x370F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3710: ; MOVE		0x3710-0x3717
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3718: ; MOVE		0x3718-0x371F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3720: ; MOVE		0x3720-0x3727
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3728: ; MOVE		0x3728-0x372F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3730: ; MOVE		0x3730-0x3737
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3738: ; MOVE		0x3738
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3739: ; MOVE		0x3739
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I373A: ; MOVE		0x373A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I373B: ; MOVE		0x373B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I373C: ; MOVE		0x373C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+3*4], byte 2
mov  ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3740: ; MOVE		0x3740-0x3747
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3748: ; MOVE		0x3748-0x374F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3750: ; MOVE		0x3750-0x3757
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3758: ; MOVE		0x3758-0x375F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3760: ; MOVE		0x3760-0x3767
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3768: ; MOVE		0x3768-0x376F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3770: ; MOVE		0x3770-0x3777
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3778: ; MOVE		0x3778
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3779: ; MOVE		0x3779
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I377A: ; MOVE		0x377A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I377B: ; MOVE		0x377B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I377C: ; MOVE		0x377C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+3*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3780: ; MOVE		0x3780-0x3787
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3788: ; MOVE		0x3788-0x378F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3790: ; MOVE		0x3790-0x3797
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3798: ; MOVE		0x3798-0x379F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37A0: ; MOVE		0x37A0-0x37A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37A8: ; MOVE		0x37A8-0x37AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37B0: ; MOVE		0x37B0-0x37B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37B8: ; MOVE		0x37B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37B9: ; MOVE		0x37B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37BA: ; MOVE		0x37BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37BB: ; MOVE		0x37BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I37BC: ; MOVE		0x37BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+3*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3800: ; MOVE		0x3800-0x3807
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3808: ; MOVE		0x3808-0x380F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3810: ; MOVE		0x3810-0x3817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3818: ; MOVE		0x3818-0x381F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3820: ; MOVE		0x3820-0x3827
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3828: ; MOVE		0x3828-0x382F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3830: ; MOVE		0x3830-0x3837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3838: ; MOVE		0x3838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3839: ; MOVE		0x3839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I383A: ; MOVE		0x383A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I383B: ; MOVE		0x383B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I383C: ; MOVE		0x383C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+4*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3880: ; MOVE		0x3880-0x3887
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3888: ; MOVE		0x3888-0x388F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3890: ; MOVE		0x3890-0x3897
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3898: ; MOVE		0x3898-0x389F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38A0: ; MOVE		0x38A0-0x38A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38A8: ; MOVE		0x38A8-0x38AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38B0: ; MOVE		0x38B0-0x38B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38B8: ; MOVE		0x38B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38B9: ; MOVE		0x38B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38BA: ; MOVE		0x38BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38BB: ; MOVE		0x38BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38BC: ; MOVE		0x38BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38C0: ; MOVE		0x38C0-0x38C7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38C8: ; MOVE		0x38C8-0x38CF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38D0: ; MOVE		0x38D0-0x38D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38D8: ; MOVE		0x38D8-0x38DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38E0: ; MOVE		0x38E0-0x38E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38E8: ; MOVE		0x38E8-0x38EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38F0: ; MOVE		0x38F0-0x38F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38F8: ; MOVE		0x38F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38F9: ; MOVE		0x38F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38FA: ; MOVE		0x38FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38FB: ; MOVE		0x38FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I38FC: ; MOVE		0x38FC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+4*4]
add  dword [a+4*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3900: ; MOVE		0x3900-0x3907
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3908: ; MOVE		0x3908-0x390F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3910: ; MOVE		0x3910-0x3917
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3918: ; MOVE		0x3918-0x391F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3920: ; MOVE		0x3920-0x3927
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3928: ; MOVE		0x3928-0x392F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3930: ; MOVE		0x3930-0x3937
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3938: ; MOVE		0x3938
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3939: ; MOVE		0x3939
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I393A: ; MOVE		0x393A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I393B: ; MOVE		0x393B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I393C: ; MOVE		0x393C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+4*4], byte 2
mov  ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3940: ; MOVE		0x3940-0x3947
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3948: ; MOVE		0x3948-0x394F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3950: ; MOVE		0x3950-0x3957
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3958: ; MOVE		0x3958-0x395F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3960: ; MOVE		0x3960-0x3967
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3968: ; MOVE		0x3968-0x396F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3970: ; MOVE		0x3970-0x3977
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3978: ; MOVE		0x3978
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3979: ; MOVE		0x3979
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I397A: ; MOVE		0x397A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I397B: ; MOVE		0x397B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I397C: ; MOVE		0x397C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+4*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3980: ; MOVE		0x3980-0x3987
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3988: ; MOVE		0x3988-0x398F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3990: ; MOVE		0x3990-0x3997
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3998: ; MOVE		0x3998-0x399F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39A0: ; MOVE		0x39A0-0x39A7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39A8: ; MOVE		0x39A8-0x39AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39B0: ; MOVE		0x39B0-0x39B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39B8: ; MOVE		0x39B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39B9: ; MOVE		0x39B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39BA: ; MOVE		0x39BA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39BB: ; MOVE		0x39BB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I39BC: ; MOVE		0x39BC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+4*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A00: ; MOVE		0x3A00-0x3A07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A08: ; MOVE		0x3A08-0x3A0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A10: ; MOVE		0x3A10-0x3A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A18: ; MOVE		0x3A18-0x3A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A20: ; MOVE		0x3A20-0x3A27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A28: ; MOVE		0x3A28-0x3A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A30: ; MOVE		0x3A30-0x3A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A38: ; MOVE		0x3A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A39: ; MOVE		0x3A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A3A: ; MOVE		0x3A3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A3B: ; MOVE		0x3A3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A3C: ; MOVE		0x3A3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+5*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A80: ; MOVE		0x3A80-0x3A87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A88: ; MOVE		0x3A88-0x3A8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A90: ; MOVE		0x3A90-0x3A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A98: ; MOVE		0x3A98-0x3A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AA0: ; MOVE		0x3AA0-0x3AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AA8: ; MOVE		0x3AA8-0x3AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AB0: ; MOVE		0x3AB0-0x3AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AB8: ; MOVE		0x3AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AB9: ; MOVE		0x3AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3ABA: ; MOVE		0x3ABA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3ABB: ; MOVE		0x3ABB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3ABC: ; MOVE		0x3ABC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AC0: ; MOVE		0x3AC0-0x3AC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AC8: ; MOVE		0x3AC8-0x3ACF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AD0: ; MOVE		0x3AD0-0x3AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AD8: ; MOVE		0x3AD8-0x3ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AE0: ; MOVE		0x3AE0-0x3AE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AE8: ; MOVE		0x3AE8-0x3AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AF0: ; MOVE		0x3AF0-0x3AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AF8: ; MOVE		0x3AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AF9: ; MOVE		0x3AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AFA: ; MOVE		0x3AFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AFB: ; MOVE		0x3AFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3AFC: ; MOVE		0x3AFC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+5*4]
add  dword [a+5*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B00: ; MOVE		0x3B00-0x3B07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B08: ; MOVE		0x3B08-0x3B0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B10: ; MOVE		0x3B10-0x3B17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B18: ; MOVE		0x3B18-0x3B1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B20: ; MOVE		0x3B20-0x3B27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B28: ; MOVE		0x3B28-0x3B2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B30: ; MOVE		0x3B30-0x3B37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B38: ; MOVE		0x3B38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B39: ; MOVE		0x3B39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B3A: ; MOVE		0x3B3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B3B: ; MOVE		0x3B3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B3C: ; MOVE		0x3B3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+5*4], byte 2
mov  ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B40: ; MOVE		0x3B40-0x3B47
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B48: ; MOVE		0x3B48-0x3B4F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B50: ; MOVE		0x3B50-0x3B57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B58: ; MOVE		0x3B58-0x3B5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B60: ; MOVE		0x3B60-0x3B67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B68: ; MOVE		0x3B68-0x3B6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B70: ; MOVE		0x3B70-0x3B77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B78: ; MOVE		0x3B78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B79: ; MOVE		0x3B79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B7A: ; MOVE		0x3B7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B7B: ; MOVE		0x3B7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B7C: ; MOVE		0x3B7C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+5*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B80: ; MOVE		0x3B80-0x3B87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B88: ; MOVE		0x3B88-0x3B8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B90: ; MOVE		0x3B90-0x3B97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3B98: ; MOVE		0x3B98-0x3B9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BA0: ; MOVE		0x3BA0-0x3BA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BA8: ; MOVE		0x3BA8-0x3BAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BB0: ; MOVE		0x3BB0-0x3BB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BB8: ; MOVE		0x3BB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BB9: ; MOVE		0x3BB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BBA: ; MOVE		0x3BBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BBB: ; MOVE		0x3BBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3BBC: ; MOVE		0x3BBC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+5*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C00: ; MOVE		0x3C00-0x3C07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C08: ; MOVE		0x3C08-0x3C0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C10: ; MOVE		0x3C10-0x3C17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C18: ; MOVE		0x3C18-0x3C1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C20: ; MOVE		0x3C20-0x3C27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C28: ; MOVE		0x3C28-0x3C2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C30: ; MOVE		0x3C30-0x3C37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C38: ; MOVE		0x3C38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C39: ; MOVE		0x3C39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C3A: ; MOVE		0x3C3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C3B: ; MOVE		0x3C3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C3C: ; MOVE		0x3C3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+6*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C80: ; MOVE		0x3C80-0x3C87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C88: ; MOVE		0x3C88-0x3C8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C90: ; MOVE		0x3C90-0x3C97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C98: ; MOVE		0x3C98-0x3C9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CA0: ; MOVE		0x3CA0-0x3CA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CA8: ; MOVE		0x3CA8-0x3CAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CB0: ; MOVE		0x3CB0-0x3CB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CB8: ; MOVE		0x3CB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CB9: ; MOVE		0x3CB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CBA: ; MOVE		0x3CBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CBB: ; MOVE		0x3CBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CBC: ; MOVE		0x3CBC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CC0: ; MOVE		0x3CC0-0x3CC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CC8: ; MOVE		0x3CC8-0x3CCF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CD0: ; MOVE		0x3CD0-0x3CD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CD8: ; MOVE		0x3CD8-0x3CDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CE0: ; MOVE		0x3CE0-0x3CE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CE8: ; MOVE		0x3CE8-0x3CEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CF0: ; MOVE		0x3CF0-0x3CF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CF8: ; MOVE		0x3CF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CF9: ; MOVE		0x3CF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CFA: ; MOVE		0x3CFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CFB: ; MOVE		0x3CFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3CFC: ; MOVE		0x3CFC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+6*4]
add  dword [a+6*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D00: ; MOVE		0x3D00-0x3D07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D08: ; MOVE		0x3D08-0x3D0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D10: ; MOVE		0x3D10-0x3D17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D18: ; MOVE		0x3D18-0x3D1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D20: ; MOVE		0x3D20-0x3D27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D28: ; MOVE		0x3D28-0x3D2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D30: ; MOVE		0x3D30-0x3D37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D38: ; MOVE		0x3D38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D39: ; MOVE		0x3D39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D3A: ; MOVE		0x3D3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D3B: ; MOVE		0x3D3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D3C: ; MOVE		0x3D3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+6*4], byte 2
mov  ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D40: ; MOVE		0x3D40-0x3D47
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D48: ; MOVE		0x3D48-0x3D4F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D50: ; MOVE		0x3D50-0x3D57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D58: ; MOVE		0x3D58-0x3D5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D60: ; MOVE		0x3D60-0x3D67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D68: ; MOVE		0x3D68-0x3D6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D70: ; MOVE		0x3D70-0x3D77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D78: ; MOVE		0x3D78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D79: ; MOVE		0x3D79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D7A: ; MOVE		0x3D7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D7B: ; MOVE		0x3D7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D7C: ; MOVE		0x3D7C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+6*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D80: ; MOVE		0x3D80-0x3D87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D88: ; MOVE		0x3D88-0x3D8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D90: ; MOVE		0x3D90-0x3D97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3D98: ; MOVE		0x3D98-0x3D9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DA0: ; MOVE		0x3DA0-0x3DA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DA8: ; MOVE		0x3DA8-0x3DAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DB0: ; MOVE		0x3DB0-0x3DB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DB8: ; MOVE		0x3DB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DB9: ; MOVE		0x3DB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DBA: ; MOVE		0x3DBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DBB: ; MOVE		0x3DBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3DBC: ; MOVE		0x3DBC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+6*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E00: ; MOVE		0x3E00-0x3E07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E08: ; MOVE		0x3E08-0x3E0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E10: ; MOVE		0x3E10-0x3E17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E18: ; MOVE		0x3E18-0x3E1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E20: ; MOVE		0x3E20-0x3E27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E28: ; MOVE		0x3E28-0x3E2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E30: ; MOVE		0x3E30-0x3E37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E38: ; MOVE		0x3E38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E39: ; MOVE		0x3E39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E3A: ; MOVE		0x3E3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E3B: ; MOVE		0x3E3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E3C: ; MOVE		0x3E3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  [d+7*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E80: ; MOVE		0x3E80-0x3E87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E88: ; MOVE		0x3E88-0x3E8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E90: ; MOVE		0x3E90-0x3E97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E98: ; MOVE		0x3E98-0x3E9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EA0: ; MOVE		0x3EA0-0x3EA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EA8: ; MOVE		0x3EA8-0x3EAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EB0: ; MOVE		0x3EB0-0x3EB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EB8: ; MOVE		0x3EB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EB9: ; MOVE		0x3EB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EBA: ; MOVE		0x3EBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EBB: ; MOVE		0x3EBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EBC: ; MOVE		0x3EBC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EC0: ; MOVE		0x3EC0-0x3EC7
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EC8: ; MOVE		0x3EC8-0x3ECF
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3ED0: ; MOVE		0x3ED0-0x3ED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3ED8: ; MOVE		0x3ED8-0x3EDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EE0: ; MOVE		0x3EE0-0x3EE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EE8: ; MOVE		0x3EE8-0x3EEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EF0: ; MOVE		0x3EF0-0x3EF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EF8: ; MOVE		0x3EF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EF9: ; MOVE		0x3EF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EFA: ; MOVE		0x3EFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EFB: ; MOVE		0x3EFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3EFC: ; MOVE		0x3EFC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
mov  ebx, [a+7*4]
add  dword [a+7*4], byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F00: ; MOVE		0x3F00-0x3F07
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F08: ; MOVE		0x3F08-0x3F0F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F10: ; MOVE		0x3F10-0x3F17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F18: ; MOVE		0x3F18-0x3F1F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F20: ; MOVE		0x3F20-0x3F27
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F28: ; MOVE		0x3F28-0x3F2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F30: ; MOVE		0x3F30-0x3F37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F38: ; MOVE		0x3F38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F39: ; MOVE		0x3F39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F3A: ; MOVE		0x3F3A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F3B: ; MOVE		0x3F3B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F3C: ; MOVE		0x3F3C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
sub  dword [a+7*4], byte 2
mov  ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F40: ; MOVE		0x3F40-0x3F47
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F48: ; MOVE		0x3F48-0x3F4F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F50: ; MOVE		0x3F50-0x3F57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F58: ; MOVE		0x3F58-0x3F5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F60: ; MOVE		0x3F60-0x3F67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F68: ; MOVE		0x3F68-0x3F6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F70: ; MOVE		0x3F70-0x3F77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F78: ; MOVE		0x3F78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F79: ; MOVE		0x3F79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F7A: ; MOVE		0x3F7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F7B: ; MOVE		0x3F7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F7C: ; MOVE		0x3F7C
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
movsx    ebx, word [esi]
add      ebx, [a+7*4]
add      esi, byte 2
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F80: ; MOVE		0x3F80-0x3F87
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F88: ; MOVE		0x3F88-0x3F8F
and  edi, byte 7
mov  edx, [a+edi*4]
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F90: ; MOVE		0x3F90-0x3F97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3F98: ; MOVE		0x3F98-0x3F9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FA0: ; MOVE		0x3FA0-0x3FA7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FA8: ; MOVE		0x3FA8-0x3FAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FB0: ; MOVE		0x3FB0-0x3FB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FB8: ; MOVE		0x3FB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FB9: ; MOVE		0x3FB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FBA: ; MOVE		0x3FBA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FBB: ; MOVE		0x3FBB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3FBC: ; MOVE		0x3FBC
mov  edx, [esi]
add  esi, byte 2
test dx, dx
lahf
xor  al, al
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ls
movsx    ebx, bx
.ls:
add      ebx, edi
add      ebx, [a+7*4]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2040: ; MOVEA		0x2040-0x2047
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2048: ; MOVEA		0x2048-0x204F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2050: ; MOVEA		0x2050-0x2057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2058: ; MOVEA		0x2058-0x205F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2060: ; MOVEA		0x2060-0x2067
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2068: ; MOVEA		0x2068-0x206F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2070: ; MOVEA		0x2070-0x2077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2078: ; MOVEA		0x2078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2079: ; MOVEA		0x2079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I207A: ; MOVEA		0x207A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I207B: ; MOVEA		0x207B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+0*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I207C: ; MOVEA		0x207C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2240: ; MOVEA		0x2240-0x2247
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2248: ; MOVEA		0x2248-0x224F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2250: ; MOVEA		0x2250-0x2257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2258: ; MOVEA		0x2258-0x225F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2260: ; MOVEA		0x2260-0x2267
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2268: ; MOVEA		0x2268-0x226F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2270: ; MOVEA		0x2270-0x2277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2278: ; MOVEA		0x2278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2279: ; MOVEA		0x2279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I227A: ; MOVEA		0x227A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I227B: ; MOVEA		0x227B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+1*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I227C: ; MOVEA		0x227C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2440: ; MOVEA		0x2440-0x2447
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2448: ; MOVEA		0x2448-0x244F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2450: ; MOVEA		0x2450-0x2457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2458: ; MOVEA		0x2458-0x245F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2460: ; MOVEA		0x2460-0x2467
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2468: ; MOVEA		0x2468-0x246F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2470: ; MOVEA		0x2470-0x2477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2478: ; MOVEA		0x2478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2479: ; MOVEA		0x2479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I247A: ; MOVEA		0x247A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I247B: ; MOVEA		0x247B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+2*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I247C: ; MOVEA		0x247C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2640: ; MOVEA		0x2640-0x2647
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2648: ; MOVEA		0x2648-0x264F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2650: ; MOVEA		0x2650-0x2657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2658: ; MOVEA		0x2658-0x265F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2660: ; MOVEA		0x2660-0x2667
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2668: ; MOVEA		0x2668-0x266F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2670: ; MOVEA		0x2670-0x2677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2678: ; MOVEA		0x2678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2679: ; MOVEA		0x2679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I267A: ; MOVEA		0x267A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I267B: ; MOVEA		0x267B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+3*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I267C: ; MOVEA		0x267C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2840: ; MOVEA		0x2840-0x2847
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2848: ; MOVEA		0x2848-0x284F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2850: ; MOVEA		0x2850-0x2857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2858: ; MOVEA		0x2858-0x285F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2860: ; MOVEA		0x2860-0x2867
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2868: ; MOVEA		0x2868-0x286F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2870: ; MOVEA		0x2870-0x2877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2878: ; MOVEA		0x2878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2879: ; MOVEA		0x2879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I287A: ; MOVEA		0x287A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I287B: ; MOVEA		0x287B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+4*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I287C: ; MOVEA		0x287C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A40: ; MOVEA		0x2A40-0x2A47
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A48: ; MOVEA		0x2A48-0x2A4F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A50: ; MOVEA		0x2A50-0x2A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A58: ; MOVEA		0x2A58-0x2A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A60: ; MOVEA		0x2A60-0x2A67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A68: ; MOVEA		0x2A68-0x2A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A70: ; MOVEA		0x2A70-0x2A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A78: ; MOVEA		0x2A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A79: ; MOVEA		0x2A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A7A: ; MOVEA		0x2A7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A7B: ; MOVEA		0x2A7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+5*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2A7C: ; MOVEA		0x2A7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C40: ; MOVEA		0x2C40-0x2C47
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C48: ; MOVEA		0x2C48-0x2C4F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C50: ; MOVEA		0x2C50-0x2C57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C58: ; MOVEA		0x2C58-0x2C5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C60: ; MOVEA		0x2C60-0x2C67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C68: ; MOVEA		0x2C68-0x2C6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C70: ; MOVEA		0x2C70-0x2C77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C78: ; MOVEA		0x2C78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C79: ; MOVEA		0x2C79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C7A: ; MOVEA		0x2C7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C7B: ; MOVEA		0x2C7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+6*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2C7C: ; MOVEA		0x2C7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E40: ; MOVEA		0x2E40-0x2E47
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [a+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E48: ; MOVEA		0x2E48-0x2E4F
and  edi, byte 7
mov  edx, [a+edi*4]
mov  [a+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E50: ; MOVEA		0x2E50-0x2E57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E58: ; MOVEA		0x2E58-0x2E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E60: ; MOVEA		0x2E60-0x2E67
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E68: ; MOVEA		0x2E68-0x2E6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E70: ; MOVEA		0x2E70-0x2E77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E78: ; MOVEA		0x2E78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E79: ; MOVEA		0x2E79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E7A: ; MOVEA		0x2E7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E7B: ; MOVEA		0x2E7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadLong
mov  [a+7*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I2E7C: ; MOVEA		0x2E7C
mov  edx, [esi]
add  esi, byte 4
ror  edx, byte 16
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3040: ; MOVEA		0x3040-0x3047
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3048: ; MOVEA		0x3048-0x304F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+0*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3050: ; MOVEA		0x3050-0x3057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3058: ; MOVEA		0x3058-0x305F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3060: ; MOVEA		0x3060-0x3067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3068: ; MOVEA		0x3068-0x306F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3070: ; MOVEA		0x3070-0x3077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3078: ; MOVEA		0x3078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3079: ; MOVEA		0x3079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I307A: ; MOVEA		0x307A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I307B: ; MOVEA		0x307B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+0*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I307C: ; MOVEA		0x307C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+0*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3240: ; MOVEA		0x3240-0x3247
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3248: ; MOVEA		0x3248-0x324F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+1*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3250: ; MOVEA		0x3250-0x3257
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3258: ; MOVEA		0x3258-0x325F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3260: ; MOVEA		0x3260-0x3267
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3268: ; MOVEA		0x3268-0x326F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3270: ; MOVEA		0x3270-0x3277
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3278: ; MOVEA		0x3278
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3279: ; MOVEA		0x3279
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I327A: ; MOVEA		0x327A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I327B: ; MOVEA		0x327B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+1*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I327C: ; MOVEA		0x327C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+1*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3440: ; MOVEA		0x3440-0x3447
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3448: ; MOVEA		0x3448-0x344F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+2*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3450: ; MOVEA		0x3450-0x3457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3458: ; MOVEA		0x3458-0x345F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3460: ; MOVEA		0x3460-0x3467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3468: ; MOVEA		0x3468-0x346F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3470: ; MOVEA		0x3470-0x3477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3478: ; MOVEA		0x3478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3479: ; MOVEA		0x3479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I347A: ; MOVEA		0x347A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I347B: ; MOVEA		0x347B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+2*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I347C: ; MOVEA		0x347C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+2*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3640: ; MOVEA		0x3640-0x3647
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3648: ; MOVEA		0x3648-0x364F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+3*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3650: ; MOVEA		0x3650-0x3657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3658: ; MOVEA		0x3658-0x365F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3660: ; MOVEA		0x3660-0x3667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3668: ; MOVEA		0x3668-0x366F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3670: ; MOVEA		0x3670-0x3677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3678: ; MOVEA		0x3678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3679: ; MOVEA		0x3679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I367A: ; MOVEA		0x367A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I367B: ; MOVEA		0x367B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+3*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I367C: ; MOVEA		0x367C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+3*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3840: ; MOVEA		0x3840-0x3847
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3848: ; MOVEA		0x3848-0x384F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+4*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3850: ; MOVEA		0x3850-0x3857
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3858: ; MOVEA		0x3858-0x385F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3860: ; MOVEA		0x3860-0x3867
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3868: ; MOVEA		0x3868-0x386F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3870: ; MOVEA		0x3870-0x3877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3878: ; MOVEA		0x3878
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3879: ; MOVEA		0x3879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I387A: ; MOVEA		0x387A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I387B: ; MOVEA		0x387B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+4*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I387C: ; MOVEA		0x387C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+4*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A40: ; MOVEA		0x3A40-0x3A47
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A48: ; MOVEA		0x3A48-0x3A4F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+5*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A50: ; MOVEA		0x3A50-0x3A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A58: ; MOVEA		0x3A58-0x3A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A60: ; MOVEA		0x3A60-0x3A67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A68: ; MOVEA		0x3A68-0x3A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A70: ; MOVEA		0x3A70-0x3A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A78: ; MOVEA		0x3A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A79: ; MOVEA		0x3A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A7A: ; MOVEA		0x3A7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A7B: ; MOVEA		0x3A7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+5*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3A7C: ; MOVEA		0x3A7C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+5*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C40: ; MOVEA		0x3C40-0x3C47
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C48: ; MOVEA		0x3C48-0x3C4F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+6*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C50: ; MOVEA		0x3C50-0x3C57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C58: ; MOVEA		0x3C58-0x3C5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C60: ; MOVEA		0x3C60-0x3C67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C68: ; MOVEA		0x3C68-0x3C6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C70: ; MOVEA		0x3C70-0x3C77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C78: ; MOVEA		0x3C78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C79: ; MOVEA		0x3C79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C7A: ; MOVEA		0x3C7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C7B: ; MOVEA		0x3C7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+6*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3C7C: ; MOVEA		0x3C7C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+6*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E40: ; MOVEA		0x3E40-0x3E47
and  edi, byte 7
movsx    edx, word [d+edi*4]
mov  [a+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E48: ; MOVEA		0x3E48-0x3E4F
and  edi, byte 7
movsx    edx, word [a+edi*4]
mov  [a+7*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E50: ; MOVEA		0x3E50-0x3E57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E58: ; MOVEA		0x3E58-0x3E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E60: ; MOVEA		0x3E60-0x3E67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E68: ; MOVEA		0x3E68-0x3E6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E70: ; MOVEA		0x3E70-0x3E77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E78: ; MOVEA		0x3E78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E79: ; MOVEA		0x3E79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E7A: ; MOVEA		0x3E7A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E7B: ; MOVEA		0x3E7B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWordSX
mov  [a+7*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I3E7C: ; MOVEA		0x3E7C
movsx    edx, word [esi]
add  esi, byte 2
mov  [a+7*4], edx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4890: ; MOVEM		0x4890-0x4897
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48A0: ; MOVEM		0x48A0-0x48A7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  [run_edi], edi
mov  esi, d+15*4
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
sub  ebx, byte 2
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
.skip:
sub  esi, byte 4
cmp  esi, d-4
jne  short .loop
mov  edi, [run_edi]
mov  [a+edi*4], ebx
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48A8: ; MOVEM		0x48A8-0x48AF
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48B0: ; MOVEM		0x48B0-0x48B7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48B8: ; MOVEM		0x48B8
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48B9: ; MOVEM		0x48B9
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48D0: ; MOVEM		0x48D0-0x48D7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48E0: ; MOVEM		0x48E0-0x48E7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  [run_edi], edi
mov  esi, d+15*4
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
sub  ebx, byte 4
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
.skip:
sub  esi, byte 4
cmp  esi, d-4
jne  short .loop
mov  edi, [run_edi]
mov  [a+edi*4], ebx
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48E8: ; MOVEM		0x48E8-0x48EF
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48F0: ; MOVEM		0x48F0-0x48F7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48F8: ; MOVEM		0x48F8
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I48F9: ; MOVEM		0x48F9
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
mov  edx, [esi]
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4C90: ; MOVEM		0x4C90-0x4C97
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4C98: ; MOVEM		0x4C98-0x4C9F
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  [run_edi], edi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  edi, [run_edi]
mov  [a+edi*4], ebx
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CA8: ; MOVEM		0x4CA8-0x4CAF
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CB0: ; MOVEM		0x4CB0-0x4CB7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CB8: ; MOVEM		0x4CB8
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CB9: ; MOVEM		0x4CB9
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CBA: ; MOVEM		0x4CBA
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CBB: ; MOVEM		0x4CBB
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 4
shr  eax, byte 1
jnc  short .skip
call ReadWordSX
mov  [esi], edx
add  ebx, byte 2
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CD0: ; MOVEM		0x4CD0-0x4CD7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CD8: ; MOVEM		0x4CD8-0x4CDF
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_esi], esi
mov  [run_edi], edi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  edi, [run_edi]
mov  [a+edi*4], ebx
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CE8: ; MOVEM		0x4CE8-0x4CEF
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CF0: ; MOVEM		0x4CF0-0x4CF7
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CF8: ; MOVEM		0x4CF8
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CF9: ; MOVEM		0x4CF9
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CFA: ; MOVEM		0x4CFA
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4CFB: ; MOVEM		0x4CFB
mov  [run_eax], eax
mov  eax, [esi]
add  esi, byte 2
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
mov  [run_esi], esi
mov  esi, d
.loop:
sub  ecx, byte 8
shr  eax, byte 1
jnc  short .skip
call ReadLong
mov  [esi], edx
add  ebx, byte 4
.skip:
add  esi, byte 4
cmp  esi, d+16*4
jne  short .loop
mov  esi, [run_esi]
mov  eax, [run_eax]
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0108: ; MOVEP		0x0108-0x010F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+0*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+0*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0148: ; MOVEP		0x0148-0x014F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+0*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+0*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+0*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+0*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0188: ; MOVEP		0x0188-0x018F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+0*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I01C8: ; MOVEP		0x01C8-0x01CF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+0*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0308: ; MOVEP		0x0308-0x030F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+1*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+1*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0348: ; MOVEP		0x0348-0x034F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+1*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+1*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+1*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+1*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0388: ; MOVEP		0x0388-0x038F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+1*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I03C8: ; MOVEP		0x03C8-0x03CF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+1*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0508: ; MOVEP		0x0508-0x050F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+2*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+2*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0548: ; MOVEP		0x0548-0x054F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+2*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+2*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+2*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+2*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0588: ; MOVEP		0x0588-0x058F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+2*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I05C8: ; MOVEP		0x05C8-0x05CF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+2*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0708: ; MOVEP		0x0708-0x070F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+3*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+3*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0748: ; MOVEP		0x0748-0x074F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+3*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+3*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+3*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+3*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0788: ; MOVEP		0x0788-0x078F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+3*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I07C8: ; MOVEP		0x07C8-0x07CF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+3*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0908: ; MOVEP		0x0908-0x090F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+4*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+4*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0948: ; MOVEP		0x0948-0x094F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+4*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+4*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+4*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+4*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0988: ; MOVEP		0x0988-0x098F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+4*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I09C8: ; MOVEP		0x09C8-0x09CF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+4*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B08: ; MOVEP		0x0B08-0x0B0F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+5*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+5*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B48: ; MOVEP		0x0B48-0x0B4F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+5*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+5*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+5*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+5*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0B88: ; MOVEP		0x0B88-0x0B8F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+5*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0BC8: ; MOVEP		0x0BC8-0x0BCF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+5*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D08: ; MOVEP		0x0D08-0x0D0F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+6*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+6*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D48: ; MOVEP		0x0D48-0x0D4F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+6*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+6*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+6*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+6*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0D88: ; MOVEP		0x0D88-0x0D8F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+6*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0DC8: ; MOVEP		0x0DC8-0x0DCF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+6*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F08: ; MOVEP		0x0F08-0x0F0F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+7*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+7*4+0], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F48: ; MOVEP		0x0F48-0x0F4F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  [d+7*4+3], dl
add  ebx, byte 2
call ReadByte
mov  [d+7*4+2], dl
add  ebx, byte 2
call ReadByte
mov  [d+7*4+1], dl
add  ebx, byte 2
call ReadByte
mov  [d+7*4+0], dl
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0F88: ; MOVEP		0x0F88-0x0F8F
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      dx, [d+7*4]
mov      eax, ebx
ror      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I0FC8: ; MOVEP		0x0FC8-0x0FCF
and      edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
mov  [run_eax], eax
mov      edx, [d+7*4]
mov      eax, ebx
rol      edx, byte 8
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
add      eax, byte 2
rol      edx, byte 8
mov      ebx, eax
call WriteByte
mov  eax, [run_eax]
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7000: ; MOVEQ		0x7000-0x70FF
mov      ebx, edi
movsx    ebx, bl
mov      [d+0*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7200: ; MOVEQ		0x7200-0x72FF
mov      ebx, edi
movsx    ebx, bl
mov      [d+1*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7400: ; MOVEQ		0x7400-0x74FF
mov      ebx, edi
movsx    ebx, bl
mov      [d+2*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7600: ; MOVEQ		0x7600-0x76FF
mov      ebx, edi
movsx    ebx, bl
mov      [d+3*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7800: ; MOVEQ		0x7800-0x78FF
mov      ebx, edi
movsx    ebx, bl
mov      [d+4*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7A00: ; MOVEQ		0x7A00-0x7AFF
mov      ebx, edi
movsx    ebx, bl
mov      [d+5*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7C00: ; MOVEQ		0x7C00-0x7CFF
mov      ebx, edi
movsx    ebx, bl
mov      [d+6*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I7E00: ; MOVEQ		0x7E00-0x7EFF
mov      ebx, edi
movsx    ebx, bl
mov      [d+7*4], ebx
test     bl, bl
lahf
xor      al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C0: ; MOVE from SR		0x40C0
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C1: ; MOVE from SR		0x40C1
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C2: ; MOVE from SR		0x40C2
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C3: ; MOVE from SR		0x40C3
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C4: ; MOVE from SR		0x40C4
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C5: ; MOVE from SR		0x40C5
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C6: ; MOVE from SR		0x40C6
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40C7: ; MOVE from SR		0x40C7
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  [d+edi*4], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D0: ; MOVE from SR		0x40D0
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D1: ; MOVE from SR		0x40D1
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D2: ; MOVE from SR		0x40D2
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D3: ; MOVE from SR		0x40D3
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D4: ; MOVE from SR		0x40D4
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D5: ; MOVE from SR		0x40D5
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D6: ; MOVE from SR		0x40D6
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D7: ; MOVE from SR		0x40D7
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D8: ; MOVE from SR		0x40D8
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40D9: ; MOVE from SR		0x40D9
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DA: ; MOVE from SR		0x40DA
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DB: ; MOVE from SR		0x40DB
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DC: ; MOVE from SR		0x40DC
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DD: ; MOVE from SR		0x40DD
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DE: ; MOVE from SR		0x40DE
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40DF: ; MOVE from SR		0x40DF
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E0: ; MOVE from SR		0x40E0
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E1: ; MOVE from SR		0x40E1
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E2: ; MOVE from SR		0x40E2
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E3: ; MOVE from SR		0x40E3
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E4: ; MOVE from SR		0x40E4
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E5: ; MOVE from SR		0x40E5
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E6: ; MOVE from SR		0x40E6
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E7: ; MOVE from SR		0x40E7
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E8: ; MOVE from SR		0x40E8
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40E9: ; MOVE from SR		0x40E9
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40EA: ; MOVE from SR		0x40EA
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40EB: ; MOVE from SR		0x40EB
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40EC: ; MOVE from SR		0x40EC
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40ED: ; MOVE from SR		0x40ED
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40EE: ; MOVE from SR		0x40EE
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40EF: ; MOVE from SR		0x40EF
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F0: ; MOVE from SR		0x40F0
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F1: ; MOVE from SR		0x40F1
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F2: ; MOVE from SR		0x40F2
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F3: ; MOVE from SR		0x40F3
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F4: ; MOVE from SR		0x40F4
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F5: ; MOVE from SR		0x40F5
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F6: ; MOVE from SR		0x40F6
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F7: ; MOVE from SR		0x40F7
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F8: ; MOVE from SR		0x40F8
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
movsx    ebx, word [esi]
add      esi, byte 2
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40F9: ; MOVE from SR		0x40F9
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [sr]
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteWord
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44C0: ; MOVE to CCR		0x44C0-0x44C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44D0: ; MOVE to CCR		0x44D0-0x44D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44D8: ; MOVE to CCR		0x44D8-0x44DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44E0: ; MOVE to CCR		0x44E0-0x44E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44E8: ; MOVE to CCR		0x44E8-0x44EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44F0: ; MOVE to CCR		0x44F0-0x44F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44F8: ; MOVE to CCR		0x44F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44F9: ; MOVE to CCR		0x44F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44FA: ; MOVE to CCR		0x44FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44FB: ; MOVE to CCR		0x44FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44FC: ; MOVE to CCR		0x44FC
mov  edx, [esi]
add  esi, byte 2
mov  [sr], dl
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46C0: ; MOVE to SR		0x46C0-0x46C7
and  edi, byte 7
mov  edx, [d+edi*4]
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46D0: ; MOVE to SR		0x46D0-0x46D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46D8: ; MOVE to SR		0x46D8-0x46DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46E0: ; MOVE to SR		0x46E0-0x46E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46E8: ; MOVE to SR		0x46E8-0x46EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46F0: ; MOVE to SR		0x46F0-0x46F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46F8: ; MOVE to SR		0x46F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46F9: ; MOVE to SR		0x46F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46FA: ; MOVE to SR		0x46FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46FB: ; MOVE to SR		0x46FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46FC: ; MOVE to SR		0x46FC
mov  edx, [esi]
add  esi, byte 2
test byte [sr+1], 0x20
jz   near exception_privilege_violation
test dh, 0x20
jnz  .still_s
mov  ebx, [__sp]
xchg [a+7*4], ebx
mov  [__sp], ebx
.still_s:
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E60: ; MOVE USP		0x4E60-0x4E67
and      edi, byte 7
test     byte [sr+1], 0x20
jz       near exception_privilege_violation
mov  edx, [a+edi*4]
mov  [__sp], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E68: ; MOVE USP		0x4E68-0x4E6F
and      edi, byte 7
test     byte [sr+1], 0x20
jz       near exception_privilege_violation
mov  edx, [__sp]
mov  [a+edi*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1C0: ; MULS		0xC1C0-0xC1C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1D0: ; MULS		0xC1D0-0xC1D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1D8: ; MULS		0xC1D8-0xC1DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1E0: ; MULS		0xC1E0-0xC1E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1E8: ; MULS		0xC1E8-0xC1EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1F0: ; MULS		0xC1F0-0xC1F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1F8: ; MULS		0xC1F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1F9: ; MULS		0xC1F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1FA: ; MULS		0xC1FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1FB: ; MULS		0xC1FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC1FC: ; MULS		0xC1FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+0*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3C0: ; MULS		0xC3C0-0xC3C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3D0: ; MULS		0xC3D0-0xC3D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3D8: ; MULS		0xC3D8-0xC3DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3E0: ; MULS		0xC3E0-0xC3E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3E8: ; MULS		0xC3E8-0xC3EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3F0: ; MULS		0xC3F0-0xC3F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3F8: ; MULS		0xC3F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3F9: ; MULS		0xC3F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3FA: ; MULS		0xC3FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3FB: ; MULS		0xC3FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC3FC: ; MULS		0xC3FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+1*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5C0: ; MULS		0xC5C0-0xC5C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5D0: ; MULS		0xC5D0-0xC5D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5D8: ; MULS		0xC5D8-0xC5DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5E0: ; MULS		0xC5E0-0xC5E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5E8: ; MULS		0xC5E8-0xC5EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5F0: ; MULS		0xC5F0-0xC5F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5F8: ; MULS		0xC5F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5F9: ; MULS		0xC5F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5FA: ; MULS		0xC5FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5FB: ; MULS		0xC5FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC5FC: ; MULS		0xC5FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+2*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7C0: ; MULS		0xC7C0-0xC7C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7D0: ; MULS		0xC7D0-0xC7D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7D8: ; MULS		0xC7D8-0xC7DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7E0: ; MULS		0xC7E0-0xC7E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7E8: ; MULS		0xC7E8-0xC7EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7F0: ; MULS		0xC7F0-0xC7F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7F8: ; MULS		0xC7F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7F9: ; MULS		0xC7F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7FA: ; MULS		0xC7FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7FB: ; MULS		0xC7FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC7FC: ; MULS		0xC7FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+3*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9C0: ; MULS		0xC9C0-0xC9C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9D0: ; MULS		0xC9D0-0xC9D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9D8: ; MULS		0xC9D8-0xC9DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9E0: ; MULS		0xC9E0-0xC9E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9E8: ; MULS		0xC9E8-0xC9EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9F0: ; MULS		0xC9F0-0xC9F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9F8: ; MULS		0xC9F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9F9: ; MULS		0xC9F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9FA: ; MULS		0xC9FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9FB: ; MULS		0xC9FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC9FC: ; MULS		0xC9FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+4*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBC0: ; MULS		0xCBC0-0xCBC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBD0: ; MULS		0xCBD0-0xCBD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBD8: ; MULS		0xCBD8-0xCBDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBE0: ; MULS		0xCBE0-0xCBE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBE8: ; MULS		0xCBE8-0xCBEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBF0: ; MULS		0xCBF0-0xCBF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBF8: ; MULS		0xCBF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBF9: ; MULS		0xCBF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBFA: ; MULS		0xCBFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBFB: ; MULS		0xCBFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICBFC: ; MULS		0xCBFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+5*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDC0: ; MULS		0xCDC0-0xCDC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDD0: ; MULS		0xCDD0-0xCDD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDD8: ; MULS		0xCDD8-0xCDDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDE0: ; MULS		0xCDE0-0xCDE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDE8: ; MULS		0xCDE8-0xCDEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDF0: ; MULS		0xCDF0-0xCDF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDF8: ; MULS		0xCDF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDF9: ; MULS		0xCDF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDFA: ; MULS		0xCDFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDFB: ; MULS		0xCDFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICDFC: ; MULS		0xCDFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+6*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFC0: ; MULS		0xCFC0-0xCFC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFD0: ; MULS		0xCFD0-0xCFD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFD8: ; MULS		0xCFD8-0xCFDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFE0: ; MULS		0xCFE0-0xCFE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFE8: ; MULS		0xCFE8-0xCFEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFF0: ; MULS		0xCFF0-0xCFF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFF8: ; MULS		0xCFF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFF9: ; MULS		0xCFF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFFA: ; MULS		0xCFFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFFB: ; MULS		0xCFFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICFFC: ; MULS		0xCFFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+7*4]
imul dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0C0: ; MULU		0xC0C0-0xC0C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0D0: ; MULU		0xC0D0-0xC0D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0D8: ; MULU		0xC0D8-0xC0DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0E0: ; MULU		0xC0E0-0xC0E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0E8: ; MULU		0xC0E8-0xC0EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0F0: ; MULU		0xC0F0-0xC0F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0F8: ; MULU		0xC0F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0F9: ; MULU		0xC0F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0FA: ; MULU		0xC0FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0FB: ; MULU		0xC0FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC0FC: ; MULU		0xC0FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+0*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+0*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2C0: ; MULU		0xC2C0-0xC2C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2D0: ; MULU		0xC2D0-0xC2D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2D8: ; MULU		0xC2D8-0xC2DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2E0: ; MULU		0xC2E0-0xC2E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2E8: ; MULU		0xC2E8-0xC2EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2F0: ; MULU		0xC2F0-0xC2F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2F8: ; MULU		0xC2F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2F9: ; MULU		0xC2F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2FA: ; MULU		0xC2FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2FB: ; MULU		0xC2FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC2FC: ; MULU		0xC2FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+1*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+1*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4C0: ; MULU		0xC4C0-0xC4C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4D0: ; MULU		0xC4D0-0xC4D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4D8: ; MULU		0xC4D8-0xC4DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4E0: ; MULU		0xC4E0-0xC4E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4E8: ; MULU		0xC4E8-0xC4EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4F0: ; MULU		0xC4F0-0xC4F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4F8: ; MULU		0xC4F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4F9: ; MULU		0xC4F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4FA: ; MULU		0xC4FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4FB: ; MULU		0xC4FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC4FC: ; MULU		0xC4FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+2*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+2*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6C0: ; MULU		0xC6C0-0xC6C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6D0: ; MULU		0xC6D0-0xC6D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6D8: ; MULU		0xC6D8-0xC6DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6E0: ; MULU		0xC6E0-0xC6E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6E8: ; MULU		0xC6E8-0xC6EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6F0: ; MULU		0xC6F0-0xC6F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6F8: ; MULU		0xC6F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6F9: ; MULU		0xC6F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6FA: ; MULU		0xC6FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6FB: ; MULU		0xC6FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC6FC: ; MULU		0xC6FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+3*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+3*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8C0: ; MULU		0xC8C0-0xC8C7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8D0: ; MULU		0xC8D0-0xC8D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8D8: ; MULU		0xC8D8-0xC8DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8E0: ; MULU		0xC8E0-0xC8E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8E8: ; MULU		0xC8E8-0xC8EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8F0: ; MULU		0xC8F0-0xC8F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8F8: ; MULU		0xC8F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8F9: ; MULU		0xC8F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8FA: ; MULU		0xC8FA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8FB: ; MULU		0xC8FB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IC8FC: ; MULU		0xC8FC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+4*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+4*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAC0: ; MULU		0xCAC0-0xCAC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAD0: ; MULU		0xCAD0-0xCAD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAD8: ; MULU		0xCAD8-0xCADF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAE0: ; MULU		0xCAE0-0xCAE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAE8: ; MULU		0xCAE8-0xCAEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAF0: ; MULU		0xCAF0-0xCAF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAF8: ; MULU		0xCAF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAF9: ; MULU		0xCAF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAFA: ; MULU		0xCAFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAFB: ; MULU		0xCAFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICAFC: ; MULU		0xCAFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+5*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+5*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCC0: ; MULU		0xCCC0-0xCCC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCD0: ; MULU		0xCCD0-0xCCD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCD8: ; MULU		0xCCD8-0xCCDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCE0: ; MULU		0xCCE0-0xCCE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCE8: ; MULU		0xCCE8-0xCCEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCF0: ; MULU		0xCCF0-0xCCF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCF8: ; MULU		0xCCF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCF9: ; MULU		0xCCF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCFA: ; MULU		0xCCFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCFB: ; MULU		0xCCFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICCFC: ; MULU		0xCCFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+6*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+6*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEC0: ; MULU		0xCEC0-0xCEC7
and  edi, byte 7
mov  edx, [d+edi*4]
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 70
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICED0: ; MULU		0xCED0-0xCED7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICED8: ; MULU		0xCED8-0xCEDF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEE0: ; MULU		0xCEE0-0xCEE7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 76
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEE8: ; MULU		0xCEE8-0xCEEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEF0: ; MULU		0xCEF0-0xCEF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEF8: ; MULU		0xCEF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEF9: ; MULU		0xCEF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 82
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEFA: ; MULU		0xCEFA
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 78
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEFB: ; MULU		0xCEFB
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
call ReadWord
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 80
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
ICEFC: ; MULU		0xCEFC
mov  edx, [esi]
add  esi, byte 2
mov  ax, [d+7*4]
mul  dx
shl  edx, byte 16
mov  dx, ax
mov  [d+7*4], edx
test edx, edx
lahf
xor  al, al
sub  ecx, byte 74
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4800: ; NBCD		0x4800-0x4807
and  edi, byte 7
xor  al, al
shr  byte [x], 1
sbb  al, [d+edi*4]
das
mov  bl, ah
lahf
setc byte [x]
jnz  short .clr
and  bl, byte 0x40
and  ah, byte 0xbf
or   ah, bl
.clr:
mov  [d+edi*4], al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4810: ; NBCD		0x4810-0x4817
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4818: ; NBCD		0x4818-0x481F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4820: ; NBCD		0x4820-0x4827
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4828: ; NBCD		0x4828-0x482F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4830: ; NBCD		0x4830-0x4837
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4838: ; NBCD		0x4838
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4839: ; NBCD		0x4839
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
xor  al, al
shr  byte [x], 1
sbb  al, dl
mov  dl, ah
das
lahf
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  dl, al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4000: ; NEGX		0x4000-0x4007
and  edi, byte 7
xor  dl, dl
shr  byte [x], 1
sbb  dl, [d+edi*4]
mov  bl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  bl, byte 0x40
and  ah, byte 0xbf
or   ah, bl
.clr:
mov  [d+edi*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4010: ; NEGX		0x4010-0x4017
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4018: ; NEGX		0x4018-0x401F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4020: ; NEGX		0x4020-0x4027
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4028: ; NEGX		0x4028-0x402F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4030: ; NEGX		0x4030-0x4037
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4038: ; NEGX		0x4038
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4039: ; NEGX		0x4039
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
mov  edi, ebx
xor  bl, bl
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4040: ; NEGX		0x4040-0x4047
and  edi, byte 7
xor  edx, edx
shr  byte [x], 1
sbb  dx, [d+edi*4]
mov  bl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  bl, byte 0x40
and  ah, byte 0xbf
or   ah, bl
.clr:
mov  [d+edi*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4050: ; NEGX		0x4050-0x4057
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4058: ; NEGX		0x4058-0x405F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4060: ; NEGX		0x4060-0x4067
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4068: ; NEGX		0x4068-0x406F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4070: ; NEGX		0x4070-0x4077
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4078: ; NEGX		0x4078
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4079: ; NEGX		0x4079
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4080: ; NEGX		0x4080-0x4087
and  edi, byte 7
xor  edx, edx
shr  byte [x], 1
sbb  edx, [d+edi*4]
mov  bl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  bl, byte 0x40
and  ah, byte 0xbf
or   ah, bl
.clr:
mov  [d+edi*4], edx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4090: ; NEGX		0x4090-0x4097
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4098: ; NEGX		0x4098-0x409F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40A0: ; NEGX		0x40A0-0x40A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40A8: ; NEGX		0x40A8-0x40AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40B0: ; NEGX		0x40B0-0x40B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40B8: ; NEGX		0x40B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I40B9: ; NEGX		0x40B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
mov  edi, ebx
xor  ebx, ebx
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
seto al
setc byte [x]
jnz  short .clr
and  dl, byte 0x40
and  ah, byte 0xbf
or   ah, dl
.clr:
mov  edx, ebx
mov  ebx, edi
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4400: ; NEG		0x4400-0x4407
and  edi, byte 7
neg  byte [d+edi*4]
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4410: ; NEG		0x4410-0x4417
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4418: ; NEG		0x4418-0x441F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4420: ; NEG		0x4420-0x4427
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4428: ; NEG		0x4428-0x442F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4430: ; NEG		0x4430-0x4437
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4438: ; NEG		0x4438
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4439: ; NEG		0x4439
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
neg  dl
lahf
seto al
setc byte [x]
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4440: ; NEG		0x4440-0x4447
and  edi, byte 7
neg  word [d+edi*4]
lahf
seto al
setc byte [x]
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4450: ; NEG		0x4450-0x4457
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4458: ; NEG		0x4458-0x445F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4460: ; NEG		0x4460-0x4467
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4468: ; NEG		0x4468-0x446F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4470: ; NEG		0x4470-0x4477
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4478: ; NEG		0x4478
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4479: ; NEG		0x4479
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
neg  dx
lahf
seto al
setc byte [x]
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4480: ; NEG		0x4480-0x4487
and  edi, byte 7
neg  dword [d+edi*4]
lahf
seto al
setc byte [x]
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4490: ; NEG		0x4490-0x4497
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4498: ; NEG		0x4498-0x449F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44A0: ; NEG		0x44A0-0x44A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44A8: ; NEG		0x44A8-0x44AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44B0: ; NEG		0x44B0-0x44B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44B8: ; NEG		0x44B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I44B9: ; NEG		0x44B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
neg  edx
lahf
seto al
setc byte [x]
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E71: ; NOP		0x4E71
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4600: ; NOT		0x4600-0x4607
and  edi, byte 7
mov  dl, [d+edi*4]
not  dl
test dl, dl
lahf
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4610: ; NOT		0x4610-0x4617
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4618: ; NOT		0x4618-0x461F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4620: ; NOT		0x4620-0x4627
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4628: ; NOT		0x4628-0x462F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4630: ; NOT		0x4630-0x4637
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4638: ; NOT		0x4638
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4639: ; NOT		0x4639
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
not  dl
test dl, dl
lahf
xor  al, al
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4640: ; NOT		0x4640-0x4647
and  edi, byte 7
mov  edx, [d+edi*4]
not  dx
test dx, dx
lahf
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4650: ; NOT		0x4650-0x4657
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4658: ; NOT		0x4658-0x465F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4660: ; NOT		0x4660-0x4667
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4668: ; NOT		0x4668-0x466F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4670: ; NOT		0x4670-0x4677
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4678: ; NOT		0x4678
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4679: ; NOT		0x4679
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
not  dx
test dx, dx
lahf
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4680: ; NOT		0x4680-0x4687
and  edi, byte 7
mov  edx, [d+edi*4]
not  edx
test edx, edx
lahf
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4690: ; NOT		0x4690-0x4697
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4698: ; NOT		0x4698-0x469F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46A0: ; NOT		0x46A0-0x46A7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46A8: ; NOT		0x46A8-0x46AF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46B0: ; NOT		0x46B0-0x46B7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46B8: ; NOT		0x46B8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I46B9: ; NOT		0x46B9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
not  edx
test edx, edx
lahf
xor  al, al
call WriteLong
sub  ecx, byte 28
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I003C: ; ORI to CCR		0x003C
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  dl, [esi]
and  dl, 0x1f
or   [sr], dl
add  esi, byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
db 66, 65, 82, 84, 33, 1
times ($$-$) & 3 nop
I007C: ; ORI to SR		0x007C
test byte [sr+1], 0x20
jz   near exception_privilege_violation
and      eax, 0xc101
mov      dh, ah
shr      dh, byte 4
shl      al, byte 1
or       dh, al
shr      ah, byte 1
adc      dh, 0
mov      ah, [x]
shl      ah, byte 4
or       dh, ah
mov      [sr], dh
mov  edx, [esi]
and  edx, 0xa71f
or   [sr], dx
add  esi, byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4850: ; PEA		0x4850-0x4857
and  edi, byte 7
mov  ebx, [a+edi*4]
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4868: ; PEA		0x4868-0x486F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4870: ; PEA		0x4870-0x4877
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4878: ; PEA		0x4878
movsx    ebx, word [esi]
add      esi, byte 2
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4879: ; PEA		0x4879
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I487A: ; PEA		0x487A
movsx    ebx, word [esi]
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I487B: ; PEA		0x487B
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, esi
sub      ebx, ebp
add      esi, byte 2
sub  dword [a+7*4], byte 4
mov  edx, ebx
mov  ebx, [a+7*4]
call WriteLong
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E70: ; RESET		0x4E70
test byte [sr+1], 0x20
jz   near exception_privilege_violation
mov  ebx, [Reset]
test ebx, ebx
jz   short .no_reset
mov  [memhandler_eax], eax
mov  [memhandler_ebp], ebp
mov  dword [pc], esi
sub  dword [pc], ebp
mov     [remaining], ecx
call    ebx
mov     esi, [pc]
mov     ecx, [remaining]
mov  eax, [memhandler_eax]
mov  ebp, [memhandler_ebp]
add     esi, ebp
xor     edi, edi
.no_reset:
sub  ecx, 132
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE018: ; ROR		0xE018-0xE01F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 8
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE038: ; ROR		0xE038-0xE03F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE058: ; ROR		0xE058-0xE05F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 8
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE078: ; ROR		0xE078-0xE07F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE098: ; ROR		0xE098-0xE09F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 8
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0B8: ; ROR		0xE0B8-0xE0BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE118: ; ROL		0xE118-0xE11F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 8
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE138: ; ROL		0xE138-0xE13F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE158: ; ROL		0xE158-0xE15F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 8
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE178: ; ROL		0xE178-0xE17F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE198: ; ROL		0xE198-0xE19F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 8
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1B8: ; ROL		0xE1B8-0xE1BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE218: ; ROR		0xE218-0xE21F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 1
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE238: ; ROR		0xE238-0xE23F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE258: ; ROR		0xE258-0xE25F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 1
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE278: ; ROR		0xE278-0xE27F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE298: ; ROR		0xE298-0xE29F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 1
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2B8: ; ROR		0xE2B8-0xE2BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE318: ; ROL		0xE318-0xE31F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 1
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE338: ; ROL		0xE338-0xE33F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE358: ; ROL		0xE358-0xE35F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 1
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE378: ; ROL		0xE378-0xE37F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE398: ; ROL		0xE398-0xE39F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 1
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3B8: ; ROL		0xE3B8-0xE3BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE418: ; ROR		0xE418-0xE41F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 2
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE438: ; ROR		0xE438-0xE43F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE458: ; ROR		0xE458-0xE45F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 2
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE478: ; ROR		0xE478-0xE47F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE498: ; ROR		0xE498-0xE49F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 2
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4B8: ; ROR		0xE4B8-0xE4BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE518: ; ROL		0xE518-0xE51F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 2
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE538: ; ROL		0xE538-0xE53F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE558: ; ROL		0xE558-0xE55F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 2
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE578: ; ROL		0xE578-0xE57F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE598: ; ROL		0xE598-0xE59F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 2
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5B8: ; ROL		0xE5B8-0xE5BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE618: ; ROR		0xE618-0xE61F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 3
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE638: ; ROR		0xE638-0xE63F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE658: ; ROR		0xE658-0xE65F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 3
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE678: ; ROR		0xE678-0xE67F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE698: ; ROR		0xE698-0xE69F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 3
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6B8: ; ROR		0xE6B8-0xE6BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE718: ; ROL		0xE718-0xE71F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 3
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE738: ; ROL		0xE738-0xE73F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE758: ; ROL		0xE758-0xE75F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 3
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE778: ; ROL		0xE778-0xE77F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE798: ; ROL		0xE798-0xE79F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 3
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7B8: ; ROL		0xE7B8-0xE7BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE818: ; ROR		0xE818-0xE81F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 4
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE838: ; ROR		0xE838-0xE83F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE858: ; ROR		0xE858-0xE85F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 4
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE878: ; ROR		0xE878-0xE87F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE898: ; ROR		0xE898-0xE89F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 4
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE8B8: ; ROR		0xE8B8-0xE8BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE918: ; ROL		0xE918-0xE91F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 4
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE938: ; ROL		0xE938-0xE93F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE958: ; ROL		0xE958-0xE95F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 4
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE978: ; ROL		0xE978-0xE97F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE998: ; ROL		0xE998-0xE99F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 4
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE9B8: ; ROL		0xE9B8-0xE9BF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA18: ; ROR		0xEA18-0xEA1F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 5
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA38: ; ROR		0xEA38-0xEA3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA58: ; ROR		0xEA58-0xEA5F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 5
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA78: ; ROR		0xEA78-0xEA7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA98: ; ROR		0xEA98-0xEA9F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 5
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEAB8: ; ROR		0xEAB8-0xEABF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB18: ; ROL		0xEB18-0xEB1F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 5
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB38: ; ROL		0xEB38-0xEB3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB58: ; ROL		0xEB58-0xEB5F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 5
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB78: ; ROL		0xEB78-0xEB7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB98: ; ROL		0xEB98-0xEB9F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 5
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEBB8: ; ROL		0xEBB8-0xEBBF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC18: ; ROR		0xEC18-0xEC1F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 6
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC38: ; ROR		0xEC38-0xEC3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC58: ; ROR		0xEC58-0xEC5F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 6
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC78: ; ROR		0xEC78-0xEC7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC98: ; ROR		0xEC98-0xEC9F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 6
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IECB8: ; ROR		0xECB8-0xECBF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED18: ; ROL		0xED18-0xED1F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 6
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED38: ; ROL		0xED38-0xED3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED58: ; ROL		0xED58-0xED5F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 6
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED78: ; ROL		0xED78-0xED7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED98: ; ROL		0xED98-0xED9F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 6
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEDB8: ; ROL		0xEDB8-0xEDBF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE18: ; ROR		0xEE18-0xEE1F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dl, byte 7
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE38: ; ROR		0xEE38-0xEE3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE58: ; ROR		0xEE58-0xEE5F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   dx, byte 7
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE78: ; ROR		0xEE78-0xEE7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE98: ; ROR		0xEE98-0xEE9F
and  edi, byte 7
mov  edx, [d+edi*4]
ror   edx, byte 7
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEEB8: ; ROR		0xEEB8-0xEEBF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
ror   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF18: ; ROL		0xEF18-0xEF1F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dl, byte 7
setc al
test dl, dl
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF38: ; ROL		0xEF38-0xEF3F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF58: ; ROL		0xEF58-0xEF5F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   dx, byte 7
setc al
test dx, dx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF78: ; ROL		0xEF78-0xEF7F
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF98: ; ROL		0xEF98-0xEF9F
and  edi, byte 7
mov  edx, [d+edi*4]
rol   edx, byte 7
setc al
test edx, edx
lahf
or       ah, al
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEFB8: ; ROL		0xEFB8-0xEFBF
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  edx, [d+edi*4]
rol   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jnz  short .not_zero
xor      al, al
.not_zero:
or   ah, al
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6D0: ; ROR		0xE6D0-0xE6D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6D8: ; ROR		0xE6D8-0xE6DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6E0: ; ROR		0xE6E0-0xE6E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6E8: ; ROR		0xE6E8-0xE6EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6F0: ; ROR		0xE6F0-0xE6F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6F8: ; ROR		0xE6F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6F9: ; ROR		0xE6F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
ror  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7D0: ; ROL		0xE7D0-0xE7D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7D8: ; ROL		0xE7D8-0xE7DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7E0: ; ROL		0xE7E0-0xE7E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7E8: ; ROL		0xE7E8-0xE7EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7F0: ; ROL		0xE7F0-0xE7F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7F8: ; ROL		0xE7F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7F9: ; ROL		0xE7F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
rol  dx, byte 1
setc al
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE010: ; ROXR		0xE010-0xE017
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 8
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE030: ; ROXR		0xE030-0xE037
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE050: ; ROXR		0xE050-0xE057
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 8
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE070: ; ROXR		0xE070-0xE077
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE090: ; ROXR		0xE090-0xE097
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 8
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE0B0: ; ROXR		0xE0B0-0xE0B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE110: ; ROXL		0xE110-0xE117
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 8
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE130: ; ROXL		0xE130-0xE137
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE150: ; ROXL		0xE150-0xE157
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 8
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE170: ; ROXL		0xE170-0xE177
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE190: ; ROXL		0xE190-0xE197
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 8
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE1B0: ; ROXL		0xE1B0-0xE1B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+0*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE210: ; ROXR		0xE210-0xE217
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 1
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE230: ; ROXR		0xE230-0xE237
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE250: ; ROXR		0xE250-0xE257
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 1
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE270: ; ROXR		0xE270-0xE277
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE290: ; ROXR		0xE290-0xE297
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 1
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE2B0: ; ROXR		0xE2B0-0xE2B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE310: ; ROXL		0xE310-0xE317
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 1
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE330: ; ROXL		0xE330-0xE337
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE350: ; ROXL		0xE350-0xE357
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 1
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE370: ; ROXL		0xE370-0xE377
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE390: ; ROXL		0xE390-0xE397
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 1
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE3B0: ; ROXL		0xE3B0-0xE3B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+1*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE410: ; ROXR		0xE410-0xE417
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 2
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE430: ; ROXR		0xE430-0xE437
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE450: ; ROXR		0xE450-0xE457
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 2
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE470: ; ROXR		0xE470-0xE477
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE490: ; ROXR		0xE490-0xE497
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 2
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4B0: ; ROXR		0xE4B0-0xE4B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE510: ; ROXL		0xE510-0xE517
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 2
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE530: ; ROXL		0xE530-0xE537
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE550: ; ROXL		0xE550-0xE557
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 2
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE570: ; ROXL		0xE570-0xE577
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE590: ; ROXL		0xE590-0xE597
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 2
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5B0: ; ROXL		0xE5B0-0xE5B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+2*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE610: ; ROXR		0xE610-0xE617
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 3
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE630: ; ROXR		0xE630-0xE637
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE650: ; ROXR		0xE650-0xE657
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 3
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE670: ; ROXR		0xE670-0xE677
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE690: ; ROXR		0xE690-0xE697
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 3
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE6B0: ; ROXR		0xE6B0-0xE6B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE710: ; ROXL		0xE710-0xE717
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 3
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE730: ; ROXL		0xE730-0xE737
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE750: ; ROXL		0xE750-0xE757
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 3
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE770: ; ROXL		0xE770-0xE777
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE790: ; ROXL		0xE790-0xE797
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 3
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE7B0: ; ROXL		0xE7B0-0xE7B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+3*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE810: ; ROXR		0xE810-0xE817
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 4
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE830: ; ROXR		0xE830-0xE837
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE850: ; ROXR		0xE850-0xE857
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 4
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE870: ; ROXR		0xE870-0xE877
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE890: ; ROXR		0xE890-0xE897
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 4
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE8B0: ; ROXR		0xE8B0-0xE8B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE910: ; ROXL		0xE910-0xE917
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 4
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE930: ; ROXL		0xE930-0xE937
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE950: ; ROXL		0xE950-0xE957
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 4
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE970: ; ROXL		0xE970-0xE977
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE990: ; ROXL		0xE990-0xE997
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 4
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE9B0: ; ROXL		0xE9B0-0xE9B7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+4*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA10: ; ROXR		0xEA10-0xEA17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 5
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA30: ; ROXR		0xEA30-0xEA37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA50: ; ROXR		0xEA50-0xEA57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 5
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA70: ; ROXR		0xEA70-0xEA77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEA90: ; ROXR		0xEA90-0xEA97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 5
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEAB0: ; ROXR		0xEAB0-0xEAB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB10: ; ROXL		0xEB10-0xEB17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 5
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB30: ; ROXL		0xEB30-0xEB37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB50: ; ROXL		0xEB50-0xEB57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 5
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB70: ; ROXL		0xEB70-0xEB77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEB90: ; ROXL		0xEB90-0xEB97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 5
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEBB0: ; ROXL		0xEBB0-0xEBB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+5*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC10: ; ROXR		0xEC10-0xEC17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 6
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC30: ; ROXR		0xEC30-0xEC37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC50: ; ROXR		0xEC50-0xEC57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 6
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC70: ; ROXR		0xEC70-0xEC77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEC90: ; ROXR		0xEC90-0xEC97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 6
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IECB0: ; ROXR		0xECB0-0xECB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED10: ; ROXL		0xED10-0xED17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 6
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED30: ; ROXL		0xED30-0xED37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED50: ; ROXL		0xED50-0xED57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 6
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED70: ; ROXL		0xED70-0xED77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IED90: ; ROXL		0xED90-0xED97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 6
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEDB0: ; ROXL		0xEDB0-0xEDB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+6*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE10: ; ROXR		0xEE10-0xEE17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, byte 7
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE30: ; ROXR		0xEE30-0xEE37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE50: ; ROXR		0xEE50-0xEE57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, byte 7
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE70: ; ROXR		0xEE70-0xEE77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEE90: ; ROXR		0xEE90-0xEE97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, byte 7
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEEB0: ; ROXR		0xEEB0-0xEEB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcr   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF10: ; ROXL		0xEF10-0xEF17
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, byte 7
setc al
test dl, dl
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dl
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF30: ; ROXL		0xEF30-0xEF37
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dl, cl
setc al
test dl, dl
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dl
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF50: ; ROXL		0xEF50-0xEF57
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, byte 7
setc al
test dx, dx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], dx
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF70: ; ROXL		0xEF70-0xEF77
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   dx, cl
setc al
test dx, dx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], dx
shl  cl, byte 1
add  cl, byte 6
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEF90: ; ROXL		0xEF90-0xEF97
and  edi, byte 7
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, byte 7
setc al
test edx, edx
lahf
or       ah, al
test     ah, 1
setnz    byte [x]
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IEFB0: ; ROXL		0xEFB0-0xEFB7
and  edi, byte 7
mov  ebx, ecx
mov  ecx, [d+7*4]
and  ecx, byte 0x3f
mov  dl, [x]
shr  dl, byte 1
mov  edx, [d+edi*4]
rcl   edx, cl
setc al
test edx, edx
lahf
test cl, cl
jz   short .zero_count
or       ah, al
test     ah, 1
setnz    byte [x]
jmp      short .skip
.zero_count:
mov  al, [x]
and  al, byte 1
and  ah, 0xfe
or   ah, al
.skip:
xor  al, al
mov  [d+edi*4], edx
shl  cl, byte 1
add  cl, byte 8
neg  ecx
add  ecx, ebx
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4D0: ; ROXR		0xE4D0-0xE4D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4D8: ; ROXR		0xE4D8-0xE4DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4E0: ; ROXR		0xE4E0-0xE4E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4E8: ; ROXR		0xE4E8-0xE4EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4F0: ; ROXR		0xE4F0-0xE4F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4F8: ; ROXR		0xE4F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE4F9: ; ROXR		0xE4F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
shr  byte [x], 1
rcr  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5D0: ; ROXL		0xE5D0-0xE5D7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5D8: ; ROXL		0xE5D8-0xE5DF
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5E0: ; ROXL		0xE5E0-0xE5E7
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5E8: ; ROXL		0xE5E8-0xE5EF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5F0: ; ROXL		0xE5F0-0xE5F7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5F8: ; ROXL		0xE5F8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IE5F9: ; ROXL		0xE5F9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
shr  byte [x], 1
rcl  dx, byte 1
setc al
setc byte [x]
test dx, dx
lahf
or   ah, al
xor  al, al
call WriteWord
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E73: ; RTE		0x4E73
test byte [sr+1], 0x20
jz   near exception_privilege_violation
mov  ebx, [a+7*4]
mov  [run_ebx], ebx
call ReadWord
mov  ebx, [run_ebx]
and  edx, 0xa71f
mov  [sr], edx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
add  ebx, byte 2
mov  [run_ebx], ebx
call ReadLong
mov  ebx, [run_ebx]
add  ebx, byte 4
mov  [a+7*4], ebx
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
test byte [sr+1], 0x20
jnz  short .in_s
mov  ebx, [a+7*4]
mov  edx, [__sp]
mov  [a+7*4], edx
mov  [__sp], ebx
.in_s:
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E77: ; RTR		0x4E77
mov  ebx, [a+7*4]
call ReadWord
and  dl, 0x1f
mov  [sr], dl
add  dword [a+7*4], byte 2
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
mov  ebx, [a+7*4]
call ReadLong
add  dword [a+7*4], byte 4
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E75: ; RTS		0x4E75
mov  ebx, [a+7*4]
call ReadLong
add  dword [a+7*4], byte 4
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8100: ; SBCD		0x8100-0x8107
and  edi, byte 7
mov  al, [d+0*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+0*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8108: ; SBCD		0x8108-0x810F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+0*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+0*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8300: ; SBCD		0x8300-0x8307
and  edi, byte 7
mov  al, [d+1*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+1*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8308: ; SBCD		0x8308-0x830F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+1*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+1*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8500: ; SBCD		0x8500-0x8507
and  edi, byte 7
mov  al, [d+2*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+2*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8508: ; SBCD		0x8508-0x850F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+2*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+2*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8700: ; SBCD		0x8700-0x8707
and  edi, byte 7
mov  al, [d+3*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+3*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8708: ; SBCD		0x8708-0x870F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+3*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+3*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8900: ; SBCD		0x8900-0x8907
and  edi, byte 7
mov  al, [d+4*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+4*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8908: ; SBCD		0x8908-0x890F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+4*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+4*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B00: ; SBCD		0x8B00-0x8B07
and  edi, byte 7
mov  al, [d+5*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+5*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8B08: ; SBCD		0x8B08-0x8B0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+5*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+5*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D00: ; SBCD		0x8D00-0x8D07
and  edi, byte 7
mov  al, [d+6*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+6*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8D08: ; SBCD		0x8D08-0x8D0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+6*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+6*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F00: ; SBCD		0x8F00-0x8F07
and  edi, byte 7
mov  al, [d+7*4]
mov  bl, [d+edi*4]
shr  byte [x], 1
sbb  al, bl
das
setc dl
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, dl
mov  [d+7*4], al
xor  al, al
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I8F08: ; SBCD		0x8F08-0x8F0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], dl
mov  ebx, [a+7*4]
call ReadByte
mov  al, dl
mov  dl, [run_edx]
shr  byte [x], 1
sbb  al, dl
das
setc byte [x]
jz   short .unchanged_z
and  ah, 0xbf
.unchanged_z:
and  ah, 0x7e
or   ah, [x]
mov  dl, al
call WriteByte
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50C0: ; ST		0x50C0-0x50C7
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50D0: ; ST		0x50D0-0x50D7
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50D8: ; ST		0x50D8-0x50DF
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50E0: ; ST		0x50E0-0x50E7
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50E8: ; ST		0x50E8-0x50EF
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50F0: ; ST		0x50F0-0x50F7
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50F8: ; ST		0x50F8
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I50F9: ; ST		0x50F9
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51C0: ; SF		0x51C0-0x51C7
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51D0: ; SF		0x51D0-0x51D7
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51D8: ; SF		0x51D8-0x51DF
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51E0: ; SF		0x51E0-0x51E7
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51E8: ; SF		0x51E8-0x51EF
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51F0: ; SF		0x51F0-0x51F7
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51F8: ; SF		0x51F8
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I51F9: ; SF		0x51F9
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52C0: ; SHI		0x52C0-0x52C7
test ah, byte 0x41
jz   short do21184
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21184:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52D0: ; SHI		0x52D0-0x52D7
test ah, byte 0x41
jz   short do21200
false21200:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21200:
true21200:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52D8: ; SHI		0x52D8-0x52DF
test ah, byte 0x41
jz   short do21208
false21208:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21208:
true21208:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52E0: ; SHI		0x52E0-0x52E7
test ah, byte 0x41
jz   short do21216
false21216:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21216:
true21216:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52E8: ; SHI		0x52E8-0x52EF
test ah, byte 0x41
jz   short do21224
false21224:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21224:
true21224:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52F0: ; SHI		0x52F0-0x52F7
test ah, byte 0x41
jz   short do21232
false21232:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21232:
true21232:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52F8: ; SHI		0x52F8
test ah, byte 0x41
jz   short do21240
false21240:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21240:
true21240:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I52F9: ; SHI		0x52F9
test ah, byte 0x41
jz   short do21241
false21241:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21241:
true21241:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53C0: ; SLS		0x53C0-0x53C7
test ah, byte 0x41
jnz  short do21440
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21440:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53D0: ; SLS		0x53D0-0x53D7
test ah, byte 0x41
jnz  short do21456
false21456:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21456:
true21456:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53D8: ; SLS		0x53D8-0x53DF
test ah, byte 0x41
jnz  short do21464
false21464:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21464:
true21464:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53E0: ; SLS		0x53E0-0x53E7
test ah, byte 0x41
jnz  short do21472
false21472:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21472:
true21472:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53E8: ; SLS		0x53E8-0x53EF
test ah, byte 0x41
jnz  short do21480
false21480:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21480:
true21480:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53F0: ; SLS		0x53F0-0x53F7
test ah, byte 0x41
jnz  short do21488
false21488:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21488:
true21488:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53F8: ; SLS		0x53F8
test ah, byte 0x41
jnz  short do21496
false21496:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21496:
true21496:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I53F9: ; SLS		0x53F9
test ah, byte 0x41
jnz  short do21497
false21497:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21497:
true21497:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54C0: ; SCC		0x54C0-0x54C7
test ah, byte 1
jz   short do21696
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21696:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54D0: ; SCC		0x54D0-0x54D7
test ah, byte 1
jz   short do21712
false21712:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21712:
true21712:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54D8: ; SCC		0x54D8-0x54DF
test ah, byte 1
jz   short do21720
false21720:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21720:
true21720:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54E0: ; SCC		0x54E0-0x54E7
test ah, byte 1
jz   short do21728
false21728:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21728:
true21728:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54E8: ; SCC		0x54E8-0x54EF
test ah, byte 1
jz   short do21736
false21736:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21736:
true21736:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54F0: ; SCC		0x54F0-0x54F7
test ah, byte 1
jz   short do21744
false21744:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21744:
true21744:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54F8: ; SCC		0x54F8
test ah, byte 1
jz   short do21752
false21752:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21752:
true21752:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I54F9: ; SCC		0x54F9
test ah, byte 1
jz   short do21753
false21753:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21753:
true21753:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55C0: ; SCS		0x55C0-0x55C7
test ah, byte 1
jnz  short do21952
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21952:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55D0: ; SCS		0x55D0-0x55D7
test ah, byte 1
jnz  short do21968
false21968:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21968:
true21968:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55D8: ; SCS		0x55D8-0x55DF
test ah, byte 1
jnz  short do21976
false21976:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21976:
true21976:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55E0: ; SCS		0x55E0-0x55E7
test ah, byte 1
jnz  short do21984
false21984:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21984:
true21984:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55E8: ; SCS		0x55E8-0x55EF
test ah, byte 1
jnz  short do21992
false21992:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do21992:
true21992:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55F0: ; SCS		0x55F0-0x55F7
test ah, byte 1
jnz  short do22000
false22000:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22000:
true22000:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55F8: ; SCS		0x55F8
test ah, byte 1
jnz  short do22008
false22008:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22008:
true22008:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I55F9: ; SCS		0x55F9
test ah, byte 1
jnz  short do22009
false22009:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22009:
true22009:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56C0: ; SNE		0x56C0-0x56C7
test ah, byte 0x40
jz   short do22208
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22208:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56D0: ; SNE		0x56D0-0x56D7
test ah, byte 0x40
jz   short do22224
false22224:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22224:
true22224:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56D8: ; SNE		0x56D8-0x56DF
test ah, byte 0x40
jz   short do22232
false22232:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22232:
true22232:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56E0: ; SNE		0x56E0-0x56E7
test ah, byte 0x40
jz   short do22240
false22240:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22240:
true22240:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56E8: ; SNE		0x56E8-0x56EF
test ah, byte 0x40
jz   short do22248
false22248:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22248:
true22248:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56F0: ; SNE		0x56F0-0x56F7
test ah, byte 0x40
jz   short do22256
false22256:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22256:
true22256:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56F8: ; SNE		0x56F8
test ah, byte 0x40
jz   short do22264
false22264:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22264:
true22264:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I56F9: ; SNE		0x56F9
test ah, byte 0x40
jz   short do22265
false22265:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22265:
true22265:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57C0: ; SEQ		0x57C0-0x57C7
test ah, byte 0x40
jnz  short do22464
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22464:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57D0: ; SEQ		0x57D0-0x57D7
test ah, byte 0x40
jnz  short do22480
false22480:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22480:
true22480:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57D8: ; SEQ		0x57D8-0x57DF
test ah, byte 0x40
jnz  short do22488
false22488:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22488:
true22488:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57E0: ; SEQ		0x57E0-0x57E7
test ah, byte 0x40
jnz  short do22496
false22496:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22496:
true22496:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57E8: ; SEQ		0x57E8-0x57EF
test ah, byte 0x40
jnz  short do22504
false22504:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22504:
true22504:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57F0: ; SEQ		0x57F0-0x57F7
test ah, byte 0x40
jnz  short do22512
false22512:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22512:
true22512:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57F8: ; SEQ		0x57F8
test ah, byte 0x40
jnz  short do22520
false22520:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22520:
true22520:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I57F9: ; SEQ		0x57F9
test ah, byte 0x40
jnz  short do22521
false22521:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22521:
true22521:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58C0: ; SVC		0x58C0-0x58C7
test al, byte 1
jz   short do22720
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22720:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58D0: ; SVC		0x58D0-0x58D7
test al, byte 1
jz   short do22736
false22736:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22736:
true22736:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58D8: ; SVC		0x58D8-0x58DF
test al, byte 1
jz   short do22744
false22744:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22744:
true22744:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58E0: ; SVC		0x58E0-0x58E7
test al, byte 1
jz   short do22752
false22752:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22752:
true22752:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58E8: ; SVC		0x58E8-0x58EF
test al, byte 1
jz   short do22760
false22760:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22760:
true22760:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58F0: ; SVC		0x58F0-0x58F7
test al, byte 1
jz   short do22768
false22768:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22768:
true22768:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58F8: ; SVC		0x58F8
test al, byte 1
jz   short do22776
false22776:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22776:
true22776:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I58F9: ; SVC		0x58F9
test al, byte 1
jz   short do22777
false22777:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22777:
true22777:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59C0: ; SVS		0x59C0-0x59C7
test al, byte 1
jnz  short do22976
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22976:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59D0: ; SVS		0x59D0-0x59D7
test al, byte 1
jnz  short do22992
false22992:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do22992:
true22992:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59D8: ; SVS		0x59D8-0x59DF
test al, byte 1
jnz  short do23000
false23000:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23000:
true23000:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59E0: ; SVS		0x59E0-0x59E7
test al, byte 1
jnz  short do23008
false23008:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23008:
true23008:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59E8: ; SVS		0x59E8-0x59EF
test al, byte 1
jnz  short do23016
false23016:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23016:
true23016:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59F0: ; SVS		0x59F0-0x59F7
test al, byte 1
jnz  short do23024
false23024:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23024:
true23024:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59F8: ; SVS		0x59F8
test al, byte 1
jnz  short do23032
false23032:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23032:
true23032:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I59F9: ; SVS		0x59F9
test al, byte 1
jnz  short do23033
false23033:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23033:
true23033:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AC0: ; SPL		0x5AC0-0x5AC7
test ah, byte 0x80
jz   short do23232
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23232:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AD0: ; SPL		0x5AD0-0x5AD7
test ah, byte 0x80
jz   short do23248
false23248:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23248:
true23248:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AD8: ; SPL		0x5AD8-0x5ADF
test ah, byte 0x80
jz   short do23256
false23256:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23256:
true23256:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AE0: ; SPL		0x5AE0-0x5AE7
test ah, byte 0x80
jz   short do23264
false23264:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23264:
true23264:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AE8: ; SPL		0x5AE8-0x5AEF
test ah, byte 0x80
jz   short do23272
false23272:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23272:
true23272:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AF0: ; SPL		0x5AF0-0x5AF7
test ah, byte 0x80
jz   short do23280
false23280:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23280:
true23280:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AF8: ; SPL		0x5AF8
test ah, byte 0x80
jz   short do23288
false23288:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23288:
true23288:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5AF9: ; SPL		0x5AF9
test ah, byte 0x80
jz   short do23289
false23289:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23289:
true23289:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BC0: ; SMI		0x5BC0-0x5BC7
test ah, byte 0x80
jnz  short do23488
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23488:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BD0: ; SMI		0x5BD0-0x5BD7
test ah, byte 0x80
jnz  short do23504
false23504:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23504:
true23504:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BD8: ; SMI		0x5BD8-0x5BDF
test ah, byte 0x80
jnz  short do23512
false23512:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23512:
true23512:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BE0: ; SMI		0x5BE0-0x5BE7
test ah, byte 0x80
jnz  short do23520
false23520:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23520:
true23520:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BE8: ; SMI		0x5BE8-0x5BEF
test ah, byte 0x80
jnz  short do23528
false23528:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23528:
true23528:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BF0: ; SMI		0x5BF0-0x5BF7
test ah, byte 0x80
jnz  short do23536
false23536:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23536:
true23536:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BF8: ; SMI		0x5BF8
test ah, byte 0x80
jnz  short do23544
false23544:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23544:
true23544:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5BF9: ; SMI		0x5BF9
test ah, byte 0x80
jnz  short do23545
false23545:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23545:
true23545:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CC0: ; SGE		0x5CC0-0x5CC7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23744
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23744:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CD0: ; SGE		0x5CD0-0x5CD7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23760
false23760:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23760:
true23760:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CD8: ; SGE		0x5CD8-0x5CDF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23768
false23768:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23768:
true23768:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CE0: ; SGE		0x5CE0-0x5CE7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23776
false23776:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23776:
true23776:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CE8: ; SGE		0x5CE8-0x5CEF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23784
false23784:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23784:
true23784:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CF0: ; SGE		0x5CF0-0x5CF7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23792
false23792:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23792:
true23792:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CF8: ; SGE		0x5CF8
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23800
false23800:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23800:
true23800:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5CF9: ; SGE		0x5CF9
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do23801
false23801:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do23801:
true23801:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DC0: ; SLT		0x5DC0-0x5DC7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24000
.dont_do:
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24000:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DD0: ; SLT		0x5DD0-0x5DD7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24016
.dont_do:
false24016:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24016:
true24016:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DD8: ; SLT		0x5DD8-0x5DDF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24024
.dont_do:
false24024:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24024:
true24024:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DE0: ; SLT		0x5DE0-0x5DE7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24032
.dont_do:
false24032:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24032:
true24032:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DE8: ; SLT		0x5DE8-0x5DEF
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24040
.dont_do:
false24040:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24040:
true24040:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DF0: ; SLT		0x5DF0-0x5DF7
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24048
.dont_do:
false24048:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24048:
true24048:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DF8: ; SLT		0x5DF8
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24056
.dont_do:
false24056:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24056:
true24056:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5DF9: ; SLT		0x5DF9
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24057
.dont_do:
false24057:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24057:
true24057:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EC0: ; SGT		0x5EC0-0x5EC7
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24256
.dont_do:
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24256:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5ED0: ; SGT		0x5ED0-0x5ED7
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24272
.dont_do:
false24272:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24272:
true24272:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5ED8: ; SGT		0x5ED8-0x5EDF
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24280
.dont_do:
false24280:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24280:
true24280:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EE0: ; SGT		0x5EE0-0x5EE7
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24288
.dont_do:
false24288:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24288:
true24288:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EE8: ; SGT		0x5EE8-0x5EEF
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24296
.dont_do:
false24296:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24296:
true24296:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EF0: ; SGT		0x5EF0-0x5EF7
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24304
.dont_do:
false24304:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24304:
true24304:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EF8: ; SGT		0x5EF8
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24312
.dont_do:
false24312:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24312:
true24312:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5EF9: ; SGT		0x5EF9
test ah, byte 0x40
jnz  short .dont_do
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpe  short do24313
.dont_do:
false24313:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24313:
true24313:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FC0: ; SLE		0x5FC0-0x5FC7
test ah, 0x40
jnz  short do24512
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24512
and  edi, byte 7
mov  byte [d+edi*4], 0
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24512:
and  edi, byte 7
mov  byte [d+edi*4], 0xff
sub  ecx, byte 6
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FD0: ; SLE		0x5FD0-0x5FD7
test ah, 0x40
jnz  short do24528
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24528
false24528:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24528:
true24528:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FD8: ; SLE		0x5FD8-0x5FDF
test ah, 0x40
jnz  short do24536
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24536
false24536:
xor  dl, dl
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24536:
true24536:
mov  dl, 0xff
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call WriteByte
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FE0: ; SLE		0x5FE0-0x5FE7
test ah, 0x40
jnz  short do24544
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24544
false24544:
xor  dl, dl
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24544:
true24544:
mov  dl, 0xff
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FE8: ; SLE		0x5FE8-0x5FEF
test ah, 0x40
jnz  short do24552
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24552
false24552:
xor  dl, dl
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24552:
true24552:
mov  dl, 0xff
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FF0: ; SLE		0x5FF0-0x5FF7
test ah, 0x40
jnz  short do24560
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24560
false24560:
xor  dl, dl
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24560:
true24560:
mov  dl, 0xff
and      edi, byte 7
mov  [memhandler_edi], edi
xor      ebx, ebx
mov      bx, [esi]
movsx    edi, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edi
mov  edi, [memhandler_edi]
add      ebx, [a+edi*4]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FF8: ; SLE		0x5FF8
test ah, 0x40
jnz  short do24568
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24568
false24568:
xor  dl, dl
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24568:
true24568:
mov  dl, 0xff
movsx    ebx, word [esi]
add      esi, byte 2
call WriteByte
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I5FF9: ; SLE		0x5FF9
test ah, 0x40
jnz  short do24569
mov  bl, al
mov  bh, ah
and  bh, 0xfe
or   bl, bh
test bl, byte 0x81
jpo  short do24569
false24569:
xor  dl, dl
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
do24569:
true24569:
mov  dl, 0xff
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E72: ; STOP		0x4E72
test byte [sr+1], 0x20
jz   near exception_privilege_violation
mov  edx, [esi]
add  esi, byte 2
and  edx, 0xa71f
mov  [sr], dx
mov      dh, [sr]
mov      ah, dh
shl      ah, byte 4
test     dh, byte 2
setnz    al
test     dh, 0x10
setnz    byte [x]
and      dh, byte 1
or       ah, dh
or   byte [status], 2
sub  ecx, byte 4
js   near Genital68KRun_done
mov  dword [remaining], 0
jmp  near Genital68KRun_done
times ($$-$) & 3 nop
I9100: ; SUBX		0x9100-0x9107
and  edi, byte 7
mov  bl, [d+0*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9108: ; SUBX		0x9108-0x910F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+0*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9140: ; SUBX		0x9140-0x9147
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9148: ; SUBX		0x9148-0x914F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+0*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9180: ; SUBX		0x9180-0x9187
and  edi, byte 7
mov  ebx, [d+0*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+0*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9188: ; SUBX		0x9188-0x918F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+0*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+0*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9300: ; SUBX		0x9300-0x9307
and  edi, byte 7
mov  bl, [d+1*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9308: ; SUBX		0x9308-0x930F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+1*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9340: ; SUBX		0x9340-0x9347
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9348: ; SUBX		0x9348-0x934F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+1*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9380: ; SUBX		0x9380-0x9387
and  edi, byte 7
mov  ebx, [d+1*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+1*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9388: ; SUBX		0x9388-0x938F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+1*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+1*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9500: ; SUBX		0x9500-0x9507
and  edi, byte 7
mov  bl, [d+2*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9508: ; SUBX		0x9508-0x950F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+2*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9540: ; SUBX		0x9540-0x9547
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9548: ; SUBX		0x9548-0x954F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+2*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9580: ; SUBX		0x9580-0x9587
and  edi, byte 7
mov  ebx, [d+2*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+2*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9588: ; SUBX		0x9588-0x958F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+2*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+2*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9700: ; SUBX		0x9700-0x9707
and  edi, byte 7
mov  bl, [d+3*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9708: ; SUBX		0x9708-0x970F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+3*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9740: ; SUBX		0x9740-0x9747
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9748: ; SUBX		0x9748-0x974F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+3*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9780: ; SUBX		0x9780-0x9787
and  edi, byte 7
mov  ebx, [d+3*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+3*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9788: ; SUBX		0x9788-0x978F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+3*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+3*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9900: ; SUBX		0x9900-0x9907
and  edi, byte 7
mov  bl, [d+4*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9908: ; SUBX		0x9908-0x990F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+4*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9940: ; SUBX		0x9940-0x9947
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9948: ; SUBX		0x9948-0x994F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+4*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9980: ; SUBX		0x9980-0x9987
and  edi, byte 7
mov  ebx, [d+4*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+4*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9988: ; SUBX		0x9988-0x998F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+4*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+4*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B00: ; SUBX		0x9B00-0x9B07
and  edi, byte 7
mov  bl, [d+5*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B08: ; SUBX		0x9B08-0x9B0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+5*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B40: ; SUBX		0x9B40-0x9B47
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B48: ; SUBX		0x9B48-0x9B4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+5*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B80: ; SUBX		0x9B80-0x9B87
and  edi, byte 7
mov  ebx, [d+5*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+5*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9B88: ; SUBX		0x9B88-0x9B8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+5*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+5*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D00: ; SUBX		0x9D00-0x9D07
and  edi, byte 7
mov  bl, [d+6*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D08: ; SUBX		0x9D08-0x9D0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
dec  dword [a+6*4]
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D40: ; SUBX		0x9D40-0x9D47
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D48: ; SUBX		0x9D48-0x9D4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+6*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D80: ; SUBX		0x9D80-0x9D87
and  edi, byte 7
mov  ebx, [d+6*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+6*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9D88: ; SUBX		0x9D88-0x9D8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+6*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+6*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F00: ; SUBX		0x9F00-0x9F07
and  edi, byte 7
mov  bl, [d+7*4]
mov  dl, [d+edi*4]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], bl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F08: ; SUBX		0x9F08-0x9F0F
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadByte
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadByte
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bl, dl
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F40: ; SUBX		0x9F40-0x9F47
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], bx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F48: ; SUBX		0x9F48-0x9F4F
and  edi, byte 7
sub  dword [a+edi*4], byte 2
sub  dword [a+7*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadWord
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  bx, dx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteWord
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F80: ; SUBX		0x9F80-0x9F87
and  edi, byte 7
mov  ebx, [d+7*4]
mov  edx, [d+edi*4]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  [d+7*4], ebx
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I9F88: ; SUBX		0x9F88-0x9F8F
and  edi, byte 7
sub  dword [a+edi*4], byte 4
sub  dword [a+7*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
mov  [run_edx], edx
mov  ebx, [a+7*4]
call ReadLong
mov  [run_ebx], ebx
mov  ebx, edx
mov  edx, [run_edx]
shr  byte [x], 1
sbb  ebx, edx
mov  dl, ah
lahf
setc byte [x]
seto al
jnz  short .z
and  dl, 0x40
and  ah, 0xbf
or   ah, dl
.z:
mov  edx, ebx
mov  ebx, [run_ebx]
call WriteLong
sub  ecx, byte 30
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4840: ; SWAP		0x4840-0x4847
and  edi, byte 7
mov  edx, [d+edi*4]
rol  edx, byte 16
test edx, edx
lahf
xor  al, al
mov  [d+edi*4], edx
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AC0: ; TAS		0x4AC0-0x4AC7
and  edi, byte 7
mov  dl, [d+edi*4]
test dl, dl
lahf
xor  al, al
or   dl, 0x80
mov  [d+edi*4], dl
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AD0: ; TAS		0x4AD0-0x4AD7
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AD8: ; TAS		0x4AD8-0x4ADF
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AE0: ; TAS		0x4AE0-0x4AE7
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AE8: ; TAS		0x4AE8-0x4AEF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AF0: ; TAS		0x4AF0-0x4AF7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 24
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AF8: ; TAS		0x4AF8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 22
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AF9: ; TAS		0x4AF9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
or   dl, 0x80
call WriteByte
sub  ecx, byte 26
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E40: ; TRAP		0x4E40-0x4E4F
and  edi, byte 0xf
mov  [run_edi], edi
mov  edx, 0x2000
xor  edx, [sr]
test edx, 0x2000
jz   .no_sp_magic
mov  edx, [__sp]
xchg [a+7*4], edx
mov  [__sp], edx
.no_sp_magic:
mov  ebx, [a+7*4]
mov  edx, esi
sub  edx, ebp
sub  ebx, byte 4
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
mov  edx, [sr]
sub  ebx, byte 2
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
or   dh, 0x20
and  edx, 0xa71f
xchg [sr], edx
mov  [a+7*4], ebx
mov  edi, [run_edi]
mov  ebx, edi
add  ebx, byte 32
shl  ebx, 2
call ReadLong
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
xor  edi, edi
sub  ecx, byte 34
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E76: ; TRAPV		0x4E76
test al, al
jz   near .no_trap
mov  edx, 0x2000
xor  edx, [sr]
test edx, 0x2000
jz   .no_sp_magic
mov  edx, [__sp]
xchg [a+7*4], edx
mov  [__sp], edx
.no_sp_magic:
mov  ebx, [a+7*4]
mov  edx, esi
sub  edx, ebp
sub  ebx, byte 4
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
mov  edx, [sr]
sub  ebx, byte 2
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
or   dh, 0x20
and  edx, 0xa71f
xchg [sr], edx
mov  [a+7*4], ebx
mov  ebx, 7*4
call ReadLong
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
sub  ecx, byte 38
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
.no_trap:
xor  edi, edi
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A00: ; TST		0x4A00-0x4A07
and  edi, byte 7
mov  edx, [d+edi*4]
test dl, dl
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A10: ; TST		0x4A10-0x4A17
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A18: ; TST		0x4A18-0x4A1F
and  edi, byte 7
mov  ebx, [a+edi*4]
cmp  edi, byte 7
cmc
adc  dword [a+edi*4], byte 1
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A20: ; TST		0x4A20-0x4A27
and  edi, byte 7
cmp  edi, byte 7
cmc
sbb  dword [a+edi*4], byte 1
mov  ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A28: ; TST		0x4A28-0x4A2F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A30: ; TST		0x4A30-0x4A37
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A38: ; TST		0x4A38
movsx    ebx, word [esi]
add      esi, byte 2
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A39: ; TST		0x4A39
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadByte
test dl, dl
lahf
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A40: ; TST		0x4A40-0x4A47
and  edi, byte 7
mov  edx, [d+edi*4]
test dx, dx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A50: ; TST		0x4A50-0x4A57
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A58: ; TST		0x4A58-0x4A5F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 8
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A60: ; TST		0x4A60-0x4A67
and  edi, byte 7
sub  dword [a+edi*4], byte 2
mov  ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 10
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A68: ; TST		0x4A68-0x4A6F
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A70: ; TST		0x4A70-0x4A77
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A78: ; TST		0x4A78
movsx    ebx, word [esi]
add      esi, byte 2
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A79: ; TST		0x4A79
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadWord
test dx, dx
lahf
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A80: ; TST		0x4A80-0x4A87
and  edi, byte 7
mov  edx, [d+edi*4]
test edx, edx
lahf
xor  al, al
sub  ecx, byte 4
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A90: ; TST		0x4A90-0x4A97
and  edi, byte 7
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4A98: ; TST		0x4A98-0x4A9F
and  edi, byte 7
mov  ebx, [a+edi*4]
add  dword [a+edi*4], byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AA0: ; TST		0x4AA0-0x4AA7
and  edi, byte 7
sub  dword [a+edi*4], byte 4
mov  ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 14
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AA8: ; TST		0x4AA8-0x4AAF
and  edi, byte 7
movsx    ebx, word [esi]
add      ebx, [a+edi*4]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AB0: ; TST		0x4AB0-0x4AB7
and      edi, byte 7
xor      ebx, ebx
mov      bx, [esi]
movsx    edx, bl
add      esi, byte 2
shr      ebx, byte 12
mov      ebx, [d+ebx*4]
jc       short .ll
movsx    ebx, bx
.ll:
add      ebx, edx
add      ebx, [a+edi*4]
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 18
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AB8: ; TST		0x4AB8
movsx    ebx, word [esi]
add      esi, byte 2
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 16
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4AB9: ; TST		0x4AB9
mov      ebx, [esi]
ror      ebx, byte 16
add      esi, byte 4
call ReadLong
test edx, edx
lahf
xor  al, al
sub  ecx, byte 20
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
I4E58: ; UNLK		0x4E58-0x4E5F
and  edi, byte 7
mov  ebx, [a+edi*4]
mov  [run_ebx], ebx
mov  [run_edi], edi
call ReadLong
mov  edi, [run_edi]
mov  ebx, [run_ebx]
mov  [a+edi*4], edx
add  ebx, byte 4
mov  [a+7*4], ebx
sub  ecx, byte 12
js   near Genital68KRun_done
mov  di, [esi]
add  esi, byte 2
jmp  dword [jmptab+edi*4]
times ($$-$) & 3 nop
IA000: ; Line 1010 Emulator		0xA000-0xAFFF
jmp  near exception_line_1010_emulator
times ($$-$) & 3 nop
IF000: ; Line 1111 Emulator		0xF000-0xFFFF
jmp  near exception_line_1111_emulator
I_Invalid:
jmp  near exception_illegal_instruction
times ($$-$) & 3 nop
exception_chk:
push dword 6*4
jmp  short do_exception
times ($$-$) & 3 nop
exception_divide_by_zero:
push dword 5*4
jmp  short do_exception
times ($$-$) & 3 nop
exception_privilege_violation:
push dword 8*4
jmp  short do_exception_fix_pc
times ($$-$) & 3 nop
exception_illegal_instruction:
push dword 4*4
jmp  short do_exception_fix_pc
times ($$-$) & 3 nop
exception_line_1010_emulator:
push dword 10*4
jmp  short do_exception_fix_pc
times ($$-$) & 3 nop
exception_line_1111_emulator:
push dword 11*4
jmp  short do_exception_fix_pc
do_exception_fix_pc:
sub  esi, byte 2
do_exception:
mov  edx, 0x2000
xor  edx, [sr]
test edx, 0x2000
jz   .no_sp_magic
mov  edx, [__sp]
xchg [a+7*4], edx
mov  [__sp], edx
.no_sp_magic:
mov  ebx, [a+7*4]
mov  edx, esi
sub  edx, ebp
sub  ebx, byte 4
mov  [run_ebx], ebx
call WriteLong
mov  ebx, [run_ebx]
mov  edx, [sr]
sub  ebx, byte 2
mov  [run_ebx], ebx
call WriteWord
mov  ebx, [run_ebx]
or   dh, 0x20
and  edx, 0xa71f
mov  [sr], edx
mov  [a+7*4], ebx
pop  ebx
call ReadLong
mov  esi, edx
mov  [fetch_esi], esi
and  esi, 0x00FFFFFF
mov     edx, [fetch]
.find_fetch_loop0:
add     edx, byte 12
cmp     dword [edx], byte -1
je      near fetch_error
cmp     esi, [edx]
jb      .find_fetch_loop0
cmp     esi, [edx+4]
ja      .find_fetch_loop0
mov     ebp, [edx+8]
mov  edx, [fetch_esi]
mov  esi, edx
and  edx, 0xFF000000
sub  ebp, edx
add     esi, ebp
xor  edi, edi
sub  ecx, byte 40
mov  di, [esi]
add  esi, byte 2
jmp  [jmptab+edi*4]
section .bss
times ($$-$) & 31 nop
jmptab       resd 65536
section .data
times ($$-$) & 3 nop
compressed_jmptab:
dw 0x8000 + 1
dd I0000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0010
dd I0018
dd I0020
dd I0028
dd I0030
dw 0xc000 + 2
dd I0038
dd I0039
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I003C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0040
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0050
dd I0058
dd I0060
dd I0068
dd I0070
dw 0xc000 + 2
dd I0078
dd I0079
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I007C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0080
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0090
dd I0098
dd I00A0
dd I00A8
dd I00B0
dw 0xc000 + 2
dd I00B8
dd I00B9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0100
dd I0108
dd I0110
dd I0118
dd I0120
dd I0128
dd I0130
dw 0xc000 + 5
dd I0138
dd I0139
dd I013A
dd I013B
dd I013C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0140
dd I0148
dd I0150
dd I0158
dd I0160
dd I0168
dd I0170
dw 0xc000 + 2
dd I0178
dd I0179
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0180
dd I0188
dd I0190
dd I0198
dd I01A0
dd I01A8
dd I01B0
dw 0xc000 + 2
dd I01B8
dd I01B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I01C0
dd I01C8
dd I01D0
dd I01D8
dd I01E0
dd I01E8
dd I01F0
dw 0xc000 + 2
dd I01F8
dd I01F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0210
dd I0218
dd I0220
dd I0228
dd I0230
dw 0xc000 + 2
dd I0238
dd I0239
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I023C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0240
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0250
dd I0258
dd I0260
dd I0268
dd I0270
dw 0xc000 + 2
dd I0278
dd I0279
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I027C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0280
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0290
dd I0298
dd I02A0
dd I02A8
dd I02B0
dw 0xc000 + 2
dd I02B8
dd I02B9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0300
dd I0308
dd I0310
dd I0318
dd I0320
dd I0328
dd I0330
dw 0xc000 + 5
dd I0338
dd I0339
dd I033A
dd I033B
dd I033C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0340
dd I0348
dd I0350
dd I0358
dd I0360
dd I0368
dd I0370
dw 0xc000 + 2
dd I0378
dd I0379
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0380
dd I0388
dd I0390
dd I0398
dd I03A0
dd I03A8
dd I03B0
dw 0xc000 + 2
dd I03B8
dd I03B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I03C0
dd I03C8
dd I03D0
dd I03D8
dd I03E0
dd I03E8
dd I03F0
dw 0xc000 + 2
dd I03F8
dd I03F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0410
dd I0418
dd I0420
dd I0428
dd I0430
dw 0xc000 + 2
dd I0438
dd I0439
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0440
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0450
dd I0458
dd I0460
dd I0468
dd I0470
dw 0xc000 + 2
dd I0478
dd I0479
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0480
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0490
dd I0498
dd I04A0
dd I04A8
dd I04B0
dw 0xc000 + 2
dd I04B8
dd I04B9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0500
dd I0508
dd I0510
dd I0518
dd I0520
dd I0528
dd I0530
dw 0xc000 + 5
dd I0538
dd I0539
dd I053A
dd I053B
dd I053C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0540
dd I0548
dd I0550
dd I0558
dd I0560
dd I0568
dd I0570
dw 0xc000 + 2
dd I0578
dd I0579
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0580
dd I0588
dd I0590
dd I0598
dd I05A0
dd I05A8
dd I05B0
dw 0xc000 + 2
dd I05B8
dd I05B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I05C0
dd I05C8
dd I05D0
dd I05D8
dd I05E0
dd I05E8
dd I05F0
dw 0xc000 + 2
dd I05F8
dd I05F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0610
dd I0618
dd I0620
dd I0628
dd I0630
dw 0xc000 + 2
dd I0638
dd I0639
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0640
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0650
dd I0658
dd I0660
dd I0668
dd I0670
dw 0xc000 + 2
dd I0678
dd I0679
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0680
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0690
dd I0698
dd I06A0
dd I06A8
dd I06B0
dw 0xc000 + 2
dd I06B8
dd I06B9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0700
dd I0708
dd I0710
dd I0718
dd I0720
dd I0728
dd I0730
dw 0xc000 + 5
dd I0738
dd I0739
dd I073A
dd I073B
dd I073C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0740
dd I0748
dd I0750
dd I0758
dd I0760
dd I0768
dd I0770
dw 0xc000 + 2
dd I0778
dd I0779
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0780
dd I0788
dd I0790
dd I0798
dd I07A0
dd I07A8
dd I07B0
dw 0xc000 + 2
dd I07B8
dd I07B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I07C0
dd I07C8
dd I07D0
dd I07D8
dd I07E0
dd I07E8
dd I07F0
dw 0xc000 + 2
dd I07F8
dd I07F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0810
dd I0818
dd I0820
dd I0828
dd I0830
dw 0xc000 + 4
dd I0838
dd I0839
dd I083A
dd I083B
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I0840
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0850
dd I0858
dd I0860
dd I0868
dd I0870
dw 0xc000 + 2
dd I0878
dd I0879
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0880
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0890
dd I0898
dd I08A0
dd I08A8
dd I08B0
dw 0xc000 + 2
dd I08B8
dd I08B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I08C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I08D0
dd I08D8
dd I08E0
dd I08E8
dd I08F0
dw 0xc000 + 2
dd I08F8
dd I08F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0900
dd I0908
dd I0910
dd I0918
dd I0920
dd I0928
dd I0930
dw 0xc000 + 5
dd I0938
dd I0939
dd I093A
dd I093B
dd I093C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0940
dd I0948
dd I0950
dd I0958
dd I0960
dd I0968
dd I0970
dw 0xc000 + 2
dd I0978
dd I0979
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0980
dd I0988
dd I0990
dd I0998
dd I09A0
dd I09A8
dd I09B0
dw 0xc000 + 2
dd I09B8
dd I09B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I09C0
dd I09C8
dd I09D0
dd I09D8
dd I09E0
dd I09E8
dd I09F0
dw 0xc000 + 2
dd I09F8
dd I09F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0A10
dd I0A18
dd I0A20
dd I0A28
dd I0A30
dw 0xc000 + 2
dd I0A38
dd I0A39
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I0A3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0A40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0A50
dd I0A58
dd I0A60
dd I0A68
dd I0A70
dw 0xc000 + 2
dd I0A78
dd I0A79
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I0A7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I0A80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0A90
dd I0A98
dd I0AA0
dd I0AA8
dd I0AB0
dw 0xc000 + 2
dd I0AB8
dd I0AB9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0B00
dd I0B08
dd I0B10
dd I0B18
dd I0B20
dd I0B28
dd I0B30
dw 0xc000 + 5
dd I0B38
dd I0B39
dd I0B3A
dd I0B3B
dd I0B3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0B40
dd I0B48
dd I0B50
dd I0B58
dd I0B60
dd I0B68
dd I0B70
dw 0xc000 + 2
dd I0B78
dd I0B79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0B80
dd I0B88
dd I0B90
dd I0B98
dd I0BA0
dd I0BA8
dd I0BB0
dw 0xc000 + 2
dd I0BB8
dd I0BB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0BC0
dd I0BC8
dd I0BD0
dd I0BD8
dd I0BE0
dd I0BE8
dd I0BF0
dw 0xc000 + 2
dd I0BF8
dd I0BF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0C00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0C10
dd I0C18
dd I0C20
dd I0C28
dd I0C30
dw 0xc000 + 2
dd I0C38
dd I0C39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0C40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0C50
dd I0C58
dd I0C60
dd I0C68
dd I0C70
dw 0xc000 + 2
dd I0C78
dd I0C79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I0C80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I0C90
dd I0C98
dd I0CA0
dd I0CA8
dd I0CB0
dw 0xc000 + 2
dd I0CB8
dd I0CB9
dw 0x0000 + 70
dd I_Invalid
dw 0x8000 + 7
dd I0D00
dd I0D08
dd I0D10
dd I0D18
dd I0D20
dd I0D28
dd I0D30
dw 0xc000 + 5
dd I0D38
dd I0D39
dd I0D3A
dd I0D3B
dd I0D3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0D40
dd I0D48
dd I0D50
dd I0D58
dd I0D60
dd I0D68
dd I0D70
dw 0xc000 + 2
dd I0D78
dd I0D79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0D80
dd I0D88
dd I0D90
dd I0D98
dd I0DA0
dd I0DA8
dd I0DB0
dw 0xc000 + 2
dd I0DB8
dd I0DB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0DC0
dd I0DC8
dd I0DD0
dd I0DD8
dd I0DE0
dd I0DE8
dd I0DF0
dw 0xc000 + 2
dd I0DF8
dd I0DF9
dw 0x0000 + 262
dd I_Invalid
dw 0x8000 + 7
dd I0F00
dd I0F08
dd I0F10
dd I0F18
dd I0F20
dd I0F28
dd I0F30
dw 0xc000 + 5
dd I0F38
dd I0F39
dd I0F3A
dd I0F3B
dd I0F3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I0F40
dd I0F48
dd I0F50
dd I0F58
dd I0F60
dd I0F68
dd I0F70
dw 0xc000 + 2
dd I0F78
dd I0F79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0F80
dd I0F88
dd I0F90
dd I0F98
dd I0FA0
dd I0FA8
dd I0FB0
dw 0xc000 + 2
dd I0FB8
dd I0FB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I0FC0
dd I0FC8
dd I0FD0
dd I0FD8
dd I0FE0
dd I0FE8
dd I0FF0
dw 0xc000 + 2
dd I0FF8
dd I0FF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I1000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1010
dd I1018
dd I1020
dd I1028
dd I1030
dw 0xc000 + 5
dd I1038
dd I1039
dd I103A
dd I103B
dd I103C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1080
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1090
dd I1098
dd I10A0
dd I10A8
dd I10B0
dw 0xc000 + 5
dd I10B8
dd I10B9
dd I10BA
dd I10BB
dd I10BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I10C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I10D0
dd I10D8
dd I10E0
dd I10E8
dd I10F0
dw 0xc000 + 5
dd I10F8
dd I10F9
dd I10FA
dd I10FB
dd I10FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1100
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1110
dd I1118
dd I1120
dd I1128
dd I1130
dw 0xc000 + 5
dd I1138
dd I1139
dd I113A
dd I113B
dd I113C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1140
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1150
dd I1158
dd I1160
dd I1168
dd I1170
dw 0xc000 + 5
dd I1178
dd I1179
dd I117A
dd I117B
dd I117C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1180
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1190
dd I1198
dd I11A0
dd I11A8
dd I11B0
dw 0xc000 + 5
dd I11B8
dd I11B9
dd I11BA
dd I11BB
dd I11BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I11C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I11D0
dd I11D8
dd I11E0
dd I11E8
dd I11F0
dw 0xc000 + 5
dd I11F8
dd I11F9
dd I11FA
dd I11FB
dd I11FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1210
dd I1218
dd I1220
dd I1228
dd I1230
dw 0xc000 + 5
dd I1238
dd I1239
dd I123A
dd I123B
dd I123C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1280
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1290
dd I1298
dd I12A0
dd I12A8
dd I12B0
dw 0xc000 + 5
dd I12B8
dd I12B9
dd I12BA
dd I12BB
dd I12BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I12C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I12D0
dd I12D8
dd I12E0
dd I12E8
dd I12F0
dw 0xc000 + 5
dd I12F8
dd I12F9
dd I12FA
dd I12FB
dd I12FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1300
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1310
dd I1318
dd I1320
dd I1328
dd I1330
dw 0xc000 + 5
dd I1338
dd I1339
dd I133A
dd I133B
dd I133C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1340
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1350
dd I1358
dd I1360
dd I1368
dd I1370
dw 0xc000 + 5
dd I1378
dd I1379
dd I137A
dd I137B
dd I137C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1380
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1390
dd I1398
dd I13A0
dd I13A8
dd I13B0
dw 0xc000 + 5
dd I13B8
dd I13B9
dd I13BA
dd I13BB
dd I13BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I13C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I13D0
dd I13D8
dd I13E0
dd I13E8
dd I13F0
dw 0xc000 + 5
dd I13F8
dd I13F9
dd I13FA
dd I13FB
dd I13FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1410
dd I1418
dd I1420
dd I1428
dd I1430
dw 0xc000 + 5
dd I1438
dd I1439
dd I143A
dd I143B
dd I143C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1480
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1490
dd I1498
dd I14A0
dd I14A8
dd I14B0
dw 0xc000 + 5
dd I14B8
dd I14B9
dd I14BA
dd I14BB
dd I14BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I14C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I14D0
dd I14D8
dd I14E0
dd I14E8
dd I14F0
dw 0xc000 + 5
dd I14F8
dd I14F9
dd I14FA
dd I14FB
dd I14FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1500
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1510
dd I1518
dd I1520
dd I1528
dd I1530
dw 0xc000 + 5
dd I1538
dd I1539
dd I153A
dd I153B
dd I153C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1540
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1550
dd I1558
dd I1560
dd I1568
dd I1570
dw 0xc000 + 5
dd I1578
dd I1579
dd I157A
dd I157B
dd I157C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1580
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1590
dd I1598
dd I15A0
dd I15A8
dd I15B0
dw 0xc000 + 5
dd I15B8
dd I15B9
dd I15BA
dd I15BB
dd I15BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1610
dd I1618
dd I1620
dd I1628
dd I1630
dw 0xc000 + 5
dd I1638
dd I1639
dd I163A
dd I163B
dd I163C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1680
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1690
dd I1698
dd I16A0
dd I16A8
dd I16B0
dw 0xc000 + 5
dd I16B8
dd I16B9
dd I16BA
dd I16BB
dd I16BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I16C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I16D0
dd I16D8
dd I16E0
dd I16E8
dd I16F0
dw 0xc000 + 5
dd I16F8
dd I16F9
dd I16FA
dd I16FB
dd I16FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1700
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1710
dd I1718
dd I1720
dd I1728
dd I1730
dw 0xc000 + 5
dd I1738
dd I1739
dd I173A
dd I173B
dd I173C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1740
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1750
dd I1758
dd I1760
dd I1768
dd I1770
dw 0xc000 + 5
dd I1778
dd I1779
dd I177A
dd I177B
dd I177C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1780
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1790
dd I1798
dd I17A0
dd I17A8
dd I17B0
dw 0xc000 + 5
dd I17B8
dd I17B9
dd I17BA
dd I17BB
dd I17BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1810
dd I1818
dd I1820
dd I1828
dd I1830
dw 0xc000 + 5
dd I1838
dd I1839
dd I183A
dd I183B
dd I183C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1880
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1890
dd I1898
dd I18A0
dd I18A8
dd I18B0
dw 0xc000 + 5
dd I18B8
dd I18B9
dd I18BA
dd I18BB
dd I18BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I18C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I18D0
dd I18D8
dd I18E0
dd I18E8
dd I18F0
dw 0xc000 + 5
dd I18F8
dd I18F9
dd I18FA
dd I18FB
dd I18FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1900
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1910
dd I1918
dd I1920
dd I1928
dd I1930
dw 0xc000 + 5
dd I1938
dd I1939
dd I193A
dd I193B
dd I193C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1940
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1950
dd I1958
dd I1960
dd I1968
dd I1970
dw 0xc000 + 5
dd I1978
dd I1979
dd I197A
dd I197B
dd I197C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1980
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1990
dd I1998
dd I19A0
dd I19A8
dd I19B0
dw 0xc000 + 5
dd I19B8
dd I19B9
dd I19BA
dd I19BB
dd I19BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1A10
dd I1A18
dd I1A20
dd I1A28
dd I1A30
dw 0xc000 + 5
dd I1A38
dd I1A39
dd I1A3A
dd I1A3B
dd I1A3C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1A80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1A90
dd I1A98
dd I1AA0
dd I1AA8
dd I1AB0
dw 0xc000 + 5
dd I1AB8
dd I1AB9
dd I1ABA
dd I1ABB
dd I1ABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1AC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1AD0
dd I1AD8
dd I1AE0
dd I1AE8
dd I1AF0
dw 0xc000 + 5
dd I1AF8
dd I1AF9
dd I1AFA
dd I1AFB
dd I1AFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1B00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1B10
dd I1B18
dd I1B20
dd I1B28
dd I1B30
dw 0xc000 + 5
dd I1B38
dd I1B39
dd I1B3A
dd I1B3B
dd I1B3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1B40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1B50
dd I1B58
dd I1B60
dd I1B68
dd I1B70
dw 0xc000 + 5
dd I1B78
dd I1B79
dd I1B7A
dd I1B7B
dd I1B7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1B80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1B90
dd I1B98
dd I1BA0
dd I1BA8
dd I1BB0
dw 0xc000 + 5
dd I1BB8
dd I1BB9
dd I1BBA
dd I1BBB
dd I1BBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1C00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1C10
dd I1C18
dd I1C20
dd I1C28
dd I1C30
dw 0xc000 + 5
dd I1C38
dd I1C39
dd I1C3A
dd I1C3B
dd I1C3C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1C80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1C90
dd I1C98
dd I1CA0
dd I1CA8
dd I1CB0
dw 0xc000 + 5
dd I1CB8
dd I1CB9
dd I1CBA
dd I1CBB
dd I1CBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1CC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1CD0
dd I1CD8
dd I1CE0
dd I1CE8
dd I1CF0
dw 0xc000 + 5
dd I1CF8
dd I1CF9
dd I1CFA
dd I1CFB
dd I1CFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1D00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1D10
dd I1D18
dd I1D20
dd I1D28
dd I1D30
dw 0xc000 + 5
dd I1D38
dd I1D39
dd I1D3A
dd I1D3B
dd I1D3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1D40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1D50
dd I1D58
dd I1D60
dd I1D68
dd I1D70
dw 0xc000 + 5
dd I1D78
dd I1D79
dd I1D7A
dd I1D7B
dd I1D7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1D80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1D90
dd I1D98
dd I1DA0
dd I1DA8
dd I1DB0
dw 0xc000 + 5
dd I1DB8
dd I1DB9
dd I1DBA
dd I1DBB
dd I1DBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1E00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1E10
dd I1E18
dd I1E20
dd I1E28
dd I1E30
dw 0xc000 + 5
dd I1E38
dd I1E39
dd I1E3A
dd I1E3B
dd I1E3C
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I1E80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1E90
dd I1E98
dd I1EA0
dd I1EA8
dd I1EB0
dw 0xc000 + 5
dd I1EB8
dd I1EB9
dd I1EBA
dd I1EBB
dd I1EBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1EC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1ED0
dd I1ED8
dd I1EE0
dd I1EE8
dd I1EF0
dw 0xc000 + 5
dd I1EF8
dd I1EF9
dd I1EFA
dd I1EFB
dd I1EFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1F00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1F10
dd I1F18
dd I1F20
dd I1F28
dd I1F30
dw 0xc000 + 5
dd I1F38
dd I1F39
dd I1F3A
dd I1F3B
dd I1F3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1F40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1F50
dd I1F58
dd I1F60
dd I1F68
dd I1F70
dw 0xc000 + 5
dd I1F78
dd I1F79
dd I1F7A
dd I1F7B
dd I1F7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I1F80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I1F90
dd I1F98
dd I1FA0
dd I1FA8
dd I1FB0
dw 0xc000 + 5
dd I1FB8
dd I1FB9
dd I1FBA
dd I1FBB
dd I1FBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2000
dd I2008
dd I2010
dd I2018
dd I2020
dd I2028
dd I2030
dw 0xc000 + 5
dd I2038
dd I2039
dd I203A
dd I203B
dd I203C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2040
dd I2048
dd I2050
dd I2058
dd I2060
dd I2068
dd I2070
dw 0xc000 + 5
dd I2078
dd I2079
dd I207A
dd I207B
dd I207C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2080
dd I2088
dd I2090
dd I2098
dd I20A0
dd I20A8
dd I20B0
dw 0xc000 + 5
dd I20B8
dd I20B9
dd I20BA
dd I20BB
dd I20BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I20C0
dd I20C8
dd I20D0
dd I20D8
dd I20E0
dd I20E8
dd I20F0
dw 0xc000 + 5
dd I20F8
dd I20F9
dd I20FA
dd I20FB
dd I20FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2100
dd I2108
dd I2110
dd I2118
dd I2120
dd I2128
dd I2130
dw 0xc000 + 5
dd I2138
dd I2139
dd I213A
dd I213B
dd I213C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2140
dd I2148
dd I2150
dd I2158
dd I2160
dd I2168
dd I2170
dw 0xc000 + 5
dd I2178
dd I2179
dd I217A
dd I217B
dd I217C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2180
dd I2188
dd I2190
dd I2198
dd I21A0
dd I21A8
dd I21B0
dw 0xc000 + 5
dd I21B8
dd I21B9
dd I21BA
dd I21BB
dd I21BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I21C0
dd I21C8
dd I21D0
dd I21D8
dd I21E0
dd I21E8
dd I21F0
dw 0xc000 + 5
dd I21F8
dd I21F9
dd I21FA
dd I21FB
dd I21FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2200
dd I2208
dd I2210
dd I2218
dd I2220
dd I2228
dd I2230
dw 0xc000 + 5
dd I2238
dd I2239
dd I223A
dd I223B
dd I223C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2240
dd I2248
dd I2250
dd I2258
dd I2260
dd I2268
dd I2270
dw 0xc000 + 5
dd I2278
dd I2279
dd I227A
dd I227B
dd I227C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2280
dd I2288
dd I2290
dd I2298
dd I22A0
dd I22A8
dd I22B0
dw 0xc000 + 5
dd I22B8
dd I22B9
dd I22BA
dd I22BB
dd I22BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I22C0
dd I22C8
dd I22D0
dd I22D8
dd I22E0
dd I22E8
dd I22F0
dw 0xc000 + 5
dd I22F8
dd I22F9
dd I22FA
dd I22FB
dd I22FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2300
dd I2308
dd I2310
dd I2318
dd I2320
dd I2328
dd I2330
dw 0xc000 + 5
dd I2338
dd I2339
dd I233A
dd I233B
dd I233C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2340
dd I2348
dd I2350
dd I2358
dd I2360
dd I2368
dd I2370
dw 0xc000 + 5
dd I2378
dd I2379
dd I237A
dd I237B
dd I237C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2380
dd I2388
dd I2390
dd I2398
dd I23A0
dd I23A8
dd I23B0
dw 0xc000 + 5
dd I23B8
dd I23B9
dd I23BA
dd I23BB
dd I23BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I23C0
dd I23C8
dd I23D0
dd I23D8
dd I23E0
dd I23E8
dd I23F0
dw 0xc000 + 5
dd I23F8
dd I23F9
dd I23FA
dd I23FB
dd I23FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2400
dd I2408
dd I2410
dd I2418
dd I2420
dd I2428
dd I2430
dw 0xc000 + 5
dd I2438
dd I2439
dd I243A
dd I243B
dd I243C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2440
dd I2448
dd I2450
dd I2458
dd I2460
dd I2468
dd I2470
dw 0xc000 + 5
dd I2478
dd I2479
dd I247A
dd I247B
dd I247C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2480
dd I2488
dd I2490
dd I2498
dd I24A0
dd I24A8
dd I24B0
dw 0xc000 + 5
dd I24B8
dd I24B9
dd I24BA
dd I24BB
dd I24BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I24C0
dd I24C8
dd I24D0
dd I24D8
dd I24E0
dd I24E8
dd I24F0
dw 0xc000 + 5
dd I24F8
dd I24F9
dd I24FA
dd I24FB
dd I24FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2500
dd I2508
dd I2510
dd I2518
dd I2520
dd I2528
dd I2530
dw 0xc000 + 5
dd I2538
dd I2539
dd I253A
dd I253B
dd I253C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2540
dd I2548
dd I2550
dd I2558
dd I2560
dd I2568
dd I2570
dw 0xc000 + 5
dd I2578
dd I2579
dd I257A
dd I257B
dd I257C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2580
dd I2588
dd I2590
dd I2598
dd I25A0
dd I25A8
dd I25B0
dw 0xc000 + 5
dd I25B8
dd I25B9
dd I25BA
dd I25BB
dd I25BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2600
dd I2608
dd I2610
dd I2618
dd I2620
dd I2628
dd I2630
dw 0xc000 + 5
dd I2638
dd I2639
dd I263A
dd I263B
dd I263C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2640
dd I2648
dd I2650
dd I2658
dd I2660
dd I2668
dd I2670
dw 0xc000 + 5
dd I2678
dd I2679
dd I267A
dd I267B
dd I267C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2680
dd I2688
dd I2690
dd I2698
dd I26A0
dd I26A8
dd I26B0
dw 0xc000 + 5
dd I26B8
dd I26B9
dd I26BA
dd I26BB
dd I26BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I26C0
dd I26C8
dd I26D0
dd I26D8
dd I26E0
dd I26E8
dd I26F0
dw 0xc000 + 5
dd I26F8
dd I26F9
dd I26FA
dd I26FB
dd I26FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2700
dd I2708
dd I2710
dd I2718
dd I2720
dd I2728
dd I2730
dw 0xc000 + 5
dd I2738
dd I2739
dd I273A
dd I273B
dd I273C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2740
dd I2748
dd I2750
dd I2758
dd I2760
dd I2768
dd I2770
dw 0xc000 + 5
dd I2778
dd I2779
dd I277A
dd I277B
dd I277C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2780
dd I2788
dd I2790
dd I2798
dd I27A0
dd I27A8
dd I27B0
dw 0xc000 + 5
dd I27B8
dd I27B9
dd I27BA
dd I27BB
dd I27BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2800
dd I2808
dd I2810
dd I2818
dd I2820
dd I2828
dd I2830
dw 0xc000 + 5
dd I2838
dd I2839
dd I283A
dd I283B
dd I283C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2840
dd I2848
dd I2850
dd I2858
dd I2860
dd I2868
dd I2870
dw 0xc000 + 5
dd I2878
dd I2879
dd I287A
dd I287B
dd I287C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2880
dd I2888
dd I2890
dd I2898
dd I28A0
dd I28A8
dd I28B0
dw 0xc000 + 5
dd I28B8
dd I28B9
dd I28BA
dd I28BB
dd I28BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I28C0
dd I28C8
dd I28D0
dd I28D8
dd I28E0
dd I28E8
dd I28F0
dw 0xc000 + 5
dd I28F8
dd I28F9
dd I28FA
dd I28FB
dd I28FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2900
dd I2908
dd I2910
dd I2918
dd I2920
dd I2928
dd I2930
dw 0xc000 + 5
dd I2938
dd I2939
dd I293A
dd I293B
dd I293C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2940
dd I2948
dd I2950
dd I2958
dd I2960
dd I2968
dd I2970
dw 0xc000 + 5
dd I2978
dd I2979
dd I297A
dd I297B
dd I297C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2980
dd I2988
dd I2990
dd I2998
dd I29A0
dd I29A8
dd I29B0
dw 0xc000 + 5
dd I29B8
dd I29B9
dd I29BA
dd I29BB
dd I29BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2A00
dd I2A08
dd I2A10
dd I2A18
dd I2A20
dd I2A28
dd I2A30
dw 0xc000 + 5
dd I2A38
dd I2A39
dd I2A3A
dd I2A3B
dd I2A3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2A40
dd I2A48
dd I2A50
dd I2A58
dd I2A60
dd I2A68
dd I2A70
dw 0xc000 + 5
dd I2A78
dd I2A79
dd I2A7A
dd I2A7B
dd I2A7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2A80
dd I2A88
dd I2A90
dd I2A98
dd I2AA0
dd I2AA8
dd I2AB0
dw 0xc000 + 5
dd I2AB8
dd I2AB9
dd I2ABA
dd I2ABB
dd I2ABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2AC0
dd I2AC8
dd I2AD0
dd I2AD8
dd I2AE0
dd I2AE8
dd I2AF0
dw 0xc000 + 5
dd I2AF8
dd I2AF9
dd I2AFA
dd I2AFB
dd I2AFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2B00
dd I2B08
dd I2B10
dd I2B18
dd I2B20
dd I2B28
dd I2B30
dw 0xc000 + 5
dd I2B38
dd I2B39
dd I2B3A
dd I2B3B
dd I2B3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2B40
dd I2B48
dd I2B50
dd I2B58
dd I2B60
dd I2B68
dd I2B70
dw 0xc000 + 5
dd I2B78
dd I2B79
dd I2B7A
dd I2B7B
dd I2B7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2B80
dd I2B88
dd I2B90
dd I2B98
dd I2BA0
dd I2BA8
dd I2BB0
dw 0xc000 + 5
dd I2BB8
dd I2BB9
dd I2BBA
dd I2BBB
dd I2BBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2C00
dd I2C08
dd I2C10
dd I2C18
dd I2C20
dd I2C28
dd I2C30
dw 0xc000 + 5
dd I2C38
dd I2C39
dd I2C3A
dd I2C3B
dd I2C3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2C40
dd I2C48
dd I2C50
dd I2C58
dd I2C60
dd I2C68
dd I2C70
dw 0xc000 + 5
dd I2C78
dd I2C79
dd I2C7A
dd I2C7B
dd I2C7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2C80
dd I2C88
dd I2C90
dd I2C98
dd I2CA0
dd I2CA8
dd I2CB0
dw 0xc000 + 5
dd I2CB8
dd I2CB9
dd I2CBA
dd I2CBB
dd I2CBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2CC0
dd I2CC8
dd I2CD0
dd I2CD8
dd I2CE0
dd I2CE8
dd I2CF0
dw 0xc000 + 5
dd I2CF8
dd I2CF9
dd I2CFA
dd I2CFB
dd I2CFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2D00
dd I2D08
dd I2D10
dd I2D18
dd I2D20
dd I2D28
dd I2D30
dw 0xc000 + 5
dd I2D38
dd I2D39
dd I2D3A
dd I2D3B
dd I2D3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2D40
dd I2D48
dd I2D50
dd I2D58
dd I2D60
dd I2D68
dd I2D70
dw 0xc000 + 5
dd I2D78
dd I2D79
dd I2D7A
dd I2D7B
dd I2D7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2D80
dd I2D88
dd I2D90
dd I2D98
dd I2DA0
dd I2DA8
dd I2DB0
dw 0xc000 + 5
dd I2DB8
dd I2DB9
dd I2DBA
dd I2DBB
dd I2DBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I2E00
dd I2E08
dd I2E10
dd I2E18
dd I2E20
dd I2E28
dd I2E30
dw 0xc000 + 5
dd I2E38
dd I2E39
dd I2E3A
dd I2E3B
dd I2E3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2E40
dd I2E48
dd I2E50
dd I2E58
dd I2E60
dd I2E68
dd I2E70
dw 0xc000 + 5
dd I2E78
dd I2E79
dd I2E7A
dd I2E7B
dd I2E7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2E80
dd I2E88
dd I2E90
dd I2E98
dd I2EA0
dd I2EA8
dd I2EB0
dw 0xc000 + 5
dd I2EB8
dd I2EB9
dd I2EBA
dd I2EBB
dd I2EBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2EC0
dd I2EC8
dd I2ED0
dd I2ED8
dd I2EE0
dd I2EE8
dd I2EF0
dw 0xc000 + 5
dd I2EF8
dd I2EF9
dd I2EFA
dd I2EFB
dd I2EFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2F00
dd I2F08
dd I2F10
dd I2F18
dd I2F20
dd I2F28
dd I2F30
dw 0xc000 + 5
dd I2F38
dd I2F39
dd I2F3A
dd I2F3B
dd I2F3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2F40
dd I2F48
dd I2F50
dd I2F58
dd I2F60
dd I2F68
dd I2F70
dw 0xc000 + 5
dd I2F78
dd I2F79
dd I2F7A
dd I2F7B
dd I2F7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I2F80
dd I2F88
dd I2F90
dd I2F98
dd I2FA0
dd I2FA8
dd I2FB0
dw 0xc000 + 5
dd I2FB8
dd I2FB9
dd I2FBA
dd I2FBB
dd I2FBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3000
dd I3008
dd I3010
dd I3018
dd I3020
dd I3028
dd I3030
dw 0xc000 + 5
dd I3038
dd I3039
dd I303A
dd I303B
dd I303C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3040
dd I3048
dd I3050
dd I3058
dd I3060
dd I3068
dd I3070
dw 0xc000 + 5
dd I3078
dd I3079
dd I307A
dd I307B
dd I307C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3080
dd I3088
dd I3090
dd I3098
dd I30A0
dd I30A8
dd I30B0
dw 0xc000 + 5
dd I30B8
dd I30B9
dd I30BA
dd I30BB
dd I30BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I30C0
dd I30C8
dd I30D0
dd I30D8
dd I30E0
dd I30E8
dd I30F0
dw 0xc000 + 5
dd I30F8
dd I30F9
dd I30FA
dd I30FB
dd I30FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3100
dd I3108
dd I3110
dd I3118
dd I3120
dd I3128
dd I3130
dw 0xc000 + 5
dd I3138
dd I3139
dd I313A
dd I313B
dd I313C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3140
dd I3148
dd I3150
dd I3158
dd I3160
dd I3168
dd I3170
dw 0xc000 + 5
dd I3178
dd I3179
dd I317A
dd I317B
dd I317C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3180
dd I3188
dd I3190
dd I3198
dd I31A0
dd I31A8
dd I31B0
dw 0xc000 + 5
dd I31B8
dd I31B9
dd I31BA
dd I31BB
dd I31BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I31C0
dd I31C8
dd I31D0
dd I31D8
dd I31E0
dd I31E8
dd I31F0
dw 0xc000 + 5
dd I31F8
dd I31F9
dd I31FA
dd I31FB
dd I31FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3200
dd I3208
dd I3210
dd I3218
dd I3220
dd I3228
dd I3230
dw 0xc000 + 5
dd I3238
dd I3239
dd I323A
dd I323B
dd I323C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3240
dd I3248
dd I3250
dd I3258
dd I3260
dd I3268
dd I3270
dw 0xc000 + 5
dd I3278
dd I3279
dd I327A
dd I327B
dd I327C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3280
dd I3288
dd I3290
dd I3298
dd I32A0
dd I32A8
dd I32B0
dw 0xc000 + 5
dd I32B8
dd I32B9
dd I32BA
dd I32BB
dd I32BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I32C0
dd I32C8
dd I32D0
dd I32D8
dd I32E0
dd I32E8
dd I32F0
dw 0xc000 + 5
dd I32F8
dd I32F9
dd I32FA
dd I32FB
dd I32FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3300
dd I3308
dd I3310
dd I3318
dd I3320
dd I3328
dd I3330
dw 0xc000 + 5
dd I3338
dd I3339
dd I333A
dd I333B
dd I333C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3340
dd I3348
dd I3350
dd I3358
dd I3360
dd I3368
dd I3370
dw 0xc000 + 5
dd I3378
dd I3379
dd I337A
dd I337B
dd I337C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3380
dd I3388
dd I3390
dd I3398
dd I33A0
dd I33A8
dd I33B0
dw 0xc000 + 5
dd I33B8
dd I33B9
dd I33BA
dd I33BB
dd I33BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I33C0
dd I33C8
dd I33D0
dd I33D8
dd I33E0
dd I33E8
dd I33F0
dw 0xc000 + 5
dd I33F8
dd I33F9
dd I33FA
dd I33FB
dd I33FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3400
dd I3408
dd I3410
dd I3418
dd I3420
dd I3428
dd I3430
dw 0xc000 + 5
dd I3438
dd I3439
dd I343A
dd I343B
dd I343C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3440
dd I3448
dd I3450
dd I3458
dd I3460
dd I3468
dd I3470
dw 0xc000 + 5
dd I3478
dd I3479
dd I347A
dd I347B
dd I347C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3480
dd I3488
dd I3490
dd I3498
dd I34A0
dd I34A8
dd I34B0
dw 0xc000 + 5
dd I34B8
dd I34B9
dd I34BA
dd I34BB
dd I34BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I34C0
dd I34C8
dd I34D0
dd I34D8
dd I34E0
dd I34E8
dd I34F0
dw 0xc000 + 5
dd I34F8
dd I34F9
dd I34FA
dd I34FB
dd I34FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3500
dd I3508
dd I3510
dd I3518
dd I3520
dd I3528
dd I3530
dw 0xc000 + 5
dd I3538
dd I3539
dd I353A
dd I353B
dd I353C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3540
dd I3548
dd I3550
dd I3558
dd I3560
dd I3568
dd I3570
dw 0xc000 + 5
dd I3578
dd I3579
dd I357A
dd I357B
dd I357C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3580
dd I3588
dd I3590
dd I3598
dd I35A0
dd I35A8
dd I35B0
dw 0xc000 + 5
dd I35B8
dd I35B9
dd I35BA
dd I35BB
dd I35BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3600
dd I3608
dd I3610
dd I3618
dd I3620
dd I3628
dd I3630
dw 0xc000 + 5
dd I3638
dd I3639
dd I363A
dd I363B
dd I363C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3640
dd I3648
dd I3650
dd I3658
dd I3660
dd I3668
dd I3670
dw 0xc000 + 5
dd I3678
dd I3679
dd I367A
dd I367B
dd I367C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3680
dd I3688
dd I3690
dd I3698
dd I36A0
dd I36A8
dd I36B0
dw 0xc000 + 5
dd I36B8
dd I36B9
dd I36BA
dd I36BB
dd I36BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I36C0
dd I36C8
dd I36D0
dd I36D8
dd I36E0
dd I36E8
dd I36F0
dw 0xc000 + 5
dd I36F8
dd I36F9
dd I36FA
dd I36FB
dd I36FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3700
dd I3708
dd I3710
dd I3718
dd I3720
dd I3728
dd I3730
dw 0xc000 + 5
dd I3738
dd I3739
dd I373A
dd I373B
dd I373C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3740
dd I3748
dd I3750
dd I3758
dd I3760
dd I3768
dd I3770
dw 0xc000 + 5
dd I3778
dd I3779
dd I377A
dd I377B
dd I377C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3780
dd I3788
dd I3790
dd I3798
dd I37A0
dd I37A8
dd I37B0
dw 0xc000 + 5
dd I37B8
dd I37B9
dd I37BA
dd I37BB
dd I37BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3800
dd I3808
dd I3810
dd I3818
dd I3820
dd I3828
dd I3830
dw 0xc000 + 5
dd I3838
dd I3839
dd I383A
dd I383B
dd I383C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3840
dd I3848
dd I3850
dd I3858
dd I3860
dd I3868
dd I3870
dw 0xc000 + 5
dd I3878
dd I3879
dd I387A
dd I387B
dd I387C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3880
dd I3888
dd I3890
dd I3898
dd I38A0
dd I38A8
dd I38B0
dw 0xc000 + 5
dd I38B8
dd I38B9
dd I38BA
dd I38BB
dd I38BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I38C0
dd I38C8
dd I38D0
dd I38D8
dd I38E0
dd I38E8
dd I38F0
dw 0xc000 + 5
dd I38F8
dd I38F9
dd I38FA
dd I38FB
dd I38FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3900
dd I3908
dd I3910
dd I3918
dd I3920
dd I3928
dd I3930
dw 0xc000 + 5
dd I3938
dd I3939
dd I393A
dd I393B
dd I393C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3940
dd I3948
dd I3950
dd I3958
dd I3960
dd I3968
dd I3970
dw 0xc000 + 5
dd I3978
dd I3979
dd I397A
dd I397B
dd I397C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3980
dd I3988
dd I3990
dd I3998
dd I39A0
dd I39A8
dd I39B0
dw 0xc000 + 5
dd I39B8
dd I39B9
dd I39BA
dd I39BB
dd I39BC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3A00
dd I3A08
dd I3A10
dd I3A18
dd I3A20
dd I3A28
dd I3A30
dw 0xc000 + 5
dd I3A38
dd I3A39
dd I3A3A
dd I3A3B
dd I3A3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3A40
dd I3A48
dd I3A50
dd I3A58
dd I3A60
dd I3A68
dd I3A70
dw 0xc000 + 5
dd I3A78
dd I3A79
dd I3A7A
dd I3A7B
dd I3A7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3A80
dd I3A88
dd I3A90
dd I3A98
dd I3AA0
dd I3AA8
dd I3AB0
dw 0xc000 + 5
dd I3AB8
dd I3AB9
dd I3ABA
dd I3ABB
dd I3ABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3AC0
dd I3AC8
dd I3AD0
dd I3AD8
dd I3AE0
dd I3AE8
dd I3AF0
dw 0xc000 + 5
dd I3AF8
dd I3AF9
dd I3AFA
dd I3AFB
dd I3AFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3B00
dd I3B08
dd I3B10
dd I3B18
dd I3B20
dd I3B28
dd I3B30
dw 0xc000 + 5
dd I3B38
dd I3B39
dd I3B3A
dd I3B3B
dd I3B3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3B40
dd I3B48
dd I3B50
dd I3B58
dd I3B60
dd I3B68
dd I3B70
dw 0xc000 + 5
dd I3B78
dd I3B79
dd I3B7A
dd I3B7B
dd I3B7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3B80
dd I3B88
dd I3B90
dd I3B98
dd I3BA0
dd I3BA8
dd I3BB0
dw 0xc000 + 5
dd I3BB8
dd I3BB9
dd I3BBA
dd I3BBB
dd I3BBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3C00
dd I3C08
dd I3C10
dd I3C18
dd I3C20
dd I3C28
dd I3C30
dw 0xc000 + 5
dd I3C38
dd I3C39
dd I3C3A
dd I3C3B
dd I3C3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3C40
dd I3C48
dd I3C50
dd I3C58
dd I3C60
dd I3C68
dd I3C70
dw 0xc000 + 5
dd I3C78
dd I3C79
dd I3C7A
dd I3C7B
dd I3C7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3C80
dd I3C88
dd I3C90
dd I3C98
dd I3CA0
dd I3CA8
dd I3CB0
dw 0xc000 + 5
dd I3CB8
dd I3CB9
dd I3CBA
dd I3CBB
dd I3CBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3CC0
dd I3CC8
dd I3CD0
dd I3CD8
dd I3CE0
dd I3CE8
dd I3CF0
dw 0xc000 + 5
dd I3CF8
dd I3CF9
dd I3CFA
dd I3CFB
dd I3CFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3D00
dd I3D08
dd I3D10
dd I3D18
dd I3D20
dd I3D28
dd I3D30
dw 0xc000 + 5
dd I3D38
dd I3D39
dd I3D3A
dd I3D3B
dd I3D3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3D40
dd I3D48
dd I3D50
dd I3D58
dd I3D60
dd I3D68
dd I3D70
dw 0xc000 + 5
dd I3D78
dd I3D79
dd I3D7A
dd I3D7B
dd I3D7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3D80
dd I3D88
dd I3D90
dd I3D98
dd I3DA0
dd I3DA8
dd I3DB0
dw 0xc000 + 5
dd I3DB8
dd I3DB9
dd I3DBA
dd I3DBB
dd I3DBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 7
dd I3E00
dd I3E08
dd I3E10
dd I3E18
dd I3E20
dd I3E28
dd I3E30
dw 0xc000 + 5
dd I3E38
dd I3E39
dd I3E3A
dd I3E3B
dd I3E3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3E40
dd I3E48
dd I3E50
dd I3E58
dd I3E60
dd I3E68
dd I3E70
dw 0xc000 + 5
dd I3E78
dd I3E79
dd I3E7A
dd I3E7B
dd I3E7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3E80
dd I3E88
dd I3E90
dd I3E98
dd I3EA0
dd I3EA8
dd I3EB0
dw 0xc000 + 5
dd I3EB8
dd I3EB9
dd I3EBA
dd I3EBB
dd I3EBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3EC0
dd I3EC8
dd I3ED0
dd I3ED8
dd I3EE0
dd I3EE8
dd I3EF0
dw 0xc000 + 5
dd I3EF8
dd I3EF9
dd I3EFA
dd I3EFB
dd I3EFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3F00
dd I3F08
dd I3F10
dd I3F18
dd I3F20
dd I3F28
dd I3F30
dw 0xc000 + 5
dd I3F38
dd I3F39
dd I3F3A
dd I3F3B
dd I3F3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3F40
dd I3F48
dd I3F50
dd I3F58
dd I3F60
dd I3F68
dd I3F70
dw 0xc000 + 5
dd I3F78
dd I3F79
dd I3F7A
dd I3F7B
dd I3F7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I3F80
dd I3F88
dd I3F90
dd I3F98
dd I3FA0
dd I3FA8
dd I3FB0
dw 0xc000 + 5
dd I3FB8
dd I3FB9
dd I3FBA
dd I3FBB
dd I3FBC
dw 0x0000 + 67
dd I_Invalid
dw 0x8000 + 1
dd I4000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4010
dd I4018
dd I4020
dd I4028
dd I4030
dw 0xc000 + 2
dd I4038
dd I4039
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4040
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4050
dd I4058
dd I4060
dd I4068
dd I4070
dw 0xc000 + 2
dd I4078
dd I4079
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4080
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4090
dd I4098
dd I40A0
dd I40A8
dd I40B0
dw 0xc000 + 2
dd I40B8
dd I40B9
dw 0x0000 + 6
dd I_Invalid
dw 0xc000 + 8
dd I40C0
dd I40C1
dd I40C2
dd I40C3
dd I40C4
dd I40C5
dd I40C6
dd I40C7
dw 0x0000 + 8
dd I_Invalid
dw 0xc000 + 42
dd I40D0
dd I40D1
dd I40D2
dd I40D3
dd I40D4
dd I40D5
dd I40D6
dd I40D7
dd I40D8
dd I40D9
dd I40DA
dd I40DB
dd I40DC
dd I40DD
dd I40DE
dd I40DF
dd I40E0
dd I40E1
dd I40E2
dd I40E3
dd I40E4
dd I40E5
dd I40E6
dd I40E7
dd I40E8
dd I40E9
dd I40EA
dd I40EB
dd I40EC
dd I40ED
dd I40EE
dd I40EF
dd I40F0
dd I40F1
dd I40F2
dd I40F3
dd I40F4
dd I40F5
dd I40F6
dd I40F7
dd I40F8
dd I40F9
dw 0x0000 + 134
dd I_Invalid
dw 0x8000 + 1
dd I4180
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4190
dd I4198
dd I41A0
dd I41A8
dd I41B0
dw 0xc000 + 5
dd I41B8
dd I41B9
dd I41BA
dd I41BB
dd I41BC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I41D0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I41E8
dd I41F0
dw 0xc000 + 4
dd I41F8
dd I41F9
dd I41FA
dd I41FB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4210
dd I4218
dd I4220
dd I4228
dd I4230
dw 0xc000 + 2
dd I4238
dd I4239
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4240
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4250
dd I4258
dd I4260
dd I4268
dd I4270
dw 0xc000 + 2
dd I4278
dd I4279
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4280
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4290
dd I4298
dd I42A0
dd I42A8
dd I42B0
dw 0xc000 + 2
dd I42B8
dd I42B9
dw 0x0000 + 198
dd I_Invalid
dw 0x8000 + 1
dd I4380
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4390
dd I4398
dd I43A0
dd I43A8
dd I43B0
dw 0xc000 + 5
dd I43B8
dd I43B9
dd I43BA
dd I43BB
dd I43BC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I43D0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I43E8
dd I43F0
dw 0xc000 + 4
dd I43F8
dd I43F9
dd I43FA
dd I43FB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4410
dd I4418
dd I4420
dd I4428
dd I4430
dw 0xc000 + 2
dd I4438
dd I4439
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4440
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4450
dd I4458
dd I4460
dd I4468
dd I4470
dw 0xc000 + 2
dd I4478
dd I4479
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4480
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4490
dd I4498
dd I44A0
dd I44A8
dd I44B0
dw 0xc000 + 2
dd I44B8
dd I44B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I44C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I44D0
dd I44D8
dd I44E0
dd I44E8
dd I44F0
dw 0xc000 + 5
dd I44F8
dd I44F9
dd I44FA
dd I44FB
dd I44FC
dw 0x0000 + 131
dd I_Invalid
dw 0x8000 + 1
dd I4580
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4590
dd I4598
dd I45A0
dd I45A8
dd I45B0
dw 0xc000 + 5
dd I45B8
dd I45B9
dd I45BA
dd I45BB
dd I45BC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I45D0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I45E8
dd I45F0
dw 0xc000 + 4
dd I45F8
dd I45F9
dd I45FA
dd I45FB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4610
dd I4618
dd I4620
dd I4628
dd I4630
dw 0xc000 + 2
dd I4638
dd I4639
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4640
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4650
dd I4658
dd I4660
dd I4668
dd I4670
dw 0xc000 + 2
dd I4678
dd I4679
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4680
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4690
dd I4698
dd I46A0
dd I46A8
dd I46B0
dw 0xc000 + 2
dd I46B8
dd I46B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I46C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I46D0
dd I46D8
dd I46E0
dd I46E8
dd I46F0
dw 0xc000 + 5
dd I46F8
dd I46F9
dd I46FA
dd I46FB
dd I46FC
dw 0x0000 + 131
dd I_Invalid
dw 0x8000 + 1
dd I4780
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4790
dd I4798
dd I47A0
dd I47A8
dd I47B0
dw 0xc000 + 5
dd I47B8
dd I47B9
dd I47BA
dd I47BB
dd I47BC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I47D0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I47E8
dd I47F0
dw 0xc000 + 4
dd I47F8
dd I47F9
dd I47FA
dd I47FB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4810
dd I4818
dd I4820
dd I4828
dd I4830
dw 0xc000 + 2
dd I4838
dd I4839
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4840
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 1
dd I4850
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4868
dd I4870
dw 0xc000 + 4
dd I4878
dd I4879
dd I487A
dd I487B
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4880
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 1
dd I4890
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 3
dd I48A0
dd I48A8
dd I48B0
dw 0xc000 + 2
dd I48B8
dd I48B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I48C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 1
dd I48D0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 3
dd I48E0
dd I48E8
dd I48F0
dw 0xc000 + 2
dd I48F8
dd I48F9
dw 0x0000 + 134
dd I_Invalid
dw 0x8000 + 1
dd I4980
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4990
dd I4998
dd I49A0
dd I49A8
dd I49B0
dw 0xc000 + 5
dd I49B8
dd I49B9
dd I49BA
dd I49BB
dd I49BC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I49D0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I49E8
dd I49F0
dw 0xc000 + 4
dd I49F8
dd I49F9
dd I49FA
dd I49FB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I4A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4A10
dd I4A18
dd I4A20
dd I4A28
dd I4A30
dw 0xc000 + 2
dd I4A38
dd I4A39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4A40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4A50
dd I4A58
dd I4A60
dd I4A68
dd I4A70
dw 0xc000 + 2
dd I4A78
dd I4A79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4A80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4A90
dd I4A98
dd I4AA0
dd I4AA8
dd I4AB0
dw 0xc000 + 2
dd I4AB8
dd I4AB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I4AC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4AD0
dd I4AD8
dd I4AE0
dd I4AE8
dd I4AF0
dw 0xc000 + 2
dd I4AF8
dd I4AF9
dw 0x0000 + 2
dd I_Invalid
dw 0xc000 + 1
dd I4AFC
dw 0x0000 + 131
dd I_Invalid
dw 0x8000 + 1
dd I4B80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4B90
dd I4B98
dd I4BA0
dd I4BA8
dd I4BB0
dw 0xc000 + 5
dd I4BB8
dd I4BB9
dd I4BBA
dd I4BBB
dd I4BBC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I4BD0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4BE8
dd I4BF0
dw 0xc000 + 4
dd I4BF8
dd I4BF9
dd I4BFA
dd I4BFB
dw 0x0000 + 148
dd I_Invalid
dw 0x8000 + 2
dd I4C90
dd I4C98
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 2
dd I4CA8
dd I4CB0
dw 0xc000 + 4
dd I4CB8
dd I4CB9
dd I4CBA
dd I4CBB
dw 0x0000 + 20
dd I_Invalid
dw 0x8000 + 2
dd I4CD0
dd I4CD8
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 2
dd I4CE8
dd I4CF0
dw 0xc000 + 4
dd I4CF8
dd I4CF9
dd I4CFA
dd I4CFB
dw 0x0000 + 132
dd I_Invalid
dw 0x8000 + 1
dd I4D80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4D90
dd I4D98
dd I4DA0
dd I4DA8
dd I4DB0
dw 0xc000 + 5
dd I4DB8
dd I4DB9
dd I4DBA
dd I4DBB
dd I4DBC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I4DD0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4DE8
dd I4DF0
dw 0xc000 + 4
dd I4DF8
dd I4DF9
dd I4DFA
dd I4DFB
dw 0x0000 + 68
dd I_Invalid
dw 16
dd I4E40
dw 0x8000 + 4
dd I4E50
dd I4E58
dd I4E60
dd I4E68
dw 0xc000 + 4
dd I4E70
dd I4E71
dd I4E72
dd I4E73
dw 0x0000 + 1
dd I_Invalid
dw 0xc000 + 3
dd I4E75
dd I4E76
dd I4E77
dw 0x0000 + 24
dd I_Invalid
dw 0x8000 + 1
dd I4E90
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4EA8
dd I4EB0
dw 0xc000 + 4
dd I4EB8
dd I4EB9
dd I4EBA
dd I4EBB
dw 0x0000 + 20
dd I_Invalid
dw 0x8000 + 1
dd I4ED0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4EE8
dd I4EF0
dw 0xc000 + 4
dd I4EF8
dd I4EF9
dd I4EFA
dd I4EFB
dw 0x0000 + 132
dd I_Invalid
dw 0x8000 + 1
dd I4F80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I4F90
dd I4F98
dd I4FA0
dd I4FA8
dd I4FB0
dw 0xc000 + 5
dd I4FB8
dd I4FB9
dd I4FBA
dd I4FBB
dd I4FBC
dw 0x0000 + 19
dd I_Invalid
dw 0x8000 + 1
dd I4FD0
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 2
dd I4FE8
dd I4FF0
dw 0xc000 + 4
dd I4FF8
dd I4FF9
dd I4FFA
dd I4FFB
dw 0x0000 + 4
dd I_Invalid
dw 0x8000 + 1
dd I5000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5010
dd I5018
dd I5020
dd I5028
dd I5030
dw 0xc000 + 2
dd I5038
dd I5039
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5040
dd I5048
dd I5050
dd I5058
dd I5060
dd I5068
dd I5070
dw 0xc000 + 2
dd I5078
dd I5079
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5080
dd I5088
dd I5090
dd I5098
dd I50A0
dd I50A8
dd I50B0
dw 0xc000 + 2
dd I50B8
dd I50B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I50C0
dd I50C8
dd I50D0
dd I50D8
dd I50E0
dd I50E8
dd I50F0
dw 0xc000 + 2
dd I50F8
dd I50F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5100
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5110
dd I5118
dd I5120
dd I5128
dd I5130
dw 0xc000 + 2
dd I5138
dd I5139
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5140
dd I5148
dd I5150
dd I5158
dd I5160
dd I5168
dd I5170
dw 0xc000 + 2
dd I5178
dd I5179
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5180
dd I5188
dd I5190
dd I5198
dd I51A0
dd I51A8
dd I51B0
dw 0xc000 + 2
dd I51B8
dd I51B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I51C0
dd I51C8
dd I51D0
dd I51D8
dd I51E0
dd I51E8
dd I51F0
dw 0xc000 + 2
dd I51F8
dd I51F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5210
dd I5218
dd I5220
dd I5228
dd I5230
dw 0xc000 + 2
dd I5238
dd I5239
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5240
dd I5248
dd I5250
dd I5258
dd I5260
dd I5268
dd I5270
dw 0xc000 + 2
dd I5278
dd I5279
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5280
dd I5288
dd I5290
dd I5298
dd I52A0
dd I52A8
dd I52B0
dw 0xc000 + 2
dd I52B8
dd I52B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I52C0
dd I52C8
dd I52D0
dd I52D8
dd I52E0
dd I52E8
dd I52F0
dw 0xc000 + 2
dd I52F8
dd I52F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5300
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5310
dd I5318
dd I5320
dd I5328
dd I5330
dw 0xc000 + 2
dd I5338
dd I5339
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5340
dd I5348
dd I5350
dd I5358
dd I5360
dd I5368
dd I5370
dw 0xc000 + 2
dd I5378
dd I5379
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5380
dd I5388
dd I5390
dd I5398
dd I53A0
dd I53A8
dd I53B0
dw 0xc000 + 2
dd I53B8
dd I53B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I53C0
dd I53C8
dd I53D0
dd I53D8
dd I53E0
dd I53E8
dd I53F0
dw 0xc000 + 2
dd I53F8
dd I53F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5410
dd I5418
dd I5420
dd I5428
dd I5430
dw 0xc000 + 2
dd I5438
dd I5439
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5440
dd I5448
dd I5450
dd I5458
dd I5460
dd I5468
dd I5470
dw 0xc000 + 2
dd I5478
dd I5479
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5480
dd I5488
dd I5490
dd I5498
dd I54A0
dd I54A8
dd I54B0
dw 0xc000 + 2
dd I54B8
dd I54B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I54C0
dd I54C8
dd I54D0
dd I54D8
dd I54E0
dd I54E8
dd I54F0
dw 0xc000 + 2
dd I54F8
dd I54F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5500
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5510
dd I5518
dd I5520
dd I5528
dd I5530
dw 0xc000 + 2
dd I5538
dd I5539
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5540
dd I5548
dd I5550
dd I5558
dd I5560
dd I5568
dd I5570
dw 0xc000 + 2
dd I5578
dd I5579
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5580
dd I5588
dd I5590
dd I5598
dd I55A0
dd I55A8
dd I55B0
dw 0xc000 + 2
dd I55B8
dd I55B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I55C0
dd I55C8
dd I55D0
dd I55D8
dd I55E0
dd I55E8
dd I55F0
dw 0xc000 + 2
dd I55F8
dd I55F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5610
dd I5618
dd I5620
dd I5628
dd I5630
dw 0xc000 + 2
dd I5638
dd I5639
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5640
dd I5648
dd I5650
dd I5658
dd I5660
dd I5668
dd I5670
dw 0xc000 + 2
dd I5678
dd I5679
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5680
dd I5688
dd I5690
dd I5698
dd I56A0
dd I56A8
dd I56B0
dw 0xc000 + 2
dd I56B8
dd I56B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I56C0
dd I56C8
dd I56D0
dd I56D8
dd I56E0
dd I56E8
dd I56F0
dw 0xc000 + 2
dd I56F8
dd I56F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5700
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5710
dd I5718
dd I5720
dd I5728
dd I5730
dw 0xc000 + 2
dd I5738
dd I5739
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5740
dd I5748
dd I5750
dd I5758
dd I5760
dd I5768
dd I5770
dw 0xc000 + 2
dd I5778
dd I5779
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5780
dd I5788
dd I5790
dd I5798
dd I57A0
dd I57A8
dd I57B0
dw 0xc000 + 2
dd I57B8
dd I57B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I57C0
dd I57C8
dd I57D0
dd I57D8
dd I57E0
dd I57E8
dd I57F0
dw 0xc000 + 2
dd I57F8
dd I57F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5810
dd I5818
dd I5820
dd I5828
dd I5830
dw 0xc000 + 2
dd I5838
dd I5839
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5840
dd I5848
dd I5850
dd I5858
dd I5860
dd I5868
dd I5870
dw 0xc000 + 2
dd I5878
dd I5879
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5880
dd I5888
dd I5890
dd I5898
dd I58A0
dd I58A8
dd I58B0
dw 0xc000 + 2
dd I58B8
dd I58B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I58C0
dd I58C8
dd I58D0
dd I58D8
dd I58E0
dd I58E8
dd I58F0
dw 0xc000 + 2
dd I58F8
dd I58F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5900
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5910
dd I5918
dd I5920
dd I5928
dd I5930
dw 0xc000 + 2
dd I5938
dd I5939
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5940
dd I5948
dd I5950
dd I5958
dd I5960
dd I5968
dd I5970
dw 0xc000 + 2
dd I5978
dd I5979
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5980
dd I5988
dd I5990
dd I5998
dd I59A0
dd I59A8
dd I59B0
dw 0xc000 + 2
dd I59B8
dd I59B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I59C0
dd I59C8
dd I59D0
dd I59D8
dd I59E0
dd I59E8
dd I59F0
dw 0xc000 + 2
dd I59F8
dd I59F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5A10
dd I5A18
dd I5A20
dd I5A28
dd I5A30
dw 0xc000 + 2
dd I5A38
dd I5A39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5A40
dd I5A48
dd I5A50
dd I5A58
dd I5A60
dd I5A68
dd I5A70
dw 0xc000 + 2
dd I5A78
dd I5A79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5A80
dd I5A88
dd I5A90
dd I5A98
dd I5AA0
dd I5AA8
dd I5AB0
dw 0xc000 + 2
dd I5AB8
dd I5AB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5AC0
dd I5AC8
dd I5AD0
dd I5AD8
dd I5AE0
dd I5AE8
dd I5AF0
dw 0xc000 + 2
dd I5AF8
dd I5AF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5B00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5B10
dd I5B18
dd I5B20
dd I5B28
dd I5B30
dw 0xc000 + 2
dd I5B38
dd I5B39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5B40
dd I5B48
dd I5B50
dd I5B58
dd I5B60
dd I5B68
dd I5B70
dw 0xc000 + 2
dd I5B78
dd I5B79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5B80
dd I5B88
dd I5B90
dd I5B98
dd I5BA0
dd I5BA8
dd I5BB0
dw 0xc000 + 2
dd I5BB8
dd I5BB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5BC0
dd I5BC8
dd I5BD0
dd I5BD8
dd I5BE0
dd I5BE8
dd I5BF0
dw 0xc000 + 2
dd I5BF8
dd I5BF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5C00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5C10
dd I5C18
dd I5C20
dd I5C28
dd I5C30
dw 0xc000 + 2
dd I5C38
dd I5C39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5C40
dd I5C48
dd I5C50
dd I5C58
dd I5C60
dd I5C68
dd I5C70
dw 0xc000 + 2
dd I5C78
dd I5C79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5C80
dd I5C88
dd I5C90
dd I5C98
dd I5CA0
dd I5CA8
dd I5CB0
dw 0xc000 + 2
dd I5CB8
dd I5CB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5CC0
dd I5CC8
dd I5CD0
dd I5CD8
dd I5CE0
dd I5CE8
dd I5CF0
dw 0xc000 + 2
dd I5CF8
dd I5CF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5D00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5D10
dd I5D18
dd I5D20
dd I5D28
dd I5D30
dw 0xc000 + 2
dd I5D38
dd I5D39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5D40
dd I5D48
dd I5D50
dd I5D58
dd I5D60
dd I5D68
dd I5D70
dw 0xc000 + 2
dd I5D78
dd I5D79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5D80
dd I5D88
dd I5D90
dd I5D98
dd I5DA0
dd I5DA8
dd I5DB0
dw 0xc000 + 2
dd I5DB8
dd I5DB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5DC0
dd I5DC8
dd I5DD0
dd I5DD8
dd I5DE0
dd I5DE8
dd I5DF0
dw 0xc000 + 2
dd I5DF8
dd I5DF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5E00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5E10
dd I5E18
dd I5E20
dd I5E28
dd I5E30
dw 0xc000 + 2
dd I5E38
dd I5E39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5E40
dd I5E48
dd I5E50
dd I5E58
dd I5E60
dd I5E68
dd I5E70
dw 0xc000 + 2
dd I5E78
dd I5E79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5E80
dd I5E88
dd I5E90
dd I5E98
dd I5EA0
dd I5EA8
dd I5EB0
dw 0xc000 + 2
dd I5EB8
dd I5EB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5EC0
dd I5EC8
dd I5ED0
dd I5ED8
dd I5EE0
dd I5EE8
dd I5EF0
dw 0xc000 + 2
dd I5EF8
dd I5EF9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I5F00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I5F10
dd I5F18
dd I5F20
dd I5F28
dd I5F30
dw 0xc000 + 2
dd I5F38
dd I5F39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5F40
dd I5F48
dd I5F50
dd I5F58
dd I5F60
dd I5F68
dd I5F70
dw 0xc000 + 2
dd I5F78
dd I5F79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5F80
dd I5F88
dd I5F90
dd I5F98
dd I5FA0
dd I5FA8
dd I5FB0
dw 0xc000 + 2
dd I5FB8
dd I5FB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I5FC0
dd I5FC8
dd I5FD0
dd I5FD8
dd I5FE0
dd I5FE8
dd I5FF0
dw 0xc000 + 2
dd I5FF8
dd I5FF9
dw 0x0000 + 6
dd I_Invalid
dw 0xc000 + 1
dd I6000
dw 255
dd I6001
dw 0xc000 + 1
dd I6100
dw 255
dd I6101
dw 0xc000 + 1
dd I6200
dw 255
dd I6201
dw 0xc000 + 1
dd I6300
dw 255
dd I6301
dw 0xc000 + 1
dd I6400
dw 255
dd I6401
dw 0xc000 + 1
dd I6500
dw 255
dd I6501
dw 0xc000 + 1
dd I6600
dw 255
dd I6601
dw 0xc000 + 1
dd I6700
dw 255
dd I6701
dw 0xc000 + 1
dd I6800
dw 255
dd I6801
dw 0xc000 + 1
dd I6900
dw 255
dd I6901
dw 0xc000 + 1
dd I6A00
dw 255
dd I6A01
dw 0xc000 + 1
dd I6B00
dw 255
dd I6B01
dw 0xc000 + 1
dd I6C00
dw 255
dd I6C01
dw 0xc000 + 1
dd I6D00
dw 255
dd I6D01
dw 0xc000 + 1
dd I6E00
dw 255
dd I6E01
dw 0xc000 + 1
dd I6F00
dw 255
dd I6F01
dw 256
dd I7000
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7200
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7400
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7600
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7800
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7A00
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7C00
dw 0x0000 + 256
dd I_Invalid
dw 256
dd I7E00
dw 0x0000 + 256
dd I_Invalid
dw 0x8000 + 1
dd I8000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8010
dd I8018
dd I8020
dd I8028
dd I8030
dw 0xc000 + 5
dd I8038
dd I8039
dd I803A
dd I803B
dd I803C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8040
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8050
dd I8058
dd I8060
dd I8068
dd I8070
dw 0xc000 + 5
dd I8078
dd I8079
dd I807A
dd I807B
dd I807C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8080
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8090
dd I8098
dd I80A0
dd I80A8
dd I80B0
dw 0xc000 + 5
dd I80B8
dd I80B9
dd I80BA
dd I80BB
dd I80BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I80C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I80D0
dd I80D8
dd I80E0
dd I80E8
dd I80F0
dw 0xc000 + 5
dd I80F8
dd I80F9
dd I80FA
dd I80FB
dd I80FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8100
dd I8108
dd I8110
dd I8118
dd I8120
dd I8128
dd I8130
dw 0xc000 + 2
dd I8138
dd I8139
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8150
dd I8158
dd I8160
dd I8168
dd I8170
dw 0xc000 + 2
dd I8178
dd I8179
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8190
dd I8198
dd I81A0
dd I81A8
dd I81B0
dw 0xc000 + 2
dd I81B8
dd I81B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I81C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I81D0
dd I81D8
dd I81E0
dd I81E8
dd I81F0
dw 0xc000 + 5
dd I81F8
dd I81F9
dd I81FA
dd I81FB
dd I81FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8210
dd I8218
dd I8220
dd I8228
dd I8230
dw 0xc000 + 5
dd I8238
dd I8239
dd I823A
dd I823B
dd I823C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8240
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8250
dd I8258
dd I8260
dd I8268
dd I8270
dw 0xc000 + 5
dd I8278
dd I8279
dd I827A
dd I827B
dd I827C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8280
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8290
dd I8298
dd I82A0
dd I82A8
dd I82B0
dw 0xc000 + 5
dd I82B8
dd I82B9
dd I82BA
dd I82BB
dd I82BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I82C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I82D0
dd I82D8
dd I82E0
dd I82E8
dd I82F0
dw 0xc000 + 5
dd I82F8
dd I82F9
dd I82FA
dd I82FB
dd I82FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8300
dd I8308
dd I8310
dd I8318
dd I8320
dd I8328
dd I8330
dw 0xc000 + 2
dd I8338
dd I8339
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8350
dd I8358
dd I8360
dd I8368
dd I8370
dw 0xc000 + 2
dd I8378
dd I8379
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8390
dd I8398
dd I83A0
dd I83A8
dd I83B0
dw 0xc000 + 2
dd I83B8
dd I83B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I83C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I83D0
dd I83D8
dd I83E0
dd I83E8
dd I83F0
dw 0xc000 + 5
dd I83F8
dd I83F9
dd I83FA
dd I83FB
dd I83FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8410
dd I8418
dd I8420
dd I8428
dd I8430
dw 0xc000 + 5
dd I8438
dd I8439
dd I843A
dd I843B
dd I843C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8440
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8450
dd I8458
dd I8460
dd I8468
dd I8470
dw 0xc000 + 5
dd I8478
dd I8479
dd I847A
dd I847B
dd I847C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8480
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8490
dd I8498
dd I84A0
dd I84A8
dd I84B0
dw 0xc000 + 5
dd I84B8
dd I84B9
dd I84BA
dd I84BB
dd I84BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I84C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I84D0
dd I84D8
dd I84E0
dd I84E8
dd I84F0
dw 0xc000 + 5
dd I84F8
dd I84F9
dd I84FA
dd I84FB
dd I84FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8500
dd I8508
dd I8510
dd I8518
dd I8520
dd I8528
dd I8530
dw 0xc000 + 2
dd I8538
dd I8539
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8550
dd I8558
dd I8560
dd I8568
dd I8570
dw 0xc000 + 2
dd I8578
dd I8579
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8590
dd I8598
dd I85A0
dd I85A8
dd I85B0
dw 0xc000 + 2
dd I85B8
dd I85B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I85C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I85D0
dd I85D8
dd I85E0
dd I85E8
dd I85F0
dw 0xc000 + 5
dd I85F8
dd I85F9
dd I85FA
dd I85FB
dd I85FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8610
dd I8618
dd I8620
dd I8628
dd I8630
dw 0xc000 + 5
dd I8638
dd I8639
dd I863A
dd I863B
dd I863C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8640
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8650
dd I8658
dd I8660
dd I8668
dd I8670
dw 0xc000 + 5
dd I8678
dd I8679
dd I867A
dd I867B
dd I867C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8680
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8690
dd I8698
dd I86A0
dd I86A8
dd I86B0
dw 0xc000 + 5
dd I86B8
dd I86B9
dd I86BA
dd I86BB
dd I86BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I86C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I86D0
dd I86D8
dd I86E0
dd I86E8
dd I86F0
dw 0xc000 + 5
dd I86F8
dd I86F9
dd I86FA
dd I86FB
dd I86FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8700
dd I8708
dd I8710
dd I8718
dd I8720
dd I8728
dd I8730
dw 0xc000 + 2
dd I8738
dd I8739
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8750
dd I8758
dd I8760
dd I8768
dd I8770
dw 0xc000 + 2
dd I8778
dd I8779
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8790
dd I8798
dd I87A0
dd I87A8
dd I87B0
dw 0xc000 + 2
dd I87B8
dd I87B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I87C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I87D0
dd I87D8
dd I87E0
dd I87E8
dd I87F0
dw 0xc000 + 5
dd I87F8
dd I87F9
dd I87FA
dd I87FB
dd I87FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8810
dd I8818
dd I8820
dd I8828
dd I8830
dw 0xc000 + 5
dd I8838
dd I8839
dd I883A
dd I883B
dd I883C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8840
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8850
dd I8858
dd I8860
dd I8868
dd I8870
dw 0xc000 + 5
dd I8878
dd I8879
dd I887A
dd I887B
dd I887C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8880
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8890
dd I8898
dd I88A0
dd I88A8
dd I88B0
dw 0xc000 + 5
dd I88B8
dd I88B9
dd I88BA
dd I88BB
dd I88BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I88C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I88D0
dd I88D8
dd I88E0
dd I88E8
dd I88F0
dw 0xc000 + 5
dd I88F8
dd I88F9
dd I88FA
dd I88FB
dd I88FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8900
dd I8908
dd I8910
dd I8918
dd I8920
dd I8928
dd I8930
dw 0xc000 + 2
dd I8938
dd I8939
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8950
dd I8958
dd I8960
dd I8968
dd I8970
dw 0xc000 + 2
dd I8978
dd I8979
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8990
dd I8998
dd I89A0
dd I89A8
dd I89B0
dw 0xc000 + 2
dd I89B8
dd I89B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I89C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I89D0
dd I89D8
dd I89E0
dd I89E8
dd I89F0
dw 0xc000 + 5
dd I89F8
dd I89F9
dd I89FA
dd I89FB
dd I89FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8A10
dd I8A18
dd I8A20
dd I8A28
dd I8A30
dw 0xc000 + 5
dd I8A38
dd I8A39
dd I8A3A
dd I8A3B
dd I8A3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8A40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8A50
dd I8A58
dd I8A60
dd I8A68
dd I8A70
dw 0xc000 + 5
dd I8A78
dd I8A79
dd I8A7A
dd I8A7B
dd I8A7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8A80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8A90
dd I8A98
dd I8AA0
dd I8AA8
dd I8AB0
dw 0xc000 + 5
dd I8AB8
dd I8AB9
dd I8ABA
dd I8ABB
dd I8ABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8AC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8AD0
dd I8AD8
dd I8AE0
dd I8AE8
dd I8AF0
dw 0xc000 + 5
dd I8AF8
dd I8AF9
dd I8AFA
dd I8AFB
dd I8AFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8B00
dd I8B08
dd I8B10
dd I8B18
dd I8B20
dd I8B28
dd I8B30
dw 0xc000 + 2
dd I8B38
dd I8B39
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8B50
dd I8B58
dd I8B60
dd I8B68
dd I8B70
dw 0xc000 + 2
dd I8B78
dd I8B79
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8B90
dd I8B98
dd I8BA0
dd I8BA8
dd I8BB0
dw 0xc000 + 2
dd I8BB8
dd I8BB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I8BC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8BD0
dd I8BD8
dd I8BE0
dd I8BE8
dd I8BF0
dw 0xc000 + 5
dd I8BF8
dd I8BF9
dd I8BFA
dd I8BFB
dd I8BFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8C00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8C10
dd I8C18
dd I8C20
dd I8C28
dd I8C30
dw 0xc000 + 5
dd I8C38
dd I8C39
dd I8C3A
dd I8C3B
dd I8C3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8C40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8C50
dd I8C58
dd I8C60
dd I8C68
dd I8C70
dw 0xc000 + 5
dd I8C78
dd I8C79
dd I8C7A
dd I8C7B
dd I8C7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8C80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8C90
dd I8C98
dd I8CA0
dd I8CA8
dd I8CB0
dw 0xc000 + 5
dd I8CB8
dd I8CB9
dd I8CBA
dd I8CBB
dd I8CBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8CC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8CD0
dd I8CD8
dd I8CE0
dd I8CE8
dd I8CF0
dw 0xc000 + 5
dd I8CF8
dd I8CF9
dd I8CFA
dd I8CFB
dd I8CFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8D00
dd I8D08
dd I8D10
dd I8D18
dd I8D20
dd I8D28
dd I8D30
dw 0xc000 + 2
dd I8D38
dd I8D39
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8D50
dd I8D58
dd I8D60
dd I8D68
dd I8D70
dw 0xc000 + 2
dd I8D78
dd I8D79
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8D90
dd I8D98
dd I8DA0
dd I8DA8
dd I8DB0
dw 0xc000 + 2
dd I8DB8
dd I8DB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I8DC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8DD0
dd I8DD8
dd I8DE0
dd I8DE8
dd I8DF0
dw 0xc000 + 5
dd I8DF8
dd I8DF9
dd I8DFA
dd I8DFB
dd I8DFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8E00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8E10
dd I8E18
dd I8E20
dd I8E28
dd I8E30
dw 0xc000 + 5
dd I8E38
dd I8E39
dd I8E3A
dd I8E3B
dd I8E3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8E40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8E50
dd I8E58
dd I8E60
dd I8E68
dd I8E70
dw 0xc000 + 5
dd I8E78
dd I8E79
dd I8E7A
dd I8E7B
dd I8E7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8E80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8E90
dd I8E98
dd I8EA0
dd I8EA8
dd I8EB0
dw 0xc000 + 5
dd I8EB8
dd I8EB9
dd I8EBA
dd I8EBB
dd I8EBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I8EC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8ED0
dd I8ED8
dd I8EE0
dd I8EE8
dd I8EF0
dw 0xc000 + 5
dd I8EF8
dd I8EF9
dd I8EFA
dd I8EFB
dd I8EFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I8F00
dd I8F08
dd I8F10
dd I8F18
dd I8F20
dd I8F28
dd I8F30
dw 0xc000 + 2
dd I8F38
dd I8F39
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8F50
dd I8F58
dd I8F60
dd I8F68
dd I8F70
dw 0xc000 + 2
dd I8F78
dd I8F79
dw 0x0000 + 22
dd I_Invalid
dw 0x8000 + 5
dd I8F90
dd I8F98
dd I8FA0
dd I8FA8
dd I8FB0
dw 0xc000 + 2
dd I8FB8
dd I8FB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd I8FC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I8FD0
dd I8FD8
dd I8FE0
dd I8FE8
dd I8FF0
dw 0xc000 + 5
dd I8FF8
dd I8FF9
dd I8FFA
dd I8FFB
dd I8FFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9010
dd I9018
dd I9020
dd I9028
dd I9030
dw 0xc000 + 5
dd I9038
dd I9039
dd I903A
dd I903B
dd I903C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9040
dd I9048
dd I9050
dd I9058
dd I9060
dd I9068
dd I9070
dw 0xc000 + 5
dd I9078
dd I9079
dd I907A
dd I907B
dd I907C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9080
dd I9088
dd I9090
dd I9098
dd I90A0
dd I90A8
dd I90B0
dw 0xc000 + 5
dd I90B8
dd I90B9
dd I90BA
dd I90BB
dd I90BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I90C0
dd I90C8
dd I90D0
dd I90D8
dd I90E0
dd I90E8
dd I90F0
dw 0xc000 + 5
dd I90F8
dd I90F9
dd I90FA
dd I90FB
dd I90FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9100
dd I9108
dd I9110
dd I9118
dd I9120
dd I9128
dd I9130
dw 0xc000 + 2
dd I9138
dd I9139
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9140
dd I9148
dd I9150
dd I9158
dd I9160
dd I9168
dd I9170
dw 0xc000 + 2
dd I9178
dd I9179
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9180
dd I9188
dd I9190
dd I9198
dd I91A0
dd I91A8
dd I91B0
dw 0xc000 + 2
dd I91B8
dd I91B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I91C0
dd I91C8
dd I91D0
dd I91D8
dd I91E0
dd I91E8
dd I91F0
dw 0xc000 + 5
dd I91F8
dd I91F9
dd I91FA
dd I91FB
dd I91FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9210
dd I9218
dd I9220
dd I9228
dd I9230
dw 0xc000 + 5
dd I9238
dd I9239
dd I923A
dd I923B
dd I923C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9240
dd I9248
dd I9250
dd I9258
dd I9260
dd I9268
dd I9270
dw 0xc000 + 5
dd I9278
dd I9279
dd I927A
dd I927B
dd I927C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9280
dd I9288
dd I9290
dd I9298
dd I92A0
dd I92A8
dd I92B0
dw 0xc000 + 5
dd I92B8
dd I92B9
dd I92BA
dd I92BB
dd I92BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I92C0
dd I92C8
dd I92D0
dd I92D8
dd I92E0
dd I92E8
dd I92F0
dw 0xc000 + 5
dd I92F8
dd I92F9
dd I92FA
dd I92FB
dd I92FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9300
dd I9308
dd I9310
dd I9318
dd I9320
dd I9328
dd I9330
dw 0xc000 + 2
dd I9338
dd I9339
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9340
dd I9348
dd I9350
dd I9358
dd I9360
dd I9368
dd I9370
dw 0xc000 + 2
dd I9378
dd I9379
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9380
dd I9388
dd I9390
dd I9398
dd I93A0
dd I93A8
dd I93B0
dw 0xc000 + 2
dd I93B8
dd I93B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I93C0
dd I93C8
dd I93D0
dd I93D8
dd I93E0
dd I93E8
dd I93F0
dw 0xc000 + 5
dd I93F8
dd I93F9
dd I93FA
dd I93FB
dd I93FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9410
dd I9418
dd I9420
dd I9428
dd I9430
dw 0xc000 + 5
dd I9438
dd I9439
dd I943A
dd I943B
dd I943C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9440
dd I9448
dd I9450
dd I9458
dd I9460
dd I9468
dd I9470
dw 0xc000 + 5
dd I9478
dd I9479
dd I947A
dd I947B
dd I947C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9480
dd I9488
dd I9490
dd I9498
dd I94A0
dd I94A8
dd I94B0
dw 0xc000 + 5
dd I94B8
dd I94B9
dd I94BA
dd I94BB
dd I94BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I94C0
dd I94C8
dd I94D0
dd I94D8
dd I94E0
dd I94E8
dd I94F0
dw 0xc000 + 5
dd I94F8
dd I94F9
dd I94FA
dd I94FB
dd I94FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9500
dd I9508
dd I9510
dd I9518
dd I9520
dd I9528
dd I9530
dw 0xc000 + 2
dd I9538
dd I9539
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9540
dd I9548
dd I9550
dd I9558
dd I9560
dd I9568
dd I9570
dw 0xc000 + 2
dd I9578
dd I9579
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9580
dd I9588
dd I9590
dd I9598
dd I95A0
dd I95A8
dd I95B0
dw 0xc000 + 2
dd I95B8
dd I95B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I95C0
dd I95C8
dd I95D0
dd I95D8
dd I95E0
dd I95E8
dd I95F0
dw 0xc000 + 5
dd I95F8
dd I95F9
dd I95FA
dd I95FB
dd I95FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9610
dd I9618
dd I9620
dd I9628
dd I9630
dw 0xc000 + 5
dd I9638
dd I9639
dd I963A
dd I963B
dd I963C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9640
dd I9648
dd I9650
dd I9658
dd I9660
dd I9668
dd I9670
dw 0xc000 + 5
dd I9678
dd I9679
dd I967A
dd I967B
dd I967C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9680
dd I9688
dd I9690
dd I9698
dd I96A0
dd I96A8
dd I96B0
dw 0xc000 + 5
dd I96B8
dd I96B9
dd I96BA
dd I96BB
dd I96BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I96C0
dd I96C8
dd I96D0
dd I96D8
dd I96E0
dd I96E8
dd I96F0
dw 0xc000 + 5
dd I96F8
dd I96F9
dd I96FA
dd I96FB
dd I96FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9700
dd I9708
dd I9710
dd I9718
dd I9720
dd I9728
dd I9730
dw 0xc000 + 2
dd I9738
dd I9739
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9740
dd I9748
dd I9750
dd I9758
dd I9760
dd I9768
dd I9770
dw 0xc000 + 2
dd I9778
dd I9779
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9780
dd I9788
dd I9790
dd I9798
dd I97A0
dd I97A8
dd I97B0
dw 0xc000 + 2
dd I97B8
dd I97B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I97C0
dd I97C8
dd I97D0
dd I97D8
dd I97E0
dd I97E8
dd I97F0
dw 0xc000 + 5
dd I97F8
dd I97F9
dd I97FA
dd I97FB
dd I97FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9810
dd I9818
dd I9820
dd I9828
dd I9830
dw 0xc000 + 5
dd I9838
dd I9839
dd I983A
dd I983B
dd I983C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9840
dd I9848
dd I9850
dd I9858
dd I9860
dd I9868
dd I9870
dw 0xc000 + 5
dd I9878
dd I9879
dd I987A
dd I987B
dd I987C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9880
dd I9888
dd I9890
dd I9898
dd I98A0
dd I98A8
dd I98B0
dw 0xc000 + 5
dd I98B8
dd I98B9
dd I98BA
dd I98BB
dd I98BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I98C0
dd I98C8
dd I98D0
dd I98D8
dd I98E0
dd I98E8
dd I98F0
dw 0xc000 + 5
dd I98F8
dd I98F9
dd I98FA
dd I98FB
dd I98FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9900
dd I9908
dd I9910
dd I9918
dd I9920
dd I9928
dd I9930
dw 0xc000 + 2
dd I9938
dd I9939
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9940
dd I9948
dd I9950
dd I9958
dd I9960
dd I9968
dd I9970
dw 0xc000 + 2
dd I9978
dd I9979
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9980
dd I9988
dd I9990
dd I9998
dd I99A0
dd I99A8
dd I99B0
dw 0xc000 + 2
dd I99B8
dd I99B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I99C0
dd I99C8
dd I99D0
dd I99D8
dd I99E0
dd I99E8
dd I99F0
dw 0xc000 + 5
dd I99F8
dd I99F9
dd I99FA
dd I99FB
dd I99FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9A00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9A10
dd I9A18
dd I9A20
dd I9A28
dd I9A30
dw 0xc000 + 5
dd I9A38
dd I9A39
dd I9A3A
dd I9A3B
dd I9A3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9A40
dd I9A48
dd I9A50
dd I9A58
dd I9A60
dd I9A68
dd I9A70
dw 0xc000 + 5
dd I9A78
dd I9A79
dd I9A7A
dd I9A7B
dd I9A7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9A80
dd I9A88
dd I9A90
dd I9A98
dd I9AA0
dd I9AA8
dd I9AB0
dw 0xc000 + 5
dd I9AB8
dd I9AB9
dd I9ABA
dd I9ABB
dd I9ABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9AC0
dd I9AC8
dd I9AD0
dd I9AD8
dd I9AE0
dd I9AE8
dd I9AF0
dw 0xc000 + 5
dd I9AF8
dd I9AF9
dd I9AFA
dd I9AFB
dd I9AFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9B00
dd I9B08
dd I9B10
dd I9B18
dd I9B20
dd I9B28
dd I9B30
dw 0xc000 + 2
dd I9B38
dd I9B39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9B40
dd I9B48
dd I9B50
dd I9B58
dd I9B60
dd I9B68
dd I9B70
dw 0xc000 + 2
dd I9B78
dd I9B79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9B80
dd I9B88
dd I9B90
dd I9B98
dd I9BA0
dd I9BA8
dd I9BB0
dw 0xc000 + 2
dd I9BB8
dd I9BB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9BC0
dd I9BC8
dd I9BD0
dd I9BD8
dd I9BE0
dd I9BE8
dd I9BF0
dw 0xc000 + 5
dd I9BF8
dd I9BF9
dd I9BFA
dd I9BFB
dd I9BFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9C00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9C10
dd I9C18
dd I9C20
dd I9C28
dd I9C30
dw 0xc000 + 5
dd I9C38
dd I9C39
dd I9C3A
dd I9C3B
dd I9C3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9C40
dd I9C48
dd I9C50
dd I9C58
dd I9C60
dd I9C68
dd I9C70
dw 0xc000 + 5
dd I9C78
dd I9C79
dd I9C7A
dd I9C7B
dd I9C7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9C80
dd I9C88
dd I9C90
dd I9C98
dd I9CA0
dd I9CA8
dd I9CB0
dw 0xc000 + 5
dd I9CB8
dd I9CB9
dd I9CBA
dd I9CBB
dd I9CBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9CC0
dd I9CC8
dd I9CD0
dd I9CD8
dd I9CE0
dd I9CE8
dd I9CF0
dw 0xc000 + 5
dd I9CF8
dd I9CF9
dd I9CFA
dd I9CFB
dd I9CFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9D00
dd I9D08
dd I9D10
dd I9D18
dd I9D20
dd I9D28
dd I9D30
dw 0xc000 + 2
dd I9D38
dd I9D39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9D40
dd I9D48
dd I9D50
dd I9D58
dd I9D60
dd I9D68
dd I9D70
dw 0xc000 + 2
dd I9D78
dd I9D79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9D80
dd I9D88
dd I9D90
dd I9D98
dd I9DA0
dd I9DA8
dd I9DB0
dw 0xc000 + 2
dd I9DB8
dd I9DB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9DC0
dd I9DC8
dd I9DD0
dd I9DD8
dd I9DE0
dd I9DE8
dd I9DF0
dw 0xc000 + 5
dd I9DF8
dd I9DF9
dd I9DFA
dd I9DFB
dd I9DFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd I9E00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd I9E10
dd I9E18
dd I9E20
dd I9E28
dd I9E30
dw 0xc000 + 5
dd I9E38
dd I9E39
dd I9E3A
dd I9E3B
dd I9E3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9E40
dd I9E48
dd I9E50
dd I9E58
dd I9E60
dd I9E68
dd I9E70
dw 0xc000 + 5
dd I9E78
dd I9E79
dd I9E7A
dd I9E7B
dd I9E7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9E80
dd I9E88
dd I9E90
dd I9E98
dd I9EA0
dd I9EA8
dd I9EB0
dw 0xc000 + 5
dd I9EB8
dd I9EB9
dd I9EBA
dd I9EBB
dd I9EBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9EC0
dd I9EC8
dd I9ED0
dd I9ED8
dd I9EE0
dd I9EE8
dd I9EF0
dw 0xc000 + 5
dd I9EF8
dd I9EF9
dd I9EFA
dd I9EFB
dd I9EFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd I9F00
dd I9F08
dd I9F10
dd I9F18
dd I9F20
dd I9F28
dd I9F30
dw 0xc000 + 2
dd I9F38
dd I9F39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9F40
dd I9F48
dd I9F50
dd I9F58
dd I9F60
dd I9F68
dd I9F70
dw 0xc000 + 2
dd I9F78
dd I9F79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9F80
dd I9F88
dd I9F90
dd I9F98
dd I9FA0
dd I9FA8
dd I9FB0
dw 0xc000 + 2
dd I9FB8
dd I9FB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd I9FC0
dd I9FC8
dd I9FD0
dd I9FD8
dd I9FE0
dd I9FE8
dd I9FF0
dw 0xc000 + 5
dd I9FF8
dd I9FF9
dd I9FFA
dd I9FFB
dd I9FFC
dw 0x0000 + 3
dd I_Invalid
dw 4096
dd IA000
dw 0x8000 + 1
dd IB000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IB010
dd IB018
dd IB020
dd IB028
dd IB030
dw 0xc000 + 5
dd IB038
dd IB039
dd IB03A
dd IB03B
dd IB03C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB040
dd IB048
dd IB050
dd IB058
dd IB060
dd IB068
dd IB070
dw 0xc000 + 5
dd IB078
dd IB079
dd IB07A
dd IB07B
dd IB07C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB080
dd IB088
dd IB090
dd IB098
dd IB0A0
dd IB0A8
dd IB0B0
dw 0xc000 + 5
dd IB0B8
dd IB0B9
dd IB0BA
dd IB0BB
dd IB0BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB0C0
dd IB0C8
dd IB0D0
dd IB0D8
dd IB0E0
dd IB0E8
dd IB0F0
dw 0xc000 + 5
dd IB0F8
dd IB0F9
dd IB0FA
dd IB0FB
dd IB0FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB100
dd IB108
dd IB110
dd IB118
dd IB120
dd IB128
dd IB130
dw 0xc000 + 2
dd IB138
dd IB139
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB140
dd IB148
dd IB150
dd IB158
dd IB160
dd IB168
dd IB170
dw 0xc000 + 2
dd IB178
dd IB179
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB180
dd IB188
dd IB190
dd IB198
dd IB1A0
dd IB1A8
dd IB1B0
dw 0xc000 + 2
dd IB1B8
dd IB1B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB1C0
dd IB1C8
dd IB1D0
dd IB1D8
dd IB1E0
dd IB1E8
dd IB1F0
dw 0xc000 + 5
dd IB1F8
dd IB1F9
dd IB1FA
dd IB1FB
dd IB1FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IB200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IB210
dd IB218
dd IB220
dd IB228
dd IB230
dw 0xc000 + 5
dd IB238
dd IB239
dd IB23A
dd IB23B
dd IB23C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB240
dd IB248
dd IB250
dd IB258
dd IB260
dd IB268
dd IB270
dw 0xc000 + 5
dd IB278
dd IB279
dd IB27A
dd IB27B
dd IB27C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB280
dd IB288
dd IB290
dd IB298
dd IB2A0
dd IB2A8
dd IB2B0
dw 0xc000 + 5
dd IB2B8
dd IB2B9
dd IB2BA
dd IB2BB
dd IB2BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB2C0
dd IB2C8
dd IB2D0
dd IB2D8
dd IB2E0
dd IB2E8
dd IB2F0
dw 0xc000 + 5
dd IB2F8
dd IB2F9
dd IB2FA
dd IB2FB
dd IB2FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB300
dd IB308
dd IB310
dd IB318
dd IB320
dd IB328
dd IB330
dw 0xc000 + 2
dd IB338
dd IB339
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB340
dd IB348
dd IB350
dd IB358
dd IB360
dd IB368
dd IB370
dw 0xc000 + 2
dd IB378
dd IB379
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB380
dd IB388
dd IB390
dd IB398
dd IB3A0
dd IB3A8
dd IB3B0
dw 0xc000 + 2
dd IB3B8
dd IB3B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB3C0
dd IB3C8
dd IB3D0
dd IB3D8
dd IB3E0
dd IB3E8
dd IB3F0
dw 0xc000 + 5
dd IB3F8
dd IB3F9
dd IB3FA
dd IB3FB
dd IB3FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IB400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IB410
dd IB418
dd IB420
dd IB428
dd IB430
dw 0xc000 + 5
dd IB438
dd IB439
dd IB43A
dd IB43B
dd IB43C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB440
dd IB448
dd IB450
dd IB458
dd IB460
dd IB468
dd IB470
dw 0xc000 + 5
dd IB478
dd IB479
dd IB47A
dd IB47B
dd IB47C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB480
dd IB488
dd IB490
dd IB498
dd IB4A0
dd IB4A8
dd IB4B0
dw 0xc000 + 5
dd IB4B8
dd IB4B9
dd IB4BA
dd IB4BB
dd IB4BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB4C0
dd IB4C8
dd IB4D0
dd IB4D8
dd IB4E0
dd IB4E8
dd IB4F0
dw 0xc000 + 5
dd IB4F8
dd IB4F9
dd IB4FA
dd IB4FB
dd IB4FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB500
dd IB508
dd IB510
dd IB518
dd IB520
dd IB528
dd IB530
dw 0xc000 + 2
dd IB538
dd IB539
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB540
dd IB548
dd IB550
dd IB558
dd IB560
dd IB568
dd IB570
dw 0xc000 + 2
dd IB578
dd IB579
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB580
dd IB588
dd IB590
dd IB598
dd IB5A0
dd IB5A8
dd IB5B0
dw 0xc000 + 2
dd IB5B8
dd IB5B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB5C0
dd IB5C8
dd IB5D0
dd IB5D8
dd IB5E0
dd IB5E8
dd IB5F0
dw 0xc000 + 5
dd IB5F8
dd IB5F9
dd IB5FA
dd IB5FB
dd IB5FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IB600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IB610
dd IB618
dd IB620
dd IB628
dd IB630
dw 0xc000 + 5
dd IB638
dd IB639
dd IB63A
dd IB63B
dd IB63C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB640
dd IB648
dd IB650
dd IB658
dd IB660
dd IB668
dd IB670
dw 0xc000 + 5
dd IB678
dd IB679
dd IB67A
dd IB67B
dd IB67C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB680
dd IB688
dd IB690
dd IB698
dd IB6A0
dd IB6A8
dd IB6B0
dw 0xc000 + 5
dd IB6B8
dd IB6B9
dd IB6BA
dd IB6BB
dd IB6BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB6C0
dd IB6C8
dd IB6D0
dd IB6D8
dd IB6E0
dd IB6E8
dd IB6F0
dw 0xc000 + 5
dd IB6F8
dd IB6F9
dd IB6FA
dd IB6FB
dd IB6FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB700
dd IB708
dd IB710
dd IB718
dd IB720
dd IB728
dd IB730
dw 0xc000 + 2
dd IB738
dd IB739
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB740
dd IB748
dd IB750
dd IB758
dd IB760
dd IB768
dd IB770
dw 0xc000 + 2
dd IB778
dd IB779
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB780
dd IB788
dd IB790
dd IB798
dd IB7A0
dd IB7A8
dd IB7B0
dw 0xc000 + 2
dd IB7B8
dd IB7B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB7C0
dd IB7C8
dd IB7D0
dd IB7D8
dd IB7E0
dd IB7E8
dd IB7F0
dw 0xc000 + 5
dd IB7F8
dd IB7F9
dd IB7FA
dd IB7FB
dd IB7FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IB800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IB810
dd IB818
dd IB820
dd IB828
dd IB830
dw 0xc000 + 5
dd IB838
dd IB839
dd IB83A
dd IB83B
dd IB83C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB840
dd IB848
dd IB850
dd IB858
dd IB860
dd IB868
dd IB870
dw 0xc000 + 5
dd IB878
dd IB879
dd IB87A
dd IB87B
dd IB87C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB880
dd IB888
dd IB890
dd IB898
dd IB8A0
dd IB8A8
dd IB8B0
dw 0xc000 + 5
dd IB8B8
dd IB8B9
dd IB8BA
dd IB8BB
dd IB8BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB8C0
dd IB8C8
dd IB8D0
dd IB8D8
dd IB8E0
dd IB8E8
dd IB8F0
dw 0xc000 + 5
dd IB8F8
dd IB8F9
dd IB8FA
dd IB8FB
dd IB8FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IB900
dd IB908
dd IB910
dd IB918
dd IB920
dd IB928
dd IB930
dw 0xc000 + 2
dd IB938
dd IB939
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB940
dd IB948
dd IB950
dd IB958
dd IB960
dd IB968
dd IB970
dw 0xc000 + 2
dd IB978
dd IB979
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB980
dd IB988
dd IB990
dd IB998
dd IB9A0
dd IB9A8
dd IB9B0
dw 0xc000 + 2
dd IB9B8
dd IB9B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IB9C0
dd IB9C8
dd IB9D0
dd IB9D8
dd IB9E0
dd IB9E8
dd IB9F0
dw 0xc000 + 5
dd IB9F8
dd IB9F9
dd IB9FA
dd IB9FB
dd IB9FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IBA00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IBA10
dd IBA18
dd IBA20
dd IBA28
dd IBA30
dw 0xc000 + 5
dd IBA38
dd IBA39
dd IBA3A
dd IBA3B
dd IBA3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBA40
dd IBA48
dd IBA50
dd IBA58
dd IBA60
dd IBA68
dd IBA70
dw 0xc000 + 5
dd IBA78
dd IBA79
dd IBA7A
dd IBA7B
dd IBA7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBA80
dd IBA88
dd IBA90
dd IBA98
dd IBAA0
dd IBAA8
dd IBAB0
dw 0xc000 + 5
dd IBAB8
dd IBAB9
dd IBABA
dd IBABB
dd IBABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBAC0
dd IBAC8
dd IBAD0
dd IBAD8
dd IBAE0
dd IBAE8
dd IBAF0
dw 0xc000 + 5
dd IBAF8
dd IBAF9
dd IBAFA
dd IBAFB
dd IBAFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBB00
dd IBB08
dd IBB10
dd IBB18
dd IBB20
dd IBB28
dd IBB30
dw 0xc000 + 2
dd IBB38
dd IBB39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBB40
dd IBB48
dd IBB50
dd IBB58
dd IBB60
dd IBB68
dd IBB70
dw 0xc000 + 2
dd IBB78
dd IBB79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBB80
dd IBB88
dd IBB90
dd IBB98
dd IBBA0
dd IBBA8
dd IBBB0
dw 0xc000 + 2
dd IBBB8
dd IBBB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBBC0
dd IBBC8
dd IBBD0
dd IBBD8
dd IBBE0
dd IBBE8
dd IBBF0
dw 0xc000 + 5
dd IBBF8
dd IBBF9
dd IBBFA
dd IBBFB
dd IBBFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IBC00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IBC10
dd IBC18
dd IBC20
dd IBC28
dd IBC30
dw 0xc000 + 5
dd IBC38
dd IBC39
dd IBC3A
dd IBC3B
dd IBC3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBC40
dd IBC48
dd IBC50
dd IBC58
dd IBC60
dd IBC68
dd IBC70
dw 0xc000 + 5
dd IBC78
dd IBC79
dd IBC7A
dd IBC7B
dd IBC7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBC80
dd IBC88
dd IBC90
dd IBC98
dd IBCA0
dd IBCA8
dd IBCB0
dw 0xc000 + 5
dd IBCB8
dd IBCB9
dd IBCBA
dd IBCBB
dd IBCBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBCC0
dd IBCC8
dd IBCD0
dd IBCD8
dd IBCE0
dd IBCE8
dd IBCF0
dw 0xc000 + 5
dd IBCF8
dd IBCF9
dd IBCFA
dd IBCFB
dd IBCFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBD00
dd IBD08
dd IBD10
dd IBD18
dd IBD20
dd IBD28
dd IBD30
dw 0xc000 + 2
dd IBD38
dd IBD39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBD40
dd IBD48
dd IBD50
dd IBD58
dd IBD60
dd IBD68
dd IBD70
dw 0xc000 + 2
dd IBD78
dd IBD79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBD80
dd IBD88
dd IBD90
dd IBD98
dd IBDA0
dd IBDA8
dd IBDB0
dw 0xc000 + 2
dd IBDB8
dd IBDB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBDC0
dd IBDC8
dd IBDD0
dd IBDD8
dd IBDE0
dd IBDE8
dd IBDF0
dw 0xc000 + 5
dd IBDF8
dd IBDF9
dd IBDFA
dd IBDFB
dd IBDFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IBE00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IBE10
dd IBE18
dd IBE20
dd IBE28
dd IBE30
dw 0xc000 + 5
dd IBE38
dd IBE39
dd IBE3A
dd IBE3B
dd IBE3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBE40
dd IBE48
dd IBE50
dd IBE58
dd IBE60
dd IBE68
dd IBE70
dw 0xc000 + 5
dd IBE78
dd IBE79
dd IBE7A
dd IBE7B
dd IBE7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBE80
dd IBE88
dd IBE90
dd IBE98
dd IBEA0
dd IBEA8
dd IBEB0
dw 0xc000 + 5
dd IBEB8
dd IBEB9
dd IBEBA
dd IBEBB
dd IBEBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBEC0
dd IBEC8
dd IBED0
dd IBED8
dd IBEE0
dd IBEE8
dd IBEF0
dw 0xc000 + 5
dd IBEF8
dd IBEF9
dd IBEFA
dd IBEFB
dd IBEFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IBF00
dd IBF08
dd IBF10
dd IBF18
dd IBF20
dd IBF28
dd IBF30
dw 0xc000 + 2
dd IBF38
dd IBF39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBF40
dd IBF48
dd IBF50
dd IBF58
dd IBF60
dd IBF68
dd IBF70
dw 0xc000 + 2
dd IBF78
dd IBF79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBF80
dd IBF88
dd IBF90
dd IBF98
dd IBFA0
dd IBFA8
dd IBFB0
dw 0xc000 + 2
dd IBFB8
dd IBFB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IBFC0
dd IBFC8
dd IBFD0
dd IBFD8
dd IBFE0
dd IBFE8
dd IBFF0
dw 0xc000 + 5
dd IBFF8
dd IBFF9
dd IBFFA
dd IBFFB
dd IBFFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC010
dd IC018
dd IC020
dd IC028
dd IC030
dw 0xc000 + 5
dd IC038
dd IC039
dd IC03A
dd IC03B
dd IC03C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC040
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC050
dd IC058
dd IC060
dd IC068
dd IC070
dw 0xc000 + 5
dd IC078
dd IC079
dd IC07A
dd IC07B
dd IC07C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC080
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC090
dd IC098
dd IC0A0
dd IC0A8
dd IC0B0
dw 0xc000 + 5
dd IC0B8
dd IC0B9
dd IC0BA
dd IC0BB
dd IC0BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC0C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC0D0
dd IC0D8
dd IC0E0
dd IC0E8
dd IC0F0
dw 0xc000 + 5
dd IC0F8
dd IC0F9
dd IC0FA
dd IC0FB
dd IC0FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IC100
dd IC108
dd IC110
dd IC118
dd IC120
dd IC128
dd IC130
dw 0xc000 + 2
dd IC138
dd IC139
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IC140
dd IC148
dd IC150
dd IC158
dd IC160
dd IC168
dd IC170
dw 0xc000 + 2
dd IC178
dd IC179
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd IC188
dd IC190
dd IC198
dd IC1A0
dd IC1A8
dd IC1B0
dw 0xc000 + 2
dd IC1B8
dd IC1B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd IC1C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC1D0
dd IC1D8
dd IC1E0
dd IC1E8
dd IC1F0
dw 0xc000 + 5
dd IC1F8
dd IC1F9
dd IC1FA
dd IC1FB
dd IC1FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC210
dd IC218
dd IC220
dd IC228
dd IC230
dw 0xc000 + 5
dd IC238
dd IC239
dd IC23A
dd IC23B
dd IC23C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC240
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC250
dd IC258
dd IC260
dd IC268
dd IC270
dw 0xc000 + 5
dd IC278
dd IC279
dd IC27A
dd IC27B
dd IC27C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC280
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC290
dd IC298
dd IC2A0
dd IC2A8
dd IC2B0
dw 0xc000 + 5
dd IC2B8
dd IC2B9
dd IC2BA
dd IC2BB
dd IC2BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC2C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC2D0
dd IC2D8
dd IC2E0
dd IC2E8
dd IC2F0
dw 0xc000 + 5
dd IC2F8
dd IC2F9
dd IC2FA
dd IC2FB
dd IC2FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IC300
dd IC308
dd IC310
dd IC318
dd IC320
dd IC328
dd IC330
dw 0xc000 + 2
dd IC338
dd IC339
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IC340
dd IC348
dd IC350
dd IC358
dd IC360
dd IC368
dd IC370
dw 0xc000 + 2
dd IC378
dd IC379
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd IC388
dd IC390
dd IC398
dd IC3A0
dd IC3A8
dd IC3B0
dw 0xc000 + 2
dd IC3B8
dd IC3B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd IC3C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC3D0
dd IC3D8
dd IC3E0
dd IC3E8
dd IC3F0
dw 0xc000 + 5
dd IC3F8
dd IC3F9
dd IC3FA
dd IC3FB
dd IC3FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC410
dd IC418
dd IC420
dd IC428
dd IC430
dw 0xc000 + 5
dd IC438
dd IC439
dd IC43A
dd IC43B
dd IC43C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC440
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC450
dd IC458
dd IC460
dd IC468
dd IC470
dw 0xc000 + 5
dd IC478
dd IC479
dd IC47A
dd IC47B
dd IC47C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC480
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC490
dd IC498
dd IC4A0
dd IC4A8
dd IC4B0
dw 0xc000 + 5
dd IC4B8
dd IC4B9
dd IC4BA
dd IC4BB
dd IC4BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC4C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC4D0
dd IC4D8
dd IC4E0
dd IC4E8
dd IC4F0
dw 0xc000 + 5
dd IC4F8
dd IC4F9
dd IC4FA
dd IC4FB
dd IC4FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IC500
dd IC508
dd IC510
dd IC518
dd IC520
dd IC528
dd IC530
dw 0xc000 + 2
dd IC538
dd IC539
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IC540
dd IC548
dd IC550
dd IC558
dd IC560
dd IC568
dd IC570
dw 0xc000 + 2
dd IC578
dd IC579
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd IC588
dd IC590
dd IC598
dd IC5A0
dd IC5A8
dd IC5B0
dw 0xc000 + 2
dd IC5B8
dd IC5B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd IC5C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC5D0
dd IC5D8
dd IC5E0
dd IC5E8
dd IC5F0
dw 0xc000 + 5
dd IC5F8
dd IC5F9
dd IC5FA
dd IC5FB
dd IC5FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC610
dd IC618
dd IC620
dd IC628
dd IC630
dw 0xc000 + 5
dd IC638
dd IC639
dd IC63A
dd IC63B
dd IC63C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC640
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC650
dd IC658
dd IC660
dd IC668
dd IC670
dw 0xc000 + 5
dd IC678
dd IC679
dd IC67A
dd IC67B
dd IC67C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC680
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC690
dd IC698
dd IC6A0
dd IC6A8
dd IC6B0
dw 0xc000 + 5
dd IC6B8
dd IC6B9
dd IC6BA
dd IC6BB
dd IC6BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC6C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC6D0
dd IC6D8
dd IC6E0
dd IC6E8
dd IC6F0
dw 0xc000 + 5
dd IC6F8
dd IC6F9
dd IC6FA
dd IC6FB
dd IC6FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IC700
dd IC708
dd IC710
dd IC718
dd IC720
dd IC728
dd IC730
dw 0xc000 + 2
dd IC738
dd IC739
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IC740
dd IC748
dd IC750
dd IC758
dd IC760
dd IC768
dd IC770
dw 0xc000 + 2
dd IC778
dd IC779
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd IC788
dd IC790
dd IC798
dd IC7A0
dd IC7A8
dd IC7B0
dw 0xc000 + 2
dd IC7B8
dd IC7B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd IC7C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC7D0
dd IC7D8
dd IC7E0
dd IC7E8
dd IC7F0
dw 0xc000 + 5
dd IC7F8
dd IC7F9
dd IC7FA
dd IC7FB
dd IC7FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC810
dd IC818
dd IC820
dd IC828
dd IC830
dw 0xc000 + 5
dd IC838
dd IC839
dd IC83A
dd IC83B
dd IC83C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC840
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC850
dd IC858
dd IC860
dd IC868
dd IC870
dw 0xc000 + 5
dd IC878
dd IC879
dd IC87A
dd IC87B
dd IC87C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC880
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC890
dd IC898
dd IC8A0
dd IC8A8
dd IC8B0
dw 0xc000 + 5
dd IC8B8
dd IC8B9
dd IC8BA
dd IC8BB
dd IC8BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IC8C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC8D0
dd IC8D8
dd IC8E0
dd IC8E8
dd IC8F0
dw 0xc000 + 5
dd IC8F8
dd IC8F9
dd IC8FA
dd IC8FB
dd IC8FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IC900
dd IC908
dd IC910
dd IC918
dd IC920
dd IC928
dd IC930
dw 0xc000 + 2
dd IC938
dd IC939
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IC940
dd IC948
dd IC950
dd IC958
dd IC960
dd IC968
dd IC970
dw 0xc000 + 2
dd IC978
dd IC979
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd IC988
dd IC990
dd IC998
dd IC9A0
dd IC9A8
dd IC9B0
dw 0xc000 + 2
dd IC9B8
dd IC9B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd IC9C0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IC9D0
dd IC9D8
dd IC9E0
dd IC9E8
dd IC9F0
dw 0xc000 + 5
dd IC9F8
dd IC9F9
dd IC9FA
dd IC9FB
dd IC9FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICA00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICA10
dd ICA18
dd ICA20
dd ICA28
dd ICA30
dw 0xc000 + 5
dd ICA38
dd ICA39
dd ICA3A
dd ICA3B
dd ICA3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICA40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICA50
dd ICA58
dd ICA60
dd ICA68
dd ICA70
dw 0xc000 + 5
dd ICA78
dd ICA79
dd ICA7A
dd ICA7B
dd ICA7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICA80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICA90
dd ICA98
dd ICAA0
dd ICAA8
dd ICAB0
dw 0xc000 + 5
dd ICAB8
dd ICAB9
dd ICABA
dd ICABB
dd ICABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICAC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICAD0
dd ICAD8
dd ICAE0
dd ICAE8
dd ICAF0
dw 0xc000 + 5
dd ICAF8
dd ICAF9
dd ICAFA
dd ICAFB
dd ICAFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ICB00
dd ICB08
dd ICB10
dd ICB18
dd ICB20
dd ICB28
dd ICB30
dw 0xc000 + 2
dd ICB38
dd ICB39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ICB40
dd ICB48
dd ICB50
dd ICB58
dd ICB60
dd ICB68
dd ICB70
dw 0xc000 + 2
dd ICB78
dd ICB79
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd ICB88
dd ICB90
dd ICB98
dd ICBA0
dd ICBA8
dd ICBB0
dw 0xc000 + 2
dd ICBB8
dd ICBB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd ICBC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICBD0
dd ICBD8
dd ICBE0
dd ICBE8
dd ICBF0
dw 0xc000 + 5
dd ICBF8
dd ICBF9
dd ICBFA
dd ICBFB
dd ICBFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICC00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICC10
dd ICC18
dd ICC20
dd ICC28
dd ICC30
dw 0xc000 + 5
dd ICC38
dd ICC39
dd ICC3A
dd ICC3B
dd ICC3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICC40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICC50
dd ICC58
dd ICC60
dd ICC68
dd ICC70
dw 0xc000 + 5
dd ICC78
dd ICC79
dd ICC7A
dd ICC7B
dd ICC7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICC80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICC90
dd ICC98
dd ICCA0
dd ICCA8
dd ICCB0
dw 0xc000 + 5
dd ICCB8
dd ICCB9
dd ICCBA
dd ICCBB
dd ICCBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICCC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICCD0
dd ICCD8
dd ICCE0
dd ICCE8
dd ICCF0
dw 0xc000 + 5
dd ICCF8
dd ICCF9
dd ICCFA
dd ICCFB
dd ICCFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ICD00
dd ICD08
dd ICD10
dd ICD18
dd ICD20
dd ICD28
dd ICD30
dw 0xc000 + 2
dd ICD38
dd ICD39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ICD40
dd ICD48
dd ICD50
dd ICD58
dd ICD60
dd ICD68
dd ICD70
dw 0xc000 + 2
dd ICD78
dd ICD79
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd ICD88
dd ICD90
dd ICD98
dd ICDA0
dd ICDA8
dd ICDB0
dw 0xc000 + 2
dd ICDB8
dd ICDB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd ICDC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICDD0
dd ICDD8
dd ICDE0
dd ICDE8
dd ICDF0
dw 0xc000 + 5
dd ICDF8
dd ICDF9
dd ICDFA
dd ICDFB
dd ICDFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICE00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICE10
dd ICE18
dd ICE20
dd ICE28
dd ICE30
dw 0xc000 + 5
dd ICE38
dd ICE39
dd ICE3A
dd ICE3B
dd ICE3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICE40
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICE50
dd ICE58
dd ICE60
dd ICE68
dd ICE70
dw 0xc000 + 5
dd ICE78
dd ICE79
dd ICE7A
dd ICE7B
dd ICE7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICE80
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICE90
dd ICE98
dd ICEA0
dd ICEA8
dd ICEB0
dw 0xc000 + 5
dd ICEB8
dd ICEB9
dd ICEBA
dd ICEBB
dd ICEBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ICEC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICED0
dd ICED8
dd ICEE0
dd ICEE8
dd ICEF0
dw 0xc000 + 5
dd ICEF8
dd ICEF9
dd ICEFA
dd ICEFB
dd ICEFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ICF00
dd ICF08
dd ICF10
dd ICF18
dd ICF20
dd ICF28
dd ICF30
dw 0xc000 + 2
dd ICF38
dd ICF39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ICF40
dd ICF48
dd ICF50
dd ICF58
dd ICF60
dd ICF68
dd ICF70
dw 0xc000 + 2
dd ICF78
dd ICF79
dw 0x0000 + 14
dd I_Invalid
dw 0x8000 + 6
dd ICF88
dd ICF90
dd ICF98
dd ICFA0
dd ICFA8
dd ICFB0
dw 0xc000 + 2
dd ICFB8
dd ICFB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 1
dd ICFC0
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ICFD0
dd ICFD8
dd ICFE0
dd ICFE8
dd ICFF0
dw 0xc000 + 5
dd ICFF8
dd ICFF9
dd ICFFA
dd ICFFB
dd ICFFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ID000
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ID010
dd ID018
dd ID020
dd ID028
dd ID030
dw 0xc000 + 5
dd ID038
dd ID039
dd ID03A
dd ID03B
dd ID03C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID040
dd ID048
dd ID050
dd ID058
dd ID060
dd ID068
dd ID070
dw 0xc000 + 5
dd ID078
dd ID079
dd ID07A
dd ID07B
dd ID07C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID080
dd ID088
dd ID090
dd ID098
dd ID0A0
dd ID0A8
dd ID0B0
dw 0xc000 + 5
dd ID0B8
dd ID0B9
dd ID0BA
dd ID0BB
dd ID0BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID0C0
dd ID0C8
dd ID0D0
dd ID0D8
dd ID0E0
dd ID0E8
dd ID0F0
dw 0xc000 + 5
dd ID0F8
dd ID0F9
dd ID0FA
dd ID0FB
dd ID0FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID100
dd ID108
dd ID110
dd ID118
dd ID120
dd ID128
dd ID130
dw 0xc000 + 2
dd ID138
dd ID139
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID140
dd ID148
dd ID150
dd ID158
dd ID160
dd ID168
dd ID170
dw 0xc000 + 2
dd ID178
dd ID179
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID180
dd ID188
dd ID190
dd ID198
dd ID1A0
dd ID1A8
dd ID1B0
dw 0xc000 + 2
dd ID1B8
dd ID1B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID1C0
dd ID1C8
dd ID1D0
dd ID1D8
dd ID1E0
dd ID1E8
dd ID1F0
dw 0xc000 + 5
dd ID1F8
dd ID1F9
dd ID1FA
dd ID1FB
dd ID1FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ID200
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ID210
dd ID218
dd ID220
dd ID228
dd ID230
dw 0xc000 + 5
dd ID238
dd ID239
dd ID23A
dd ID23B
dd ID23C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID240
dd ID248
dd ID250
dd ID258
dd ID260
dd ID268
dd ID270
dw 0xc000 + 5
dd ID278
dd ID279
dd ID27A
dd ID27B
dd ID27C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID280
dd ID288
dd ID290
dd ID298
dd ID2A0
dd ID2A8
dd ID2B0
dw 0xc000 + 5
dd ID2B8
dd ID2B9
dd ID2BA
dd ID2BB
dd ID2BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID2C0
dd ID2C8
dd ID2D0
dd ID2D8
dd ID2E0
dd ID2E8
dd ID2F0
dw 0xc000 + 5
dd ID2F8
dd ID2F9
dd ID2FA
dd ID2FB
dd ID2FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID300
dd ID308
dd ID310
dd ID318
dd ID320
dd ID328
dd ID330
dw 0xc000 + 2
dd ID338
dd ID339
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID340
dd ID348
dd ID350
dd ID358
dd ID360
dd ID368
dd ID370
dw 0xc000 + 2
dd ID378
dd ID379
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID380
dd ID388
dd ID390
dd ID398
dd ID3A0
dd ID3A8
dd ID3B0
dw 0xc000 + 2
dd ID3B8
dd ID3B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID3C0
dd ID3C8
dd ID3D0
dd ID3D8
dd ID3E0
dd ID3E8
dd ID3F0
dw 0xc000 + 5
dd ID3F8
dd ID3F9
dd ID3FA
dd ID3FB
dd ID3FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ID400
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ID410
dd ID418
dd ID420
dd ID428
dd ID430
dw 0xc000 + 5
dd ID438
dd ID439
dd ID43A
dd ID43B
dd ID43C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID440
dd ID448
dd ID450
dd ID458
dd ID460
dd ID468
dd ID470
dw 0xc000 + 5
dd ID478
dd ID479
dd ID47A
dd ID47B
dd ID47C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID480
dd ID488
dd ID490
dd ID498
dd ID4A0
dd ID4A8
dd ID4B0
dw 0xc000 + 5
dd ID4B8
dd ID4B9
dd ID4BA
dd ID4BB
dd ID4BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID4C0
dd ID4C8
dd ID4D0
dd ID4D8
dd ID4E0
dd ID4E8
dd ID4F0
dw 0xc000 + 5
dd ID4F8
dd ID4F9
dd ID4FA
dd ID4FB
dd ID4FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID500
dd ID508
dd ID510
dd ID518
dd ID520
dd ID528
dd ID530
dw 0xc000 + 2
dd ID538
dd ID539
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID540
dd ID548
dd ID550
dd ID558
dd ID560
dd ID568
dd ID570
dw 0xc000 + 2
dd ID578
dd ID579
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID580
dd ID588
dd ID590
dd ID598
dd ID5A0
dd ID5A8
dd ID5B0
dw 0xc000 + 2
dd ID5B8
dd ID5B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID5C0
dd ID5C8
dd ID5D0
dd ID5D8
dd ID5E0
dd ID5E8
dd ID5F0
dw 0xc000 + 5
dd ID5F8
dd ID5F9
dd ID5FA
dd ID5FB
dd ID5FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ID600
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ID610
dd ID618
dd ID620
dd ID628
dd ID630
dw 0xc000 + 5
dd ID638
dd ID639
dd ID63A
dd ID63B
dd ID63C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID640
dd ID648
dd ID650
dd ID658
dd ID660
dd ID668
dd ID670
dw 0xc000 + 5
dd ID678
dd ID679
dd ID67A
dd ID67B
dd ID67C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID680
dd ID688
dd ID690
dd ID698
dd ID6A0
dd ID6A8
dd ID6B0
dw 0xc000 + 5
dd ID6B8
dd ID6B9
dd ID6BA
dd ID6BB
dd ID6BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID6C0
dd ID6C8
dd ID6D0
dd ID6D8
dd ID6E0
dd ID6E8
dd ID6F0
dw 0xc000 + 5
dd ID6F8
dd ID6F9
dd ID6FA
dd ID6FB
dd ID6FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID700
dd ID708
dd ID710
dd ID718
dd ID720
dd ID728
dd ID730
dw 0xc000 + 2
dd ID738
dd ID739
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID740
dd ID748
dd ID750
dd ID758
dd ID760
dd ID768
dd ID770
dw 0xc000 + 2
dd ID778
dd ID779
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID780
dd ID788
dd ID790
dd ID798
dd ID7A0
dd ID7A8
dd ID7B0
dw 0xc000 + 2
dd ID7B8
dd ID7B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID7C0
dd ID7C8
dd ID7D0
dd ID7D8
dd ID7E0
dd ID7E8
dd ID7F0
dw 0xc000 + 5
dd ID7F8
dd ID7F9
dd ID7FA
dd ID7FB
dd ID7FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd ID800
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd ID810
dd ID818
dd ID820
dd ID828
dd ID830
dw 0xc000 + 5
dd ID838
dd ID839
dd ID83A
dd ID83B
dd ID83C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID840
dd ID848
dd ID850
dd ID858
dd ID860
dd ID868
dd ID870
dw 0xc000 + 5
dd ID878
dd ID879
dd ID87A
dd ID87B
dd ID87C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID880
dd ID888
dd ID890
dd ID898
dd ID8A0
dd ID8A8
dd ID8B0
dw 0xc000 + 5
dd ID8B8
dd ID8B9
dd ID8BA
dd ID8BB
dd ID8BC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID8C0
dd ID8C8
dd ID8D0
dd ID8D8
dd ID8E0
dd ID8E8
dd ID8F0
dw 0xc000 + 5
dd ID8F8
dd ID8F9
dd ID8FA
dd ID8FB
dd ID8FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd ID900
dd ID908
dd ID910
dd ID918
dd ID920
dd ID928
dd ID930
dw 0xc000 + 2
dd ID938
dd ID939
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID940
dd ID948
dd ID950
dd ID958
dd ID960
dd ID968
dd ID970
dw 0xc000 + 2
dd ID978
dd ID979
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID980
dd ID988
dd ID990
dd ID998
dd ID9A0
dd ID9A8
dd ID9B0
dw 0xc000 + 2
dd ID9B8
dd ID9B9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd ID9C0
dd ID9C8
dd ID9D0
dd ID9D8
dd ID9E0
dd ID9E8
dd ID9F0
dw 0xc000 + 5
dd ID9F8
dd ID9F9
dd ID9FA
dd ID9FB
dd ID9FC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IDA00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IDA10
dd IDA18
dd IDA20
dd IDA28
dd IDA30
dw 0xc000 + 5
dd IDA38
dd IDA39
dd IDA3A
dd IDA3B
dd IDA3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDA40
dd IDA48
dd IDA50
dd IDA58
dd IDA60
dd IDA68
dd IDA70
dw 0xc000 + 5
dd IDA78
dd IDA79
dd IDA7A
dd IDA7B
dd IDA7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDA80
dd IDA88
dd IDA90
dd IDA98
dd IDAA0
dd IDAA8
dd IDAB0
dw 0xc000 + 5
dd IDAB8
dd IDAB9
dd IDABA
dd IDABB
dd IDABC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDAC0
dd IDAC8
dd IDAD0
dd IDAD8
dd IDAE0
dd IDAE8
dd IDAF0
dw 0xc000 + 5
dd IDAF8
dd IDAF9
dd IDAFA
dd IDAFB
dd IDAFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDB00
dd IDB08
dd IDB10
dd IDB18
dd IDB20
dd IDB28
dd IDB30
dw 0xc000 + 2
dd IDB38
dd IDB39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDB40
dd IDB48
dd IDB50
dd IDB58
dd IDB60
dd IDB68
dd IDB70
dw 0xc000 + 2
dd IDB78
dd IDB79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDB80
dd IDB88
dd IDB90
dd IDB98
dd IDBA0
dd IDBA8
dd IDBB0
dw 0xc000 + 2
dd IDBB8
dd IDBB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDBC0
dd IDBC8
dd IDBD0
dd IDBD8
dd IDBE0
dd IDBE8
dd IDBF0
dw 0xc000 + 5
dd IDBF8
dd IDBF9
dd IDBFA
dd IDBFB
dd IDBFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IDC00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IDC10
dd IDC18
dd IDC20
dd IDC28
dd IDC30
dw 0xc000 + 5
dd IDC38
dd IDC39
dd IDC3A
dd IDC3B
dd IDC3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDC40
dd IDC48
dd IDC50
dd IDC58
dd IDC60
dd IDC68
dd IDC70
dw 0xc000 + 5
dd IDC78
dd IDC79
dd IDC7A
dd IDC7B
dd IDC7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDC80
dd IDC88
dd IDC90
dd IDC98
dd IDCA0
dd IDCA8
dd IDCB0
dw 0xc000 + 5
dd IDCB8
dd IDCB9
dd IDCBA
dd IDCBB
dd IDCBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDCC0
dd IDCC8
dd IDCD0
dd IDCD8
dd IDCE0
dd IDCE8
dd IDCF0
dw 0xc000 + 5
dd IDCF8
dd IDCF9
dd IDCFA
dd IDCFB
dd IDCFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDD00
dd IDD08
dd IDD10
dd IDD18
dd IDD20
dd IDD28
dd IDD30
dw 0xc000 + 2
dd IDD38
dd IDD39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDD40
dd IDD48
dd IDD50
dd IDD58
dd IDD60
dd IDD68
dd IDD70
dw 0xc000 + 2
dd IDD78
dd IDD79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDD80
dd IDD88
dd IDD90
dd IDD98
dd IDDA0
dd IDDA8
dd IDDB0
dw 0xc000 + 2
dd IDDB8
dd IDDB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDDC0
dd IDDC8
dd IDDD0
dd IDDD8
dd IDDE0
dd IDDE8
dd IDDF0
dw 0xc000 + 5
dd IDDF8
dd IDDF9
dd IDDFA
dd IDDFB
dd IDDFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 1
dd IDE00
dw 1
dd I_Invalid
dw 0x0000 + 7
dd I_Invalid
dw 0x8000 + 5
dd IDE10
dd IDE18
dd IDE20
dd IDE28
dd IDE30
dw 0xc000 + 5
dd IDE38
dd IDE39
dd IDE3A
dd IDE3B
dd IDE3C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDE40
dd IDE48
dd IDE50
dd IDE58
dd IDE60
dd IDE68
dd IDE70
dw 0xc000 + 5
dd IDE78
dd IDE79
dd IDE7A
dd IDE7B
dd IDE7C
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDE80
dd IDE88
dd IDE90
dd IDE98
dd IDEA0
dd IDEA8
dd IDEB0
dw 0xc000 + 5
dd IDEB8
dd IDEB9
dd IDEBA
dd IDEBB
dd IDEBC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDEC0
dd IDEC8
dd IDED0
dd IDED8
dd IDEE0
dd IDEE8
dd IDEF0
dw 0xc000 + 5
dd IDEF8
dd IDEF9
dd IDEFA
dd IDEFB
dd IDEFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 7
dd IDF00
dd IDF08
dd IDF10
dd IDF18
dd IDF20
dd IDF28
dd IDF30
dw 0xc000 + 2
dd IDF38
dd IDF39
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDF40
dd IDF48
dd IDF50
dd IDF58
dd IDF60
dd IDF68
dd IDF70
dw 0xc000 + 2
dd IDF78
dd IDF79
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDF80
dd IDF88
dd IDF90
dd IDF98
dd IDFA0
dd IDFA8
dd IDFB0
dw 0xc000 + 2
dd IDFB8
dd IDFB9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 7
dd IDFC0
dd IDFC8
dd IDFD0
dd IDFD8
dd IDFE0
dd IDFE8
dd IDFF0
dw 0xc000 + 5
dd IDFF8
dd IDFF9
dd IDFFA
dd IDFFB
dd IDFFC
dw 0x0000 + 3
dd I_Invalid
dw 0x8000 + 24
dd IE000
dd IE008
dd IE010
dd IE018
dd IE020
dd IE028
dd IE030
dd IE038
dd IE040
dd IE048
dd IE050
dd IE058
dd IE060
dd IE068
dd IE070
dd IE078
dd IE080
dd IE088
dd IE090
dd IE098
dd IE0A0
dd IE0A8
dd IE0B0
dd IE0B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE0D0
dd IE0D8
dd IE0E0
dd IE0E8
dd IE0F0
dw 0xc000 + 2
dd IE0F8
dd IE0F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE100
dd IE108
dd IE110
dd IE118
dd IE120
dd IE128
dd IE130
dd IE138
dd IE140
dd IE148
dd IE150
dd IE158
dd IE160
dd IE168
dd IE170
dd IE178
dd IE180
dd IE188
dd IE190
dd IE198
dd IE1A0
dd IE1A8
dd IE1B0
dd IE1B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE1D0
dd IE1D8
dd IE1E0
dd IE1E8
dd IE1F0
dw 0xc000 + 2
dd IE1F8
dd IE1F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE200
dd IE208
dd IE210
dd IE218
dd IE220
dd IE228
dd IE230
dd IE238
dd IE240
dd IE248
dd IE250
dd IE258
dd IE260
dd IE268
dd IE270
dd IE278
dd IE280
dd IE288
dd IE290
dd IE298
dd IE2A0
dd IE2A8
dd IE2B0
dd IE2B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE2D0
dd IE2D8
dd IE2E0
dd IE2E8
dd IE2F0
dw 0xc000 + 2
dd IE2F8
dd IE2F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE300
dd IE308
dd IE310
dd IE318
dd IE320
dd IE328
dd IE330
dd IE338
dd IE340
dd IE348
dd IE350
dd IE358
dd IE360
dd IE368
dd IE370
dd IE378
dd IE380
dd IE388
dd IE390
dd IE398
dd IE3A0
dd IE3A8
dd IE3B0
dd IE3B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE3D0
dd IE3D8
dd IE3E0
dd IE3E8
dd IE3F0
dw 0xc000 + 2
dd IE3F8
dd IE3F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE400
dd IE408
dd IE410
dd IE418
dd IE420
dd IE428
dd IE430
dd IE438
dd IE440
dd IE448
dd IE450
dd IE458
dd IE460
dd IE468
dd IE470
dd IE478
dd IE480
dd IE488
dd IE490
dd IE498
dd IE4A0
dd IE4A8
dd IE4B0
dd IE4B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE4D0
dd IE4D8
dd IE4E0
dd IE4E8
dd IE4F0
dw 0xc000 + 2
dd IE4F8
dd IE4F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE500
dd IE508
dd IE510
dd IE518
dd IE520
dd IE528
dd IE530
dd IE538
dd IE540
dd IE548
dd IE550
dd IE558
dd IE560
dd IE568
dd IE570
dd IE578
dd IE580
dd IE588
dd IE590
dd IE598
dd IE5A0
dd IE5A8
dd IE5B0
dd IE5B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE5D0
dd IE5D8
dd IE5E0
dd IE5E8
dd IE5F0
dw 0xc000 + 2
dd IE5F8
dd IE5F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE600
dd IE608
dd IE610
dd IE618
dd IE620
dd IE628
dd IE630
dd IE638
dd IE640
dd IE648
dd IE650
dd IE658
dd IE660
dd IE668
dd IE670
dd IE678
dd IE680
dd IE688
dd IE690
dd IE698
dd IE6A0
dd IE6A8
dd IE6B0
dd IE6B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE6D0
dd IE6D8
dd IE6E0
dd IE6E8
dd IE6F0
dw 0xc000 + 2
dd IE6F8
dd IE6F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE700
dd IE708
dd IE710
dd IE718
dd IE720
dd IE728
dd IE730
dd IE738
dd IE740
dd IE748
dd IE750
dd IE758
dd IE760
dd IE768
dd IE770
dd IE778
dd IE780
dd IE788
dd IE790
dd IE798
dd IE7A0
dd IE7A8
dd IE7B0
dd IE7B8
dw 1
dd I_Invalid
dw 0x0000 + 15
dd I_Invalid
dw 0x8000 + 5
dd IE7D0
dd IE7D8
dd IE7E0
dd IE7E8
dd IE7F0
dw 0xc000 + 2
dd IE7F8
dd IE7F9
dw 0x0000 + 6
dd I_Invalid
dw 0x8000 + 24
dd IE800
dd IE808
dd IE810
dd IE818
dd IE820
dd IE828
dd IE830
dd IE838
dd IE840
dd IE848
dd IE850
dd IE858
dd IE860
dd IE868
dd IE870
dd IE878
dd IE880
dd IE888
dd IE890
dd IE898
dd IE8A0
dd IE8A8
dd IE8B0
dd IE8B8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IE900
dd IE908
dd IE910
dd IE918
dd IE920
dd IE928
dd IE930
dd IE938
dd IE940
dd IE948
dd IE950
dd IE958
dd IE960
dd IE968
dd IE970
dd IE978
dd IE980
dd IE988
dd IE990
dd IE998
dd IE9A0
dd IE9A8
dd IE9B0
dd IE9B8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IEA00
dd IEA08
dd IEA10
dd IEA18
dd IEA20
dd IEA28
dd IEA30
dd IEA38
dd IEA40
dd IEA48
dd IEA50
dd IEA58
dd IEA60
dd IEA68
dd IEA70
dd IEA78
dd IEA80
dd IEA88
dd IEA90
dd IEA98
dd IEAA0
dd IEAA8
dd IEAB0
dd IEAB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IEB00
dd IEB08
dd IEB10
dd IEB18
dd IEB20
dd IEB28
dd IEB30
dd IEB38
dd IEB40
dd IEB48
dd IEB50
dd IEB58
dd IEB60
dd IEB68
dd IEB70
dd IEB78
dd IEB80
dd IEB88
dd IEB90
dd IEB98
dd IEBA0
dd IEBA8
dd IEBB0
dd IEBB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IEC00
dd IEC08
dd IEC10
dd IEC18
dd IEC20
dd IEC28
dd IEC30
dd IEC38
dd IEC40
dd IEC48
dd IEC50
dd IEC58
dd IEC60
dd IEC68
dd IEC70
dd IEC78
dd IEC80
dd IEC88
dd IEC90
dd IEC98
dd IECA0
dd IECA8
dd IECB0
dd IECB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IED00
dd IED08
dd IED10
dd IED18
dd IED20
dd IED28
dd IED30
dd IED38
dd IED40
dd IED48
dd IED50
dd IED58
dd IED60
dd IED68
dd IED70
dd IED78
dd IED80
dd IED88
dd IED90
dd IED98
dd IEDA0
dd IEDA8
dd IEDB0
dd IEDB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IEE00
dd IEE08
dd IEE10
dd IEE18
dd IEE20
dd IEE28
dd IEE30
dd IEE38
dd IEE40
dd IEE48
dd IEE50
dd IEE58
dd IEE60
dd IEE68
dd IEE70
dd IEE78
dd IEE80
dd IEE88
dd IEE90
dd IEE98
dd IEEA0
dd IEEA8
dd IEEB0
dd IEEB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 0x8000 + 24
dd IEF00
dd IEF08
dd IEF10
dd IEF18
dd IEF20
dd IEF28
dd IEF30
dd IEF38
dd IEF40
dd IEF48
dd IEF50
dd IEF58
dd IEF60
dd IEF68
dd IEF70
dd IEF78
dd IEF80
dd IEF88
dd IEF90
dd IEF98
dd IEFA0
dd IEFA8
dd IEFB0
dd IEFB8
dw 1
dd I_Invalid
dw 0x0000 + 63
dd I_Invalid
dw 4096
dd IF000
dd -1
